@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -32,6 +32,85 @@
 
 #include "helpers.h"
 #include "catch.hpp"
+#include "tiledb/common/logger.h"
+#include "tiledb/sm/cpp_api/tiledb"
+#include "tiledb/sm/enums/encryption_type.h"
+#include "tiledb/sm/global_state/unit_test_config.h"
+#include "tiledb/sm/misc/constants.h"
+#include "tiledb/sm/misc/uri.h"
+#include "tiledb/sm/subarray/subarray_partitioner.h"
+
+std::mutex catch2_macro_mutex;
+
+namespace tiledb {
+namespace test {
+
+// Command line arguments.
+extern std::string g_vfs;
+
+bool use_refactored_dense_reader() {
+  const char* value = nullptr;
+  tiledb_config_t* cfg;
+  tiledb_error_t* err = nullptr;
+  auto rc = tiledb_config_alloc(&cfg, &err);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(err == nullptr);
+
+  rc = tiledb_config_get(cfg, "sm.query.dense.reader", &value, &err);
+  CHECK(rc == TILEDB_OK);
+  CHECK(err == nullptr);
+
+  bool use_refactored_readers = strcmp(value, "refactored") == 0;
+
+  tiledb_config_free(&cfg);
+
+  return use_refactored_readers;
+}
+
+bool use_refactored_sparse_global_order_reader() {
+  const char* value = nullptr;
+  tiledb_config_t* cfg;
+  tiledb_error_t* err = nullptr;
+  auto rc = tiledb_config_alloc(&cfg, &err);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(err == nullptr);
+
+  rc = tiledb_config_get(
+      cfg, "sm.query.sparse_global_order.reader", &value, &err);
+  CHECK(rc == TILEDB_OK);
+  CHECK(err == nullptr);
+
+  bool use_refactored_readers = strcmp(value, "refactored") == 0;
+
+  tiledb_config_free(&cfg);
+
+  return use_refactored_readers;
+}
+
+bool use_refactored_sparse_unordered_with_dups_reader() {
+  const char* value = nullptr;
+  tiledb_config_t* cfg;
+  tiledb_error_t* err = nullptr;
+  auto rc = tiledb_config_alloc(&cfg, &err);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(err == nullptr);
+
+  rc = tiledb_config_get(
+      cfg, "sm.query.sparse_unordered_with_dups.reader", &value, &err);
+  CHECK(rc == TILEDB_OK);
+  CHECK(err == nullptr);
+
+  bool use_refactored_readers = strcmp(value, "refactored") == 0;
+
+  tiledb_config_free(&cfg);
+
+  return use_refactored_readers;
+}
+
+tdb_shared_ptr<Logger> g_helper_logger(void) {
+  static tdb_shared_ptr<Logger> g_helper_logger = tdb_make_shared(Logger, "");
+  return g_helper_logger;
+}
 
 template <class T>
 void check_partitions(
@@ -66,8 +145,8 @@ void check_partitions(
     CHECK(unsplittable);
   } else {
     CHECK(!unsplittable);
-    //    CHECK(partitioner.next(&unsplittable).ok());
-    //    CHECK(!unsplittable);
+    CHECK(partitioner.next(&unsplittable).ok());
+    CHECK(!unsplittable);
     CHECK(partitioner.done());
   }
 }
@@ -92,14 +171,16 @@ void check_subarray(
 
   // Check ranges
   uint64_t dim_range_num = 0;
-  const T* range;
+  const sm::Range* range;
   for (unsigned i = 0; i < dim_num; ++i) {
     CHECK(subarray.get_range_num(i, &dim_range_num).ok());
     CHECK(dim_range_num == ranges[i].size() / 2);
     for (uint64_t j = 0; j < dim_range_num; ++j) {
-      subarray.get_range(i, j, (const void**)&range);
-      CHECK(range[0] == ranges[i][2 * j]);
-      CHECK(range[1] == ranges[i][2 * j + 1]);
+      subarray.get_range(i, j, &range);
+      auto r = (const T*)range->data();
+
+      CHECK(r[0] == ranges[i][2 * j]);
+      CHECK(r[1] == ranges[i][2 * j + 1]);
     }
   }
 }
@@ -109,6 +190,69 @@ void close_array(tiledb_ctx_t* ctx, tiledb_array_t* array) {
   CHECK(rc == TILEDB_OK);
 }
 
+int array_create_wrapper(
+    tiledb_ctx_t* ctx,
+    const std::string& path,
+    tiledb_array_schema_t* array_schema,
+    bool serialize_array_schema) {
+#ifndef TILEDB_SERIALIZATION
+  return tiledb_array_create(ctx, path.c_str(), array_schema);
+#endif
+
+  if (!serialize_array_schema) {
+    return tiledb_array_create(ctx, path.c_str(), array_schema);
+  }
+
+  // Serialize the array
+  tiledb_buffer_t* buff;
+  REQUIRE(
+      tiledb_serialize_array_schema(
+          ctx,
+          array_schema,
+          (tiledb_serialization_type_t)tiledb::sm::SerializationType::CAPNP,
+          1,
+          &buff) == TILEDB_OK);
+
+  // Load array schema from the rest server
+  tiledb_array_schema_t* new_array_schema = nullptr;
+  REQUIRE(
+      tiledb_deserialize_array_schema(
+          ctx,
+          buff,
+          (tiledb_serialization_type_t)tiledb::sm::SerializationType::CAPNP,
+          0,
+          &new_array_schema) == TILEDB_OK);
+
+  // Create array from new schema
+  int rc = tiledb_array_create(ctx, path.c_str(), new_array_schema);
+
+  // Serialize the new array schema and deserialize into the original array
+  // schema.
+  tiledb_buffer_t* buff2;
+  REQUIRE(
+      tiledb_serialize_array_schema(
+          ctx,
+          new_array_schema,
+          (tiledb_serialization_type_t)tiledb::sm::SerializationType::CAPNP,
+          0,
+          &buff2) == TILEDB_OK);
+  REQUIRE(
+      tiledb_deserialize_array_schema(
+          ctx,
+          buff2,
+          (tiledb_serialization_type_t)tiledb::sm::SerializationType::CAPNP,
+          1,
+          &array_schema) == TILEDB_OK);
+
+  // Clean up.
+  tiledb_array_schema_free(&array_schema);
+  tiledb_array_schema_free(&new_array_schema);
+  tiledb_buffer_free(&buff);
+  tiledb_buffer_free(&buff2);
+
+  return rc;
+}
+
 void create_array(
     tiledb_ctx_t* ctx,
     const std::string& array_name,
@@ -123,7 +267,9 @@ void create_array(
     const std::vector<std::pair<tiledb_filter_type_t, int>>& compressors,
     tiledb_layout_t tile_order,
     tiledb_layout_t cell_order,
-    uint64_t capacity) {
+    uint64_t capacity,
+    bool allows_dups,
+    bool serialize_array_schema) {
   // For easy reference
   auto dim_num = dim_names.size();
   auto attr_num = attr_names.size();
@@ -146,6 +292,8 @@ void create_array(
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_schema_set_capacity(ctx, array_schema, capacity);
   REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_schema_set_allows_dups(ctx, array_schema, (int)allows_dups);
+  REQUIRE(rc == TILEDB_OK);
 
   // Create dimensions and domain
   tiledb_domain_t* domain;
@@ -191,7 +339,8 @@ void create_array(
   REQUIRE(rc == TILEDB_OK);
 
   // Create array
-  rc = tiledb_array_create(ctx, array_name.c_str(), array_schema);
+  rc = array_create_wrapper(
+      ctx, array_name, array_schema, serialize_array_schema);
   REQUIRE(rc == TILEDB_OK);
 
   // Clean up
@@ -283,12 +432,29 @@ void create_array(
   REQUIRE(rc == TILEDB_OK);
 
   // Create array
-  rc = tiledb_array_create_with_key(
-      ctx, array_name.c_str(), array_schema, enc_type, key, key_len);
+  tiledb_config_t* config;
+  tiledb_error_t* error = nullptr;
+  rc = tiledb_config_alloc(&config, &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  std::string encryption_type_string =
+      encryption_type_str((tiledb::sm::EncryptionType)enc_type);
+  rc = tiledb_config_set(
+      config, "sm.encryption_type", encryption_type_string.c_str(), &error);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.encryption_key", key, &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  tiledb::sm::UnitTestConfig::instance().array_encryption_key_length.set(
+      key_len);
+  tiledb_ctx_t* ctx_array;
+  REQUIRE(tiledb_ctx_alloc(config, &ctx_array) == TILEDB_OK);
+  rc = tiledb_array_create(ctx_array, array_name.c_str(), array_schema);
   REQUIRE(rc == TILEDB_OK);
 
   // Clean up
   tiledb_array_schema_free(&array_schema);
+  tiledb_ctx_free(&ctx_array);
 }
 
 void create_s3_bucket(
@@ -308,8 +474,29 @@ void create_s3_bucket(
   }
 }
 
+void create_azure_container(
+    const std::string& container_name,
+    bool azure_supported,
+    tiledb_ctx_t* ctx,
+    tiledb_vfs_t* vfs) {
+  if (azure_supported) {
+    // Create container if it does not exist
+    int is_container = 0;
+    int rc =
+        tiledb_vfs_is_bucket(ctx, vfs, container_name.c_str(), &is_container);
+    REQUIRE(rc == TILEDB_OK);
+    if (!is_container) {
+      rc = tiledb_vfs_create_bucket(ctx, vfs, container_name.c_str());
+      REQUIRE(rc == TILEDB_OK);
+    }
+  }
+}
+
 void create_ctx_and_vfs(
-    bool s3_supported, tiledb_ctx_t** ctx, tiledb_vfs_t** vfs) {
+    bool s3_supported,
+    bool azure_supported,
+    tiledb_ctx_t** ctx,
+    tiledb_vfs_t** vfs) {
   // Create TileDB context
   tiledb_config_t* config = nullptr;
   tiledb_error_t* error = nullptr;
@@ -334,6 +521,30 @@ void create_ctx_and_vfs(
     REQUIRE(error == nullptr);
 #endif
   }
+  if (azure_supported) {
+    REQUIRE(
+        tiledb_config_set(
+            config,
+            "vfs.azure.storage_account_name",
+            "devstoreaccount1",
+            &error) == TILEDB_OK);
+    REQUIRE(
+        tiledb_config_set(
+            config,
+            "vfs.azure.storage_account_key",
+            "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/"
+            "K1SZFPTOtr/KBHBeksoGMGw==",
+            &error) == TILEDB_OK);
+    REQUIRE(
+        tiledb_config_set(
+            config,
+            "vfs.azure.blob_endpoint",
+            "127.0.0.1:10000/devstoreaccount1",
+            &error) == TILEDB_OK);
+    REQUIRE(
+        tiledb_config_set(config, "vfs.azure.use_https", "false", &error) ==
+        TILEDB_OK);
+  }
   REQUIRE(tiledb_ctx_alloc(config, ctx) == TILEDB_OK);
   REQUIRE(error == nullptr);
 
@@ -353,21 +564,28 @@ void create_subarray(
     tiledb::sm::Array* array,
     const SubarrayRanges<T>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray) {
-  tiledb::sm::Subarray ret(array, layout);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges) {
+  tiledb::sm::Subarray ret(
+      array, layout, &g_helper_stats, g_helper_logger(), coalesce_ranges);
 
   auto dim_num = (unsigned)ranges.size();
-  for (unsigned i = 0; i < dim_num; ++i) {
-    auto dim_range_num = ranges[i].size() / 2;
+  for (unsigned d = 0; d < dim_num; ++d) {
+    auto dim_range_num = ranges[d].size() / 2;
     for (size_t j = 0; j < dim_range_num; ++j) {
-      ret.add_range(i, &ranges[i][2 * j]);
+      sm::Range range(&ranges[d][2 * j], 2 * sizeof(T));
+      ret.add_range(d, std::move(range), true);
     }
   }
 
   *subarray = ret;
 }
 
-void get_supported_fs(bool* s3_supported, bool* hdfs_supported) {
+void get_supported_fs(
+    bool* s3_supported,
+    bool* hdfs_supported,
+    bool* azure_supported,
+    bool* gcs_supported) {
   tiledb_ctx_t* ctx = nullptr;
   REQUIRE(tiledb_ctx_alloc(nullptr, &ctx) == TILEDB_OK);
 
@@ -378,6 +596,54 @@ void get_supported_fs(bool* s3_supported, bool* hdfs_supported) {
   rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_HDFS, &is_supported);
   REQUIRE(rc == TILEDB_OK);
   *hdfs_supported = (bool)is_supported;
+  rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_AZURE, &is_supported);
+  REQUIRE(rc == TILEDB_OK);
+  *azure_supported = (bool)is_supported;
+  rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_GCS, &is_supported);
+  REQUIRE(rc == TILEDB_OK);
+  *gcs_supported = (bool)is_supported;
+
+  // Override VFS support if the user used the '--vfs' command line argument.
+  if (!g_vfs.empty()) {
+    REQUIRE(
+        (g_vfs == "native" || g_vfs == "s3" || g_vfs == "hdfs" ||
+         g_vfs == "azure" || g_vfs == "gcs"));
+
+    if (g_vfs == "native") {
+      *s3_supported = false;
+      *hdfs_supported = false;
+      *azure_supported = false;
+      *gcs_supported = false;
+    }
+
+    if (g_vfs == "s3") {
+      *s3_supported = true;
+      *hdfs_supported = false;
+      *azure_supported = false;
+      *gcs_supported = false;
+    }
+
+    if (g_vfs == "hdfs") {
+      *s3_supported = false;
+      *hdfs_supported = true;
+      *azure_supported = false;
+      *gcs_supported = false;
+    }
+
+    if (g_vfs == "azure") {
+      *s3_supported = false;
+      *hdfs_supported = false;
+      *azure_supported = true;
+      *gcs_supported = false;
+    }
+
+    if (g_vfs == "gcs") {
+      *s3_supported = false;
+      *hdfs_supported = false;
+      *azure_supported = false;
+      *gcs_supported = true;
+    }
+  }
 
   tiledb_ctx_free(&ctx);
 }
@@ -388,7 +654,7 @@ void open_array(
   CHECK(rc == TILEDB_OK);
 }
 
-std::string random_bucket_name(const std::string& prefix) {
+std::string random_name(const std::string& prefix) {
   std::stringstream ss;
   ss << prefix << "-" << std::this_thread::get_id() << "-"
      << TILEDB_TIMESTAMP_NOW_MS;
@@ -449,11 +715,181 @@ void write_array(
     tiledb_ctx_t* ctx,
     const std::string& array_name,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers) {
-  // Open array
+    const QueryBuffers& buffers) {
+  write_array(
+      ctx, array_name, TILEDB_TIMESTAMP_NOW_MS, nullptr, layout, buffers);
+}
+
+void write_array(
+    tiledb_ctx_t* ctx,
+    const std::string& array_name,
+    uint64_t timestamp,
+    tiledb_layout_t layout,
+    const QueryBuffers& buffers) {
+  write_array(ctx, array_name, timestamp, nullptr, layout, buffers);
+}
+
+void write_array(
+    tiledb_ctx_t* ctx,
+    const std::string& array_name,
+    tiledb_encryption_type_t encryption_type,
+    const char* key,
+    uint64_t key_len,
+    uint64_t timestamp,
+    tiledb_layout_t layout,
+    const QueryBuffers& buffers) {
+  write_array(
+      ctx,
+      array_name,
+      encryption_type,
+      key,
+      key_len,
+      timestamp,
+      nullptr,
+      layout,
+      buffers);
+}
+
+void write_array(
+    tiledb_ctx_t* ctx,
+    const std::string& array_name,
+    const void* subarray,
+    tiledb_layout_t layout,
+    const QueryBuffers& buffers) {
+  write_array(
+      ctx, array_name, TILEDB_TIMESTAMP_NOW_MS, subarray, layout, buffers);
+}
+
+void write_array(
+    tiledb_ctx_t* ctx,
+    const std::string& array_name,
+    uint64_t timestamp,
+    const void* subarray,
+    tiledb_layout_t layout,
+    const QueryBuffers& buffers) {
+  std::string uri;
+  write_array(ctx, array_name, timestamp, subarray, layout, buffers, &uri);
+  (void)uri;
+}
+
+void write_array(
+    tiledb_ctx_t* ctx,
+    const std::string& array_name,
+    tiledb_encryption_type_t encryption_type,
+    const char* key,
+    uint64_t key_len,
+    uint64_t timestamp,
+    const void* subarray,
+    tiledb_layout_t layout,
+    const QueryBuffers& buffers) {
+  std::string uri;
+  write_array(
+      ctx,
+      array_name,
+      encryption_type,
+      key,
+      key_len,
+      timestamp,
+      subarray,
+      layout,
+      buffers,
+      &uri);
+  (void)uri;
+}
+
+void write_array(
+    tiledb_ctx_t* ctx,
+    const std::string& array_name,
+    uint64_t timestamp,
+    tiledb_layout_t layout,
+    const QueryBuffers& buffers,
+    std::string* uri) {
+  write_array(ctx, array_name, timestamp, nullptr, layout, buffers, uri);
+}
+
+void write_array(
+    tiledb_ctx_t* ctx,
+    const std::string& array_name,
+    tiledb_encryption_type_t encryption_type,
+    const char* key,
+    uint64_t key_len,
+    uint64_t timestamp,
+    tiledb_layout_t layout,
+    const QueryBuffers& buffers,
+    std::string* uri) {
+  write_array(
+      ctx,
+      array_name,
+      encryption_type,
+      key,
+      key_len,
+      timestamp,
+      nullptr,
+      layout,
+      buffers,
+      uri);
+}
+
+void write_array(
+    tiledb_ctx_t* ctx,
+    const std::string& array_name,
+    uint64_t timestamp,
+    const void* subarray,
+    tiledb_layout_t layout,
+    const QueryBuffers& buffers,
+    std::string* uri) {
+  write_array(
+      ctx,
+      array_name,
+      TILEDB_NO_ENCRYPTION,
+      nullptr,
+      0,
+      timestamp,
+      subarray,
+      layout,
+      buffers,
+      uri);
+}
+
+void write_array(
+    tiledb_ctx_t* ctx,
+    const std::string& array_name,
+    tiledb_encryption_type_t encryption_type,
+    const char* key,
+    uint64_t key_len,
+    uint64_t timestamp,
+    const void* subarray,
+    tiledb_layout_t layout,
+    const QueryBuffers& buffers,
+    std::string* uri) {
+  // Set array configuration
   tiledb_array_t* array;
   int rc = tiledb_array_alloc(ctx, array_name.c_str(), &array);
   CHECK(rc == TILEDB_OK);
+  tiledb_config_t* cfg;
+  tiledb_error_t* err = nullptr;
+  REQUIRE(tiledb_config_alloc(&cfg, &err) == TILEDB_OK);
+  REQUIRE(err == nullptr);
+
+  rc = tiledb_array_set_open_timestamp_end(ctx, array, timestamp);
+  REQUIRE(rc == TILEDB_OK);
+
+  // Open array
+  if (encryption_type != TILEDB_NO_ENCRYPTION) {
+    std::string encryption_type_string =
+        encryption_type_str((tiledb::sm::EncryptionType)encryption_type);
+    rc = tiledb_config_set(
+        cfg, "sm.encryption_type", encryption_type_string.c_str(), &err);
+    REQUIRE(rc == TILEDB_OK);
+    REQUIRE(err == nullptr);
+    rc = tiledb_config_set(cfg, "sm.encryption_key", key, &err);
+    REQUIRE(rc == TILEDB_OK);
+    REQUIRE(err == nullptr);
+    rc = tiledb_array_set_config(ctx, array, cfg);
+    REQUIRE(rc == TILEDB_OK);
+    tiledb::sm::UnitTestConfig::instance().array_encryption_key_length.set(
+        key_len);
+  }
   rc = tiledb_array_open(ctx, array, TILEDB_WRITE);
   CHECK(rc == TILEDB_OK);
 
@@ -461,13 +897,17 @@ void write_array(
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx, array, TILEDB_WRITE, &query);
   CHECK(rc == TILEDB_OK);
+  if (subarray != nullptr) {
+    rc = tiledb_query_set_subarray(ctx, query, subarray);
+    CHECK(rc == TILEDB_OK);
+  }
   rc = tiledb_query_set_layout(ctx, query, layout);
   CHECK(rc == TILEDB_OK);
 
   // Set buffers
-  for (const auto& b : attr_buffers) {
+  for (const auto& b : buffers) {
     if (b.second.var_ == nullptr) {  // Fixed-sized
-      rc = tiledb_query_set_buffer(
+      rc = tiledb_query_set_data_buffer(
           ctx,
           query,
           b.first.c_str(),
@@ -475,14 +915,19 @@ void write_array(
           (uint64_t*)&(b.second.fixed_size_));
       CHECK(rc == TILEDB_OK);
     } else {  // Var-sized
-      rc = tiledb_query_set_buffer_var(
+      rc = tiledb_query_set_data_buffer(
+          ctx,
+          query,
+          b.first.c_str(),
+          b.second.var_,
+          (uint64_t*)&(b.second.var_size_));
+      CHECK(rc == TILEDB_OK);
+      rc = tiledb_query_set_offsets_buffer(
           ctx,
           query,
           b.first.c_str(),
           (uint64_t*)b.second.fixed_,
-          (uint64_t*)&(b.second.fixed_size_),
-          b.second.var_,
-          (uint64_t*)&(b.second.var_size_));
+          (uint64_t*)&(b.second.fixed_size_));
       CHECK(rc == TILEDB_OK);
     }
   }
@@ -495,6 +940,12 @@ void write_array(
   rc = tiledb_query_finalize(ctx, query);
   CHECK(rc == TILEDB_OK);
 
+  // Get fragment uri
+  const char* temp_uri;
+  rc = tiledb_query_get_fragment_uri(ctx, query, 0, &temp_uri);
+  REQUIRE(rc == TILEDB_OK);
+  *uri = std::string(temp_uri);
+
   // Close array
   rc = tiledb_array_close(ctx, array);
   CHECK(rc == TILEDB_OK);
@@ -502,6 +953,7 @@ void write_array(
   // Clean up
   tiledb_array_free(&array);
   tiledb_query_free(&query);
+  tiledb_config_free(&cfg);
 }
 
 template <class T>
@@ -510,7 +962,7 @@ void read_array(
     tiledb_array_t* array,
     const SubarrayRanges<T>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers) {
+    const QueryBuffers& buffers) {
   // Create query
   tiledb_query_t* query;
   int rc = tiledb_query_alloc(ctx, array, TILEDB_READ, &query);
@@ -529,9 +981,9 @@ void read_array(
   }
 
   // Set buffers
-  for (const auto& b : attr_buffers) {
+  for (const auto& b : buffers) {
     if (b.second.var_ == nullptr) {  // Fixed-sized
-      rc = tiledb_query_set_buffer(
+      rc = tiledb_query_set_data_buffer(
           ctx,
           query,
           b.first.c_str(),
@@ -539,14 +991,19 @@ void read_array(
           (uint64_t*)&(b.second.fixed_size_));
       CHECK(rc == TILEDB_OK);
     } else {  // Var-sized
-      rc = tiledb_query_set_buffer_var(
+      rc = tiledb_query_set_data_buffer(
+          ctx,
+          query,
+          b.first.c_str(),
+          b.second.var_,
+          (uint64_t*)&(b.second.var_size_));
+      CHECK(rc == TILEDB_OK);
+      rc = tiledb_query_set_offsets_buffer(
           ctx,
           query,
           b.first.c_str(),
           (uint64_t*)b.second.fixed_,
-          (uint64_t*)&(b.second.fixed_size_),
-          b.second.var_,
-          (uint64_t*)&(b.second.var_size_));
+          (uint64_t*)&(b.second.fixed_size_));
       CHECK(rc == TILEDB_OK);
     }
   }
@@ -565,7 +1022,25 @@ void read_array(
   tiledb_query_free(&query);
 }
 
-// Explicit template instantiations
+int32_t num_fragments(const std::string& array_name) {
+  Context ctx;
+  VFS vfs(ctx);
+
+  // Get all URIs in the array directory
+  auto uris = vfs.ls(array_name);
+
+  // Exclude '__meta' folder and any file with a suffix
+  int ret = 0;
+  for (const auto& uri : uris) {
+    auto name = tiledb::sm::URI(uri).remove_trailing_slash().last_path_part();
+    if (name != tiledb::sm::constants::array_metadata_folder_name &&
+        name != tiledb::sm::constants::array_schema_folder_name &&
+        name.find_first_of('.') == std::string::npos)
+      ++ret;
+  }
+
+  return ret;
+}
 
 template void check_subarray<int8_t>(
     tiledb::sm::Subarray& subarray, const SubarrayRanges<int8_t>& ranges);
@@ -592,52 +1067,62 @@ template void create_subarray<int8_t>(
     tiledb::sm::Array* array,
     const SubarrayRanges<int8_t>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 template void create_subarray<uint8_t>(
     tiledb::sm::Array* array,
     const SubarrayRanges<uint8_t>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 template void create_subarray<int16_t>(
     tiledb::sm::Array* array,
     const SubarrayRanges<int16_t>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 template void create_subarray<uint16_t>(
     tiledb::sm::Array* array,
     const SubarrayRanges<uint16_t>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 template void create_subarray<int32_t>(
     tiledb::sm::Array* array,
     const SubarrayRanges<int32_t>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 template void create_subarray<uint32_t>(
     tiledb::sm::Array* array,
     const SubarrayRanges<uint32_t>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 template void create_subarray<int64_t>(
     tiledb::sm::Array* array,
     const SubarrayRanges<int64_t>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 template void create_subarray<uint64_t>(
     tiledb::sm::Array* array,
     const SubarrayRanges<uint64_t>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 template void create_subarray<float>(
     tiledb::sm::Array* array,
     const SubarrayRanges<float>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 template void create_subarray<double>(
     tiledb::sm::Array* array,
     const SubarrayRanges<double>& ranges,
     tiledb::sm::Layout layout,
-    tiledb::sm::Subarray* subarray);
+    tiledb::sm::Subarray* subarray,
+    bool coalesce_ranges);
 
 template void check_partitions<int8_t>(
     tiledb::sm::SubarrayPartitioner& partitioner,
@@ -685,58 +1170,62 @@ template void read_array<int8_t>(
     tiledb_array_t* array,
     const SubarrayRanges<int8_t>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
 template void read_array<uint8_t>(
     tiledb_ctx_t* ctx,
     tiledb_array_t* array,
     const SubarrayRanges<uint8_t>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
 template void read_array<int16_t>(
     tiledb_ctx_t* ctx,
     tiledb_array_t* array,
     const SubarrayRanges<int16_t>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
 template void read_array<uint16_t>(
     tiledb_ctx_t* ctx,
     tiledb_array_t* array,
     const SubarrayRanges<uint16_t>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
 template void read_array<int32_t>(
     tiledb_ctx_t* ctx,
     tiledb_array_t* array,
     const SubarrayRanges<int32_t>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
 template void read_array<uint32_t>(
     tiledb_ctx_t* ctx,
     tiledb_array_t* array,
     const SubarrayRanges<uint32_t>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
 template void read_array<int64_t>(
     tiledb_ctx_t* ctx,
     tiledb_array_t* array,
     const SubarrayRanges<int64_t>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
 template void read_array<uint64_t>(
     tiledb_ctx_t* ctx,
     tiledb_array_t* array,
     const SubarrayRanges<uint64_t>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
 template void read_array<float>(
     tiledb_ctx_t* ctx,
     tiledb_array_t* array,
     const SubarrayRanges<float>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
 template void read_array<double>(
     tiledb_ctx_t* ctx,
     tiledb_array_t* array,
     const SubarrayRanges<double>& ranges,
     tiledb_layout_t layout,
-    const AttrBuffers& attr_buffers);
+    const QueryBuffers& buffers);
+
+}  // End of namespace test
+
+}  // End of namespace tiledb
