@@ -32,6 +32,7 @@
 
 #include "test/src/helpers.h"
 #include "test/src/vfs_helpers.h"
+#include "tiledb/common/heap_memory.h"
 #include "tiledb/sm/array_schema/tile_domain.h"
 #include "tiledb/sm/c_api/tiledb_struct_def.h"
 #include "tiledb/sm/query/read_cell_slab_iter.h"
@@ -71,6 +72,7 @@ struct ReadCellSlabIterFx {
       const std::vector<std::vector<uint64_t>>& c_result_cell_slabs);
   template <class T>
   void create_result_space_tiles(
+      const std::vector<tdb_shared_ptr<FragmentMetadata>>& fragments,
       const Domain* dom,
       const NDRange& dom_ndrange,
       Layout layout,
@@ -135,6 +137,7 @@ void ReadCellSlabIterFx::check_iter(
 
 template <class T>
 void ReadCellSlabIterFx::create_result_space_tiles(
+    const std::vector<tdb_shared_ptr<FragmentMetadata>>& fragments,
     const Domain* dom,
     const NDRange& dom_ndrange,
     Layout layout,
@@ -146,7 +149,7 @@ void ReadCellSlabIterFx::create_result_space_tiles(
   std::vector<TileDomain<T>> frag_tile_domains;
   for (size_t i = 0; i < domain_slices.size(); ++i) {
     frag_tile_domains.emplace_back(
-        (unsigned)(domain_slices.size() - i),
+        (unsigned)(domain_slices.size() - i - 1),
         domain,
         domain_slices[i],
         tile_extents,
@@ -155,7 +158,7 @@ void ReadCellSlabIterFx::create_result_space_tiles(
   TileDomain<T> array_tile_domain(
       UINT32_MAX, domain, dom_ndrange, tile_extents, layout);
   Reader::compute_result_space_tiles<T>(
-      dom,
+      fragments,
       tile_coords,
       array_tile_domain,
       frag_tile_domains,
@@ -218,7 +221,19 @@ TEST_CASE_METHOD(
   const auto& tile_coords = subarray.tile_coords();
   std::map<const uint64_t*, ResultSpaceTile<uint64_t>> result_space_tiles;
   auto dom = array_->array_->array_schema_latest()->domain();
+
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragments;
+  tdb_shared_ptr<FragmentMetadata> fragment = tdb_make_shared(
+      FragmentMetadata,
+      nullptr,
+      array_->array_->array_schema_latest(),
+      URI(),
+      std::make_pair<uint64_t, uint64_t>(0, 0),
+      true);
+  fragments.emplace_back(std::move(fragment));
+
   create_result_space_tiles(
+      fragments,
       dom,
       dom->domain(),
       subarray_layout,
@@ -231,8 +246,8 @@ TEST_CASE_METHOD(
   ReadCellSlabIter<uint64_t> iter(
       &subarray, &result_space_tiles, &result_coords);
   std::vector<std::vector<uint64_t>> c_result_cell_slabs = {
-      {1, 0, 4, 6},
-      {1, 1, 0, 5},
+      {0, 0, 4, 6},
+      {0, 1, 0, 5},
   };
   check_iter<uint64_t>(&iter, c_result_cell_slabs);
 
@@ -278,7 +293,19 @@ TEST_CASE_METHOD(
   const auto& tile_coords = subarray.tile_coords();
   std::map<const uint64_t*, ResultSpaceTile<uint64_t>> result_space_tiles;
   auto dom = array_->array_->array_schema_latest()->domain();
+
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragments;
+  tdb_shared_ptr<FragmentMetadata> fragment = tdb_make_shared(
+      FragmentMetadata,
+      nullptr,
+      array_->array_->array_schema_latest(),
+      URI(),
+      std::make_pair<uint64_t, uint64_t>(0, 0),
+      true);
+  fragments.emplace_back(std::move(fragment));
+
   create_result_space_tiles(
+      fragments,
       dom,
       dom->domain(),
       subarray_layout,
@@ -341,7 +368,21 @@ TEST_CASE_METHOD(
   const auto& tile_coords = subarray.tile_coords();
   std::map<const uint64_t*, ResultSpaceTile<uint64_t>> result_space_tiles;
   auto dom = array_->array_->array_schema_latest()->domain();
+
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragments;
+  for (uint64_t i = 0; i < 2; i++) {
+    tdb_shared_ptr<FragmentMetadata> fragment = tdb_make_shared(
+        FragmentMetadata,
+        nullptr,
+        array_->array_->array_schema_latest(),
+        URI(),
+        std::make_pair<uint64_t, uint64_t>(0, 0),
+        true);
+    fragments.emplace_back(std::move(fragment));
+  }
+
   create_result_space_tiles(
+      fragments,
       dom,
       dom->domain(),
       subarray_layout,
@@ -354,14 +395,14 @@ TEST_CASE_METHOD(
   ReadCellSlabIter<uint64_t> iter(
       &subarray, &result_space_tiles, &result_coords);
   std::vector<std::vector<uint64_t>> c_result_cell_slabs = {
-      {2, 0, 4, 6},
-      {2, 1, 0, 2},
-      {1, 1, 2, 3},
+      {1, 0, 4, 6},
+      {1, 1, 0, 2},
+      {0, 1, 2, 3},
       {UINT64_MAX, 0, 2, 1},
-      {1, 0, 3, 1},
-      {2, 0, 4, 1},
-      {2, 1, 0, 2},
-      {1, 1, 2, 2},
+      {0, 0, 3, 1},
+      {1, 0, 4, 1},
+      {1, 1, 0, 2},
+      {0, 1, 2, 2},
   };
   check_iter<uint64_t>(&iter, c_result_cell_slabs);
 
@@ -408,7 +449,21 @@ TEST_CASE_METHOD(
   const auto& tile_coords = subarray.tile_coords();
   std::map<const uint64_t*, ResultSpaceTile<uint64_t>> result_space_tiles;
   auto dom = array_->array_->array_schema_latest()->domain();
+
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragments;
+  for (uint64_t i = 0; i < 2; i++) {
+    tdb_shared_ptr<FragmentMetadata> fragment = tdb_make_shared(
+        FragmentMetadata,
+        nullptr,
+        array_->array_->array_schema_latest(),
+        URI(),
+        std::make_pair<uint64_t, uint64_t>(0, 0),
+        true);
+    fragments.emplace_back(std::move(fragment));
+  }
+
   create_result_space_tiles(
+      fragments,
       dom,
       dom->domain(),
       subarray_layout,
@@ -418,9 +473,9 @@ TEST_CASE_METHOD(
 
   // Create result coordinates
   std::vector<ResultCoords> result_coords;
-  ResultTile result_tile_2_0(2, 0, dom);
-  ResultTile result_tile_3_0(3, 0, dom);
-  ResultTile result_tile_3_1(3, 1, dom);
+  ResultTile result_tile_2_0(1, 0, array_->array_->array_schema_latest());
+  ResultTile result_tile_3_0(2, 0, array_->array_->array_schema_latest());
+  ResultTile result_tile_3_1(2, 1, array_->array_->array_schema_latest());
 
   result_tile_2_0.init_coord_tile("d", 0);
   result_tile_3_0.init_coord_tile("d", 0);
@@ -459,17 +514,17 @@ TEST_CASE_METHOD(
   ReadCellSlabIter<uint64_t> iter(
       &subarray, &result_space_tiles, &result_coords);
   std::vector<std::vector<uint64_t>> c_result_cell_slabs = {
-      {2, 0, 1, 1},
+      {1, 0, 1, 1},
+      {0, 0, 3, 1},
       {1, 0, 3, 1},
-      {2, 0, 3, 1},
-      {1, 0, 5, 2},
-      {3, 0, 2, 1},
-      {1, 0, 8, 2},
-      {1, 1, 0, 1},
-      {3, 1, 1, 1},
+      {0, 0, 5, 2},
+      {2, 0, 2, 1},
+      {0, 0, 8, 2},
+      {0, 1, 0, 1},
+      {2, 1, 1, 1},
       {UINT64_MAX, 1, 2, 3},
       {UINT64_MAX, 1, 7, 1},
-      {3, 1, 2, 1},
+      {2, 1, 2, 1},
       {UINT64_MAX, 1, 9, 1},
   };
   check_iter<uint64_t>(&iter, c_result_cell_slabs);
@@ -494,10 +549,10 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::ROW_MAJOR;
     tile_domain_layout = Layout::ROW_MAJOR;
     c_result_cell_slabs = {
-        {1, 0, 4, 2},
-        {1, 1, 3, 3},
-        {1, 0, 7, 2},
-        {1, 1, 6, 3},
+        {0, 0, 4, 2},
+        {0, 1, 3, 3},
+        {0, 0, 7, 2},
+        {0, 1, 6, 3},
     };
   }
 
@@ -507,10 +562,10 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::ROW_MAJOR;
     tile_domain_layout = Layout::ROW_MAJOR;
     c_result_cell_slabs = {
-        {1, 0, 4, 2},
-        {1, 1, 1, 3},
-        {1, 0, 5, 2},
-        {1, 1, 2, 3},
+        {0, 0, 4, 2},
+        {0, 1, 1, 3},
+        {0, 0, 5, 2},
+        {0, 1, 2, 3},
     };
   }
 
@@ -520,10 +575,10 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::ROW_MAJOR;
     tile_domain_layout = Layout::COL_MAJOR;
     c_result_cell_slabs = {
-        {1, 0, 4, 2},
-        {1, 2, 3, 3},
-        {1, 0, 7, 2},
-        {1, 2, 6, 3},
+        {0, 0, 4, 2},
+        {0, 2, 3, 3},
+        {0, 0, 7, 2},
+        {0, 2, 6, 3},
     };
   }
 
@@ -533,10 +588,10 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::ROW_MAJOR;
     tile_domain_layout = Layout::COL_MAJOR;
     c_result_cell_slabs = {
-        {1, 0, 4, 2},
-        {1, 2, 1, 3},
-        {1, 0, 5, 2},
-        {1, 2, 2, 3},
+        {0, 0, 4, 2},
+        {0, 2, 1, 3},
+        {0, 0, 5, 2},
+        {0, 2, 2, 3},
     };
   }
 
@@ -546,11 +601,11 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::COL_MAJOR;
     tile_domain_layout = Layout::ROW_MAJOR;
     c_result_cell_slabs = {
-        {1, 0, 4, 2},
-        {1, 0, 5, 2},
-        {1, 1, 3, 2},
-        {1, 1, 4, 2},
-        {1, 1, 5, 2},
+        {0, 0, 4, 2},
+        {0, 0, 5, 2},
+        {0, 1, 3, 2},
+        {0, 1, 4, 2},
+        {0, 1, 5, 2},
     };
   }
 
@@ -560,11 +615,11 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::COL_MAJOR;
     tile_domain_layout = Layout::ROW_MAJOR;
     c_result_cell_slabs = {
-        {1, 0, 4, 2},
-        {1, 0, 7, 2},
-        {1, 1, 1, 2},
-        {1, 1, 4, 2},
-        {1, 1, 7, 2},
+        {0, 0, 4, 2},
+        {0, 0, 7, 2},
+        {0, 1, 1, 2},
+        {0, 1, 4, 2},
+        {0, 1, 7, 2},
     };
   }
 
@@ -574,11 +629,11 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::COL_MAJOR;
     tile_domain_layout = Layout::COL_MAJOR;
     c_result_cell_slabs = {
-        {1, 0, 4, 2},
-        {1, 0, 5, 2},
-        {1, 2, 3, 2},
-        {1, 2, 4, 2},
-        {1, 2, 5, 2},
+        {0, 0, 4, 2},
+        {0, 0, 5, 2},
+        {0, 2, 3, 2},
+        {0, 2, 4, 2},
+        {0, 2, 5, 2},
     };
   }
 
@@ -588,11 +643,11 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::COL_MAJOR;
     tile_domain_layout = Layout::COL_MAJOR;
     c_result_cell_slabs = {
-        {1, 0, 4, 2},
-        {1, 0, 7, 2},
-        {1, 2, 1, 2},
-        {1, 2, 4, 2},
-        {1, 2, 7, 2},
+        {0, 0, 4, 2},
+        {0, 0, 7, 2},
+        {0, 2, 1, 2},
+        {0, 2, 4, 2},
+        {0, 2, 7, 2},
     };
   }
 
@@ -631,7 +686,19 @@ TEST_CASE_METHOD(
   const auto& tile_coords = subarray.tile_coords();
   std::map<const uint64_t*, ResultSpaceTile<uint64_t>> result_space_tiles;
   auto dom = array_->array_->array_schema_latest()->domain();
+
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragments;
+  tdb_shared_ptr<FragmentMetadata> fragment = tdb_make_shared(
+      FragmentMetadata,
+      nullptr,
+      array_->array_->array_schema_latest(),
+      URI(),
+      std::make_pair<uint64_t, uint64_t>(0, 0),
+      true);
+  fragments.emplace_back(std::move(fragment));
+
   create_result_space_tiles(
+      fragments,
       dom,
       dom->domain(),
       tile_domain_layout,
@@ -803,7 +870,19 @@ TEST_CASE_METHOD(
   const auto& tile_coords = subarray.tile_coords();
   std::map<const uint64_t*, ResultSpaceTile<uint64_t>> result_space_tiles;
   auto dom = array_->array_->array_schema_latest()->domain();
+
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragments;
+  tdb_shared_ptr<FragmentMetadata> fragment = tdb_make_shared(
+      FragmentMetadata,
+      nullptr,
+      array_->array_->array_schema_latest(),
+      URI(),
+      std::make_pair<uint64_t, uint64_t>(0, 0),
+      true);
+  fragments.emplace_back(std::move(fragment));
+
   create_result_space_tiles(
+      fragments,
       dom,
       dom->domain(),
       tile_domain_layout,
@@ -843,7 +922,7 @@ TEST_CASE_METHOD(
         {UINT64_MAX, 1, 3, 3},
         {UINT64_MAX, 0, 7, 2},
         {UINT64_MAX, 1, 6, 1},
-        {1, 0, 7, 2},
+        {0, 0, 7, 2},
     };
   }
 
@@ -857,7 +936,7 @@ TEST_CASE_METHOD(
         {UINT64_MAX, 1, 1, 3},
         {UINT64_MAX, 0, 5, 2},
         {UINT64_MAX, 1, 2, 1},
-        {1, 0, 5, 2},
+        {0, 0, 5, 2},
     };
   }
 
@@ -871,7 +950,7 @@ TEST_CASE_METHOD(
         {UINT64_MAX, 2, 3, 3},
         {UINT64_MAX, 0, 7, 2},
         {UINT64_MAX, 2, 6, 1},
-        {1, 0, 7, 2},
+        {0, 0, 7, 2},
     };
   }
 
@@ -885,7 +964,7 @@ TEST_CASE_METHOD(
         {UINT64_MAX, 2, 1, 3},
         {UINT64_MAX, 0, 5, 2},
         {UINT64_MAX, 2, 2, 1},
-        {1, 0, 5, 2},
+        {0, 0, 5, 2},
     };
   }
 
@@ -899,9 +978,9 @@ TEST_CASE_METHOD(
         {UINT64_MAX, 0, 5, 2},
         {UINT64_MAX, 1, 3, 2},
         {UINT64_MAX, 1, 4, 1},
-        {1, 0, 7, 1},
+        {0, 0, 7, 1},
         {UINT64_MAX, 1, 5, 1},
-        {1, 0, 8, 1},
+        {0, 0, 8, 1},
     };
   }
 
@@ -915,9 +994,9 @@ TEST_CASE_METHOD(
         {UINT64_MAX, 0, 7, 2},
         {UINT64_MAX, 1, 1, 2},
         {UINT64_MAX, 1, 4, 1},
-        {1, 0, 5, 1},
+        {0, 0, 5, 1},
         {UINT64_MAX, 1, 7, 1},
-        {1, 0, 8, 1},
+        {0, 0, 8, 1},
     };
   }
 
@@ -931,9 +1010,9 @@ TEST_CASE_METHOD(
         {UINT64_MAX, 0, 5, 2},
         {UINT64_MAX, 2, 3, 2},
         {UINT64_MAX, 2, 4, 1},
-        {1, 0, 7, 1},
+        {0, 0, 7, 1},
         {UINT64_MAX, 2, 5, 1},
-        {1, 0, 8, 1},
+        {0, 0, 8, 1},
     };
   }
 
@@ -947,9 +1026,9 @@ TEST_CASE_METHOD(
         {UINT64_MAX, 0, 7, 2},
         {UINT64_MAX, 2, 1, 2},
         {UINT64_MAX, 2, 4, 1},
-        {1, 0, 5, 1},
+        {0, 0, 5, 1},
         {UINT64_MAX, 2, 7, 1},
-        {1, 0, 8, 1},
+        {0, 0, 8, 1},
     };
   }
 
@@ -988,7 +1067,19 @@ TEST_CASE_METHOD(
   const auto& tile_coords = subarray.tile_coords();
   std::map<const uint64_t*, ResultSpaceTile<uint64_t>> result_space_tiles;
   auto dom = array_->array_->array_schema_latest()->domain();
+
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragments;
+  tdb_shared_ptr<FragmentMetadata> fragment = tdb_make_shared(
+      FragmentMetadata,
+      nullptr,
+      array_->array_->array_schema_latest(),
+      URI(),
+      std::make_pair<uint64_t, uint64_t>(0, 0),
+      true);
+  fragments.emplace_back(std::move(fragment));
+
   create_result_space_tiles(
+      fragments,
       dom,
       dom->domain(),
       tile_domain_layout,
@@ -1024,17 +1115,17 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::ROW_MAJOR;
     tile_domain_layout = Layout::ROW_MAJOR;
     c_result_cell_slabs = {
-        {2, 0, 7, 1},
-        {3, 0, 1, 1},
-        {2, 1, 6, 1},
-        {1, 1, 7, 2},
-        {2, 2, 1, 2},
-        {2, 3, 0, 1},
+        {1, 0, 7, 1},
+        {2, 0, 1, 1},
+        {1, 1, 6, 1},
+        {0, 1, 7, 2},
+        {1, 2, 1, 2},
+        {1, 3, 0, 1},
         {UINT64_MAX, 3, 1, 2},
-        {2, 2, 4, 2},
-        {2, 3, 3, 1},
-        {3, 1, 0, 1},
-        {3, 1, 2, 1},
+        {1, 2, 4, 2},
+        {1, 3, 3, 1},
+        {2, 1, 0, 1},
+        {2, 1, 2, 1},
     };
   }
 
@@ -1044,17 +1135,17 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::ROW_MAJOR;
     tile_domain_layout = Layout::ROW_MAJOR;
     c_result_cell_slabs = {
-        {2, 0, 5, 1},
-        {3, 0, 1, 1},
-        {2, 1, 2, 1},
-        {1, 1, 5, 2},
-        {2, 2, 3, 2},
-        {2, 3, 0, 1},
+        {1, 0, 5, 1},
+        {2, 0, 1, 1},
+        {1, 1, 2, 1},
+        {0, 1, 5, 2},
+        {1, 2, 3, 2},
+        {1, 3, 0, 1},
         {UINT64_MAX, 3, 3, 2},
-        {2, 2, 4, 2},
-        {2, 3, 1, 1},
-        {3, 1, 0, 1},
-        {3, 1, 2, 1},
+        {1, 2, 4, 2},
+        {1, 3, 1, 1},
+        {2, 1, 0, 1},
+        {2, 1, 2, 1},
     };
   }
 
@@ -1064,17 +1155,17 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::ROW_MAJOR;
     tile_domain_layout = Layout::COL_MAJOR;
     c_result_cell_slabs = {
-        {2, 0, 7, 1},
-        {3, 0, 1, 1},
-        {2, 2, 6, 1},
-        {1, 1, 7, 2},
-        {2, 1, 1, 2},
-        {2, 3, 0, 1},
+        {1, 0, 7, 1},
+        {2, 0, 1, 1},
+        {1, 2, 6, 1},
+        {0, 1, 7, 2},
+        {1, 1, 1, 2},
+        {1, 3, 0, 1},
         {UINT64_MAX, 3, 1, 2},
-        {2, 1, 4, 2},
-        {2, 3, 3, 1},
-        {3, 1, 0, 1},
-        {3, 1, 2, 1},
+        {1, 1, 4, 2},
+        {1, 3, 3, 1},
+        {2, 1, 0, 1},
+        {2, 1, 2, 1},
     };
   }
 
@@ -1084,17 +1175,17 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::ROW_MAJOR;
     tile_domain_layout = Layout::COL_MAJOR;
     c_result_cell_slabs = {
-        {2, 0, 5, 1},
-        {3, 0, 1, 1},
-        {2, 2, 2, 1},
-        {1, 1, 5, 2},
-        {2, 1, 3, 2},
-        {2, 3, 0, 1},
+        {1, 0, 5, 1},
+        {2, 0, 1, 1},
+        {1, 2, 2, 1},
+        {0, 1, 5, 2},
+        {1, 1, 3, 2},
+        {1, 3, 0, 1},
         {UINT64_MAX, 3, 3, 2},
-        {2, 1, 4, 2},
-        {2, 3, 1, 1},
-        {3, 1, 0, 1},
-        {3, 1, 2, 1},
+        {1, 1, 4, 2},
+        {1, 3, 1, 1},
+        {2, 1, 0, 1},
+        {2, 1, 2, 1},
     };
   }
 
@@ -1104,18 +1195,18 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::COL_MAJOR;
     tile_domain_layout = Layout::ROW_MAJOR;
     c_result_cell_slabs = {
-        {2, 0, 7, 1},
-        {2, 2, 1, 2},
-        {3, 0, 1, 1},
-        {2, 2, 2, 2},
-        {2, 1, 6, 1},
-        {2, 3, 0, 2},
-        {1, 1, 7, 1},
+        {1, 0, 7, 1},
+        {1, 2, 1, 2},
+        {2, 0, 1, 1},
+        {1, 2, 2, 2},
+        {1, 1, 6, 1},
+        {1, 3, 0, 2},
+        {0, 1, 7, 1},
         {UINT64_MAX, 3, 1, 1},
-        {3, 1, 0, 1},
-        {1, 1, 8, 1},
+        {2, 1, 0, 1},
+        {0, 1, 8, 1},
         {UINT64_MAX, 3, 2, 1},
-        {3, 1, 2, 1},
+        {2, 1, 2, 1},
     };
   }
 
@@ -1125,18 +1216,18 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::COL_MAJOR;
     tile_domain_layout = Layout::ROW_MAJOR;
     c_result_cell_slabs = {
-        {2, 0, 5, 1},
-        {2, 2, 3, 2},
-        {3, 0, 1, 1},
-        {2, 2, 6, 2},
-        {2, 1, 2, 1},
-        {2, 3, 0, 2},
-        {1, 1, 5, 1},
+        {1, 0, 5, 1},
+        {1, 2, 3, 2},
+        {2, 0, 1, 1},
+        {1, 2, 6, 2},
+        {1, 1, 2, 1},
+        {1, 3, 0, 2},
+        {0, 1, 5, 1},
         {UINT64_MAX, 3, 3, 1},
-        {3, 1, 0, 1},
-        {1, 1, 8, 1},
+        {2, 1, 0, 1},
+        {0, 1, 8, 1},
         {UINT64_MAX, 3, 6, 1},
-        {3, 1, 2, 1},
+        {2, 1, 2, 1},
     };
   }
 
@@ -1146,18 +1237,18 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::COL_MAJOR;
     tile_domain_layout = Layout::COL_MAJOR;
     c_result_cell_slabs = {
-        {2, 0, 7, 1},
-        {2, 1, 1, 2},
-        {3, 0, 1, 1},
-        {2, 1, 2, 2},
-        {2, 2, 6, 1},
-        {2, 3, 0, 2},
-        {1, 1, 7, 1},
+        {1, 0, 7, 1},
+        {1, 1, 1, 2},
+        {2, 0, 1, 1},
+        {1, 1, 2, 2},
+        {1, 2, 6, 1},
+        {1, 3, 0, 2},
+        {0, 1, 7, 1},
         {UINT64_MAX, 3, 1, 1},
-        {3, 1, 0, 1},
-        {1, 1, 8, 1},
+        {2, 1, 0, 1},
+        {0, 1, 8, 1},
         {UINT64_MAX, 3, 2, 1},
-        {3, 1, 2, 1},
+        {2, 1, 2, 1},
     };
   }
 
@@ -1167,18 +1258,18 @@ TEST_CASE_METHOD(
     subarray_layout = Layout::COL_MAJOR;
     tile_domain_layout = Layout::COL_MAJOR;
     c_result_cell_slabs = {
-        {2, 0, 5, 1},
-        {2, 1, 3, 2},
-        {3, 0, 1, 1},
-        {2, 1, 6, 2},
-        {2, 2, 2, 1},
-        {2, 3, 0, 2},
-        {1, 1, 5, 1},
+        {1, 0, 5, 1},
+        {1, 1, 3, 2},
+        {2, 0, 1, 1},
+        {1, 1, 6, 2},
+        {1, 2, 2, 1},
+        {1, 3, 0, 2},
+        {0, 1, 5, 1},
         {UINT64_MAX, 3, 3, 1},
-        {3, 1, 0, 1},
-        {1, 1, 8, 1},
+        {2, 1, 0, 1},
+        {0, 1, 8, 1},
         {UINT64_MAX, 3, 6, 1},
-        {3, 1, 2, 1},
+        {2, 1, 2, 1},
     };
   }
 
@@ -1219,7 +1310,21 @@ TEST_CASE_METHOD(
   const auto& tile_coords = subarray.tile_coords();
   std::map<const uint64_t*, ResultSpaceTile<uint64_t>> result_space_tiles;
   auto dom = array_->array_->array_schema_latest()->domain();
+
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragments;
+  for (uint64_t i = 0; i < 2; i++) {
+    tdb_shared_ptr<FragmentMetadata> fragment = tdb_make_shared(
+        FragmentMetadata,
+        nullptr,
+        array_->array_->array_schema_latest(),
+        URI(),
+        std::make_pair<uint64_t, uint64_t>(0, 0),
+        true);
+    fragments.emplace_back(std::move(fragment));
+  }
+
   create_result_space_tiles(
+      fragments,
       dom,
       dom->domain(),
       tile_domain_layout,
@@ -1229,8 +1334,8 @@ TEST_CASE_METHOD(
 
   // Create result coordinates
   std::vector<ResultCoords> result_coords;
-  ResultTile result_tile_3_0(3, 0, dom);
-  ResultTile result_tile_3_1(3, 1, dom);
+  ResultTile result_tile_3_0(2, 0, array_->array_->array_schema_latest());
+  ResultTile result_tile_3_1(2, 1, array_->array_->array_schema_latest());
 
   result_tile_3_0.init_coord_tile("d1", 0);
   result_tile_3_0.init_coord_tile("d2", 1);
@@ -1275,4 +1380,4 @@ TEST_CASE_METHOD(
   check_iter<uint64_t>(&iter, c_result_cell_slabs);
 
   close_array(ctx_, array_);
-}
\ No newline at end of file
+}
