@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,6 +31,7 @@
  */
 
 #include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #include "tiledb/sm/c_api/tiledb_struct_def.h"
 #include "tiledb/sm/subarray/subarray_partitioner.h"
 
@@ -43,7 +44,9 @@
 #include <catch.hpp>
 #include <iostream>
 
+using namespace tiledb::common;
 using namespace tiledb::sm;
+using namespace tiledb::test;
 
 /* ********************************* */
 /*         STRUCT DEFINITION         */
@@ -52,10 +55,8 @@ using namespace tiledb::sm;
 struct SubarrayPartitionerDenseFx {
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
-  bool s3_supported_, hdfs_supported_;
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
   std::string temp_dir_;
-  const std::string s3_bucket_name_ =
-      "s3://" + random_bucket_name("tiledb") + "/";
   std::string array_name_;
   const char* ARRAY_NAME = "subarray_partitioner_dense";
   tiledb_array_t* array_ = nullptr;
@@ -112,27 +113,19 @@ struct SubarrayPartitionerDenseFx {
       bool unsplittable = false);
 };
 
-SubarrayPartitionerDenseFx::SubarrayPartitionerDenseFx() {
-  ctx_ = nullptr;
-  vfs_ = nullptr;
-  hdfs_supported_ = false;
-  s3_supported_ = false;
-
-  get_supported_fs(&s3_supported_, &hdfs_supported_);
-  create_ctx_and_vfs(s3_supported_, &ctx_, &vfs_);
-  create_s3_bucket(s3_bucket_name_, s3_supported_, ctx_, vfs_);
+SubarrayPartitionerDenseFx::SubarrayPartitionerDenseFx()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 
 // Create temporary directory based on the supported filesystem
 #ifdef _WIN32
-  temp_dir_ = tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
+  SupportedFsLocal windows_fs;
+  temp_dir_ = windows_fs.file_prefix() + windows_fs.temp_dir();
 #else
-  temp_dir_ = "file://" + tiledb::sm::Posix::current_dir() + "/tiledb_test/";
+  SupportedFsLocal posix_fs;
+  temp_dir_ = posix_fs.file_prefix() + posix_fs.temp_dir();
 #endif
-  if (s3_supported_)
-    temp_dir_ = s3_bucket_name_ + "tiledb/test/";
-  // Disabling this for now, as this test takes a long time on HDFS
-  // if (hdfs_supported_)
-  //  temp_dir_ = "hdfs:///tiledb_test/";
 
   create_dir(temp_dir_, ctx_, vfs_);
 
@@ -163,8 +156,8 @@ void SubarrayPartitionerDenseFx::create_default_1d_array(
       {"a", "b"},
       {TILEDB_INT32, TILEDB_INT32},
       {1, TILEDB_VAR_NUM},
-      {::Compressor(TILEDB_FILTER_LZ4, -1),
-       ::Compressor(TILEDB_FILTER_LZ4, -1)},
+      {tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1)},
       tile_order,
       cell_order,
       2);
@@ -185,15 +178,15 @@ void SubarrayPartitionerDenseFx::create_default_2d_array(
       {"a", "b"},
       {TILEDB_INT32, TILEDB_INT32},
       {1, TILEDB_VAR_NUM},
-      {::Compressor(TILEDB_FILTER_LZ4, -1),
-       ::Compressor(TILEDB_FILTER_LZ4, -1)},
+      {tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1)},
       tile_order,
       cell_order,
       2);
 }
 
 void SubarrayPartitionerDenseFx::write_default_1d_array() {
-  AttrBuffers attr_buffers;
+  tiledb::test::QueryBuffers buffers;
   std::vector<int> a = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
   uint64_t a_size = a.size() * sizeof(int);
   std::vector<uint64_t> b_off = {0,
@@ -210,14 +203,14 @@ void SubarrayPartitionerDenseFx::write_default_1d_array() {
   std::vector<int> b_val = {
       1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 6, 7, 8, 9, 10};
   uint64_t b_val_size = b_val.size() * sizeof(int);
-  attr_buffers["a"] = AttrBuffer({&a[0], a_size, nullptr, 0});
-  attr_buffers["b"] =
-      AttrBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
-  write_array(ctx_, array_name_, TILEDB_GLOBAL_ORDER, attr_buffers);
+  buffers["a"] = tiledb::test::QueryBuffer({&a[0], a_size, nullptr, 0});
+  buffers["b"] =
+      tiledb::test::QueryBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
+  write_array(ctx_, array_name_, TILEDB_GLOBAL_ORDER, buffers);
 }
 
 void SubarrayPartitionerDenseFx::write_default_2d_array() {
-  AttrBuffers attr_buffers;
+  tiledb::test::QueryBuffers buffers;
   std::vector<int> a = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};
   uint64_t a_size = a.size() * sizeof(int);
   std::vector<uint64_t> b_off = {0,
@@ -241,10 +234,10 @@ void SubarrayPartitionerDenseFx::write_default_2d_array() {
                             6,  6,  6,  6,  7,  7,  8,  8,  8, 9, 10,
                             10, 11, 12, 13, 13, 14, 15, 16, 16};
   uint64_t b_val_size = b_val.size() * sizeof(int);
-  attr_buffers["a"] = AttrBuffer({&a[0], a_size, nullptr, 0});
-  attr_buffers["b"] =
-      AttrBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
-  write_array(ctx_, array_name_, TILEDB_GLOBAL_ORDER, attr_buffers);
+  buffers["a"] = tiledb::test::QueryBuffer({&a[0], a_size, nullptr, 0});
+  buffers["b"] =
+      tiledb::test::QueryBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
+  write_array(ctx_, array_name_, TILEDB_GLOBAL_ORDER, buffers);
 }
 
 template <class T>
@@ -258,8 +251,18 @@ void SubarrayPartitionerDenseFx::test_subarray_partitioner(
   Subarray subarray;
   create_subarray(array_->array_, ranges, subarray_layout, &subarray);
 
+  ThreadPool tp;
+  CHECK(tp.init(4).ok());
+  Config config;
   SubarrayPartitioner subarray_partitioner(
-      subarray, memory_budget_, memory_budget_var_);
+      &config,
+      subarray,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
   auto st = subarray_partitioner.set_result_budget(attr.c_str(), budget);
   CHECK(st.ok());
 
@@ -277,8 +280,18 @@ void SubarrayPartitionerDenseFx::test_subarray_partitioner(
   Subarray subarray;
   create_subarray(array_->array_, ranges, subarray_layout, &subarray);
 
+  ThreadPool tp;
+  CHECK(tp.init(4).ok());
+  Config config;
   SubarrayPartitioner subarray_partitioner(
-      subarray, memory_budget_, memory_budget_var_);
+      &config,
+      subarray,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
 
   // Note: this is necessary, otherwise the subarray partitioner does
   // not check if the memory budget is exceeded for attributes whose
@@ -290,7 +303,7 @@ void SubarrayPartitionerDenseFx::test_subarray_partitioner(
   st = subarray_partitioner.set_result_budget("b", 1000000, 1000000);
   CHECK(st.ok());
 
-  st = subarray_partitioner.set_memory_budget(budget, budget_var);
+  st = subarray_partitioner.set_memory_budget(budget, budget_var, 0);
   CHECK(st.ok());
 
   check_partitions(subarray_partitioner, partitions, unsplittable);
@@ -549,8 +562,18 @@ TEST_CASE_METHOD(
 
   create_subarray(array_->array_, ranges, subarray_layout, &subarray);
 
+  ThreadPool tp;
+  CHECK(tp.init(4).ok());
+  Config config;
   SubarrayPartitioner subarray_partitioner(
-      subarray, memory_budget_, memory_budget_var_);
+      &config,
+      subarray,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
   auto st = subarray_partitioner.set_result_budget("a", 100 * sizeof(int));
   CHECK(st.ok());
   st = subarray_partitioner.set_result_budget("b", 1, 1);
