@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,6 +31,7 @@
  */
 
 #include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #include "tiledb/sm/c_api/tiledb_struct_def.h"
 #include "tiledb/sm/subarray/subarray_partitioner.h"
 
@@ -43,7 +44,9 @@
 #include <catch.hpp>
 #include <iostream>
 
+using namespace tiledb::common;
 using namespace tiledb::sm;
+using namespace tiledb::test;
 
 /* ********************************* */
 /*         STRUCT DEFINITION         */
@@ -52,10 +55,8 @@ using namespace tiledb::sm;
 struct SubarrayPartitionerSparseFx {
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
-  bool s3_supported_, hdfs_supported_;
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
   std::string temp_dir_;
-  const std::string s3_bucket_name_ =
-      "s3://" + random_bucket_name("tiledb") + "/";
   std::string array_name_;
   const char* ARRAY_NAME = "subarray_partitioner_sparse";
   tiledb_array_t* array_ = nullptr;
@@ -110,6 +111,20 @@ struct SubarrayPartitionerSparseFx {
       uint64_t budget,
       bool unsplittable = false);
 
+  /**
+   * Helper function to test the subarray partitioner for the given arguments.
+   */
+  template <class T>
+  void test_subarray_partitioner(
+      Layout subarray_layout,
+      const SubarrayRanges<T>& ranges,
+      const std::vector<SubarrayRanges<T>>& partitions,
+      const std::string& attr,  // Attribute to set the budget for
+      uint64_t result_budget,
+      uint64_t memory_budget,
+      uint64_t memory_budget_var,
+      bool unsplittable = false);
+
   /**
    * Helper function to test the subarray partitioner for the given arguments.
    * This is different from the above in that it tests the memory
@@ -125,27 +140,19 @@ struct SubarrayPartitionerSparseFx {
       bool unsplittable = false);
 };
 
-SubarrayPartitionerSparseFx::SubarrayPartitionerSparseFx() {
-  ctx_ = nullptr;
-  vfs_ = nullptr;
-  hdfs_supported_ = false;
-  s3_supported_ = false;
-
-  get_supported_fs(&s3_supported_, &hdfs_supported_);
-  create_ctx_and_vfs(s3_supported_, &ctx_, &vfs_);
-  create_s3_bucket(s3_bucket_name_, s3_supported_, ctx_, vfs_);
+SubarrayPartitionerSparseFx::SubarrayPartitionerSparseFx()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 
 // Create temporary directory based on the supported filesystem
 #ifdef _WIN32
-  temp_dir_ = tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
+  SupportedFsLocal windows_fs;
+  temp_dir_ = windows_fs.file_prefix() + windows_fs.temp_dir();
 #else
-  temp_dir_ = "file://" + tiledb::sm::Posix::current_dir() + "/tiledb_test/";
+  SupportedFsLocal posix_fs;
+  temp_dir_ = posix_fs.file_prefix() + posix_fs.temp_dir();
 #endif
-  if (s3_supported_)
-    temp_dir_ = s3_bucket_name_ + "tiledb/test/";
-  // Disabling this for now, as this test takes a long time on HDFS
-  // if (hdfs_supported_)
-  //  temp_dir_ = "hdfs:///tiledb_test/";
 
   create_dir(temp_dir_, ctx_, vfs_);
 
@@ -176,8 +183,8 @@ void SubarrayPartitionerSparseFx::create_default_1d_array(
       {"a", "b"},
       {TILEDB_INT32, TILEDB_INT32},
       {1, TILEDB_VAR_NUM},
-      {::Compressor(TILEDB_FILTER_LZ4, -1),
-       ::Compressor(TILEDB_FILTER_LZ4, -1)},
+      {tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1)},
       tile_order,
       cell_order,
       2);
@@ -198,8 +205,8 @@ void SubarrayPartitionerSparseFx::create_default_1d_float_array(
       {"a", "b"},
       {TILEDB_INT32, TILEDB_INT32},
       {1, TILEDB_VAR_NUM},
-      {::Compressor(TILEDB_FILTER_LZ4, -1),
-       ::Compressor(TILEDB_FILTER_LZ4, -1)},
+      {tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1)},
       tile_order,
       cell_order,
       2);
@@ -220,15 +227,15 @@ void SubarrayPartitionerSparseFx::create_default_2d_array(
       {"a", "b"},
       {TILEDB_INT32, TILEDB_INT32},
       {1, TILEDB_VAR_NUM},
-      {::Compressor(TILEDB_FILTER_LZ4, -1),
-       ::Compressor(TILEDB_FILTER_LZ4, -1)},
+      {tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1)},
       tile_order,
       cell_order,
       2);
 }
 
 void SubarrayPartitionerSparseFx::write_default_1d_array() {
-  AttrBuffers attr_buffers;
+  tiledb::test::QueryBuffers buffers;
   std::vector<uint64_t> coords = {2, 4, 5, 10, 12, 18};
   uint64_t coords_size = coords.size() * sizeof(uint64_t);
   std::vector<int> a = {1, 2, 3, 4, 5, 6};
@@ -242,16 +249,16 @@ void SubarrayPartitionerSparseFx::write_default_1d_array() {
   uint64_t b_off_size = b_off.size() * sizeof(uint64_t);
   std::vector<int> b_val = {1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 6};
   uint64_t b_val_size = b_val.size() * sizeof(int);
-  attr_buffers[TILEDB_COORDS] =
-      AttrBuffer({&coords[0], coords_size, nullptr, 0});
-  attr_buffers["a"] = AttrBuffer({&a[0], a_size, nullptr, 0});
-  attr_buffers["b"] =
-      AttrBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
-  write_array(ctx_, array_name_, TILEDB_UNORDERED, attr_buffers);
+  buffers[TILEDB_COORDS] =
+      tiledb::test::QueryBuffer({&coords[0], coords_size, nullptr, 0});
+  buffers["a"] = tiledb::test::QueryBuffer({&a[0], a_size, nullptr, 0});
+  buffers["b"] =
+      tiledb::test::QueryBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
+  write_array(ctx_, array_name_, TILEDB_UNORDERED, buffers);
 }
 
 void SubarrayPartitionerSparseFx::write_default_1d_array_2() {
-  AttrBuffers attr_buffers;
+  tiledb::test::QueryBuffers buffers;
   std::vector<uint64_t> coords = {2, 4, 5, 10, 12, 18, 25, 27, 33, 40};
   uint64_t coords_size = coords.size() * sizeof(uint64_t);
   std::vector<int> a = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
@@ -270,16 +277,16 @@ void SubarrayPartitionerSparseFx::write_default_1d_array_2() {
   std::vector<int> b_val = {
       1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 6, 7, 8, 9, 10};
   uint64_t b_val_size = b_val.size() * sizeof(int);
-  attr_buffers[TILEDB_COORDS] =
-      AttrBuffer({&coords[0], coords_size, nullptr, 0});
-  attr_buffers["a"] = AttrBuffer({&a[0], a_size, nullptr, 0});
-  attr_buffers["b"] =
-      AttrBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
-  write_array(ctx_, array_name_, TILEDB_UNORDERED, attr_buffers);
+  buffers[TILEDB_COORDS] =
+      tiledb::test::QueryBuffer({&coords[0], coords_size, nullptr, 0});
+  buffers["a"] = tiledb::test::QueryBuffer({&a[0], a_size, nullptr, 0});
+  buffers["b"] =
+      tiledb::test::QueryBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
+  write_array(ctx_, array_name_, TILEDB_UNORDERED, buffers);
 }
 
 void SubarrayPartitionerSparseFx::write_default_1d_float_array() {
-  AttrBuffers attr_buffers;
+  tiledb::test::QueryBuffers buffers;
   std::vector<float> coords = {2.0f, 4.0f, 5.0f, 10.0f, 12.0f, 18.0f};
   uint64_t coords_size = coords.size() * sizeof(float);
   std::vector<int> a = {1, 2, 3, 4, 5, 6};
@@ -293,16 +300,16 @@ void SubarrayPartitionerSparseFx::write_default_1d_float_array() {
   uint64_t b_off_size = b_off.size() * sizeof(uint64_t);
   std::vector<int> b_val = {1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 6};
   uint64_t b_val_size = b_val.size() * sizeof(int);
-  attr_buffers[TILEDB_COORDS] =
-      AttrBuffer({&coords[0], coords_size, nullptr, 0});
-  attr_buffers["a"] = AttrBuffer({&a[0], a_size, nullptr, 0});
-  attr_buffers["b"] =
-      AttrBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
-  write_array(ctx_, array_name_, TILEDB_UNORDERED, attr_buffers);
+  buffers[TILEDB_COORDS] =
+      tiledb::test::QueryBuffer({&coords[0], coords_size, nullptr, 0});
+  buffers["a"] = tiledb::test::QueryBuffer({&a[0], a_size, nullptr, 0});
+  buffers["b"] =
+      tiledb::test::QueryBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
+  write_array(ctx_, array_name_, TILEDB_UNORDERED, buffers);
 }
 
 void SubarrayPartitionerSparseFx::write_default_2d_array() {
-  AttrBuffers attr_buffers;
+  tiledb::test::QueryBuffers buffers;
   std::vector<uint64_t> coords = {1, 2, 2, 5, 3, 3, 3, 9, 4, 1, 4, 7};
   uint64_t coords_size = coords.size() * sizeof(uint64_t);
   std::vector<int> a = {1, 2, 3, 4, 5, 6};
@@ -316,12 +323,12 @@ void SubarrayPartitionerSparseFx::write_default_2d_array() {
   uint64_t b_off_size = b_off.size() * sizeof(uint64_t);
   std::vector<int> b_val = {1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 6};
   uint64_t b_val_size = b_val.size() * sizeof(int);
-  attr_buffers[TILEDB_COORDS] =
-      AttrBuffer({&coords[0], coords_size, nullptr, 0});
-  attr_buffers["a"] = AttrBuffer({&a[0], a_size, nullptr, 0});
-  attr_buffers["b"] =
-      AttrBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
-  write_array(ctx_, array_name_, TILEDB_UNORDERED, attr_buffers);
+  buffers[TILEDB_COORDS] =
+      tiledb::test::QueryBuffer({&coords[0], coords_size, nullptr, 0});
+  buffers["a"] = tiledb::test::QueryBuffer({&a[0], a_size, nullptr, 0});
+  buffers["b"] =
+      tiledb::test::QueryBuffer({&b_off[0], b_off_size, &b_val[0], b_val_size});
+  write_array(ctx_, array_name_, TILEDB_UNORDERED, buffers);
 }
 
 template <class T>
@@ -335,14 +342,55 @@ void SubarrayPartitionerSparseFx::test_subarray_partitioner(
   Subarray subarray;
   create_subarray(array_->array_, ranges, subarray_layout, &subarray);
 
+  ThreadPool tp;
+  CHECK(tp.init(4).ok());
+  Config config;
   SubarrayPartitioner subarray_partitioner(
-      subarray, memory_budget_, memory_budget_var_);
+      &config,
+      subarray,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
   auto st = subarray_partitioner.set_result_budget(attr.c_str(), budget);
   CHECK(st.ok());
 
   check_partitions(subarray_partitioner, partitions, unsplittable);
 }
 
+template <class T>
+void SubarrayPartitionerSparseFx::test_subarray_partitioner(
+    Layout subarray_layout,
+    const SubarrayRanges<T>& ranges,
+    const std::vector<SubarrayRanges<T>>& partitions,
+    const std::string& attr,
+    uint64_t result_budget,
+    uint64_t memory_budget,
+    uint64_t memory_budget_var,
+    bool unsplittable) {
+  Subarray subarray;
+  create_subarray(array_->array_, ranges, subarray_layout, &subarray);
+
+  ThreadPool tp;
+  CHECK(tp.init(4).ok());
+  Config config;
+  SubarrayPartitioner subarray_partitioner(
+      &config,
+      subarray,
+      memory_budget,
+      memory_budget_var,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
+  auto st = subarray_partitioner.set_result_budget(attr.c_str(), result_budget);
+  CHECK(st.ok());
+
+  check_partitions(subarray_partitioner, partitions, unsplittable);
+}
+
 template <class T>
 void SubarrayPartitionerSparseFx::test_subarray_partitioner(
     Layout subarray_layout,
@@ -354,8 +402,18 @@ void SubarrayPartitionerSparseFx::test_subarray_partitioner(
   Subarray subarray;
   create_subarray(array_->array_, ranges, subarray_layout, &subarray);
 
+  ThreadPool tp;
+  CHECK(tp.init(4).ok());
+  Config config;
   SubarrayPartitioner subarray_partitioner(
-      subarray, memory_budget_, memory_budget_var_);
+      &config,
+      subarray,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
 
   // Note: this is necessary, otherwise the subarray partitioner does
   // not check if the memory budget is exceeded for attributes whose
@@ -367,7 +425,7 @@ void SubarrayPartitionerSparseFx::test_subarray_partitioner(
   st = subarray_partitioner.set_result_budget("b", 1000000, 1000000);
   CHECK(st.ok());
 
-  st = subarray_partitioner.set_memory_budget(budget, budget_var);
+  st = subarray_partitioner.set_memory_budget(budget, budget_var, 0);
   CHECK(st.ok());
 
   check_partitions(subarray_partitioner, partitions, unsplittable);
@@ -626,8 +684,18 @@ TEST_CASE_METHOD(
 
   create_subarray(array_->array_, ranges, subarray_layout, &subarray);
 
+  ThreadPool tp;
+  CHECK(tp.init(4).ok());
+  Config config;
   SubarrayPartitioner subarray_partitioner(
-      subarray, memory_budget_, memory_budget_var_);
+      &config,
+      subarray,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
   auto st = subarray_partitioner.set_result_budget("a", 100);
   CHECK(st.ok());
   st = subarray_partitioner.set_result_budget("b", 1, 1);
@@ -1600,8 +1668,7 @@ TEST_CASE_METHOD(
     // subarray: col
     subarray_layout = Layout::COL_MAJOR;
     ranges = {{1, 2, 3, 4}, {1, 2, 3, 5, 7, 9}};
-    partitions = {
-        {{1, 2, 3, 4}, {1, 2}}, {{1, 2, 3, 4}, {3, 5}}, {{1, 2, 3, 4}, {7, 9}}};
+    partitions = {{{1, 2, 3, 4}, {1, 2, 3, 5}}, {{1, 2, 3, 4}, {7, 9}}};
     test_subarray_partitioner(
         subarray_layout, ranges, partitions, attr, budget);
 
@@ -1628,16 +1695,14 @@ TEST_CASE_METHOD(
     // subarray: col
     subarray_layout = Layout::COL_MAJOR;
     ranges = {{1, 2, 3, 4}, {1, 2, 3, 5, 7, 9}};
-    partitions = {
-        {{1, 2, 3, 4}, {1, 2}}, {{1, 2, 3, 4}, {3, 5}}, {{1, 2, 3, 4}, {7, 9}}};
+    partitions = {{{1, 2, 3, 4}, {1, 2, 3, 5}}, {{1, 2, 3, 4}, {7, 9}}};
     test_subarray_partitioner(
         subarray_layout, ranges, partitions, attr, budget);
 
     // subarray: unordered
     subarray_layout = Layout::UNORDERED;
     ranges = {{1, 2, 3, 4}, {1, 2, 3, 5, 7, 9}};
-    partitions = {
-        {{1, 2, 3, 4}, {1, 2}}, {{1, 2, 3, 4}, {3, 5}}, {{1, 2, 3, 4}, {7, 9}}};
+    partitions = {{{1, 2, 3, 4}, {1, 2, 3, 5}}, {{1, 2, 3, 4}, {7, 9}}};
     test_subarray_partitioner(
         subarray_layout, ranges, partitions, attr, budget);
   }
@@ -2132,3 +2197,430 @@ TEST_CASE_METHOD(
 
   close_array(ctx_, array_);
 }
+
+TEST_CASE_METHOD(
+    SubarrayPartitionerSparseFx,
+    "SubarrayPartitioner (Sparse): 1D, single-range, string dimension",
+    "[SubarrayPartitioner][sparse][1D][SR][string-dims][basic]") {
+  // Create array
+  create_array(
+      ctx_,
+      array_name_,
+      TILEDB_SPARSE,
+      {"d"},
+      {TILEDB_STRING_ASCII},
+      {nullptr},
+      {nullptr},
+      {"a"},
+      {TILEDB_INT32},
+      {1},
+      {tiledb::test::Compressor(TILEDB_FILTER_NONE, -1)},
+      TILEDB_ROW_MAJOR,
+      TILEDB_ROW_MAJOR,
+      2,
+      false,
+      false);
+
+  // ####### WRITE #######
+
+  // Open array
+  tiledb_array_t* array;
+  int rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
+  CHECK(rc == TILEDB_OK);
+
+  // Create and submit query
+  tiledb_query_t* query;
+  rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
+  REQUIRE(rc == TILEDB_OK);
+
+  char d_data[] = "ccbbccddaa";
+  uint64_t d_data_size = sizeof(d_data) - 1;  // Ignore '\0'
+  uint64_t d_off[] = {0, 2, 4, 8};
+  uint64_t d_off_size = sizeof(d_off);
+  int32_t a_data[] = {3, 2, 4, 1};
+  uint64_t a_size = sizeof(a_data);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "d", d_data, &d_data_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_offsets_buffer(ctx_, query, "d", d_off, &d_off_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a_data, &a_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_layout(ctx_, query, TILEDB_UNORDERED);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_submit(ctx_, query);
+  REQUIRE(rc == TILEDB_OK);
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  CHECK(rc == TILEDB_OK);
+  tiledb_array_free(&array);
+
+  // #### PARTITIONER ####
+
+  // Open array
+  rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
+  CHECK(rc == TILEDB_OK);
+
+  Layout layout = Layout::ROW_MAJOR;
+
+  SECTION("Global order") {
+    layout = Layout::GLOBAL_ORDER;
+  }
+
+  SECTION("Row-major") {
+    layout = Layout::ROW_MAJOR;
+  }
+
+  SECTION("Col-major") {
+    layout = Layout::COL_MAJOR;
+  }
+
+  SECTION("Unordered") {
+    layout = Layout::UNORDERED;
+  }
+
+  // Check unsplittable
+  tiledb::sm::Subarray subarray(
+      array->array_, layout, &g_helper_stats, g_helper_logger());
+  tiledb::sm::Range r;
+  r.set_str_range("bb", "bb");
+  subarray.add_range(0, std::move(r), true);
+  ThreadPool tp;
+  CHECK(tp.init(4).ok());
+  Config config;
+  SubarrayPartitioner partitioner(
+      &config,
+      subarray,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
+  auto st = partitioner.set_result_budget("d", 10);
+  CHECK(!st.ok());
+  uint64_t budget = 0;
+  CHECK(!partitioner.get_result_budget("d", &budget).ok());
+  st = partitioner.set_result_budget("d", 10, 1);
+  CHECK(st.ok());
+  uint64_t budget_off = 0, budget_val = 0;
+  CHECK(partitioner.get_result_budget("d", &budget_off, &budget_val).ok());
+  CHECK(budget_off == 10);
+  CHECK(budget_val == 1);
+  bool unsplittable = true;
+  CHECK(partitioner.next(&unsplittable).ok());
+  CHECK(unsplittable);
+  auto partition = partitioner.current();
+  CHECK(partition.range_num() == 1);
+  const Range* range = nullptr;
+  partition.get_range(0, 0, &range);
+  CHECK(range != nullptr);
+  CHECK(range->start_str() == std::string("bb", 2));
+  CHECK(range->end_str() == std::string("bb", 2));
+
+  // Check full
+  tiledb::sm::Subarray subarray_full(
+      array->array_, layout, &g_helper_stats, g_helper_logger());
+  r.set_str_range("a", "bb");
+  subarray_full.add_range(0, std::move(r), true);
+  SubarrayPartitioner partitioner_full(
+      &config,
+      subarray_full,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
+  st = partitioner_full.set_result_budget("d", 16, 4);
+  CHECK(st.ok());
+  CHECK(partitioner_full.get_result_budget("d", &budget_off, &budget_val).ok());
+  CHECK(budget_off == 16);
+  CHECK(budget_val == 4);
+  CHECK(partitioner_full.next(&unsplittable).ok());
+  CHECK(!unsplittable);
+  partition = partitioner_full.current();
+  CHECK(partition.range_num() == 1);
+  partition.get_range(0, 0, &range);
+  CHECK(range != nullptr);
+  CHECK(range->start_str() == std::string("a", 1));
+  CHECK(range->end_str() == std::string("bb", 2));
+
+  // Check split
+  tiledb::sm::Subarray subarray_split(
+      array->array_, layout, &g_helper_stats, g_helper_logger());
+  r.set_str_range("a", "bb");
+  subarray_split.add_range(0, std::move(r), true);
+  SubarrayPartitioner partitioner_split(
+      &config,
+      subarray_split,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
+  st = partitioner_split.set_result_budget("d", 10, 4);
+  CHECK(st.ok());
+  CHECK(
+      partitioner_split.get_result_budget("d", &budget_off, &budget_val).ok());
+  CHECK(budget_off == 10);
+  CHECK(budget_val == 4);
+  CHECK(partitioner_split.next(&unsplittable).ok());
+  CHECK(!unsplittable);
+  partition = partitioner_split.current();
+  CHECK(partition.range_num() == 1);
+  partition.get_range(0, 0, &range);
+  CHECK(range != nullptr);
+  CHECK(range->start_str() == std::string("a", 1));
+  CHECK(range->end_str() == std::string("a\x7F", 2));
+  CHECK(partitioner_split.next(&unsplittable).ok());
+  CHECK(!unsplittable);
+  partition = partitioner_split.current();
+  CHECK(partition.range_num() == 1);
+  partition.get_range(0, 0, &range);
+  CHECK(range != nullptr);
+  CHECK(range->start_str() == std::string("b", 1));
+  CHECK(range->end_str() == std::string("bb", 2));
+  CHECK(partitioner_split.done());
+
+  // Check no split 2 MBRs
+  tiledb::sm::Subarray subarray_no_split(
+      array->array_, layout, &g_helper_stats, g_helper_logger());
+  r.set_str_range("bb", "cc");
+  subarray_no_split.add_range(0, std::move(r), true);
+  SubarrayPartitioner partitioner_no_split(
+      &config,
+      subarray_no_split,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
+  st = partitioner_no_split.set_result_budget("d", 16, 10);
+  CHECK(st.ok());
+  CHECK(partitioner_no_split.get_result_budget("d", &budget_off, &budget_val)
+            .ok());
+  CHECK(budget_off == 16);
+  CHECK(budget_val == 10);
+  CHECK(partitioner_no_split.next(&unsplittable).ok());
+  CHECK(partitioner_no_split.done());
+  CHECK(!unsplittable);
+  partition = partitioner_no_split.current();
+  CHECK(partition.range_num() == 1);
+  partition.get_range(0, 0, &range);
+  CHECK(range != nullptr);
+  CHECK(range->start_str() == std::string("bb", 2));
+  CHECK(range->end_str() == std::string("cc", 2));
+
+  // Check split 2 MBRs
+  tiledb::sm::Subarray subarray_split_2(
+      array->array_, layout, &g_helper_stats, g_helper_logger());
+  r.set_str_range("bb", "cc");
+  subarray_split_2.add_range(0, std::move(r), true);
+  SubarrayPartitioner partitioner_split_2(
+      &config,
+      subarray_split_2,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
+  st = partitioner_split_2.set_result_budget("d", 8, 10);
+  CHECK(st.ok());
+  CHECK(partitioner_split_2.get_result_budget("d", &budget_off, &budget_val)
+            .ok());
+  CHECK(budget_off == 8);
+  CHECK(budget_val == 10);
+  CHECK(partitioner_split_2.next(&unsplittable).ok());
+  CHECK(!partitioner_split_2.done());
+  CHECK(!unsplittable);
+  partition = partitioner_split_2.current();
+  CHECK(partition.range_num() == 1);
+  partition.get_range(0, 0, &range);
+  CHECK(range != nullptr);
+  CHECK(range->start_str() == std::string("bb", 2));
+  CHECK(range->end_str() == std::string("b\x7F", 2));
+  CHECK(partitioner_split_2.next(&unsplittable).ok());
+  CHECK(!unsplittable);
+  partition = partitioner_split_2.current();
+  CHECK(partition.range_num() == 1);
+  partition.get_range(0, 0, &range);
+  CHECK(range != nullptr);
+  CHECK(range->start_str() == std::string("c", 1));
+  CHECK(range->end_str() == std::string("cc", 2));
+  CHECK(partitioner_split_2.done());
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  CHECK(rc == TILEDB_OK);
+
+  // Clean up
+  tiledb_array_free(&array);
+  tiledb_query_free(&query);
+}
+
+TEST_CASE_METHOD(
+    SubarrayPartitionerSparseFx,
+    "SubarrayPartitioner (Sparse): 1D, single-range, string dimension, edge "
+    "split",
+    "[SubarrayPartitioner][sparse][1D][SR][string-dims][edge-split]") {
+  // Create array
+  create_array(
+      ctx_,
+      array_name_,
+      TILEDB_SPARSE,
+      {"d"},
+      {TILEDB_STRING_ASCII},
+      {nullptr},
+      {nullptr},
+      {"a"},
+      {TILEDB_INT32},
+      {1},
+      {tiledb::test::Compressor(TILEDB_FILTER_NONE, -1)},
+      TILEDB_ROW_MAJOR,
+      TILEDB_ROW_MAJOR,
+      4,
+      false,
+      false);
+
+  // ####### WRITE #######
+
+  // Open array
+  tiledb_array_t* array;
+  int rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
+  CHECK(rc == TILEDB_OK);
+
+  // Create and submit query
+  tiledb_query_t* query;
+  rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
+  REQUIRE(rc == TILEDB_OK);
+
+  char d_data[] = "ccbbccddaa";
+  uint64_t d_data_size = sizeof(d_data) - 1;  // Ignore '\0'
+  uint64_t d_off[] = {0, 2, 4, 8};
+  uint64_t d_off_size = sizeof(d_off);
+  int32_t a_data[] = {3, 2, 4, 1};
+  uint64_t a_size = sizeof(a_data);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "d", d_data, &d_data_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_offsets_buffer(ctx_, query, "d", d_off, &d_off_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a_data, &a_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_layout(ctx_, query, TILEDB_UNORDERED);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_submit(ctx_, query);
+  REQUIRE(rc == TILEDB_OK);
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  CHECK(rc == TILEDB_OK);
+
+  // Free array
+  tiledb_array_free(&array);
+
+  // #### PARTITIONER ####
+
+  // Open array
+  rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
+  CHECK(rc == TILEDB_OK);
+
+  // Unsplittable
+  bool unsplittable = false;
+  uint64_t budget_off = 0, budget_val = 0;
+
+  Layout layout = Layout::ROW_MAJOR;
+
+  SECTION("Global order") {
+    layout = Layout::GLOBAL_ORDER;
+  }
+
+  SECTION("Row-major") {
+    layout = Layout::ROW_MAJOR;
+  }
+
+  SECTION("Col-major") {
+    layout = Layout::COL_MAJOR;
+  }
+
+  SECTION("Unordered") {
+    layout = Layout::UNORDERED;
+  }
+
+  tiledb::sm::Subarray subarray(
+      array->array_, layout, &g_helper_stats, g_helper_logger());
+  tiledb::sm::Range r;
+  r.set_str_range("cc", "ccd");
+  subarray.add_range(0, std::move(r), true);
+  ThreadPool tp;
+  CHECK(tp.init(4).ok());
+  Config config;
+  SubarrayPartitioner partitioner(
+      &config,
+      subarray,
+      memory_budget_,
+      memory_budget_var_,
+      0,
+      &tp,
+      &g_helper_stats,
+      g_helper_logger());
+  auto st = partitioner.set_result_budget("d", 10, 4);
+  CHECK(st.ok());
+  CHECK(partitioner.get_result_budget("d", &budget_off, &budget_val).ok());
+  CHECK(budget_off == 10);
+  CHECK(budget_val == 4);
+  CHECK(partitioner.next(&unsplittable).ok());
+  CHECK(unsplittable);
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  CHECK(rc == TILEDB_OK);
+
+  // Clean up
+  tiledb_array_free(&array);
+  tiledb_query_free(&query);
+}
+
+TEST_CASE_METHOD(
+    SubarrayPartitionerSparseFx,
+    "SubarrayPartitioner (Sparse): 2D, multi-range, test memory budget "
+    "calibration",
+    "[SubarrayPartitioner][sparse][2D][MR][calibrate-memory_budget]") {
+  Layout subarray_layout = Layout::ROW_MAJOR;
+  std::string attr = "a";
+
+  create_default_2d_array(TILEDB_ROW_MAJOR, TILEDB_ROW_MAJOR);
+  write_default_2d_array();
+  open_array(ctx_, array_, TILEDB_READ);
+
+  subarray_layout = Layout::ROW_MAJOR;
+  SubarrayRanges<uint64_t> ranges = {{1, 1}, {2, 2, 2, 2, 2, 2}};
+  std::vector<SubarrayRanges<uint64_t>> partitions = {
+      {{1, 1}, {2, 2, 2, 2, 2, 2}},
+  };
+
+  uint64_t result_budget = 3 * sizeof(int);
+  uint64_t memory_budget = 2 * sizeof(int);
+  uint64_t memory_budget_var = 20 * sizeof(int);
+  test_subarray_partitioner(
+      subarray_layout,
+      ranges,
+      partitions,
+      attr,
+      result_budget,
+      memory_budget,
+      memory_budget_var);
+
+  close_array(ctx_, array_);
+}
\ No newline at end of file
