@@ -258,14 +258,11 @@ void check_save_to_file() {
   ss << "sm.mem.reader.sparse_global_order.ratio_coords 0.5\n";
   ss << "sm.mem.reader.sparse_global_order.ratio_query_condition 0.25\n";
   ss << "sm.mem.reader.sparse_global_order.ratio_rcs 0.05\n";
-  ss << "sm.mem.reader.sparse_global_order.ratio_result_tiles 0.05\n";
   ss << "sm.mem.reader.sparse_global_order.ratio_tile_ranges 0.1\n";
   ss << "sm.mem.reader.sparse_unordered_with_dups.ratio_array_data 0.1\n";
   ss << "sm.mem.reader.sparse_unordered_with_dups.ratio_coords 0.5\n";
   ss << "sm.mem.reader.sparse_unordered_with_dups.ratio_query_condition "
         "0.25\n";
-  ss << "sm.mem.reader.sparse_unordered_with_dups.ratio_rcs 0.05\n";
-  ss << "sm.mem.reader.sparse_unordered_with_dups.ratio_result_tiles 0.05\n";
   ss << "sm.mem.reader.sparse_unordered_with_dups.ratio_tile_ranges 0.1\n";
   ss << "sm.mem.total_budget 10737418240\n";
   ss << "sm.memory_budget 5368709120\n";
@@ -582,8 +579,6 @@ TEST_CASE("C API: Test config iter", "[capi][config]") {
       "0.1";
   all_param_values["sm.mem.reader.sparse_global_order.ratio_array_data"] =
       "0.1";
-  all_param_values["sm.mem.reader.sparse_global_order.ratio_result_tiles"] =
-      "0.05";
   all_param_values["sm.mem.reader.sparse_global_order.ratio_rcs"] = "0.05";
   all_param_values["sm.mem.reader.sparse_unordered_with_dups.ratio_coords"] =
       "0.5";
@@ -594,10 +589,6 @@ TEST_CASE("C API: Test config iter", "[capi][config]") {
       ["sm.mem.reader.sparse_unordered_with_dups.ratio_tile_ranges"] = "0.1";
   all_param_values
       ["sm.mem.reader.sparse_unordered_with_dups.ratio_array_data"] = "0.1";
-  all_param_values
-      ["sm.mem.reader.sparse_unordered_with_dups.ratio_result_tiles"] = "0.05";
-  all_param_values["sm.mem.reader.sparse_unordered_with_dups.ratio_rcs"] =
-      "0.05";
   all_param_values["sm.enable_signal_handlers"] = "true";
   all_param_values["sm.compute_concurrency_level"] =
       std::to_string(std::thread::hardware_concurrency());
