@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2018 TileDB Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,6 +31,8 @@
  */
 
 #include "catch.hpp"
+#include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #ifdef _WIN32
 #include "tiledb/sm/filesystem/win.h"
 #else
@@ -43,33 +45,21 @@
 #include <sstream>
 #include <thread>
 
+using namespace tiledb::test;
+
 struct DenseVectorFx {
   std::string ATTR_NAME = "val";
   const tiledb_datatype_t ATTR_TYPE = TILEDB_INT64;
   const char* DIM0_NAME = "dim0";
   const tiledb_datatype_t DIM_TYPE = TILEDB_INT64;
-  const std::string HDFS_TEMP_DIR = "hdfs:///tiledb_test/";
-  const std::string S3_PREFIX = "s3://";
-  const std::string S3_BUCKET = S3_PREFIX + random_bucket_name("tiledb") + "/";
-  const std::string S3_TEMP_DIR = S3_BUCKET + "tiledb_test/";
-#ifdef _WIN32
-  const std::string FILE_URI_PREFIX = "";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
-#else
-  const std::string FILE_URI_PREFIX = "file://";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Posix::current_dir() + "/tiledb_test/";
-#endif
   const std::string VECTOR = "vector";
 
   // TileDB context
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
 
-  // Supported filesystems
-  bool supports_s3_;
-  bool supports_hdfs_;
+  // Vector of supported filsystems
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
 
   // Functions
   DenseVectorFx();
@@ -80,90 +70,24 @@ struct DenseVectorFx {
       tiledb_layout_t tile_order);
   void check_read(const std::string& path, tiledb_layout_t layout);
   void check_update(const std::string& path);
-  void check_duplicate_coords(const std::string& path);
   void create_temp_dir(const std::string& path);
   void remove_temp_dir(const std::string& path);
-  static std::string random_bucket_name(const std::string& prefix);
-  void set_supported_fs();
+  static std::string random_name(const std::string& prefix);
 };
 
-DenseVectorFx::DenseVectorFx() {
-  // Supported filesystems
-  set_supported_fs();
-
-  // Create TileDB context
-  tiledb_config_t* config = nullptr;
-  tiledb_error_t* error = nullptr;
-  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  if (supports_s3_) {
-#ifndef TILEDB_TESTS_AWS_S3_CONFIG
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.endpoint_override", "localhost:9999", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.scheme", "https", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.use_virtual_addressing", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.verify_ssl", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(error == nullptr);
-#endif
-  }
-  REQUIRE(tiledb_ctx_alloc(config, &ctx_) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  vfs_ = nullptr;
-  REQUIRE(tiledb_vfs_alloc(ctx_, config, &vfs_) == TILEDB_OK);
-  tiledb_config_free(&config);
-
-  // Connect to S3
-  if (supports_s3_) {
-    // Create bucket if it does not exist
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    REQUIRE(rc == TILEDB_OK);
-    if (!is_bucket) {
-      rc = tiledb_vfs_create_bucket(ctx_, vfs_, S3_BUCKET.c_str());
-      REQUIRE(rc == TILEDB_OK);
-    }
-  }
+DenseVectorFx::DenseVectorFx()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 }
 
 DenseVectorFx::~DenseVectorFx() {
-  if (supports_s3_) {
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    CHECK(rc == TILEDB_OK);
-    if (is_bucket) {
-      CHECK(
-          tiledb_vfs_remove_bucket(ctx_, vfs_, S3_BUCKET.c_str()) == TILEDB_OK);
-    }
-  }
-
+  // Close vfs test
+  REQUIRE(vfs_test_close(fs_vec_, ctx_, vfs_).ok());
   tiledb_vfs_free(&vfs_);
   tiledb_ctx_free(&ctx_);
 }
 
-void DenseVectorFx::set_supported_fs() {
-  tiledb_ctx_t* ctx = nullptr;
-  REQUIRE(tiledb_ctx_alloc(nullptr, &ctx) == TILEDB_OK);
-
-  int is_supported = 0;
-  int rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_S3, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_s3_ = (bool)is_supported;
-  rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_HDFS, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_hdfs_ = (bool)is_supported;
-
-  tiledb_ctx_free(&ctx);
-}
-
 void DenseVectorFx::create_temp_dir(const std::string& path) {
   remove_temp_dir(path);
   REQUIRE(tiledb_vfs_create_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
@@ -176,7 +100,7 @@ void DenseVectorFx::remove_temp_dir(const std::string& path) {
     REQUIRE(tiledb_vfs_remove_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
 }
 
-std::string DenseVectorFx::random_bucket_name(const std::string& prefix) {
+std::string DenseVectorFx::random_name(const std::string& prefix) {
   std::stringstream ss;
   ss << prefix << "-" << std::this_thread::get_id() << "-"
      << TILEDB_TIMESTAMP_NOW_MS;
@@ -197,7 +121,7 @@ void DenseVectorFx::create_dense_vector(
   REQUIRE(rc == TILEDB_OK);
   tiledb_dimension_t* dim;
   rc = tiledb_dimension_alloc(
-      ctx_, DIM0_NAME, TILEDB_INT64, dim_domain, &tile_extent, &dim);
+      ctx_, "d", TILEDB_INT64, dim_domain, &tile_extent, &dim);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_domain_add_dimension(ctx_, domain, dim);
   REQUIRE(rc == TILEDB_OK);
@@ -228,6 +152,7 @@ void DenseVectorFx::create_dense_vector(
   REQUIRE(rc == TILEDB_OK);
   tiledb_attribute_free(&attr);
   tiledb_dimension_free(&dim);
+  tiledb_domain_free(&domain);
   tiledb_array_schema_free(&array_schema);
 
   const char* attributes[] = {ATTR_NAME.c_str()};
@@ -246,7 +171,7 @@ void DenseVectorFx::create_dense_vector(
   tiledb_query_t* write_query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &write_query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
+  rc = tiledb_query_set_data_buffer(
       ctx_,
       write_query,
       attributes[0],
@@ -288,7 +213,7 @@ void DenseVectorFx::check_read(
 
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &read_query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
+  rc = tiledb_query_set_data_buffer(
       ctx_, read_query, attributes[0], read_buffers[0], &read_buffer_sizes[0]);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, read_query, layout);
@@ -331,7 +256,7 @@ void DenseVectorFx::check_update(const std::string& path) {
   tiledb_query_t* update_query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &update_query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
+  rc = tiledb_query_set_data_buffer(
       ctx_,
       update_query,
       attributes[0],
@@ -365,7 +290,7 @@ void DenseVectorFx::check_update(const std::string& path) {
   tiledb_query_t* read_query = nullptr;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &read_query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
+  rc = tiledb_query_set_data_buffer(
       ctx_, read_query, attributes[0], read_buffers[0], &read_buffer_sizes[0]);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, read_query, TILEDB_COL_MAJOR);
@@ -388,151 +313,47 @@ void DenseVectorFx::check_update(const std::string& path) {
   CHECK((buffer[0] == 9 && buffer[1] == 8 && buffer[2] == 7));
 }
 
-void DenseVectorFx::check_duplicate_coords(const std::string& path) {
-  // Open array
-  tiledb_array_t* array;
-  int rc = tiledb_array_alloc(ctx_, path.c_str(), &array);
-  CHECK(rc == TILEDB_OK);
-  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
-  CHECK(rc == TILEDB_OK);
-
-  const int64_t num_writes = 5;
-  for (int64_t write_num = 0; write_num < num_writes; write_num++) {
-    const char* attributes[] = {ATTR_NAME.c_str(), TILEDB_COORDS};
-    int64_t update_buffer[] = {write_num, write_num, write_num};
-    int64_t coords_buffer[] = {7, 8, 9};
-    void* update_buffers[] = {update_buffer, coords_buffer};
-    uint64_t update_buffer_sizes[] = {sizeof(update_buffer),
-                                      sizeof(coords_buffer)};
-
-    // Update
-    tiledb_query_t* update_query;
-    rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &update_query);
-    REQUIRE(rc == TILEDB_OK);
-    rc = tiledb_query_set_buffer(
-        ctx_,
-        update_query,
-        attributes[0],
-        update_buffers[0],
-        &update_buffer_sizes[0]);
-    REQUIRE(rc == TILEDB_OK);
-    rc = tiledb_query_set_buffer(
-        ctx_,
-        update_query,
-        attributes[1],
-        update_buffers[1],
-        &update_buffer_sizes[1]);
-    REQUIRE(rc == TILEDB_OK);
-    rc = tiledb_query_set_layout(ctx_, update_query, TILEDB_UNORDERED);
-    REQUIRE(rc == TILEDB_OK);
-    rc = tiledb_query_submit(ctx_, update_query);
-    REQUIRE(rc == TILEDB_OK);
-    rc = tiledb_query_finalize(ctx_, update_query);
-    REQUIRE(rc == TILEDB_OK);
-    tiledb_query_free(&update_query);
-  }
-
-  // Close array
-  rc = tiledb_array_close(ctx_, array);
-  CHECK(rc == TILEDB_OK);
-
-  // Open array
-  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
-  CHECK(rc == TILEDB_OK);
-
-  // Read
-  uint64_t subarray[] = {7, 9};
-  const char* attributes[] = {ATTR_NAME.c_str()};
-  int64_t buffer[3] = {0};
-  void* read_buffers[] = {buffer};
-  uint64_t read_buffer_sizes[] = {sizeof(buffer)};
-  tiledb_query_t* read_query = nullptr;
-  rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &read_query);
-  REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
-      ctx_, read_query, attributes[0], read_buffers[0], &read_buffer_sizes[0]);
-  REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_layout(ctx_, read_query, TILEDB_GLOBAL_ORDER);
-  REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_subarray(ctx_, read_query, subarray);
-  REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_submit(ctx_, read_query);
-  REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_finalize(ctx_, read_query);
-  REQUIRE(rc == TILEDB_OK);
-
-  // Close array
-  rc = tiledb_array_close(ctx_, array);
-  CHECK(rc == TILEDB_OK);
-
-  // Clean up
-  tiledb_array_free(&array);
-  tiledb_query_free(&read_query);
-
-  CHECK(
-      (buffer[0] == num_writes - 1 && buffer[1] == num_writes - 1 &&
-       buffer[2] == num_writes - 1));
-}
-
 TEST_CASE_METHOD(
-    DenseVectorFx, "C API: Test 1d dense vector", "[capi], [dense-vector]") {
+    DenseVectorFx, "C API: Test 1d dense vector", "[capi][dense-vector]") {
   std::string vector_name;
 
-  if (supports_s3_) {
-    // S3
-    create_temp_dir(S3_TEMP_DIR);
-    vector_name = S3_TEMP_DIR + VECTOR;
-    create_dense_vector(vector_name, TILEDB_ROW_MAJOR, TILEDB_ROW_MAJOR);
-    check_read(vector_name, TILEDB_ROW_MAJOR);
-    check_read(vector_name, TILEDB_COL_MAJOR);
-    check_update(vector_name);
-    check_duplicate_coords(vector_name);
-    remove_temp_dir(S3_TEMP_DIR);
-  } else if (supports_hdfs_) {
-    // HDFS
-    create_temp_dir(HDFS_TEMP_DIR);
-    vector_name = HDFS_TEMP_DIR + VECTOR;
-    create_dense_vector(vector_name, TILEDB_ROW_MAJOR, TILEDB_ROW_MAJOR);
-    check_read(vector_name, TILEDB_ROW_MAJOR);
-    check_read(vector_name, TILEDB_COL_MAJOR);
-    check_update(vector_name);
-    check_duplicate_coords(vector_name);
-    remove_temp_dir(HDFS_TEMP_DIR);
-  } else {
-    // File
-    create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
-    vector_name = FILE_URI_PREFIX + FILE_TEMP_DIR + VECTOR;
-    create_dense_vector(vector_name, TILEDB_ROW_MAJOR, TILEDB_ROW_MAJOR);
-    check_read(vector_name, TILEDB_ROW_MAJOR);
-    check_read(vector_name, TILEDB_COL_MAJOR);
-    check_update(vector_name);
-    check_duplicate_coords(vector_name);
-    remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
-  }
+  // TODO: refactor for each supported FS.
+  std::string temp_dir = fs_vec_[0]->temp_dir();
+  create_temp_dir(temp_dir);
+  vector_name = temp_dir + VECTOR;
+  create_dense_vector(vector_name, TILEDB_ROW_MAJOR, TILEDB_ROW_MAJOR);
+  check_read(vector_name, TILEDB_ROW_MAJOR);
+  check_read(vector_name, TILEDB_COL_MAJOR);
+  check_update(vector_name);
+  remove_temp_dir(temp_dir);
 }
 
 TEST_CASE_METHOD(
     DenseVectorFx,
     "C API: Test 1d dense vector with anonymous attribute",
-    "[capi], [dense-vector], [anon-attr]") {
+    "[capi][dense-vector][anon-attr]") {
   ATTR_NAME = "";
 
-  std::string vector_name = FILE_URI_PREFIX + FILE_TEMP_DIR + VECTOR;
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  SupportedFsLocal local_fs;
+
+  std::string vector_name =
+      local_fs.file_prefix() + local_fs.temp_dir() + VECTOR;
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
   create_dense_vector(vector_name, TILEDB_ROW_MAJOR, TILEDB_ROW_MAJOR);
   check_read(vector_name, TILEDB_ROW_MAJOR);
   check_read(vector_name, TILEDB_COL_MAJOR);
   check_update(vector_name);
-  check_duplicate_coords(vector_name);
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
 
 TEST_CASE_METHOD(
     DenseVectorFx,
     "C API: Test 1d dense vector cell/tile layout",
-    "[capi], [dense-vector], [dense-vector-layout]") {
-  std::string vector_name = FILE_URI_PREFIX + FILE_TEMP_DIR + VECTOR;
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+    "[capi][dense-vector][dense-vector-layout]") {
+  SupportedFsLocal local_fs;
+  std::string vector_name =
+      local_fs.file_prefix() + local_fs.temp_dir() + VECTOR;
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
   create_dense_vector(vector_name, TILEDB_COL_MAJOR, TILEDB_COL_MAJOR);
 
   // Check layout
@@ -548,15 +369,17 @@ TEST_CASE_METHOD(
   CHECK(tile_order == TILEDB_COL_MAJOR);
   tiledb_array_schema_free(&array_schema);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
 
 TEST_CASE_METHOD(
     DenseVectorFx,
     "C API: Test 1d dense vector, update",
-    "[capi], [dense-vector], [dense-vector-update]") {
-  std::string vector_name = FILE_URI_PREFIX + FILE_TEMP_DIR + VECTOR;
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+    "[capi][dense-vector][dense-vector-update]") {
+  SupportedFsLocal local_fs;
+  std::string vector_name =
+      local_fs.file_prefix() + local_fs.temp_dir() + VECTOR;
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
   int rc;
 
   // --- Create array ----
@@ -600,6 +423,7 @@ TEST_CASE_METHOD(
   REQUIRE(rc == TILEDB_OK);
   tiledb_attribute_free(&attr);
   tiledb_dimension_free(&dim);
+  tiledb_domain_free(&domain);
   tiledb_array_schema_free(&array_schema);
 
   // --- Zero write ----
@@ -620,7 +444,7 @@ TEST_CASE_METHOD(
   tiledb_query_t* write_query_0;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &write_query_0);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
+  rc = tiledb_query_set_data_buffer(
       ctx_,
       write_query_0,
       ATTR_NAME.c_str(),
@@ -661,7 +485,7 @@ TEST_CASE_METHOD(
   tiledb_query_t* write_query_1;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &write_query_1);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
+  rc = tiledb_query_set_data_buffer(
       ctx_,
       write_query_1,
       ATTR_NAME.c_str(),
@@ -701,7 +525,7 @@ TEST_CASE_METHOD(
   tiledb_query_t* write_query_2;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &write_query_2);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
+  rc = tiledb_query_set_data_buffer(
       ctx_,
       write_query_2,
       ATTR_NAME.c_str(),
@@ -740,7 +564,7 @@ TEST_CASE_METHOD(
   tiledb_query_t* write_query_3;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &write_query_3);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
+  rc = tiledb_query_set_data_buffer(
       ctx_,
       write_query_3,
       ATTR_NAME.c_str(),
@@ -780,7 +604,7 @@ TEST_CASE_METHOD(
 
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &read_query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(
+  rc = tiledb_query_set_data_buffer(
       ctx_, read_query, ATTR_NAME.c_str(), read_buffer, &read_buffer_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, read_query, TILEDB_ROW_MAJOR);
@@ -808,5 +632,5 @@ TEST_CASE_METHOD(
   tiledb_array_free(&array);
   tiledb_query_free(&read_query);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
