@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,8 +31,11 @@
  */
 
 #include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #include "tiledb/sm/c_api/tiledb.h"
 #include "tiledb/sm/c_api/tiledb_struct_def.h"
+#include "tiledb/sm/enums/encryption_type.h"
+#include "tiledb/sm/global_state/unit_test_config.h"
 
 #ifdef _WIN32
 #include "tiledb/sm/filesystem/win.h"
@@ -46,6 +49,7 @@
 #include <thread>
 
 using namespace tiledb::sm;
+using namespace tiledb::test;
 
 /* ********************************* */
 /*         STRUCT DEFINITION         */
@@ -54,10 +58,8 @@ using namespace tiledb::sm;
 struct CMetadataFx {
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
-  bool s3_supported_, hdfs_supported_;
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
   std::string temp_dir_;
-  const std::string s3_bucket_name_ =
-      "s3://" + random_bucket_name("tiledb") + "/";
   std::string array_name_;
   const char* ARRAY_NAME = "test_metadata";
   tiledb_array_t* array_ = nullptr;
@@ -69,30 +71,31 @@ struct CMetadataFx {
   void create_default_array_1d();
   void create_default_array_1d_with_key();
 
+  // Used to get the number of directories or files of another directory
+  struct get_num_struct {
+    int num;
+  };
+
+  static int get_meta_num(const char* path, void* data);
+
   CMetadataFx();
   ~CMetadataFx();
 };
 
-CMetadataFx::CMetadataFx() {
-  ctx_ = nullptr;
-  vfs_ = nullptr;
-  hdfs_supported_ = false;
-  s3_supported_ = false;
-
-  get_supported_fs(&s3_supported_, &hdfs_supported_);
-  create_ctx_and_vfs(s3_supported_, &ctx_, &vfs_);
-  create_s3_bucket(s3_bucket_name_, s3_supported_, ctx_, vfs_);
+CMetadataFx::CMetadataFx()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 
 // Create temporary directory based on the supported filesystem
 #ifdef _WIN32
-  temp_dir_ = tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
+  SupportedFsLocal windows_fs;
+  temp_dir_ = windows_fs.file_prefix() + windows_fs.temp_dir();
 #else
-  temp_dir_ = "file://" + tiledb::sm::Posix::current_dir() + "/tiledb_test/";
+  SupportedFsLocal posix_fs;
+  temp_dir_ = posix_fs.file_prefix() + posix_fs.temp_dir();
 #endif
-  if (s3_supported_)
-    temp_dir_ = s3_bucket_name_ + "tiledb/test/";
-  if (hdfs_supported_)
-    temp_dir_ = "hdfs:///tiledb_test/";
+
   create_dir(temp_dir_, ctx_, vfs_);
 
   array_name_ = temp_dir_ + ARRAY_NAME;
@@ -107,6 +110,14 @@ CMetadataFx::~CMetadataFx() {
   tiledb_vfs_free(&vfs_);
 }
 
+int CMetadataFx::get_meta_num(const char* path, void* data) {
+  (void)path;
+  auto data_struct = (CMetadataFx::get_num_struct*)data;
+  ++data_struct->num;
+
+  return 1;
+}
+
 void CMetadataFx::create_default_array_1d() {
   uint64_t domain[] = {1, 10};
   uint64_t tile_extent = 5;
@@ -121,9 +132,9 @@ void CMetadataFx::create_default_array_1d() {
       {"a", "b", "c"},
       {TILEDB_INT32, TILEDB_CHAR, TILEDB_FLOAT32},
       {1, TILEDB_VAR_NUM, 2},
-      {::Compressor(TILEDB_FILTER_NONE, -1),
-       ::Compressor(TILEDB_FILTER_ZSTD, -1),
-       ::Compressor(TILEDB_FILTER_LZ4, -1)},
+      {tiledb::test::Compressor(TILEDB_FILTER_NONE, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_ZSTD, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1)},
       TILEDB_ROW_MAJOR,
       TILEDB_ROW_MAJOR,
       2);
@@ -146,9 +157,9 @@ void CMetadataFx::create_default_array_1d_with_key() {
       {"a", "b", "c"},
       {TILEDB_INT32, TILEDB_CHAR, TILEDB_FLOAT32},
       {1, TILEDB_VAR_NUM, 2},
-      {::Compressor(TILEDB_FILTER_NONE, -1),
-       ::Compressor(TILEDB_FILTER_ZSTD, -1),
-       ::Compressor(TILEDB_FILTER_LZ4, -1)},
+      {tiledb::test::Compressor(TILEDB_FILTER_NONE, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_ZSTD, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1)},
       TILEDB_ROW_MAJOR,
       TILEDB_ROW_MAJOR,
       2);
@@ -204,12 +215,30 @@ TEST_CASE_METHOD(
   REQUIRE(rc == TILEDB_OK);
 
   // Open with key
-  rc = tiledb_array_open_with_key(
-      ctx_, array, TILEDB_READ, enc_type_, key_, key_len_);
+  tiledb_config_t* config;
+  tiledb_error_t* error = nullptr;
+  rc = tiledb_config_alloc(&config, &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  std::string encryption_type_string =
+      encryption_type_str((tiledb::sm::EncryptionType)enc_type_);
+  rc = tiledb_config_set(
+      config, "sm.encryption_type", encryption_type_string.c_str(), &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.encryption_key", key_, &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  tiledb::sm::UnitTestConfig::instance().array_encryption_key_length.set(
+      key_len_);
+  rc = tiledb_array_set_config(ctx_, array, config);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
   CHECK(rc == TILEDB_ERR);
 
   // Clean up
   tiledb_array_free(&array);
+  tiledb_config_free(&config);
 }
 
 TEST_CASE_METHOD(
@@ -257,8 +286,8 @@ TEST_CASE_METHOD(
   CHECK(rc == TILEDB_OK);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
 
   rc = tiledb_array_get_metadata(ctx_, array, "foo", &v_type, &v_num, &v_r);
   CHECK(rc == TILEDB_OK);
@@ -280,8 +309,8 @@ TEST_CASE_METHOD(
   CHECK(rc == TILEDB_OK);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
   CHECK(key_len == strlen("bb"));
   CHECK(!strncmp(key, "bb", strlen("bb")));
 
@@ -421,8 +450,8 @@ TEST_CASE_METHOD(
   CHECK(rc == TILEDB_OK);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
 
   rc = tiledb_array_get_metadata(ctx_, array, "foo", &v_type, &v_num, &v_r);
   CHECK(rc == TILEDB_OK);
@@ -440,8 +469,8 @@ TEST_CASE_METHOD(
   CHECK(rc == TILEDB_OK);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
   CHECK(key_len == strlen("bb"));
   CHECK(!strncmp(key, "bb", strlen("bb")));
 
@@ -453,7 +482,7 @@ TEST_CASE_METHOD(
 
 TEST_CASE_METHOD(
     CMetadataFx,
-    "C API: Metadata, multiple metadata and cosnolidate",
+    "C API: Metadata, multiple metadata and consolidate",
     "[capi][metadata][multiple][consolidation]") {
   // Create default array
   create_default_array_1d();
@@ -462,6 +491,8 @@ TEST_CASE_METHOD(
   tiledb_array_t* array;
   int rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 1);
+  REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
   REQUIRE(rc == TILEDB_OK);
 
@@ -478,12 +509,11 @@ TEST_CASE_METHOD(
   REQUIRE(rc == TILEDB_OK);
   tiledb_array_free(&array);
 
-  // Prevent array metadata filename/timestamp conflicts
-  std::this_thread::sleep_for(std::chrono::milliseconds(1));
-
   // Update
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 2);
+  REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_delete_metadata(ctx_, array, "aaa");
@@ -513,8 +543,8 @@ TEST_CASE_METHOD(
   CHECK(rc == TILEDB_OK);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
 
   rc = tiledb_array_get_metadata(ctx_, array, "cccc", &v_type, &v_num, &v_r);
   CHECK(rc == TILEDB_OK);
@@ -534,8 +564,8 @@ TEST_CASE_METHOD(
   CHECK(rc == TILEDB_OK);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
   CHECK(key_len == strlen("bb"));
   CHECK(!strncmp(key, "bb", strlen("bb")));
 
@@ -545,8 +575,62 @@ TEST_CASE_METHOD(
   tiledb_array_free(&array);
 
   // Consolidate
-  rc = tiledb_array_consolidate_metadata(ctx_, array_name_.c_str(), nullptr);
+  SECTION("tiledb_array_consolidate") {
+    // Configuration for consolidating array metadata
+    tiledb_config_t* config = nullptr;
+    tiledb_error_t* error = nullptr;
+    REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
+    REQUIRE(error == nullptr);
+    int rc = tiledb_config_set(
+        config, "sm.consolidation.mode", "array_meta", &error);
+    REQUIRE(rc == TILEDB_OK);
+    REQUIRE(error == nullptr);
+    rc = tiledb_array_consolidate(ctx_, array_name_.c_str(), config);
+    CHECK(rc == TILEDB_OK);
+    tiledb_config_free(&config);
+  }
+
+  // Check number of metadata files
+  get_num_struct data = {0};
+  auto meta_folder =
+      array_name_ + "/" + tiledb::sm::constants::array_metadata_folder_name;
+  rc = tiledb_vfs_ls(ctx_, vfs_, meta_folder.c_str(), &get_meta_num, &data);
   CHECK(rc == TILEDB_OK);
+  CHECK(data.num == 4);
+
+  // Read at timestamp 1
+  rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 1);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_get_metadata(ctx_, array, "aaa", &v_type, &v_num, &v_r);
+  CHECK(rc == TILEDB_OK);
+  CHECK(v_type == TILEDB_INT32);
+  CHECK(*(const int32_t*)v_r == 5);
+  CHECK(v_num == 1);
+  rc = tiledb_array_close(ctx_, array);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb_array_free(&array);
+
+  // Vacuum
+  tiledb_config_t* config_v = nullptr;
+  tiledb_error_t* error_v = nullptr;
+  REQUIRE(tiledb_config_alloc(&config_v, &error_v) == TILEDB_OK);
+  REQUIRE(error_v == nullptr);
+  rc = tiledb_config_set(config_v, "sm.vacuum.mode", "array_meta", &error_v);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error_v == nullptr);
+  rc = tiledb_array_vacuum(ctx_, array_name_.c_str(), config_v);
+  CHECK(rc == TILEDB_OK);
+  tiledb_config_free(&config_v);
+
+  // Check number of metadata files
+  data = {0};
+  rc = tiledb_vfs_ls(ctx_, vfs_, meta_folder.c_str(), &get_meta_num, &data);
+  CHECK(rc == TILEDB_OK);
+  CHECK(data.num == 1);
 
   // Open the array in read mode
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
@@ -581,8 +665,17 @@ TEST_CASE_METHOD(
   tiledb_array_free(&array);
 
   // Consolidate again
-  rc = tiledb_array_consolidate_metadata(ctx_, array_name_.c_str(), nullptr);
+  tiledb_config_t* config = nullptr;
+  tiledb_error_t* error = nullptr;
+  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.consolidation.mode", "array_meta", &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_array_consolidate(ctx_, array_name_.c_str(), config);
   CHECK(rc == TILEDB_OK);
+  tiledb_config_free(&config);
+  tiledb_array_free(&array);
 
   // Open the array in read mode
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
@@ -622,6 +715,8 @@ TEST_CASE_METHOD(
   tiledb_array_t* array;
   int rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 1);
+  REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
   REQUIRE(rc == TILEDB_OK);
 
@@ -638,13 +733,11 @@ TEST_CASE_METHOD(
   REQUIRE(rc == TILEDB_OK);
   tiledb_array_free(&array);
 
-  // Prevent array metadata filename/timestamp conflicts
-  auto timestamp = tiledb::sm::utils::time::timestamp_now_ms();
-  std::this_thread::sleep_for(std::chrono::milliseconds(1));
-
   // Update
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 2);
+  REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_delete_metadata(ctx_, array, "aaa");
@@ -656,7 +749,9 @@ TEST_CASE_METHOD(
   // Open the array in read mode at a timestamp
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_array_open_at(ctx_, array, TILEDB_READ, timestamp);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 1);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
   REQUIRE(rc == TILEDB_OK);
 
   // Read
@@ -689,6 +784,8 @@ TEST_CASE_METHOD(
   tiledb_array_t* array;
   int rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 1);
+  REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
   REQUIRE(rc == TILEDB_OK);
 
@@ -705,13 +802,11 @@ TEST_CASE_METHOD(
   REQUIRE(rc == TILEDB_OK);
   tiledb_array_free(&array);
 
-  // Prevent array metadata filename/timestamp conflicts
-  auto timestamp = tiledb::sm::utils::time::timestamp_now_ms();
-  std::this_thread::sleep_for(std::chrono::milliseconds(1));
-
   // Update
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 2);
+  REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_delete_metadata(ctx_, array, "aaa");
@@ -723,7 +818,9 @@ TEST_CASE_METHOD(
   // Open the array in read mode at a timestamp
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_array_open_at(ctx_, array, TILEDB_READ, timestamp);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 1);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
   REQUIRE(rc == TILEDB_OK);
 
   // Read
@@ -742,6 +839,9 @@ TEST_CASE_METHOD(
   CHECK(num == 2);
 
   // Reopen
+  rc = tiledb_array_set_open_timestamp_end(
+      ctx_, array, tiledb::sm::utils::time::timestamp_now_ms());
+  REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_reopen(ctx_, array);
   CHECK(rc == TILEDB_OK);
 
@@ -761,6 +861,147 @@ TEST_CASE_METHOD(
   tiledb_array_free(&array);
 }
 
+TEST_CASE_METHOD(
+    CMetadataFx,
+    "C API: Metadata, timestamp_end",
+    "[capi][metadata][timestamp-end]") {
+  // Create default array
+  create_default_array_1d();
+
+  // Create and open array in write mode
+  tiledb_array_t* array;
+  int rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 1);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
+  REQUIRE(rc == TILEDB_OK);
+
+  // Write items
+  int32_t v1 = 4;
+  rc = tiledb_array_put_metadata(ctx_, array, "aaa", TILEDB_INT32, 1, &v1);
+  CHECK(rc == TILEDB_OK);
+  float f1[] = {1.0f, 1.2f};
+  rc = tiledb_array_put_metadata(ctx_, array, "bb", TILEDB_FLOAT32, 2, f1);
+  CHECK(rc == TILEDB_OK);
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb_array_free(&array);
+
+  // Update
+  rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 2);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_delete_metadata(ctx_, array, "aaa");
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_array_close(ctx_, array);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb_array_free(&array);
+
+  // Create and open array in write mode
+  rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 3);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
+  REQUIRE(rc == TILEDB_OK);
+
+  // Write items
+  int32_t v2 = 5;
+  rc = tiledb_array_put_metadata(ctx_, array, "aaa", TILEDB_INT32, 1, &v2);
+  CHECK(rc == TILEDB_OK);
+  float f2[] = {1.1f, 1.2f};
+  rc = tiledb_array_put_metadata(ctx_, array, "bb", TILEDB_FLOAT32, 2, f2);
+  CHECK(rc == TILEDB_OK);
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb_array_free(&array);
+
+  // Update
+  rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 4);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_delete_metadata(ctx_, array, "aaa");
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_array_close(ctx_, array);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb_array_free(&array);
+
+  // Create and open array in write mode
+  rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 5);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
+  REQUIRE(rc == TILEDB_OK);
+
+  // Write items
+  int32_t v3 = 6;
+  rc = tiledb_array_put_metadata(ctx_, array, "aaa", TILEDB_INT32, 1, &v3);
+  CHECK(rc == TILEDB_OK);
+  float f3[] = {1.2f, 1.3f};
+  rc = tiledb_array_put_metadata(ctx_, array, "bb", TILEDB_FLOAT32, 2, f3);
+  CHECK(rc == TILEDB_OK);
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb_array_free(&array);
+
+  // Update
+  rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 6);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_delete_metadata(ctx_, array, "aaa");
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_array_close(ctx_, array);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb_array_free(&array);
+
+  // Open the array in read mode between timestamp1 and timestamp2
+  rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
+  rc = tiledb_array_set_open_timestamp_start(ctx_, array, 2);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_set_open_timestamp_end(ctx_, array, 3);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
+  REQUIRE(rc == TILEDB_OK);
+
+  // Read - ensure that the data is only that written between
+  // timestamp1 and timestamp2
+  const void* v_r;
+  tiledb_datatype_t v_type;
+  uint32_t v_num;
+  rc = tiledb_array_get_metadata(ctx_, array, "aaa", &v_type, &v_num, &v_r);
+  CHECK(rc == TILEDB_OK);
+  CHECK(v_type == TILEDB_INT32);
+  CHECK(v_num == 1);
+  CHECK(*((const int32_t*)v_r) == 5);
+
+  uint64_t num = 0;
+  rc = tiledb_array_get_metadata_num(ctx_, array, &num);
+  CHECK(rc == TILEDB_OK);
+  CHECK(num == 2);
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb_array_free(&array);
+}
+
 TEST_CASE_METHOD(
     CMetadataFx,
     "C API: Metadata, encryption",
@@ -772,8 +1013,25 @@ TEST_CASE_METHOD(
   tiledb_array_t* array;
   int rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_array_open_with_key(
-      ctx_, array, TILEDB_WRITE, enc_type_, key_, key_len_);
+  tiledb_config_t* config;
+  tiledb_error_t* error = nullptr;
+  rc = tiledb_config_alloc(&config, &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  std::string encryption_type_string =
+      encryption_type_str((tiledb::sm::EncryptionType)enc_type_);
+  rc = tiledb_config_set(
+      config, "sm.encryption_type", encryption_type_string.c_str(), &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.encryption_key", key_, &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_array_set_config(ctx_, array, config);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb::sm::UnitTestConfig::instance().array_encryption_key_length.set(
+      key_len_);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
   REQUIRE(rc == TILEDB_OK);
 
   // Write items
@@ -795,8 +1053,9 @@ TEST_CASE_METHOD(
   // Update
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_array_open_with_key(
-      ctx_, array, TILEDB_WRITE, enc_type_, key_, key_len_);
+  rc = tiledb_array_set_config(ctx_, array, config);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_array_delete_metadata(ctx_, array, "aaa");
   CHECK(rc == TILEDB_OK);
@@ -810,8 +1069,9 @@ TEST_CASE_METHOD(
   // Open the array in read mode
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_array_open_with_key(
-      ctx_, array, TILEDB_READ, enc_type_, key_, key_len_);
+  rc = tiledb_array_set_config(ctx_, array, config);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
   REQUIRE(rc == TILEDB_OK);
 
   // Read
@@ -826,8 +1086,8 @@ TEST_CASE_METHOD(
   CHECK(rc == TILEDB_OK);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
 
   rc = tiledb_array_get_metadata(ctx_, array, "cccc", &v_type, &v_num, &v_r);
   CHECK(rc == TILEDB_OK);
@@ -847,8 +1107,8 @@ TEST_CASE_METHOD(
   CHECK(rc == TILEDB_OK);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
   CHECK(key_len == strlen("bb"));
   CHECK(!strncmp(key, "bb", strlen("bb")));
 
@@ -858,19 +1118,53 @@ TEST_CASE_METHOD(
   tiledb_array_free(&array);
 
   // Consolidate without key - error
-  rc = tiledb_array_consolidate_metadata(ctx_, array_name_.c_str(), nullptr);
+  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.consolidation.mode", "array_meta", &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_array_consolidate(ctx_, array_name_.c_str(), config);
   CHECK(rc == TILEDB_ERR);
+  tiledb_config_free(&config);
 
   // Consolidate with key - ok
-  rc = tiledb_array_consolidate_metadata_with_key(
-      ctx_, array_name_.c_str(), enc_type_, key_, key_len_, nullptr);
+  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.consolidation.mode", "array_meta", &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  encryption_type_string =
+      encryption_type_str((tiledb::sm::EncryptionType)enc_type_);
+  rc = tiledb_config_set(
+      config, "sm.encryption_type", encryption_type_string.c_str(), &error);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.encryption_key", key_, &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_array_consolidate(ctx_, array_name_.c_str(), config);
   CHECK(rc == TILEDB_OK);
+  tiledb_config_free(&config);
+  tiledb_array_free(&array);
 
   // Open the array in read mode
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_array_open_with_key(
-      ctx_, array, TILEDB_READ, enc_type_, key_, key_len_);
+  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  encryption_type_string =
+      encryption_type_str((tiledb::sm::EncryptionType)enc_type_);
+  rc = tiledb_config_set(
+      config, "sm.encryption_type", encryption_type_string.c_str(), &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.encryption_key", key_, &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_array_set_config(ctx_, array, config);
+  REQUIRE(rc == TILEDB_OK);
+  tiledb::sm::UnitTestConfig::instance().array_encryption_key_length.set(
+      key_len_);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
   REQUIRE(rc == TILEDB_OK);
 
   num = 0;
@@ -886,8 +1180,9 @@ TEST_CASE_METHOD(
   // Write once more
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_array_open_with_key(
-      ctx_, array, TILEDB_WRITE, enc_type_, key_, key_len_);
+  rc = tiledb_array_set_config(ctx_, array, config);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
   REQUIRE(rc == TILEDB_OK);
 
   // Write items
@@ -901,15 +1196,28 @@ TEST_CASE_METHOD(
   tiledb_array_free(&array);
 
   // Consolidate again
-  rc = tiledb_array_consolidate_metadata_with_key(
-      ctx_, array_name_.c_str(), enc_type_, key_, key_len_, nullptr);
+  rc = tiledb_config_set(config, "sm.consolidation.mode", "array_meta", &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  encryption_type_string =
+      encryption_type_str((tiledb::sm::EncryptionType)enc_type_);
+  rc = tiledb_config_set(
+      config, "sm.encryption_type", encryption_type_string.c_str(), &error);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.encryption_key", key_, &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+
+  rc = tiledb_array_consolidate(ctx_, array_name_.c_str(), config);
   CHECK(rc == TILEDB_OK);
+  tiledb_array_free(&array);
 
   // Open the array in read mode
   rc = tiledb_array_alloc(ctx_, array_name_.c_str(), &array);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_array_open_with_key(
-      ctx_, array, TILEDB_READ, enc_type_, key_, key_len_);
+  rc = tiledb_array_set_config(ctx_, array, config);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
   REQUIRE(rc == TILEDB_OK);
 
   num = 0;
@@ -933,6 +1241,7 @@ TEST_CASE_METHOD(
   rc = tiledb_array_close(ctx_, array);
   REQUIRE(rc == TILEDB_OK);
   tiledb_array_free(&array);
+  tiledb_config_free(&config);
 }
 
 TEST_CASE_METHOD(
@@ -972,6 +1281,7 @@ TEST_CASE_METHOD(
   CHECK(*((int32_t*)vback_ptr) == 10);
   rc = tiledb_array_close(ctx_, array);
   CHECK(rc == TILEDB_OK);
+  tiledb_array_free(&array);
 
   // Prevent array metadata filename/timestamp conflicts
   std::this_thread::sleep_for(std::chrono::milliseconds(1));
@@ -997,6 +1307,8 @@ TEST_CASE_METHOD(
   CHECK(*((int32_t*)vback_ptr) == 20);
   rc = tiledb_array_close(ctx_, array);
   CHECK(rc == TILEDB_OK);
+
+  tiledb_array_free(&array);
 }
 
 TEST_CASE_METHOD(
