@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2018 TileDB Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,7 +31,8 @@
  */
 
 #include "catch.hpp"
-
+#include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #ifdef _WIN32
 #include "tiledb/sm/filesystem/win.h"
 #else
@@ -45,23 +46,9 @@
 #include <sstream>
 #include <thread>
 
+using namespace tiledb::test;
+
 struct ObjectMgmtFx {
-  const std::string HDFS_TEMP_DIR = "hdfs:///tiledb_test/";
-  const std::string HDFS_FULL_TEMP_DIR = "hdfs://localhost:9000/tiledb_test/";
-  const std::string S3_PREFIX = "s3://";
-  const std::string S3_BUCKET = S3_PREFIX + random_bucket_name("tiledb") + "/";
-  const std::string S3_TEMP_DIR = S3_BUCKET + "tiledb_test/";
-#if defined(_WIN32)
-  const std::string FILE_URI_PREFIX = "";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
-  const std::string FILE_FULL_TEMP_DIR = FILE_TEMP_DIR;
-#else
-  const std::string FILE_URI_PREFIX = "file://";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Posix::current_dir() + "/tiledb_test/";
-  const std::string FILE_FULL_TEMP_DIR = std::string("file://") + FILE_TEMP_DIR;
-#endif
   const std::string GROUP = "group/";
   const std::string ARRAY = "array/";
 
@@ -69,9 +56,8 @@ struct ObjectMgmtFx {
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
 
-  // Supported filesystems
-  bool supports_s3_;
-  bool supports_hdfs_;
+  // Vector of supported filsystems
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
 
   // Functions
   ObjectMgmtFx();
@@ -87,89 +73,22 @@ struct ObjectMgmtFx {
   std::string get_golden_walk(const std::string& path);
   std::string get_golden_ls(const std::string& path);
   static int write_path(const char* path, tiledb_object_t type, void* data);
-  static std::string random_bucket_name(const std::string& prefix);
-  void set_supported_fs();
+  static std::string random_name(const std::string& prefix);
 };
 
-ObjectMgmtFx::ObjectMgmtFx() {
-  // Supported filesystems
-  set_supported_fs();
-
-  // Create TileDB context
-  tiledb_config_t* config = nullptr;
-  tiledb_error_t* error = nullptr;
-  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-
-  if (supports_s3_) {
-#ifndef TILEDB_TESTS_AWS_S3_CONFIG
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.endpoint_override", "localhost:9999", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.scheme", "https", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.use_virtual_addressing", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.verify_ssl", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(error == nullptr);
-#endif
-  }
-
-  REQUIRE(tiledb_ctx_alloc(config, &ctx_) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  vfs_ = nullptr;
-  REQUIRE(tiledb_vfs_alloc(ctx_, config, &vfs_) == TILEDB_OK);
-  tiledb_config_free(&config);
-
-  // Connect to S3
-  if (supports_s3_) {
-    // Create bucket if it does not exist
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    REQUIRE(rc == TILEDB_OK);
-    if (!is_bucket) {
-      rc = tiledb_vfs_create_bucket(ctx_, vfs_, S3_BUCKET.c_str());
-      REQUIRE(rc == TILEDB_OK);
-    }
-  }
+ObjectMgmtFx::ObjectMgmtFx()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 }
 
 ObjectMgmtFx::~ObjectMgmtFx() {
-  if (supports_s3_) {
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    CHECK(rc == TILEDB_OK);
-    if (is_bucket) {
-      CHECK(
-          tiledb_vfs_remove_bucket(ctx_, vfs_, S3_BUCKET.c_str()) == TILEDB_OK);
-    }
-  }
-
+  // Close vfs test
+  REQUIRE(vfs_test_close(fs_vec_, ctx_, vfs_).ok());
   tiledb_vfs_free(&vfs_);
   tiledb_ctx_free(&ctx_);
 }
 
-void ObjectMgmtFx::set_supported_fs() {
-  tiledb_ctx_t* ctx = nullptr;
-  REQUIRE(tiledb_ctx_alloc(nullptr, &ctx) == TILEDB_OK);
-
-  int is_supported = 0;
-  int rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_S3, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_s3_ = (bool)is_supported;
-  rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_HDFS, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_hdfs_ = (bool)is_supported;
-
-  tiledb_ctx_free(&ctx);
-}
-
 void ObjectMgmtFx::create_temp_dir(const std::string& path) {
   remove_temp_dir(path);
   REQUIRE(tiledb_vfs_create_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
@@ -211,7 +130,12 @@ void ObjectMgmtFx::create_array(const std::string& path) {
 
   // Create array
   REQUIRE(tiledb_array_create(ctx_, path.c_str(), array_schema) == TILEDB_OK);
+
+  // Free objects
+  tiledb_attribute_free(&a1);
   tiledb_dimension_free(&d1);
+  tiledb_domain_free(&domain);
+  tiledb_array_schema_free(&array_schema);
 }
 
 void ObjectMgmtFx::check_object_type(const std::string& path) {
@@ -432,7 +356,7 @@ int ObjectMgmtFx::write_path(
   return 1;
 }
 
-std::string ObjectMgmtFx::random_bucket_name(const std::string& prefix) {
+std::string ObjectMgmtFx::random_name(const std::string& prefix) {
   std::stringstream ss;
   ss << prefix << "-" << std::this_thread::get_id() << "-"
      << TILEDB_TIMESTAMP_NOW_MS;
@@ -442,126 +366,105 @@ std::string ObjectMgmtFx::random_bucket_name(const std::string& prefix) {
 TEST_CASE_METHOD(
     ObjectMgmtFx,
     "C API: Test object management methods: object_type, delete, move",
-    "[capi], [object], [delete], [move], [object_type]") {
-  if (supports_s3_) {
-    // S3
-    create_temp_dir(S3_TEMP_DIR);
-    check_object_type(S3_TEMP_DIR);
-    check_delete(S3_TEMP_DIR);
-    check_move(S3_TEMP_DIR);
-    remove_temp_dir(S3_TEMP_DIR);
-  } else if (supports_hdfs_) {
-    // HDFS
-    create_temp_dir(HDFS_TEMP_DIR);
-    check_object_type(HDFS_TEMP_DIR);
-    check_delete(HDFS_TEMP_DIR);
-    check_move(HDFS_TEMP_DIR);
-    remove_temp_dir(HDFS_TEMP_DIR);
-  } else {
-    // File
-    create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
-    check_object_type(FILE_URI_PREFIX + FILE_TEMP_DIR);
-    check_delete(FILE_URI_PREFIX + FILE_TEMP_DIR);
-    check_move(FILE_URI_PREFIX + FILE_TEMP_DIR);
-    remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
-  }
+    "[capi][object][delete][move][object_type]") {
+  // TODO: refactor for each supported FS.
+  std::string temp_dir = fs_vec_[0]->temp_dir();
+  create_temp_dir(temp_dir);
+  check_object_type(temp_dir);
+  check_delete(temp_dir);
+  check_move(temp_dir);
+  remove_temp_dir(temp_dir);
 }
 
 TEST_CASE_METHOD(
     ObjectMgmtFx,
     "C API: Test object management methods: walk, ls",
-    "[capi], [object], [walk], [ls]") {
+    "[capi][object][walk][ls]") {
   std::string golden_walk, golden_ls;
   std::string walk_str, ls_str;
   int rc;
 
-  if (supports_s3_) {
-    // S3
-    remove_temp_dir(S3_TEMP_DIR);
-    create_hierarchy(S3_TEMP_DIR);
-    golden_walk = get_golden_walk(S3_TEMP_DIR);
-    golden_ls = get_golden_ls(S3_TEMP_DIR);
-    walk_str.clear();
-    ls_str.clear();
-    rc = tiledb_object_walk(
-        ctx_, S3_TEMP_DIR.c_str(), TILEDB_PREORDER, write_path, &walk_str);
-    CHECK(rc == TILEDB_OK);
-    rc = tiledb_object_walk(
-        ctx_, S3_TEMP_DIR.c_str(), TILEDB_POSTORDER, write_path, &walk_str);
-    CHECK(rc == TILEDB_OK);
-    CHECK_THAT(golden_walk, Catch::Equals(walk_str));
-    rc = tiledb_object_ls(ctx_, S3_TEMP_DIR.c_str(), write_path, &ls_str);
-    CHECK(rc == TILEDB_OK);
-    CHECK_THAT(golden_ls, Catch::Equals(ls_str));
-    remove_temp_dir(S3_TEMP_DIR);
-  } else if (supports_hdfs_) {
-    // HDFS
-    remove_temp_dir(HDFS_TEMP_DIR);
-    create_hierarchy(HDFS_TEMP_DIR);
-    golden_walk = get_golden_walk(HDFS_FULL_TEMP_DIR);
-    golden_ls = get_golden_ls(HDFS_FULL_TEMP_DIR);
-    walk_str.clear();
-    ls_str.clear();
-    rc = tiledb_object_walk(
-        ctx_, HDFS_TEMP_DIR.c_str(), TILEDB_PREORDER, write_path, &walk_str);
-    CHECK(rc == TILEDB_OK);
-    rc = tiledb_object_walk(
-        ctx_, HDFS_TEMP_DIR.c_str(), TILEDB_POSTORDER, write_path, &walk_str);
-    CHECK(rc == TILEDB_OK);
-    CHECK_THAT(golden_walk, Catch::Equals(walk_str));
-    rc = tiledb_object_ls(ctx_, HDFS_TEMP_DIR.c_str(), write_path, &ls_str);
-    CHECK(rc == TILEDB_OK);
-    CHECK_THAT(golden_ls, Catch::Equals(ls_str));
-    remove_temp_dir(HDFS_TEMP_DIR);
-  } else {
-    // File
-    remove_temp_dir(FILE_FULL_TEMP_DIR);
-    create_hierarchy(FILE_FULL_TEMP_DIR);
+  SupportedFs* const fs = fs_vec_[0].get();
+  if (dynamic_cast<SupportedFsLocal*>(fs) != nullptr) {
+    SupportedFsLocal local_fs;
+    std::string local_dir = local_fs.file_prefix() + local_fs.temp_dir();
+    remove_temp_dir(local_dir);
+    create_hierarchy(local_dir);
 #ifdef _WIN32
     // `VFS::ls(...)` returns `file:///` URIs instead of Windows paths.
-    golden_walk =
-        get_golden_walk(tiledb::sm::Win::uri_from_path(FILE_FULL_TEMP_DIR));
-    golden_ls =
-        get_golden_ls(tiledb::sm::Win::uri_from_path(FILE_FULL_TEMP_DIR));
+    SupportedFsLocal windows_fs;
+    std::string temp_dir = windows_fs.file_prefix() + windows_fs.temp_dir();
+    golden_walk = get_golden_walk(tiledb::sm::Win::uri_from_path(temp_dir));
+    golden_ls = get_golden_ls(tiledb::sm::Win::uri_from_path(temp_dir));
 #else
-    golden_walk = get_golden_walk(FILE_FULL_TEMP_DIR);
-    golden_ls = get_golden_ls(FILE_FULL_TEMP_DIR);
+    SupportedFsLocal posix_fs;
+    std::string temp_dir = posix_fs.file_prefix() + posix_fs.temp_dir();
+    golden_walk = get_golden_walk(temp_dir);
+    golden_ls = get_golden_ls(temp_dir);
 #endif
 
     walk_str.clear();
     ls_str.clear();
     rc = tiledb_object_walk(
-        ctx_,
-        FILE_FULL_TEMP_DIR.c_str(),
-        TILEDB_PREORDER,
-        write_path,
-        &walk_str);
+        ctx_, temp_dir.c_str(), TILEDB_PREORDER, write_path, &walk_str);
     CHECK(rc == TILEDB_OK);
     rc = tiledb_object_walk(
-        ctx_,
-        FILE_FULL_TEMP_DIR.c_str(),
-        TILEDB_POSTORDER,
-        write_path,
-        &walk_str);
+        ctx_, temp_dir.c_str(), TILEDB_POSTORDER, write_path, &walk_str);
     CHECK(rc == TILEDB_OK);
     CHECK_THAT(golden_walk, Catch::Equals(walk_str));
-    rc =
-        tiledb_object_ls(ctx_, FILE_FULL_TEMP_DIR.c_str(), write_path, &ls_str);
+    rc = tiledb_object_ls(ctx_, temp_dir.c_str(), write_path, &ls_str);
     CHECK(rc == TILEDB_OK);
     CHECK_THAT(golden_ls, Catch::Equals(ls_str));
-    remove_temp_dir(FILE_FULL_TEMP_DIR);
+    remove_temp_dir(temp_dir);
+  } else if (dynamic_cast<SupportedFsHDFS*>(fs) != nullptr) {
+    std::string temp_dir = "hdfs://localhost:9000/tiledb_test/";
+    remove_temp_dir(temp_dir);
+    create_hierarchy(temp_dir);
+    golden_walk = get_golden_walk(temp_dir);
+    golden_ls = get_golden_ls(temp_dir);
+    walk_str.clear();
+    ls_str.clear();
+    rc = tiledb_object_walk(
+        ctx_, temp_dir.c_str(), TILEDB_PREORDER, write_path, &walk_str);
+    CHECK(rc == TILEDB_OK);
+    rc = tiledb_object_walk(
+        ctx_, temp_dir.c_str(), TILEDB_POSTORDER, write_path, &walk_str);
+    CHECK(rc == TILEDB_OK);
+    CHECK_THAT(golden_walk, Catch::Equals(walk_str));
+    rc = tiledb_object_ls(ctx_, temp_dir.c_str(), write_path, &ls_str);
+    CHECK(rc == TILEDB_OK);
+    CHECK_THAT(golden_ls, Catch::Equals(ls_str));
+    remove_temp_dir(temp_dir);
+  } else {
+    // TODO: refactor for each supported FS.
+    std::string temp_dir = fs->temp_dir();
+    remove_temp_dir(temp_dir);
+    create_hierarchy(temp_dir);
+    golden_walk = get_golden_walk(temp_dir);
+    golden_ls = get_golden_ls(temp_dir);
+    walk_str.clear();
+    ls_str.clear();
+    rc = tiledb_object_walk(
+        ctx_, temp_dir.c_str(), TILEDB_PREORDER, write_path, &walk_str);
+    CHECK(rc == TILEDB_OK);
+    rc = tiledb_object_walk(
+        ctx_, temp_dir.c_str(), TILEDB_POSTORDER, write_path, &walk_str);
+    CHECK(rc == TILEDB_OK);
+    CHECK_THAT(golden_walk, Catch::Equals(walk_str));
+    rc = tiledb_object_ls(ctx_, temp_dir.c_str(), write_path, &ls_str);
+    CHECK(rc == TILEDB_OK);
+    CHECK_THAT(golden_ls, Catch::Equals(ls_str));
+    remove_temp_dir(temp_dir);
   }
 }
 
 TEST_CASE_METHOD(
     ObjectMgmtFx,
     "C API: Test listing directory with >1000 objects on S3",
-    "[capi], [object], [ls-1000]") {
-  // Applicable only to S3
-  if (!supports_s3_)
-    return;
-
-  create_temp_dir(S3_TEMP_DIR);
-  check_ls_1000(S3_TEMP_DIR);
-  remove_temp_dir(S3_TEMP_DIR);
+    "[capi][object][ls-1000]") {
+  // TODO: refactor for each supported FS.
+  std::string temp_dir = fs_vec_[0]->temp_dir();
+  create_temp_dir(temp_dir);
+  check_ls_1000(temp_dir);
+  remove_temp_dir(temp_dir);
 }
