@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  * @copyright Copyright (c) 2016 MIT and Intel Corporation
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -31,6 +31,7 @@
  * Tests for the C API tiledb_query_t spec.
  */
 
+#include <tiledb/sm/c_api/tiledb_struct_def.h>
 #include <cassert>
 #include <cstring>
 #include <iostream>
@@ -38,6 +39,8 @@
 #include <thread>
 
 #include "catch.hpp"
+#include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #ifdef _WIN32
 #include "tiledb/sm/filesystem/win.h"
 #else
@@ -46,29 +49,15 @@
 #include "tiledb/sm/c_api/tiledb.h"
 #include "tiledb/sm/misc/utils.h"
 
-struct QueryFx {
-  // Filesystem related
-  const std::string HDFS_TEMP_DIR = "hdfs:///tiledb_test/";
-  const std::string S3_PREFIX = "s3://";
-  const std::string S3_BUCKET = S3_PREFIX + random_bucket_name("tiledb") + "/";
-  const std::string S3_TEMP_DIR = S3_BUCKET + "tiledb_test/";
-#ifdef _WIN32
-  const std::string FILE_URI_PREFIX = "";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
-#else
-  const std::string FILE_URI_PREFIX = "file://";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Posix::current_dir() + "/tiledb_test/";
-#endif
+using namespace tiledb::test;
 
+struct QueryFx {
   // TileDB context and vfs
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
 
-  // Supported filesystems
-  bool supports_s3_;
-  bool supports_hdfs_;
+  // Vector of supported filsystems
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
 
   // Functions
   QueryFx();
@@ -77,98 +66,32 @@ struct QueryFx {
   void create_temp_dir(const std::string& path);
   void create_array(const std::string& path);
   void test_get_buffer_write(const std::string& path);
+  void test_get_buffer_write_decoupled(const std::string& path);
   void test_get_buffer_read(const std::string& path);
-  static std::string random_bucket_name(const std::string& prefix);
-  void set_supported_fs();
+  void test_get_buffer_read_decoupled(const std::string& path);
+  static std::string random_name(const std::string& prefix);
 };
 
-QueryFx::QueryFx() {
-  // Supported filesystems
-  set_supported_fs();
-
-  // Create TileDB context
-  tiledb_config_t* config = nullptr;
-  tiledb_error_t* error = nullptr;
-  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-
-  if (supports_s3_) {
-#ifndef TILEDB_TESTS_AWS_S3_CONFIG
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.endpoint_override", "localhost:9999", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.scheme", "https", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.use_virtual_addressing", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.verify_ssl", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(error == nullptr);
-#endif
-  }
-
-  ctx_ = nullptr;
-  REQUIRE(tiledb_ctx_alloc(config, &ctx_) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  vfs_ = nullptr;
-  REQUIRE(tiledb_vfs_alloc(ctx_, config, &vfs_) == TILEDB_OK);
-  tiledb_config_free(&config);
-
-  // Connect to S3
-  if (supports_s3_) {
-    // Create bucket if it does not exist
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    REQUIRE(rc == TILEDB_OK);
-    if (!is_bucket) {
-      rc = tiledb_vfs_create_bucket(ctx_, vfs_, S3_BUCKET.c_str());
-      REQUIRE(rc == TILEDB_OK);
-    }
-  }
+QueryFx::QueryFx()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 }
 
 QueryFx::~QueryFx() {
-  if (supports_s3_) {
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    CHECK(rc == TILEDB_OK);
-    if (is_bucket) {
-      CHECK(
-          tiledb_vfs_remove_bucket(ctx_, vfs_, S3_BUCKET.c_str()) == TILEDB_OK);
-    }
-  }
-
+  // Close vfs test
+  REQUIRE(vfs_test_close(fs_vec_, ctx_, vfs_).ok());
   tiledb_vfs_free(&vfs_);
   tiledb_ctx_free(&ctx_);
 }
 
-std::string QueryFx::random_bucket_name(const std::string& prefix) {
+std::string QueryFx::random_name(const std::string& prefix) {
   std::stringstream ss;
   ss << prefix << "-" << std::this_thread::get_id() << "-"
      << TILEDB_TIMESTAMP_NOW_MS;
   return ss.str();
 }
 
-void QueryFx::set_supported_fs() {
-  tiledb_ctx_t* ctx = nullptr;
-  REQUIRE(tiledb_ctx_alloc(nullptr, &ctx) == TILEDB_OK);
-
-  int is_supported = 0;
-  int rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_S3, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_s3_ = (bool)is_supported;
-  rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_HDFS, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_hdfs_ = (bool)is_supported;
-
-  tiledb_ctx_free(&ctx);
-}
-
 void QueryFx::create_temp_dir(const std::string& path) {
   remove_temp_dir(path);
   REQUIRE(tiledb_vfs_create_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
@@ -280,16 +203,13 @@ void QueryFx::test_get_buffer_write(const std::string& path) {
   uint64_t* a2_off_got_size;
   void* a2_val_got;
   uint64_t* a2_val_got_size;
-  rc = tiledb_query_get_buffer(ctx_, query, "", &a1_got, &a1_got_size);
+  rc = tiledb_query_get_data_buffer(ctx_, query, "", &a1_got, &a1_got_size);
   CHECK(rc == TILEDB_OK);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      "a2",
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a2", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a2", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_OK);
   CHECK(a1_got == nullptr);
   CHECK(a1_got_size == nullptr);
@@ -299,65 +219,148 @@ void QueryFx::test_get_buffer_write(const std::string& path) {
   CHECK(a2_val_got_size == nullptr);
 
   // Set buffers
-  rc = tiledb_query_set_buffer(ctx_, query, "", a1, &a1_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "", a1, &a1_size);
   CHECK(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer_var(
-      ctx_, query, "a2", a2_off, &a2_off_size, a2_val, &a2_val_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a2", a2_val, &a2_val_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_set_offsets_buffer(ctx_, query, "a2", a2_off, &a2_off_size);
   CHECK(rc == TILEDB_OK);
 
-  // Get invalid fixed-sized / var-sized buffer
-  rc = tiledb_query_get_buffer(ctx_, query, "a2", &a1_got, &a1_got_size);
+  // Get invalid var-sized buffer
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a1", &a2_val_got, &a2_val_got_size);
   CHECK(rc == TILEDB_ERR);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      "a1",
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a1", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_ERR);
 
   // Test getting the coords buffer
-  rc = tiledb_query_get_buffer(
-      ctx_, query, TILEDB_COORDS, &a1_got, &a1_got_size);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "dim_1", &a2_val_got, &a2_val_got_size);
   CHECK(rc == TILEDB_OK);
   CHECK(a1_got == nullptr);
   CHECK(a1_got_size == nullptr);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      TILEDB_COORDS,
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "dim_1", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_ERR);
 
   // Test invalid attribute
-  rc = tiledb_query_get_buffer(ctx_, query, "foo", &a1_got, &a1_got_size);
+  rc = tiledb_query_get_data_buffer(ctx_, query, "foo", &a1_got, &a1_got_size);
   CHECK(rc == TILEDB_ERR);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      "foo-var",
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "foo-var", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_ERR);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "foo-var", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_ERR);
 
   // Test correctly got buffers
-  rc = tiledb_query_get_buffer(ctx_, query, "", &a1_got, &a1_got_size);
+  rc = tiledb_query_get_data_buffer(ctx_, query, "", &a1_got, &a1_got_size);
   CHECK(rc == TILEDB_OK);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      "a2",
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a2", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a2", &a2_off_got, &a2_off_got_size);
+  CHECK(rc == TILEDB_OK);
+  CHECK(a1_got == a1);
+  CHECK(a1_got_size == &a1_size);
+  CHECK(a2_off_got == a2_off);
+  CHECK(a2_off_got_size == &a2_off_size);
+  CHECK(a2_val_got == a2_val);
+  CHECK(a2_val_got_size == &a2_val_size);
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  CHECK(rc == TILEDB_OK);
+
+  // Clean up
+  tiledb_array_free(&array);
+  tiledb_query_free(&query);
+}
+
+void QueryFx::test_get_buffer_write_decoupled(const std::string& path) {
+  // Open array
+  tiledb_array_t* array;
+  int rc = tiledb_array_alloc(ctx_, path.c_str(), &array);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_WRITE);
+  CHECK(rc == TILEDB_OK);
+
+  // Prepare subarray and buffers
+  uint64_t subarray[] = {1, 2, 1, 2};
+  int a1[] = {1, 2, 3, 4};
+  uint64_t a1_size = sizeof(a1);
+  uint64_t a2_off[] = {0, 4, 8, 12};
+  uint64_t a2_off_size = sizeof(a2_off);
+  int a2_val[] = {1, 2, 3, 4};
+  uint64_t a2_val_size = sizeof(a2_val);
+
+  // Prepare query
+  tiledb_query_t* query;
+  rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_subarray(ctx_, query, subarray);
+  CHECK(rc == TILEDB_OK);
+
+  // Get unset buffers
+  void* a1_got;
+  uint64_t* a1_got_size;
+  uint64_t* a2_off_got;
+  uint64_t* a2_off_got_size;
+  void* a2_val_got;
+  uint64_t* a2_val_got_size;
+  rc = tiledb_query_get_data_buffer(ctx_, query, "", &a1_got, &a1_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a2", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a2", &a2_off_got, &a2_off_got_size);
+  CHECK(rc == TILEDB_OK);
+  CHECK(a1_got == nullptr);
+  CHECK(a1_got_size == nullptr);
+  CHECK(a2_off_got == nullptr);
+  CHECK(a2_off_got_size == nullptr);
+  CHECK(a2_val_got == nullptr);
+  CHECK(a2_val_got_size == nullptr);
+
+  // Set buffers
+  rc = tiledb_query_set_data_buffer(ctx_, query, "", a1, &a1_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a2", a2_val, &a2_val_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_set_offsets_buffer(ctx_, query, "a2", a2_off, &a2_off_size);
+  CHECK(rc == TILEDB_OK);
+
+  // Test getting the coords buffer
+  rc =
+      tiledb_query_get_data_buffer(ctx_, query, "dim_1", &a1_got, &a1_got_size);
+  CHECK(rc == TILEDB_OK);
+  CHECK(a1_got == nullptr);
+  CHECK(a1_got_size == nullptr);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "dim_1", &a2_off_got, &a2_off_got_size);
+  CHECK(rc == TILEDB_ERR);
+
+  // Test invalid attribute
+  rc = tiledb_query_get_data_buffer(ctx_, query, "foo", &a1_got, &a1_got_size);
+  CHECK(rc == TILEDB_ERR);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "foo-var", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_ERR);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "foo-var", &a2_off_got, &a2_off_got_size);
+  CHECK(rc == TILEDB_ERR);
+
+  // Test correctly got buffers
+  rc = tiledb_query_get_data_buffer(ctx_, query, "", &a1_got, &a1_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a2", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a2", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_OK);
   CHECK(a1_got == a1);
   CHECK(a1_got_size == &a1_size);
@@ -406,16 +409,13 @@ void QueryFx::test_get_buffer_read(const std::string& path) {
   uint64_t* a2_off_got_size;
   void* a2_val_got;
   uint64_t* a2_val_got_size;
-  rc = tiledb_query_get_buffer(ctx_, query, "", &a1_got, &a1_got_size);
+  rc = tiledb_query_get_data_buffer(ctx_, query, "", &a1_got, &a1_got_size);
   CHECK(rc == TILEDB_OK);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      "a2",
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a2", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a2", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_OK);
   CHECK(a1_got == nullptr);
   CHECK(a1_got_size == nullptr);
@@ -425,65 +425,157 @@ void QueryFx::test_get_buffer_read(const std::string& path) {
   CHECK(a2_val_got_size == nullptr);
 
   // Set buffers
-  rc = tiledb_query_set_buffer(ctx_, query, "", a1, &a1_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "", a1, &a1_size);
   CHECK(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer_var(
-      ctx_, query, "a2", a2_off, &a2_off_size, a2_val, &a2_val_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a2", a2_val, &a2_val_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_set_offsets_buffer(ctx_, query, "a2", a2_off, &a2_off_size);
   CHECK(rc == TILEDB_OK);
 
   // Get invalid fixed-sized / var-sized buffer
-  rc = tiledb_query_get_buffer(ctx_, query, "a2", &a1_got, &a1_got_size);
+  rc = tiledb_query_get_data_buffer(ctx_, query, "a2", &a1_got, &a1_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a1", &a2_val_got, &a2_val_got_size);
   CHECK(rc == TILEDB_ERR);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      "a1",
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a1", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_ERR);
 
   // Test getting the coords buffer
-  rc = tiledb_query_get_buffer(
-      ctx_, query, TILEDB_COORDS, &a1_got, &a1_got_size);
+  rc =
+      tiledb_query_get_data_buffer(ctx_, query, "dim_1", &a1_got, &a1_got_size);
   CHECK(rc == TILEDB_OK);
   CHECK(a1_got == nullptr);
   CHECK(a1_got_size == nullptr);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      TILEDB_COORDS,
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "dim_1", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "dim_1", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_ERR);
 
   // Test invalid attribute
-  rc = tiledb_query_get_buffer(ctx_, query, "foo", &a1_got, &a1_got_size);
+  rc = tiledb_query_get_data_buffer(ctx_, query, "foo", &a1_got, &a1_got_size);
   CHECK(rc == TILEDB_ERR);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      "foo-var",
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "foo-var", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_ERR);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "foo-var", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_ERR);
 
   // Test correctly got buffers
-  rc = tiledb_query_get_buffer(ctx_, query, "", &a1_got, &a1_got_size);
+  rc = tiledb_query_get_data_buffer(ctx_, query, "", &a1_got, &a1_got_size);
   CHECK(rc == TILEDB_OK);
-  rc = tiledb_query_get_buffer_var(
-      ctx_,
-      query,
-      "a2",
-      &a2_off_got,
-      &a2_off_got_size,
-      &a2_val_got,
-      &a2_val_got_size);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a2", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a2", &a2_off_got, &a2_off_got_size);
+  CHECK(rc == TILEDB_OK);
+  CHECK(a1_got == a1);
+  CHECK(a1_got_size == &a1_size);
+  CHECK(a2_off_got == a2_off);
+  CHECK(a2_off_got_size == &a2_off_size);
+  CHECK(a2_val_got == a2_val);
+  CHECK(a2_val_got_size == &a2_val_size);
+
+  // Close array
+  rc = tiledb_array_close(ctx_, array);
+  CHECK(rc == TILEDB_OK);
+
+  // Clean up
+  tiledb_array_free(&array);
+  tiledb_query_free(&query);
+}
+
+void QueryFx::test_get_buffer_read_decoupled(const std::string& path) {
+  // Open array
+  tiledb_array_t* array;
+  int rc = tiledb_array_alloc(ctx_, path.c_str(), &array);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
+  CHECK(rc == TILEDB_OK);
+
+  // Prepare subarray and buffers
+  uint64_t subarray[] = {1, 2, 1, 2};
+  int a1[4];
+  uint64_t a1_size = sizeof(a1);
+  uint64_t a2_off[4];
+  uint64_t a2_off_size = sizeof(a2_off);
+  int a2_val[4];
+  uint64_t a2_val_size = sizeof(a2_val);
+
+  // Prepare query
+  tiledb_query_t* query;
+  rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_subarray(ctx_, query, subarray);
+  CHECK(rc == TILEDB_OK);
+
+  // Get unset buffers
+  void* a1_got;
+  uint64_t* a1_got_size;
+  uint64_t* a2_off_got;
+  uint64_t* a2_off_got_size;
+  void* a2_val_got;
+  uint64_t* a2_val_got_size;
+  rc = tiledb_query_get_data_buffer(ctx_, query, "", &a1_got, &a1_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a2", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a2", &a2_off_got, &a2_off_got_size);
+  CHECK(rc == TILEDB_OK);
+  CHECK(a1_got == nullptr);
+  CHECK(a1_got_size == nullptr);
+  CHECK(a2_off_got == nullptr);
+  CHECK(a2_off_got_size == nullptr);
+  CHECK(a2_val_got == nullptr);
+  CHECK(a2_val_got_size == nullptr);
+
+  // Set buffers
+  rc = tiledb_query_set_data_buffer(ctx_, query, "", a1, &a1_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a2", a2_val, &a2_val_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_set_offsets_buffer(ctx_, query, "a2", a2_off, &a2_off_size);
+  CHECK(rc == TILEDB_OK);
+
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a1", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_ERR);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a1", &a2_off_got, &a2_off_got_size);
+  CHECK(rc == TILEDB_ERR);
+
+  // Test getting the coords buffer
+  rc =
+      tiledb_query_get_data_buffer(ctx_, query, "dim_1", &a1_got, &a1_got_size);
+  CHECK(rc == TILEDB_OK);
+  CHECK(a1_got == nullptr);
+  CHECK(a1_got_size == nullptr);
+
+  // Test invalid attribute
+  rc = tiledb_query_get_data_buffer(ctx_, query, "foo", &a1_got, &a1_got_size);
+  CHECK(rc == TILEDB_ERR);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "foo-var", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_ERR);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "foo-var", &a2_off_got, &a2_off_got_size);
+  CHECK(rc == TILEDB_ERR);
+
+  // Test correctly got buffers
+  rc = tiledb_query_get_data_buffer(ctx_, query, "", &a1_got, &a1_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_data_buffer(
+      ctx_, query, "a2", &a2_val_got, &a2_val_got_size);
+  CHECK(rc == TILEDB_OK);
+  rc = tiledb_query_get_offsets_buffer(
+      ctx_, query, "a2", &a2_off_got, &a2_off_got_size);
   CHECK(rc == TILEDB_OK);
   CHECK(a1_got == a1);
   CHECK(a1_got_size == &a1_size);
@@ -504,33 +596,27 @@ void QueryFx::test_get_buffer_read(const std::string& path) {
 TEST_CASE_METHOD(
     QueryFx,
     "C API: Test query get buffer",
-    "[capi], [query], [query-get-buffer]") {
-  std::string temp_dir;
-
-  // S3
-  if (supports_s3_) {
-    temp_dir = S3_TEMP_DIR;
-  } else if (supports_hdfs_) {
-    // HDFS
-    temp_dir = HDFS_TEMP_DIR;
-  } else {
-    // File
-    temp_dir = FILE_URI_PREFIX + FILE_TEMP_DIR;
-  }
+    "[capi][query][query-get-buffer]") {
+  // TODO: refactor for each supported FS.
+  SupportedFsLocal local_fs;
+  std::string temp_dir = local_fs.file_prefix() + local_fs.temp_dir();
 
   std::string array_name = temp_dir + "query_get_buffer";
   create_temp_dir(temp_dir);
   create_array(array_name);
   test_get_buffer_write(array_name);
+  test_get_buffer_write_decoupled(array_name);
   test_get_buffer_read(array_name);
+  test_get_buffer_read_decoupled(array_name);
   remove_temp_dir(temp_dir);
 }
 
 TEST_CASE_METHOD(
     QueryFx,
     "C API: Test query get layout",
-    "[capi], [query], [query-get-layout]") {
-  std::string temp_dir = FILE_URI_PREFIX + FILE_TEMP_DIR;
+    "[capi][query][query-get-layout]") {
+  SupportedFsLocal local_fs;
+  std::string temp_dir = local_fs.file_prefix() + local_fs.temp_dir();
   std::string array_name = temp_dir + "query_get_layout";
   create_temp_dir(temp_dir);
   create_array(array_name);
@@ -558,11 +644,6 @@ TEST_CASE_METHOD(
   rc = tiledb_query_get_layout(ctx_, query, &layout);
   REQUIRE(rc == TILEDB_OK);
   REQUIRE(layout == TILEDB_GLOBAL_ORDER);
-  rc = tiledb_query_set_layout(ctx_, query, TILEDB_UNORDERED);
-  REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_get_layout(ctx_, query, &layout);
-  REQUIRE(rc == TILEDB_OK);
-  REQUIRE(layout == TILEDB_UNORDERED);
 
   rc = tiledb_array_close(ctx_, array);
   REQUIRE(rc == TILEDB_OK);
@@ -571,3 +652,55 @@ TEST_CASE_METHOD(
   tiledb_array_free(&array);
   remove_temp_dir(temp_dir);
 }
+
+TEST_CASE_METHOD(
+    QueryFx, "C API: Test query get array", "[capi][query][query-get-array]") {
+  SupportedFsLocal local_fs;
+  std::string temp_dir = local_fs.file_prefix() + local_fs.temp_dir();
+  std::string array_name = temp_dir + "query_get_array";
+  create_temp_dir(temp_dir);
+  create_array(array_name);
+
+  tiledb_array_t* array;
+  int rc = tiledb_array_alloc(ctx_, array_name.c_str(), &array);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_array_open(ctx_, array, TILEDB_READ);
+  REQUIRE(rc == TILEDB_OK);
+
+  tiledb_query_t* query;
+  rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
+
+  REQUIRE(rc == TILEDB_OK);
+  tiledb_array_t* rarray;
+  rc = tiledb_query_get_array(ctx_, query, &rarray);
+  REQUIRE(rc == TILEDB_OK);
+  CHECK(rarray->array_ != array->array_);
+
+  tiledb_array_schema_t* rschema;
+  rc = tiledb_array_get_schema(ctx_, rarray, &rschema);
+  REQUIRE(rc == TILEDB_OK);
+
+  // Get schema members
+  uint64_t rcapacity;
+  rc = tiledb_array_schema_get_capacity(ctx_, rschema, &rcapacity);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(rcapacity == 10000);
+
+  tiledb_layout_t layout;
+  rc = tiledb_array_schema_get_cell_order(ctx_, rschema, &layout);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(layout == TILEDB_ROW_MAJOR);
+
+  rc = tiledb_array_schema_get_tile_order(ctx_, rschema, &layout);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(layout == TILEDB_ROW_MAJOR);
+
+  rc = tiledb_array_close(ctx_, array);
+  REQUIRE(rc == TILEDB_OK);
+
+  tiledb_array_schema_free(&rschema);
+  tiledb_query_free(&query);
+  tiledb_array_free(&array);
+  tiledb_array_free(&rarray);
+  remove_temp_dir(temp_dir);
+}
