@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2018 TileDB Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,6 +31,8 @@
  */
 
 #include "catch.hpp"
+#include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #ifdef _WIN32
 #include "tiledb/sm/filesystem/win.h"
 #else
@@ -43,28 +45,15 @@
 #include <sstream>
 #include <thread>
 
-struct SparseNegFx {
-  const std::string HDFS_TEMP_DIR = "hdfs:///tiledb_test/";
-  const std::string S3_PREFIX = "s3://";
-  const std::string S3_BUCKET = S3_PREFIX + random_bucket_name("tiledb") + "/";
-  const std::string S3_TEMP_DIR = S3_BUCKET + "tiledb_test/";
-#ifdef _WIN32
-  const std::string FILE_URI_PREFIX = "";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
-#else
-  const std::string FILE_URI_PREFIX = "file://";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Posix::current_dir() + "/tiledb_test/";
-#endif
+using namespace tiledb::test;
 
+struct SparseNegFx {
   // TileDB context
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
 
-  // Supported filesystems
-  bool supports_s3_;
-  bool supports_hdfs_;
+  // Vector of supported filsystems
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
 
   // Functions
   SparseNegFx();
@@ -79,87 +68,22 @@ struct SparseNegFx {
   void read_sparse_array_global(const std::string& path);
   void read_sparse_array_row(const std::string& path);
   void read_sparse_array_col(const std::string& path);
-  static std::string random_bucket_name(const std::string& prefix);
-  void set_supported_fs();
+  static std::string random_name(const std::string& prefix);
 };
 
-SparseNegFx::SparseNegFx() {
-  // Supported filesystems
-  set_supported_fs();
-
-  // Create TileDB context
-  tiledb_config_t* config = nullptr;
-  tiledb_error_t* error = nullptr;
-  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  if (supports_s3_) {
-#ifndef TILEDB_TESTS_AWS_S3_CONFIG
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.endpoint_override", "localhost:9999", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.scheme", "https", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.use_virtual_addressing", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.verify_ssl", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(error == nullptr);
-#endif
-  }
-  REQUIRE(tiledb_ctx_alloc(config, &ctx_) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  vfs_ = nullptr;
-  REQUIRE(tiledb_vfs_alloc(ctx_, config, &vfs_) == TILEDB_OK);
-  tiledb_config_free(&config);
-
-  // Connect to S3
-  if (supports_s3_) {
-    // Create bucket if it does not exist
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    REQUIRE(rc == TILEDB_OK);
-    if (!is_bucket) {
-      rc = tiledb_vfs_create_bucket(ctx_, vfs_, S3_BUCKET.c_str());
-      REQUIRE(rc == TILEDB_OK);
-    }
-  }
+SparseNegFx::SparseNegFx()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 }
 
 SparseNegFx::~SparseNegFx() {
-  if (supports_s3_) {
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    CHECK(rc == TILEDB_OK);
-    if (is_bucket) {
-      CHECK(
-          tiledb_vfs_remove_bucket(ctx_, vfs_, S3_BUCKET.c_str()) == TILEDB_OK);
-    }
-  }
-
+  // Close vfs test
+  REQUIRE(vfs_test_close(fs_vec_, ctx_, vfs_).ok());
   tiledb_vfs_free(&vfs_);
   tiledb_ctx_free(&ctx_);
 }
 
-void SparseNegFx::set_supported_fs() {
-  tiledb_ctx_t* ctx = nullptr;
-  REQUIRE(tiledb_ctx_alloc(nullptr, &ctx) == TILEDB_OK);
-
-  int is_supported = 0;
-  int rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_S3, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_s3_ = (bool)is_supported;
-  rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_HDFS, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_hdfs_ = (bool)is_supported;
-
-  tiledb_ctx_free(&ctx);
-}
-
 void SparseNegFx::create_temp_dir(const std::string& path) {
   remove_temp_dir(path);
   REQUIRE(tiledb_vfs_create_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
@@ -172,7 +96,7 @@ void SparseNegFx::remove_temp_dir(const std::string& path) {
     REQUIRE(tiledb_vfs_remove_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
 }
 
-std::string SparseNegFx::random_bucket_name(const std::string& prefix) {
+std::string SparseNegFx::random_name(const std::string& prefix) {
   std::stringstream ss;
   ss << prefix << "-" << std::this_thread::get_id() << "-"
      << TILEDB_TIMESTAMP_NOW_MS;
@@ -221,6 +145,7 @@ void SparseNegFx::create_sparse_vector(const std::string& path) {
   REQUIRE(rc == TILEDB_OK);
   tiledb_attribute_free(&attr);
   tiledb_dimension_free(&dim);
+  tiledb_domain_free(&domain);
   tiledb_array_schema_free(&array_schema);
 }
 
@@ -283,6 +208,8 @@ void SparseNegFx::create_sparse_array(const std::string& path) {
   CHECK(rc == TILEDB_OK);
 
   // Clean up
+  tiledb_filter_free(&filter);
+  tiledb_filter_list_free(&list);
   tiledb_attribute_free(&a);
   tiledb_dimension_free(&d1);
   tiledb_dimension_free(&d2);
@@ -305,10 +232,9 @@ void SparseNegFx::write_sparse_vector(const std::string& path) {
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "d0", coords, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_GLOBAL_ORDER);
   REQUIRE(rc == TILEDB_OK);
@@ -336,15 +262,19 @@ void SparseNegFx::write_sparse_array(const std::string& path) {
 
   int a[] = {1, 2, 3, 4};
   uint64_t a_size = sizeof(a);
-  int64_t coords[] = {-2, 0, 1, 1, -1, -1, 1, -1};
-  uint64_t coords_size = sizeof(coords);
+  int64_t coords_dim1[] = {-2, 1, -1, 1};
+  int64_t coords_dim2[] = {0, 1, -1, -1};
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_UNORDERED);
   REQUIRE(rc == TILEDB_OK);
@@ -380,11 +310,11 @@ void SparseNegFx::read_sparse_vector(const std::string& path) {
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_subarray(ctx_, query, subarray);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "d0", coords, &coords_size);
   REQUIRE(rc == TILEDB_OK);
+  ;
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_ROW_MAJOR);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_submit(ctx_, query);
@@ -418,15 +348,19 @@ void SparseNegFx::read_sparse_array_global(const std::string& path) {
 
   int a[4];
   uint64_t a_size = sizeof(a);
-  int64_t coords[8];
-  uint64_t coords_size = sizeof(coords);
+  int64_t coords_dim1[4];
+  int64_t coords_dim2[4];
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_GLOBAL_ORDER);
   REQUIRE(rc == TILEDB_OK);
@@ -436,11 +370,13 @@ void SparseNegFx::read_sparse_array_global(const std::string& path) {
   REQUIRE(rc == TILEDB_OK);
 
   int a_c[] = {3, 1, 4, 2};
-  int64_t coords_c[] = {-1, -1, -2, 0, 1, -1, 1, 1};
+  int64_t coords_c_dim1[] = {-1, -2, 1, 1};
+  int64_t coords_c_dim2[] = {-1, 0, -1, 1};
   CHECK(a_size == sizeof(a_c));
-  CHECK(coords_size == sizeof(coords_c));
+  CHECK(coords_size == sizeof(coords_c_dim1));
   CHECK(!memcmp(a, a_c, sizeof(a_c)));
-  CHECK(!memcmp(coords, coords_c, sizeof(coords_c)));
+  CHECK(!memcmp(coords_dim1, coords_c_dim1, sizeof(coords_c_dim1)));
+  CHECK(!memcmp(coords_dim2, coords_c_dim2, sizeof(coords_c_dim2)));
 
   // Close array
   rc = tiledb_array_close(ctx_, array);
@@ -461,15 +397,19 @@ void SparseNegFx::read_sparse_array_row(const std::string& path) {
 
   int a[4];
   uint64_t a_size = sizeof(a);
-  int64_t coords[8];
-  uint64_t coords_size = sizeof(coords);
+  int64_t coords_dim1[4];
+  int64_t coords_dim2[4];
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_ROW_MAJOR);
   REQUIRE(rc == TILEDB_OK);
@@ -479,11 +419,13 @@ void SparseNegFx::read_sparse_array_row(const std::string& path) {
   REQUIRE(rc == TILEDB_OK);
 
   int a_c[] = {1, 3, 4, 2};
-  int64_t coords_c[] = {-2, 0, -1, -1, 1, -1, 1, 1};
+  int64_t coords_c_dim1[] = {-2, -1, 1, 1};
+  int64_t coords_c_dim2[] = {0, -1, -1, 1};
   CHECK(a_size == sizeof(a_c));
-  CHECK(coords_size == sizeof(coords_c));
+  CHECK(coords_size == sizeof(coords_c_dim1));
   CHECK(!memcmp(a, a_c, sizeof(a_c)));
-  CHECK(!memcmp(coords, coords_c, sizeof(coords_c)));
+  CHECK(!memcmp(coords_dim1, coords_c_dim1, sizeof(coords_c_dim1)));
+  CHECK(!memcmp(coords_dim2, coords_c_dim2, sizeof(coords_c_dim2)));
 
   // Close array
   rc = tiledb_array_close(ctx_, array);
@@ -504,15 +446,19 @@ void SparseNegFx::read_sparse_array_col(const std::string& path) {
 
   int a[4];
   uint64_t a_size = sizeof(a);
-  int64_t coords[8];
-  uint64_t coords_size = sizeof(coords);
+  int64_t coords_dim1[4];
+  int64_t coords_dim2[4];
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_COL_MAJOR);
   REQUIRE(rc == TILEDB_OK);
@@ -522,11 +468,13 @@ void SparseNegFx::read_sparse_array_col(const std::string& path) {
   REQUIRE(rc == TILEDB_OK);
 
   int a_c[] = {3, 4, 1, 2};
-  int64_t coords_c[] = {-1, -1, 1, -1, -2, 0, 1, 1};
+  int64_t coords_c_dim1[] = {-1, 1, -2, 1};
+  int64_t coords_c_dim2[] = {-1, -1, 0, 1};
   CHECK(a_size == sizeof(a_c));
-  CHECK(coords_size == sizeof(coords_c));
+  CHECK(coords_size == sizeof(coords_c_dim1));
   CHECK(!memcmp(a, a_c, sizeof(a_c)));
-  CHECK(!memcmp(coords, coords_c, sizeof(coords_c)));
+  CHECK(!memcmp(coords_dim1, coords_c_dim1, sizeof(coords_c_dim1)));
+  CHECK(!memcmp(coords_dim2, coords_c_dim2, sizeof(coords_c_dim2)));
 
   // Close array
   rc = tiledb_array_close(ctx_, array);
@@ -540,25 +488,27 @@ void SparseNegFx::read_sparse_array_col(const std::string& path) {
 TEST_CASE_METHOD(
     SparseNegFx,
     "C API: Test 1d sparse vector with negative domain",
-    "[capi], [sparse-neg], [sparse-neg-vector]") {
+    "[capi][sparse-neg][sparse-neg-vector]") {
+  SupportedFsLocal local_fs;
   std::string vector_name =
-      FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_neg_vector";
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+      local_fs.file_prefix() + local_fs.temp_dir() + "sparse_neg_vector";
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 
   create_sparse_vector(vector_name);
   write_sparse_vector(vector_name);
   read_sparse_vector(vector_name);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
 
 TEST_CASE_METHOD(
     SparseNegFx,
     "C API: Test 2d sparse array with negative domain",
-    "[capi], [sparse-neg], [sparse-neg-array]") {
+    "[capi][sparse-neg][sparse-neg-array]") {
+  SupportedFsLocal local_fs;
   std::string vector_name =
-      FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_neg_array";
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+      local_fs.file_prefix() + local_fs.temp_dir() + "sparse_neg_array";
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 
   create_sparse_array(vector_name);
   write_sparse_array(vector_name);
@@ -566,5 +516,5 @@ TEST_CASE_METHOD(
   read_sparse_array_row(vector_name);
   read_sparse_array_col(vector_name);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
