@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2018 TileDB Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,6 +31,8 @@
  */
 
 #include "catch.hpp"
+#include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #ifdef _WIN32
 #include "tiledb/sm/filesystem/win.h"
 #else
@@ -43,28 +45,15 @@
 #include <sstream>
 #include <thread>
 
-struct SparseRealFx {
-  const std::string HDFS_TEMP_DIR = "hdfs:///tiledb_test/";
-  const std::string S3_PREFIX = "s3://";
-  const std::string S3_BUCKET = S3_PREFIX + random_bucket_name("tiledb") + "/";
-  const std::string S3_TEMP_DIR = S3_BUCKET + "tiledb_test/";
-#ifdef _WIN32
-  const std::string FILE_URI_PREFIX = "";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
-#else
-  const std::string FILE_URI_PREFIX = "file://";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Posix::current_dir() + "/tiledb_test/";
-#endif
+using namespace tiledb::test;
 
+struct SparseRealFx {
   // TileDB context
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
 
-  // Supported filesystems
-  bool supports_s3_;
-  bool supports_hdfs_;
+  // Vector of supported filsystems
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
 
   // Functions
   SparseRealFx();
@@ -77,87 +66,22 @@ struct SparseRealFx {
   void write_sparse_array_next_partition_bug(const std::string& path);
   void read_sparse_array(const std::string& path);
   void read_sparse_array_next_partition_bug(const std::string& path);
-  static std::string random_bucket_name(const std::string& prefix);
-  void set_supported_fs();
+  static std::string random_name(const std::string& prefix);
 };
 
-SparseRealFx::SparseRealFx() {
-  // Supported filesystems
-  set_supported_fs();
-
-  // Create TileDB context
-  tiledb_config_t* config = nullptr;
-  tiledb_error_t* error = nullptr;
-  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  if (supports_s3_) {
-#ifndef TILEDB_TESTS_AWS_S3_CONFIG
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.endpoint_override", "localhost:9999", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.scheme", "https", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.use_virtual_addressing", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.verify_ssl", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(error == nullptr);
-#endif
-  }
-  REQUIRE(tiledb_ctx_alloc(config, &ctx_) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  vfs_ = nullptr;
-  REQUIRE(tiledb_vfs_alloc(ctx_, config, &vfs_) == TILEDB_OK);
-  tiledb_config_free(&config);
-
-  // Connect to S3
-  if (supports_s3_) {
-    // Create bucket if it does not exist
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    REQUIRE(rc == TILEDB_OK);
-    if (!is_bucket) {
-      rc = tiledb_vfs_create_bucket(ctx_, vfs_, S3_BUCKET.c_str());
-      REQUIRE(rc == TILEDB_OK);
-    }
-  }
+SparseRealFx::SparseRealFx()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 }
 
 SparseRealFx::~SparseRealFx() {
-  if (supports_s3_) {
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    CHECK(rc == TILEDB_OK);
-    if (is_bucket) {
-      CHECK(
-          tiledb_vfs_remove_bucket(ctx_, vfs_, S3_BUCKET.c_str()) == TILEDB_OK);
-    }
-  }
-
+  // Close vfs test
+  REQUIRE(vfs_test_close(fs_vec_, ctx_, vfs_).ok());
   tiledb_vfs_free(&vfs_);
   tiledb_ctx_free(&ctx_);
 }
 
-void SparseRealFx::set_supported_fs() {
-  tiledb_ctx_t* ctx = nullptr;
-  REQUIRE(tiledb_ctx_alloc(nullptr, &ctx) == TILEDB_OK);
-
-  int is_supported = 0;
-  int rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_S3, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_s3_ = (bool)is_supported;
-  rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_HDFS, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_hdfs_ = (bool)is_supported;
-
-  tiledb_ctx_free(&ctx);
-}
-
 void SparseRealFx::create_temp_dir(const std::string& path) {
   remove_temp_dir(path);
   REQUIRE(tiledb_vfs_create_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
@@ -170,7 +94,7 @@ void SparseRealFx::remove_temp_dir(const std::string& path) {
     REQUIRE(tiledb_vfs_remove_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
 }
 
-std::string SparseRealFx::random_bucket_name(const std::string& prefix) {
+std::string SparseRealFx::random_name(const std::string& prefix) {
   std::stringstream ss;
   ss << prefix << "-" << std::this_thread::get_id() << "-"
      << TILEDB_TIMESTAMP_NOW_MS;
@@ -236,6 +160,8 @@ void SparseRealFx::create_sparse_array(const std::string& path) {
   CHECK(rc == TILEDB_OK);
 
   // Clean up
+  tiledb_filter_free(&filter);
+  tiledb_filter_list_free(&list);
   tiledb_attribute_free(&a);
   tiledb_dimension_free(&d1);
   tiledb_dimension_free(&d2);
@@ -253,16 +179,19 @@ void SparseRealFx::write_sparse_array(const std::string& path) {
 
   int a[] = {1, 2, 3, 4, 5};
   uint64_t a_size = sizeof(a);
-  float coords[] = {
-      -23.5f, -20.0f, 43.56f, 80.0f, 66.2f, -0.3f, -160.1f, 89.1f, 1.0f, 1.0f};
-  uint64_t coords_size = sizeof(coords);
+  float coords_dim1[] = {-23.5f, 43.56f, 66.2f, -160.1f, 1.0f};
+  float coords_dim2[] = {-20.0f, 80.0f, -0.3f, 89.1f, 1.0f};
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_UNORDERED);
   REQUIRE(rc == TILEDB_OK);
@@ -291,15 +220,19 @@ void SparseRealFx::write_sparse_array_next_partition_bug(
 
   int a[] = {1, 2};
   uint64_t a_size = sizeof(a);
-  float coords[] = {-180.0f, 1.0f, -180.0f, 2.0f};
-  uint64_t coords_size = sizeof(coords);
+  float coords_dim1[] = {-180.0f, -180.0f};
+  float coords_dim2[] = {1.0f, 2.0f};
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_UNORDERED);
   REQUIRE(rc == TILEDB_OK);
@@ -327,18 +260,22 @@ void SparseRealFx::read_sparse_array(const std::string& path) {
 
   int a[16];
   uint64_t a_size = sizeof(a);
-  float coords[32];
-  uint64_t coords_size = sizeof(coords);
+  float coords_dim1[16];
+  float coords_dim2[16];
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   float subarray[] = {-180.0f, 180.0f, -90.0f, 90.0f};
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_subarray(ctx_, query, subarray);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_ROW_MAJOR);
   REQUIRE(rc == TILEDB_OK);
@@ -348,12 +285,13 @@ void SparseRealFx::read_sparse_array(const std::string& path) {
   REQUIRE(rc == TILEDB_OK);
 
   int a_c[] = {4, 1, 5, 2, 3};
-  float coords_c[] = {
-      -160.1f, 89.1f, -23.5f, -20.0f, 1.0f, 1.0f, 43.56f, 80.0f, 66.2f, -0.3f};
+  float coords_c_dim1[] = {-160.1f, -23.5f, 1.0f, 43.56f, 66.2f};
+  float coords_c_dim2[] = {89.1f, -20.0f, 1.0f, 80.0f, -0.3f};
   CHECK(a_size == sizeof(a_c));
   CHECK(!memcmp(a, a_c, sizeof(a_c)));
-  CHECK(coords_size == sizeof(coords_c));
-  CHECK(!memcmp(coords, coords_c, sizeof(coords_c)));
+  CHECK(coords_size == sizeof(coords_c_dim1));
+  CHECK(!memcmp(coords_dim1, coords_c_dim1, sizeof(coords_c_dim1)));
+  CHECK(!memcmp(coords_dim2, coords_c_dim2, sizeof(coords_c_dim2)));
 
   // Close array
   rc = tiledb_array_close(ctx_, array);
@@ -375,18 +313,22 @@ void SparseRealFx::read_sparse_array_next_partition_bug(
 
   int a[1];
   uint64_t a_size = sizeof(a);
-  float coords[8];
-  uint64_t coords_size = sizeof(coords);
+  float coords_dim1[4];
+  float coords_dim2[4];
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   float subarray[] = {-180.0f, 180.0f, -90.0f, 90.0f};
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_subarray(ctx_, query, subarray);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_ROW_MAJOR);
   REQUIRE(rc == TILEDB_OK);
@@ -397,8 +339,8 @@ void SparseRealFx::read_sparse_array_next_partition_bug(
 
   CHECK(a_size == sizeof(int));
   CHECK(a[0] == 1);
-  CHECK(coords[0] == -180.0f);
-  CHECK(coords[1] == 1.0f);
+  CHECK(coords_dim1[0] == -180.0f);
+  CHECK(coords_dim2[0] == 1.0f);
 
   // Close array
   rc = tiledb_array_close(ctx_, array);
@@ -468,6 +410,8 @@ void SparseRealFx::create_sparse_array_double(const std::string& path) {
   CHECK(rc == TILEDB_OK);
 
   // Clean up
+  tiledb_filter_free(&filter);
+  tiledb_filter_list_free(&list);
   tiledb_attribute_free(&a);
   tiledb_dimension_free(&d1);
   tiledb_dimension_free(&d2);
@@ -478,39 +422,43 @@ void SparseRealFx::create_sparse_array_double(const std::string& path) {
 TEST_CASE_METHOD(
     SparseRealFx,
     "C API: Test 2d sparse array with real domain",
-    "[capi], [sparse-real]") {
-  std::string vector_name = FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_real";
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+    "[capi][sparse-real]") {
+  SupportedFsLocal local_fs;
+  std::string vector_name =
+      local_fs.file_prefix() + local_fs.temp_dir() + "sparse_real";
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 
   create_sparse_array(vector_name);
   write_sparse_array(vector_name);
   read_sparse_array(vector_name);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
 
 TEST_CASE_METHOD(
     SparseRealFx,
     "C API: Test 2d sparse array with real domain, next subarray partition bug",
     "[capi][sparse-real][sparse-real-next-partition-bug]") {
-  std::string array_name =
-      FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_real_next_partition_bug";
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  SupportedFsLocal local_fs;
+  std::string array_name = local_fs.file_prefix() + local_fs.temp_dir() +
+                           "sparse_real_next_partition_bug";
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 
   create_sparse_array(array_name);
   write_sparse_array_next_partition_bug(array_name);
   read_sparse_array_next_partition_bug(array_name);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
 
 TEST_CASE_METHOD(
     SparseRealFx,
     "C API: Test 2d sparse array with real domain, NaN in subarray",
     "[capi][sparse-real][sparse-real-nan-subarray]") {
+  SupportedFsLocal local_fs;
   std::string array_name =
-      FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_real_nan_subarray";
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+      local_fs.file_prefix() + local_fs.temp_dir() + "sparse_real_nan_subarray";
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 
   create_sparse_array(array_name);
   write_sparse_array(array_name);
@@ -526,6 +474,17 @@ TEST_CASE_METHOD(
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
 
+  // Set config for `sm.read_range_oob` = `error`
+  tiledb_config_t* config = nullptr;
+  tiledb_error_t* error = nullptr;
+  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.read_range_oob", "error", &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_query_set_config(ctx_, query, config);
+  REQUIRE(rc == TILEDB_OK);
+
   // Check Nan
   float subarray[] = {
       -180.0f, std::numeric_limits<float>::quiet_NaN(), -90.0f, 90.0f};
@@ -541,18 +500,20 @@ TEST_CASE_METHOD(
   rc = tiledb_array_close(ctx_, array);
   CHECK(rc == TILEDB_OK);
   tiledb_array_free(&array);
+  tiledb_config_free(&config);
   tiledb_query_free(&query);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
 
 TEST_CASE_METHOD(
     SparseRealFx,
     "C API: Test 2d sparse array with real domain, small gap point-query bug",
     "[capi][sparse-real][sparse-real-small-gap-point-query-bug]") {
-  std::string array_name =
-      FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_real_small_gap_point_query_bug";
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  SupportedFsLocal local_fs;
+  std::string array_name = local_fs.file_prefix() + local_fs.temp_dir() +
+                           "sparse_real_small_gap_point_query_bug";
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 
   create_sparse_array_double(array_name);
 
@@ -566,16 +527,20 @@ TEST_CASE_METHOD(
 
     int a[] = {1, 2};
     uint64_t a_size = sizeof(a);
-    double coords[] = {-180.0, 1.0, -179.99999999999997, 0.9999999999999999};
+    double coords_dim1[] = {-180.0, -179.99999999999997};
+    double coords_dim2[] = {1.0, 0.9999999999999999};
 
-    uint64_t coords_size = sizeof(coords);
+    uint64_t coords_size = sizeof(coords_dim1);
     tiledb_query_t* query;
     rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
     REQUIRE(rc == TILEDB_OK);
-    rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+    rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
     REQUIRE(rc == TILEDB_OK);
-    rc = tiledb_query_set_buffer(
-        ctx_, query, TILEDB_COORDS, coords, &coords_size);
+    rc = tiledb_query_set_data_buffer(
+        ctx_, query, "d1", coords_dim1, &coords_size);
+    REQUIRE(rc == TILEDB_OK);
+    rc = tiledb_query_set_data_buffer(
+        ctx_, query, "d2", coords_dim2, &coords_size);
     REQUIRE(rc == TILEDB_OK);
     rc = tiledb_query_set_layout(ctx_, query, TILEDB_UNORDERED);
     REQUIRE(rc == TILEDB_OK);
@@ -603,8 +568,9 @@ TEST_CASE_METHOD(
 
     int a[1];
     uint64_t a_size = sizeof(a);
-    double coords[2];
-    uint64_t coords_size = sizeof(coords);
+    double coords_dim1[1];
+    double coords_dim2[1];
+    uint64_t coords_size = sizeof(coords_dim1);
     tiledb_query_t* query;
     double subarray[] = {-180.0, -180.0, 1.0, 1.0};
 
@@ -612,10 +578,13 @@ TEST_CASE_METHOD(
     REQUIRE(rc == TILEDB_OK);
     rc = tiledb_query_set_subarray(ctx_, query, subarray);
     REQUIRE(rc == TILEDB_OK);
-    rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+    rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
     REQUIRE(rc == TILEDB_OK);
-    rc = tiledb_query_set_buffer(
-        ctx_, query, TILEDB_COORDS, coords, &coords_size);
+    rc = tiledb_query_set_data_buffer(
+        ctx_, query, "d1", coords_dim1, &coords_size);
+    REQUIRE(rc == TILEDB_OK);
+    rc = tiledb_query_set_data_buffer(
+        ctx_, query, "d2", coords_dim2, &coords_size);
     REQUIRE(rc == TILEDB_OK);
     rc = tiledb_query_set_layout(ctx_, query, TILEDB_ROW_MAJOR);
     REQUIRE(rc == TILEDB_OK);
@@ -626,8 +595,8 @@ TEST_CASE_METHOD(
 
     CHECK(a_size == sizeof(int));
     CHECK(a[0] == 1);
-    CHECK(coords[0] == -180.0);
-    CHECK(coords[1] == 1.0f);
+    CHECK(coords_dim1[0] == -180.0);
+    CHECK(coords_dim2[0] == 1.0f);
 
     // Close array
     rc = tiledb_array_close(ctx_, array);
@@ -637,5 +606,5 @@ TEST_CASE_METHOD(
     tiledb_array_free(&array);
     tiledb_query_free(&query);
   }
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
