@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2018 TileDB Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,6 +31,8 @@
  */
 
 #include "catch.hpp"
+#include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #ifdef _WIN32
 #include "tiledb/sm/filesystem/win.h"
 #else
@@ -43,28 +45,15 @@
 #include <sstream>
 #include <thread>
 
-struct SparseRealFx2 {
-  const std::string HDFS_TEMP_DIR = "hdfs:///tiledb_test/";
-  const std::string S3_PREFIX = "s3://";
-  const std::string S3_BUCKET = S3_PREFIX + random_bucket_name("tiledb") + "/";
-  const std::string S3_TEMP_DIR = S3_BUCKET + "tiledb_test/";
-#ifdef _WIN32
-  const std::string FILE_URI_PREFIX = "";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
-#else
-  const std::string FILE_URI_PREFIX = "file://";
-  const std::string FILE_TEMP_DIR =
-      tiledb::sm::Posix::current_dir() + "/tiledb_test/";
-#endif
+using namespace tiledb::test;
 
+struct SparseRealFx2 {
   // TileDB context
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
 
-  // Supported filesystems
-  bool supports_s3_;
-  bool supports_hdfs_;
+  // Vector of supported filsystems
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
 
   // Functions
   SparseRealFx2();
@@ -76,87 +65,22 @@ struct SparseRealFx2 {
   void write_sparse_array_next_partition_bug(const std::string& path);
   void read_sparse_array(const std::string& path);
   void read_sparse_array_next_partition_bug(const std::string& path);
-  static std::string random_bucket_name(const std::string& prefix);
-  void set_supported_fs();
+  static std::string random_name(const std::string& prefix);
 };
 
-SparseRealFx2::SparseRealFx2() {
-  // Supported filesystems
-  set_supported_fs();
-
-  // Create TileDB context
-  tiledb_config_t* config = nullptr;
-  tiledb_error_t* error = nullptr;
-  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  if (supports_s3_) {
-#ifndef TILEDB_TESTS_AWS_S3_CONFIG
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.endpoint_override", "localhost:9999", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.scheme", "https", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(
-            config, "vfs.s3.use_virtual_addressing", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(
-        tiledb_config_set(config, "vfs.s3.verify_ssl", "false", &error) ==
-        TILEDB_OK);
-    REQUIRE(error == nullptr);
-#endif
-  }
-  REQUIRE(tiledb_ctx_alloc(config, &ctx_) == TILEDB_OK);
-  REQUIRE(error == nullptr);
-  vfs_ = nullptr;
-  REQUIRE(tiledb_vfs_alloc(ctx_, config, &vfs_) == TILEDB_OK);
-  tiledb_config_free(&config);
-
-  // Connect to S3
-  if (supports_s3_) {
-    // Create bucket if it does not exist
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    REQUIRE(rc == TILEDB_OK);
-    if (!is_bucket) {
-      rc = tiledb_vfs_create_bucket(ctx_, vfs_, S3_BUCKET.c_str());
-      REQUIRE(rc == TILEDB_OK);
-    }
-  }
+SparseRealFx2::SparseRealFx2()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 }
 
 SparseRealFx2::~SparseRealFx2() {
-  if (supports_s3_) {
-    int is_bucket = 0;
-    int rc = tiledb_vfs_is_bucket(ctx_, vfs_, S3_BUCKET.c_str(), &is_bucket);
-    CHECK(rc == TILEDB_OK);
-    if (is_bucket) {
-      CHECK(
-          tiledb_vfs_remove_bucket(ctx_, vfs_, S3_BUCKET.c_str()) == TILEDB_OK);
-    }
-  }
-
+  // Close vfs test
+  REQUIRE(vfs_test_close(fs_vec_, ctx_, vfs_).ok());
   tiledb_vfs_free(&vfs_);
   tiledb_ctx_free(&ctx_);
 }
 
-void SparseRealFx2::set_supported_fs() {
-  tiledb_ctx_t* ctx = nullptr;
-  REQUIRE(tiledb_ctx_alloc(nullptr, &ctx) == TILEDB_OK);
-
-  int is_supported = 0;
-  int rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_S3, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_s3_ = (bool)is_supported;
-  rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_HDFS, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_hdfs_ = (bool)is_supported;
-
-  tiledb_ctx_free(&ctx);
-}
-
 void SparseRealFx2::create_temp_dir(const std::string& path) {
   remove_temp_dir(path);
   REQUIRE(tiledb_vfs_create_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
@@ -169,7 +93,7 @@ void SparseRealFx2::remove_temp_dir(const std::string& path) {
     REQUIRE(tiledb_vfs_remove_dir(ctx_, vfs_, path.c_str()) == TILEDB_OK);
 }
 
-std::string SparseRealFx2::random_bucket_name(const std::string& prefix) {
+std::string SparseRealFx2::random_name(const std::string& prefix) {
   std::stringstream ss;
   ss << prefix << "-" << std::this_thread::get_id() << "-"
      << TILEDB_TIMESTAMP_NOW_MS;
@@ -235,6 +159,8 @@ void SparseRealFx2::create_sparse_array(const std::string& path) {
   CHECK(rc == TILEDB_OK);
 
   // Clean up
+  tiledb_filter_free(&filter);
+  tiledb_filter_list_free(&list);
   tiledb_attribute_free(&a);
   tiledb_dimension_free(&d1);
   tiledb_dimension_free(&d2);
@@ -252,16 +178,19 @@ void SparseRealFx2::write_sparse_array(const std::string& path) {
 
   int a[] = {1, 2, 3, 4, 5};
   uint64_t a_size = sizeof(a);
-  float coords[] = {
-      -23.5f, -20.0f, 43.56f, 80.0f, 66.2f, -0.3f, -160.1f, 89.1f, 1.0f, 1.0f};
-  uint64_t coords_size = sizeof(coords);
+  float coords_dim1[] = {-23.5f, 43.56f, 66.2f, -160.1f, 1.0f};
+  float coords_dim2[] = {-20.0f, 80.0f, -0.3f, 89.1f, 1.0f};
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_UNORDERED);
   REQUIRE(rc == TILEDB_OK);
@@ -290,15 +219,19 @@ void SparseRealFx2::write_sparse_array_next_partition_bug(
 
   int a[] = {1, 2};
   uint64_t a_size = sizeof(a);
-  float coords[] = {-180.0f, 1.0f, -180.0f, 2.0f};
-  uint64_t coords_size = sizeof(coords);
+  float coords_dim1[] = {-180.0f, -180.0f};
+  float coords_dim2[] = {1.0f, 2.0f};
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_WRITE, &query);
   REQUIRE(rc == TILEDB_OK);
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_set_layout(ctx_, query, TILEDB_UNORDERED);
   REQUIRE(rc == TILEDB_OK);
@@ -326,8 +259,9 @@ void SparseRealFx2::read_sparse_array(const std::string& path) {
 
   int a[16];
   uint64_t a_size = sizeof(a);
-  float coords[32];
-  uint64_t coords_size = sizeof(coords);
+  float coords_dim1[16];
+  float coords_dim2[16];
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
@@ -342,10 +276,13 @@ void SparseRealFx2::read_sparse_array(const std::string& path) {
   rc = tiledb_query_add_range(ctx_, query, 1, &s1[0], &s1[1], nullptr);
   REQUIRE(rc == TILEDB_OK);
 
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_submit(ctx_, query);
   REQUIRE(rc == TILEDB_OK);
@@ -353,12 +290,13 @@ void SparseRealFx2::read_sparse_array(const std::string& path) {
   REQUIRE(rc == TILEDB_OK);
 
   int a_c[] = {4, 1, 5, 2, 3};
-  float coords_c[] = {
-      -160.1f, 89.1f, -23.5f, -20.0f, 1.0f, 1.0f, 43.56f, 80.0f, 66.2f, -0.3f};
+  float coords_c_dim1[] = {-160.1f, -23.5f, 1.0f, 43.56f, 66.2f};
+  float coords_c_dim2[] = {89.1f, -20.0f, 1.0f, 80.0f, -0.3f};
   CHECK(a_size == sizeof(a_c));
   CHECK(!memcmp(a, a_c, sizeof(a_c)));
-  CHECK(coords_size == sizeof(coords_c));
-  CHECK(!memcmp(coords, coords_c, sizeof(coords_c)));
+  CHECK(coords_size == sizeof(coords_c_dim1));
+  CHECK(!memcmp(coords_dim1, coords_c_dim1, sizeof(coords_c_dim1)));
+  CHECK(!memcmp(coords_dim2, coords_c_dim2, sizeof(coords_c_dim2)));
 
   // Close array
   rc = tiledb_array_close(ctx_, array);
@@ -380,8 +318,9 @@ void SparseRealFx2::read_sparse_array_next_partition_bug(
 
   int a[1];
   uint64_t a_size = sizeof(a);
-  float coords[8];
-  uint64_t coords_size = sizeof(coords);
+  float coords_dim1[4];
+  float coords_dim2[4];
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
@@ -396,18 +335,21 @@ void SparseRealFx2::read_sparse_array_next_partition_bug(
   rc = tiledb_query_add_range(ctx_, query, 1, &s1[0], &s1[1], nullptr);
   REQUIRE(rc == TILEDB_OK);
 
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_submit(ctx_, query);
   REQUIRE(rc == TILEDB_OK);
 
   CHECK(a_size == sizeof(int));
   CHECK(a[0] == 1);
-  CHECK(coords[0] == -180.0f);
-  CHECK(coords[1] == 1.0f);
+  CHECK(coords_dim1[0] == -180.0f);
+  CHECK(coords_dim2[0] == 1.0f);
 
   // Close array
   rc = tiledb_array_close(ctx_, array);
@@ -421,15 +363,17 @@ void SparseRealFx2::read_sparse_array_next_partition_bug(
 TEST_CASE_METHOD(
     SparseRealFx2,
     "C API: Test 2d sparse array with real domain 2",
-    "[capi], [sparse-real-2]") {
-  std::string vector_name = FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_real";
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+    "[capi][sparse-real-2]") {
+  SupportedFsLocal local_fs;
+  std::string vector_name =
+      local_fs.file_prefix() + local_fs.temp_dir() + "sparse_real";
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 
   create_sparse_array(vector_name);
   write_sparse_array(vector_name);
   read_sparse_array(vector_name);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
 
 TEST_CASE_METHOD(
@@ -437,24 +381,26 @@ TEST_CASE_METHOD(
     "C API: Test 2d sparse array with real domain, next subarray partition bug "
     "2",
     "[capi][sparse-real-2][sparse-real-next-partition-bug-2]") {
-  std::string array_name =
-      FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_real_next_partition_bug";
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  SupportedFsLocal local_fs;
+  std::string array_name = local_fs.file_prefix() + local_fs.temp_dir() +
+                           "sparse_real_next_partition_bug";
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 
   create_sparse_array(array_name);
   write_sparse_array_next_partition_bug(array_name);
   read_sparse_array_next_partition_bug(array_name);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
 
 TEST_CASE_METHOD(
     SparseRealFx2,
     "C API: Test 2d sparse array with real domain, NaN in subarray 2",
     "[capi][sparse-real-2][sparse-real-nan-subarray-2]") {
+  SupportedFsLocal local_fs;
   std::string array_name =
-      FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_real_nan_subarray";
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+      local_fs.file_prefix() + local_fs.temp_dir() + "sparse_real_nan_subarray";
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 
   create_sparse_array(array_name);
   write_sparse_array(array_name);
@@ -470,6 +416,17 @@ TEST_CASE_METHOD(
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
 
+  // Set config for `sm.read_range_oob` = `error`
+  tiledb_config_t* config = nullptr;
+  tiledb_error_t* error = nullptr;
+  REQUIRE(tiledb_config_alloc(&config, &error) == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_config_set(config, "sm.read_range_oob", "error", &error);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(error == nullptr);
+  rc = tiledb_query_set_config(ctx_, query, config);
+  REQUIRE(rc == TILEDB_OK);
+
   // Create some subarray
   float s0[] = {-180.0f, std::numeric_limits<float>::quiet_NaN()};
   float s1[] = {-90.0f, std::numeric_limits<float>::infinity()};
@@ -482,19 +439,21 @@ TEST_CASE_METHOD(
   rc = tiledb_array_close(ctx_, array);
   CHECK(rc == TILEDB_OK);
   tiledb_array_free(&array);
+  tiledb_config_free(&config);
   tiledb_query_free(&query);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
 
 TEST_CASE_METHOD(
     SparseRealFx2,
     "C API: Test 2d sparse array with real domain 2, unary range",
-    "[capi], [sparse-real-2]") {
+    "[capi][sparse-real-2]") {
+  SupportedFsLocal local_fs;
   std::string array_name =
-      FILE_URI_PREFIX + FILE_TEMP_DIR + "sparse_real_unary";
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
-  create_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+      local_fs.file_prefix() + local_fs.temp_dir() + "sparse_real_unary";
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
+  create_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
   create_sparse_array(array_name);
 
   // Write twice (2 fragments)
@@ -510,8 +469,9 @@ TEST_CASE_METHOD(
 
   int a[1];
   uint64_t a_size = sizeof(a);
-  float coords[2];
-  uint64_t coords_size = sizeof(coords);
+  float coords_dim1[1];
+  float coords_dim2[1];
+  uint64_t coords_size = sizeof(coords_dim1);
   tiledb_query_t* query;
   rc = tiledb_query_alloc(ctx_, array, TILEDB_READ, &query);
   REQUIRE(rc == TILEDB_OK);
@@ -526,10 +486,13 @@ TEST_CASE_METHOD(
   rc = tiledb_query_add_range(ctx_, query, 1, &s1[0], &s1[1], nullptr);
   REQUIRE(rc == TILEDB_OK);
 
-  rc = tiledb_query_set_buffer(ctx_, query, "a", a, &a_size);
+  rc = tiledb_query_set_data_buffer(ctx_, query, "a", a, &a_size);
   REQUIRE(rc == TILEDB_OK);
-  rc =
-      tiledb_query_set_buffer(ctx_, query, TILEDB_COORDS, coords, &coords_size);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d1", coords_dim1, &coords_size);
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_query_set_data_buffer(
+      ctx_, query, "d2", coords_dim2, &coords_size);
   REQUIRE(rc == TILEDB_OK);
   rc = tiledb_query_submit(ctx_, query);
   REQUIRE(rc == TILEDB_OK);
@@ -537,11 +500,13 @@ TEST_CASE_METHOD(
   REQUIRE(rc == TILEDB_OK);
 
   int a_c[] = {1};
-  float coords_c[] = {-23.5f, -20.0f};
+  float coords_c_dim1[] = {-23.5f};
+  float coords_c_dim2[] = {-20.0f};
   CHECK(a_size == sizeof(a_c));
   CHECK(!memcmp(a, a_c, sizeof(a_c)));
-  CHECK(coords_size == sizeof(coords_c));
-  CHECK(!memcmp(coords, coords_c, sizeof(coords_c)));
+  CHECK(coords_size == sizeof(coords_c_dim1));
+  CHECK(!memcmp(coords_dim1, coords_c_dim1, sizeof(coords_c_dim1)));
+  CHECK(!memcmp(coords_dim2, coords_c_dim2, sizeof(coords_c_dim2)));
 
   // Close array
   rc = tiledb_array_close(ctx_, array);
@@ -551,5 +516,5 @@ TEST_CASE_METHOD(
   tiledb_array_free(&array);
   tiledb_query_free(&query);
 
-  remove_temp_dir(FILE_URI_PREFIX + FILE_TEMP_DIR);
+  remove_temp_dir(local_fs.file_prefix() + local_fs.temp_dir());
 }
