@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,9 +31,12 @@
  */
 
 #include "test/src/helpers.h"
+#include "test/src/vfs_helpers.h"
 #include "tiledb/sm/c_api/tiledb.h"
 #include "tiledb/sm/c_api/tiledb_struct_def.h"
+#include "tiledb/sm/config/config.h"
 #include "tiledb/sm/cpp_api/tiledb"
+#include "tiledb/sm/enums/encryption_type.h"
 
 #ifdef _WIN32
 #include "tiledb/sm/filesystem/win.h"
@@ -47,6 +50,7 @@
 #include <thread>
 
 using namespace tiledb;
+using namespace tiledb::test;
 
 /* ********************************* */
 /*         STRUCT DEFINITION         */
@@ -55,10 +59,8 @@ using namespace tiledb;
 struct CPPMetadataFx {
   tiledb_ctx_t* ctx_;
   tiledb_vfs_t* vfs_;
-  bool s3_supported_, hdfs_supported_;
+  const std::vector<std::unique_ptr<SupportedFs>> fs_vec_;
   std::string temp_dir_;
-  const std::string s3_bucket_name_ =
-      "s3://" + random_bucket_name("tiledb") + "/";
   std::string array_name_;
   const char* ARRAY_NAME = "test_metadata";
   tiledb_array_t* array_ = nullptr;
@@ -74,22 +76,20 @@ struct CPPMetadataFx {
   ~CPPMetadataFx();
 };
 
-CPPMetadataFx::CPPMetadataFx() {
-  ctx_ = nullptr;
-  vfs_ = nullptr;
-  hdfs_supported_ = false;
-  s3_supported_ = false;
-
-  get_supported_fs(&s3_supported_, &hdfs_supported_);
-  create_ctx_and_vfs(s3_supported_, &ctx_, &vfs_);
-  create_s3_bucket(s3_bucket_name_, s3_supported_, ctx_, vfs_);
+CPPMetadataFx::CPPMetadataFx()
+    : fs_vec_(vfs_test_get_fs_vec()) {
+  // Initialize vfs test
+  REQUIRE(vfs_test_init(fs_vec_, &ctx_, &vfs_).ok());
 
 // Create temporary directory based on the supported filesystem
 #ifdef _WIN32
-  temp_dir_ = tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
+  SupportedFsLocal windows_fs;
+  temp_dir_ = windows_fs.file_prefix() + windows_fs.temp_dir();
 #else
-  temp_dir_ = "file://" + tiledb::sm::Posix::current_dir() + "/tiledb_test/";
+  SupportedFsLocal posix_fs;
+  temp_dir_ = posix_fs.file_prefix() + posix_fs.temp_dir();
 #endif
+
   create_dir(temp_dir_, ctx_, vfs_);
 
   array_name_ = temp_dir_ + ARRAY_NAME;
@@ -118,9 +118,9 @@ void CPPMetadataFx::create_default_array_1d() {
       {"a", "b", "c"},
       {TILEDB_INT32, TILEDB_CHAR, TILEDB_FLOAT32},
       {1, TILEDB_VAR_NUM, 2},
-      {::Compressor(TILEDB_FILTER_NONE, -1),
-       ::Compressor(TILEDB_FILTER_ZSTD, -1),
-       ::Compressor(TILEDB_FILTER_LZ4, -1)},
+      {tiledb::test::Compressor(TILEDB_FILTER_NONE, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_ZSTD, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1)},
       TILEDB_ROW_MAJOR,
       TILEDB_ROW_MAJOR,
       2);
@@ -143,9 +143,9 @@ void CPPMetadataFx::create_default_array_1d_with_key() {
       {"a", "b", "c"},
       {TILEDB_INT32, TILEDB_CHAR, TILEDB_FLOAT32},
       {1, TILEDB_VAR_NUM, 2},
-      {::Compressor(TILEDB_FILTER_NONE, -1),
-       ::Compressor(TILEDB_FILTER_ZSTD, -1),
-       ::Compressor(TILEDB_FILTER_LZ4, -1)},
+      {tiledb::test::Compressor(TILEDB_FILTER_NONE, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_ZSTD, -1),
+       tiledb::test::Compressor(TILEDB_FILTER_LZ4, -1)},
       TILEDB_ROW_MAJOR,
       TILEDB_ROW_MAJOR,
       2);
@@ -180,9 +180,6 @@ TEST_CASE_METHOD(
 
   // Close array
   array.close();
-
-  // Open with key
-  CHECK_THROWS(array.open(TILEDB_READ, enc_type_, key_, key_len_));
 }
 
 TEST_CASE_METHOD(
@@ -223,8 +220,8 @@ TEST_CASE_METHOD(
   array.get_metadata("bb", &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
 
   array.get_metadata("zero_val", &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_FLOAT32);
@@ -243,11 +240,17 @@ TEST_CASE_METHOD(
   array.get_metadata_from_index(1, &key, &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
   CHECK(key.size() == strlen("bb"));
   CHECK(!strncmp(key.data(), "bb", strlen("bb")));
 
+  // idx 2 is 'zero_val'
+  array.get_metadata_from_index(2, &key, &v_type, &v_num, &v_r);
+  CHECK(v_type == TILEDB_FLOAT32);
+  CHECK(v_num == 1);
+  CHECK(v_r == nullptr);
+
   // Check has_key
   bool has_key;
   v_type = (tiledb_datatype_t)std::numeric_limits<int32_t>::max();
@@ -259,7 +262,7 @@ TEST_CASE_METHOD(
   v_type = (tiledb_datatype_t)std::numeric_limits<int32_t>::max();
   has_key = array.has_metadata("non-existent-key", &v_type);
   CHECK(has_key == false);
-  CHECK(v_type == std::numeric_limits<int32_t>::max());
+  CHECK(v_type == (tiledb_datatype_t)std::numeric_limits<int32_t>::max());
 
   // Close array
   array.close();
@@ -312,7 +315,7 @@ TEST_CASE_METHOD(
 
   // Create and open array in write mode
   Context ctx;
-  Array array(ctx, std::string(array_name_), TILEDB_WRITE);
+  Array array(ctx, std::string(array_name_), TILEDB_WRITE, 1);
 
   // Write items
   int32_t v = 5;
@@ -323,11 +326,8 @@ TEST_CASE_METHOD(
   // Close array
   array.close();
 
-  // Prevent array metadata filename/timestamp conflicts
-  std::this_thread::sleep_for(std::chrono::milliseconds(1));
-
   // Delete an item that exists and one that does not exist
-  array.open(TILEDB_WRITE);
+  array.open(TILEDB_WRITE, 2);
   array.delete_metadata("aaa");
   array.delete_metadata("foo");
   array.close();
@@ -341,12 +341,11 @@ TEST_CASE_METHOD(
   uint32_t v_num;
   array.get_metadata("aaa", &v_type, &v_num, &v_r);
   CHECK(v_r == nullptr);
-
   array.get_metadata("bb", &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
 
   array.get_metadata("foo", &v_type, &v_num, &v_r);
   CHECK(v_r == nullptr);
@@ -358,8 +357,8 @@ TEST_CASE_METHOD(
   array.get_metadata_from_index(0, &key, &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
   CHECK(key.size() == strlen("bb"));
   CHECK(!strncmp(key.data(), "bb", strlen("bb")));
 
@@ -410,8 +409,8 @@ TEST_CASE_METHOD(
   array.get_metadata("bb", &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
 
   array.get_metadata("cccc", &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_INT32);
@@ -425,8 +424,8 @@ TEST_CASE_METHOD(
   array.get_metadata_from_index(0, &key, &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
   CHECK(key.size() == strlen("bb"));
   CHECK(!strncmp(key.data(), "bb", strlen("bb")));
 
@@ -434,7 +433,9 @@ TEST_CASE_METHOD(
   array.close();
 
   // Consolidate
-  Array::consolidate_metadata(ctx, array_name_, nullptr);
+  Config consolidation_cfg;
+  consolidation_cfg["sm.consolidation.mode"] = "array_meta";
+  Array::consolidate(ctx, array_name_, &consolidation_cfg);
 
   // Open the array in read mode
   array.open(TILEDB_READ);
@@ -456,7 +457,7 @@ TEST_CASE_METHOD(
   array.close();
 
   // Consolidate again
-  Array::consolidate_metadata(ctx, array_name_, nullptr);
+  Array::consolidate(ctx, array_name_, &consolidation_cfg);
 
   // Open the array in read mode
   array.open(TILEDB_READ);
@@ -506,7 +507,8 @@ TEST_CASE_METHOD(
   array.close();
 
   // Open the array in read mode at a timestamp
-  array.open(TILEDB_READ, timestamp);
+  array.set_open_timestamp_end(timestamp);
+  array.open(TILEDB_READ);
 
   // Read
   const void* v_r;
@@ -552,7 +554,8 @@ TEST_CASE_METHOD(
   array.close();
 
   // Open the array in read mode at a timestamp
-  array.open(TILEDB_READ, timestamp);
+  array.set_open_timestamp_end(timestamp);
+  array.open(TILEDB_READ);
 
   // Read
   const void* v_r;
@@ -588,8 +591,13 @@ TEST_CASE_METHOD(
   create_default_array_1d_with_key();
 
   // Create and open array in write mode
-  Context ctx;
-  Array array(ctx, array_name_, TILEDB_WRITE, enc_type_, key_, key_len_);
+  tiledb::Config cfg;
+  std::string enc_type_str =
+      encryption_type_str((tiledb::sm::EncryptionType)enc_type_);
+  cfg["sm.encryption_type"] = enc_type_str.c_str();
+  cfg["sm.encryption_key"] = key_;
+  Context ctx(cfg);
+  Array array(ctx, array_name_, TILEDB_WRITE);
 
   // Write items
   int32_t v = 5;
@@ -604,14 +612,14 @@ TEST_CASE_METHOD(
   std::this_thread::sleep_for(std::chrono::milliseconds(1));
 
   // Update
-  array.open(TILEDB_WRITE, enc_type_, key_, key_len_);
+  array.open(TILEDB_WRITE);
   array.delete_metadata("aaa");
   v = 10;
   array.put_metadata("cccc", TILEDB_INT32, 1, &v);
   array.close();
 
   // Open the array in read mode
-  array.open(TILEDB_READ, enc_type_, key_, key_len_);
+  array.open(TILEDB_READ);
 
   // Read
   const void* v_r;
@@ -623,8 +631,8 @@ TEST_CASE_METHOD(
   array.get_metadata("bb", &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
 
   array.get_metadata("cccc", &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_INT32);
@@ -638,8 +646,8 @@ TEST_CASE_METHOD(
   array.get_metadata_from_index(0, &key, &v_type, &v_num, &v_r);
   CHECK(v_type == TILEDB_FLOAT32);
   CHECK(v_num == 2);
-  CHECK(((const float_t*)v_r)[0] == 1.1f);
-  CHECK(((const float_t*)v_r)[1] == 1.2f);
+  CHECK(((const float*)v_r)[0] == 1.1f);
+  CHECK(((const float*)v_r)[1] == 1.2f);
   CHECK(key.size() == strlen("bb"));
   CHECK(!strncmp(key.data(), "bb", strlen("bb")));
 
@@ -647,14 +655,18 @@ TEST_CASE_METHOD(
   array.close();
 
   // Consolidate without key - error
-  CHECK_THROWS(Array::consolidate_metadata(ctx, array_name_, nullptr));
+  Config consolidate_without_key;
+  Context ctx_without_key(consolidate_without_key);
+  CHECK_THROWS(Array::consolidate(
+      ctx_without_key, array_name_, &consolidate_without_key));
 
   // Consolidate with key - ok
-  Array::consolidate_metadata(
-      ctx, array_name_, enc_type_, key_, key_len_, nullptr);
+  Config consolidation_cfg;
+  consolidation_cfg["sm.consolidation.mode"] = "array_meta";
+  Array::consolidate(ctx, array_name_, &consolidation_cfg);
 
   // Open the array in read mode
-  array.open(TILEDB_READ, enc_type_, key_, key_len_);
+  array.open(TILEDB_READ);
 
   num = array.metadata_num();
   CHECK(num == 2);
@@ -663,7 +675,7 @@ TEST_CASE_METHOD(
   array.close();
 
   // Write once more
-  array.open(TILEDB_WRITE, enc_type_, key_, key_len_);
+  array.open(TILEDB_WRITE);
 
   // Write items
   v = 50;
@@ -673,11 +685,10 @@ TEST_CASE_METHOD(
   array.close();
 
   // Consolidate again
-  Array::consolidate_metadata(
-      ctx, array_name_, enc_type_, key_, key_len_, nullptr);
+  Array::consolidate_metadata(ctx, array_name_, &consolidation_cfg);
 
   // Open the array in read mode
-  array.open(TILEDB_READ, enc_type_, key_, key_len_);
+  array.open(TILEDB_READ);
 
   num = array.metadata_num();
   CHECK(num == 3);
