@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -33,16 +33,26 @@
 #ifndef TILEDB_ARRAY_H
 #define TILEDB_ARRAY_H
 
-#include "tiledb/sm/encryption/encryption_key.h"
+#include <atomic>
+#include <unordered_map>
+#include <vector>
+
+#include "tiledb/common/status.h"
+#include "tiledb/sm/crypto/encryption_key.h"
+#include "tiledb/sm/fragment/fragment_info.h"
 #include "tiledb/sm/metadata/metadata.h"
-#include "tiledb/sm/misc/status.h"
-#include "tiledb/sm/storage_manager/open_array.h"
-#include "tiledb/sm/storage_manager/storage_manager.h"
-#include "tiledb/sm/subarray/subarray.h"
+
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
 
+class ArraySchema;
+class SchemaEvolution;
+class FragmentMetadata;
+class StorageManager;
+enum class QueryType : uint8_t;
+
 /**
  * An array object to be opened for reads/writes. An ``Array`` instance
  * is associated with the timestamp it is opened at.
@@ -56,41 +66,35 @@ class Array {
   /** Constructor. */
   Array(const URI& array_uri, StorageManager* storage_manager);
 
+  /** Copy constructor. */
+  Array(const Array& rhs);
+
   /** Destructor. */
-  ~Array();
+  ~Array() = default;
 
   /* ********************************* */
   /*                API                */
   /* ********************************* */
 
-  /** Returns the array schema. */
-  ArraySchema* array_schema() const;
+  /** Returns the latest array schema. */
+  ArraySchema* array_schema_latest() const;
+
+  /** Returns array schemas map. */
+  inline const std::unordered_map<std::string, tdb_shared_ptr<ArraySchema>>&
+  array_schemas_all() const {
+    return array_schemas_all_;
+  }
 
   /** Returns the array URI. */
   const URI& array_uri() const;
 
-  /**
-   * Computes an upper bound on the buffer sizes required for a read
-   * query, for a given subarray and set of attributes.
-   *
-   * @param subarray The subarray to focus on. Note that it must have the same
-   *     underlying type as the array domain.
-   * @param attributes The attributes to focus on.
-   * @param buffer_sizes The buffer sizes to be retrieved. This is a map from
-   *     the attribute (or coordinates) name to a pair of sizes (in bytes).
-   *     For fixed-sized attributes, the second size is ignored. For var-sized
-   *     attributes, the first is the offsets size and the second is the
-   *     values size.
-   * @return Status
-   */
-  Status compute_max_buffer_sizes(
-      const void* subarray,
-      const std::vector<std::string>& attributes,
-      std::unordered_map<std::string, std::pair<uint64_t, uint64_t>>*
-          max_buffer_sizes) const;
+  /** Returns the serialized array URI, this is for backwards compatibility with
+   * serialization in pre TileDB 2.4 */
+  const URI& array_uri_serialized() const;
 
   /**
-   * Opens the array for reading/writing.
+   * Opens the array for reading at a timestamp retrieved from the config
+   * or for writing.
    *
    * @param query_type The mode in which the array is opened.
    * @param encryption_type The encryption type of the array
@@ -106,13 +110,8 @@ class Array {
       uint32_t key_length);
 
   /**
-   * Opens the array for reading, loading only the input fragments.
-   * Note that the order of the input fragments matters; later
-   * fragments in the list may overwrite earlier ones.
+   * Opens the array for reading without fragments.
    *
-   * @param query_type The query type. This should always be READ. It
-   *    is here only for sanity check.
-   * @param fragments The fragments to open the array with.
    * @param encryption_type The encryption type of the array
    * @param encryption_key If the array is encrypted, the private encryption
    *    key. For unencrypted arrays, pass `nullptr`.
@@ -121,19 +120,26 @@ class Array {
    *
    * @note Applicable only to reads.
    */
-  Status open(
-      QueryType query_type,
-      const std::vector<FragmentInfo>& fragments,
+  Status open_without_fragments(
       EncryptionType encryption_type,
       const void* encryption_key,
       uint32_t key_length);
 
   /**
-   * Opens the array for reading at a given timestamp.
+   * Reload the array with the specified fragments.
+   *
+   * @param fragments_to_load The list of fragments to load.
+   * @return Status
+   */
+  Status load_fragments(const std::vector<TimestampedURI>& fragments_to_load);
+
+  /**
+   * Opens the array for reading.
    *
    * @param query_type The query type. This should always be READ. It
    *    is here only for sanity check.
-   * @param timestamp The timestamp at which to open the array.
+   * @param timestamp_start The start timestamp at which to open the array.
+   * @param timestamp_end The end timestamp at which to open the array.
    * @param encryption_type The encryption type of the array
    * @param encryption_key If the array is encrypted, the private encryption
    *    key. For unencrypted arrays, pass `nullptr`.
@@ -144,7 +150,8 @@ class Array {
    */
   Status open(
       QueryType query_type,
-      uint64_t timestamp,
+      uint64_t timestamp_start,
+      uint64_t timestamp_end,
       EncryptionType encryption_type,
       const void* encryption_key,
       uint32_t key_length);
@@ -159,7 +166,7 @@ class Array {
    * Returns the fragment metadata of the array. If the array is not open,
    * an empty vector is returned.
    */
-  std::vector<FragmentMetadata*> fragment_metadata() const;
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragment_metadata() const;
 
   /**
    * Returns `true` if the array is empty at the time it is opened.
@@ -180,18 +187,18 @@ class Array {
   Status get_query_type(QueryType* qyery_type) const;
 
   /**
-   * Returns the max buffer size given a fixed-sized attribute and
+   * Returns the max buffer size given a fixed-sized attribute/dimension and
    * a subarray. Errors if the array is not open.
    */
   Status get_max_buffer_size(
-      const char* attribute, const void* subarray, uint64_t* buffer_size);
+      const char* name, const void* subarray, uint64_t* buffer_size);
 
   /**
-   * Returns the max buffer size given a var-sized attribute and
+   * Returns the max buffer size given a var-sized attribute/dimension and
    * a subarray. Errors if the array is not open.
    */
   Status get_max_buffer_size(
-      const char* attribute,
+      const char* name,
       const void* subarray,
       uint64_t* buffer_off_size,
       uint64_t* buffer_val_size);
@@ -212,22 +219,41 @@ class Array {
   Status reopen();
 
   /**
-   * Re-opens the array at a specific timestamp.
+   * Re-opens the array between two specific timestamps.
    *
    * @note Applicable only for reads, it errors if the array was opened
    *     for writes.
    */
-  Status reopen(uint64_t timestamp);
+  Status reopen(uint64_t timestamp_start, uint64_t timestamp_end);
+
+  /** Returns the start timestamp. */
+  uint64_t timestamp_start() const;
+
+  /** Returns the end timestamp. */
+  uint64_t timestamp_end() const;
 
   /** Returns the timestamp at which the array was opened. */
-  uint64_t timestamp() const;
+  uint64_t timestamp_end_opened_at() const;
 
-  /** Directly set the timestamp value. */
-  Status set_timestamp(uint64_t timestamp);
+  /** Directly set the timestamp start value. */
+  Status set_timestamp_start(uint64_t timestamp_start);
+
+  /** Directly set the timestamp end value. */
+  Status set_timestamp_end(uint64_t timestamp_end);
+
+  /** Directly set the array config. */
+  Status set_config(Config config);
+
+  /** Retrieves a reference to the array config. */
+  Config config() const;
 
   /** Directly set the array URI. */
   Status set_uri(const std::string& uri);
 
+  /** Directly set the array URI for serialized compatibility with pre
+   * TileDB 2.5 clients */
+  Status set_uri_serialized(const std::string& uri);
+
   /**
    * Deletes metadata from an array opened in WRITE mode.
    *
@@ -309,34 +335,57 @@ class Array {
    * @note This is potentially an unsafe operation
    * it could have contention with locks from lazy loading of metadata.
    * This should only be used by the serialization class
-   * (tiledb/sm/serialization/array_schema.cc). In that class we need to fetch
-   * the underlying Metadata object to set the values we are loading from REST.
-   * A lock should already by taken before load_metadata is called.
+   * (tiledb/sm/serialization/array_schema_latest.cc). In that class we need to
+   * fetch the underlying Metadata object to set the values we are loading from
+   * REST. A lock should already by taken before load_metadata is called.
    */
   Metadata* metadata();
 
+  /** Returns the non-empty domain of the opened array.
+   *  If the non_empty_domain has not been computed or loaded
+   *  it will be loaded first
+   * */
+  const NDRange& non_empty_domain();
+
+  /** Returns the non-empty domain of the opened array. */
+  void set_non_empty_domain(const NDRange& non_empty_domain);
+
  private:
   /* ********************************* */
   /*         PRIVATE ATTRIBUTES        */
   /* ********************************* */
 
-  /** The array schema. */
-  ArraySchema* array_schema_;
+  /** The latest array schema. */
+  ArraySchema* array_schema_latest_;
+
+  /**
+   * A map of all array_schemas_all
+   */
+  std::unordered_map<std::string, tdb_shared_ptr<ArraySchema>>
+      array_schemas_all_;
 
   /** The array URI. */
   URI array_uri_;
 
+  /** This is a backwards compatible URI from serialization
+   *  In TileDB 2.5 we removed sending the URI but 2.4 and older were
+   * unconditionally setting the URI, so things got set to an empty stirng Now
+   * we store the serialized URI so we avoid the empty string with older
+   * clients.
+   */
+  URI array_uri_serialized_;
+
   /**
    * The private encryption key used to encrypt the array.
    *
-   * Note: this is the only place in TileDB where the user's private key bytes
-   * should be stored. Wherever a key is needed, a pointer to this memory region
-   * should be passed instead of a copy of the bytes.
+   * Note: This is the only place in TileDB where the user's private key
+   * bytes should be stored. Whenever a key is needed, a pointer to this
+   * memory region should be passed instead of a copy of the bytes.
    */
-  EncryptionKey encryption_key_;
+  tdb_shared_ptr<EncryptionKey> encryption_key_;
 
   /** The metadata of the fragments the array was opened with. */
-  std::vector<FragmentMetadata*> fragment_metadata_;
+  std::vector<tdb_shared_ptr<FragmentMetadata>> fragment_metadata_;
 
   /** `True` if the array has been opened. */
   std::atomic<bool> is_open_;
@@ -345,14 +394,35 @@ class Array {
   QueryType query_type_;
 
   /**
-   * The timestamp at which the `open_array_` got opened. In TileDB,
-   * timestamps are in ms elapsed since 1970-01-01 00:00:00 +0000 (UTC).
+   * The starting timestamp between to open `open_array_` at.
+   * In TileDB, timestamps are in ms elapsed since
+   * 1970-01-01 00:00:00 +0000 (UTC).
    */
-  uint64_t timestamp_;
+  uint64_t timestamp_start_;
+
+  /**
+   * The ending timestamp between to open `open_array_` at.
+   * In TileDB, timestamps are in ms elapsed since
+   * 1970-01-01 00:00:00 +0000 (UTC). A value of UINT64_T
+   * will be interpretted as the current timestamp.
+   */
+  uint64_t timestamp_end_;
+
+  /**
+   * The ending timestamp that the array was last opened
+   * at. This is useful when `timestamp_end_` has been
+   * set to UINT64_T. In this scenario, this variable will
+   * store the timestamp for the time that the array was
+   * opened.
+   */
+  uint64_t timestamp_end_opened_at_;
 
   /** TileDB storage manager. */
   StorageManager* storage_manager_;
 
+  /** The array config. */
+  Config config_;
+
   /** Stores the max buffer sizes requested last time by the user .*/
   std::unordered_map<std::string, std::pair<uint64_t, uint64_t>>
       last_max_buffer_sizes_;
@@ -361,7 +431,7 @@ class Array {
    * This is the last subarray used by the user to retrieve the
    * max buffer sizes.
    */
-  void* last_max_buffer_sizes_subarray_;
+  std::vector<uint8_t> last_max_buffer_sizes_subarray_;
 
   /** Mutex for thread-safety. */
   mutable std::mutex mtx_;
@@ -375,6 +445,12 @@ class Array {
   /** True if the array metadata is loaded. */
   bool metadata_loaded_;
 
+  /** True if the non_empty_domain has been computed */
+  bool non_empty_domain_computed_;
+
+  /** The non-empty domain of the array. */
+  NDRange non_empty_domain_;
+
   /* ********************************* */
   /*          PRIVATE METHODS          */
   /* ********************************* */
@@ -407,31 +483,20 @@ class Array {
       std::unordered_map<std::string, std::pair<uint64_t, uint64_t>>*
           max_buffer_sizes_) const;
 
-  /**
-   * Computes an upper bound on the buffer sizes required for a read
-   * query, for a given subarray and set of attributes. Note that
-   * the attributes are already set in `max_buffer_sizes`
-   *
-   * @tparam T The domain type
-   * @param subarray The subarray to focus on. Note that it must have the same
-   *     underlying type as the array domain.
-   * @param max_buffer_sizes The buffer sizes to be retrieved. This is a map
-   *     from an attribute to a size pair. For fixed-sized attributes, only
-   *     the first size is useful. For var-sized attributes, the first size
-   *     is the offsets size, and the second size is the values size.
-   * @return Status
-   */
-  template <class T>
-  Status compute_max_buffer_sizes(
-      const T* subarray,
-      std::unordered_map<std::string, std::pair<uint64_t, uint64_t>>*
-          max_buffer_sizes) const;
-
   /**
    * Load array metadata, handles remote arrays vs non-remote arrays
    * @return  Status
    */
   Status load_metadata();
+
+  /**
+   * Loads non empty domain from remote array
+   * @return Status
+   */
+  Status load_remote_non_empty_domain();
+
+  /** Computes the non-empty domain of the array. */
+  Status compute_non_empty_domain();
 };
 
 }  // namespace sm
