@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -33,19 +33,26 @@
 #ifndef TILEDB_DOMAIN_H
 #define TILEDB_DOMAIN_H
 
-#include "tiledb/sm/array_schema/dimension.h"
-#include "tiledb/sm/buffer/buffer.h"
-#include "tiledb/sm/enums/array_type.h"
-#include "tiledb/sm/enums/datatype.h"
-#include "tiledb/sm/enums/layout.h"
-#include "tiledb/sm/misc/status.h"
+#include "tiledb/common/macros.h"
+#include "tiledb/common/status.h"
+#include "tiledb/sm/misc/types.h"
+#include "tiledb/sm/query/query_buffer.h"
+#include "tiledb/sm/query/result_coords.h"
 
-#include <iostream>
 #include <vector>
 
+using namespace tiledb::common;
+
 namespace tiledb {
 namespace sm {
 
+class Buffer;
+class ConstBuffer;
+class Dimension;
+
+enum class Datatype : uint8_t;
+enum class Layout : uint8_t;
+
 /** Defines an array domain, which consists of dimensions. */
 class Domain {
  public:
@@ -56,13 +63,6 @@ class Domain {
   /** Empty constructor. */
   Domain();
 
-  /**
-   * Constructor.
-   *
-   * @param type The type of dimensions.
-   */
-  explicit Domain(Datatype type);
-
   /**
    * Constructor that clones the input domain.
    *
@@ -70,106 +70,29 @@ class Domain {
    */
   explicit Domain(const Domain* domain);
 
+  /** Copy constructor. */
+  DISABLE_COPY(Domain);
+
+  /** Move constructor. */
+  Domain(Domain&& rhs);
+
   /** Destructor. */
-  ~Domain();
+  ~Domain() = default;
+
+  /* ********************************* */
+  /*             OPERATORS             */
+  /* ********************************* */
+
+  /** Copy-assignment operator. */
+  DISABLE_COPY_ASSIGN(Domain);
+
+  /** Move-assignment operator. */
+  Domain& operator=(Domain&& rhs);
 
   /* ********************************* */
   /*                 API               */
   /* ********************************* */
 
-  /**
-   * Floors the value such that it coincides with the largest start of a tile
-   * that is smaller than value, on a given dimension. If there are no tile
-   * extents, then the returned value is the start of the domain on the input
-   * dimension.
-   *
-   * @tparam T The domain type.
-   * @param value The value to be floored.
-   * @param dim_idx The targeted dimension.
-   * @return The floored value.
-   */
-  template <class T>
-  T floor_to_tile(T value, unsigned dim_idx) const;
-
-  /**
-   * Splits the input subarray in half, in a way that the input layout is
-   * respected. This means that if the two resulting subarrays were to
-   * be issued as consecutive queries with the input layout, the retrieved
-   * result would be correct (i.e., the resulting cells would respect the
-   * input layout).
-   *
-   * @param subarray The input subarray.
-   * @param layout The query layout.
-   * @param subarray_1 The first subarray resulting from the split.
-   * @param subarray_2 The second subarray resulting from the split.
-   * @return Status
-   */
-  Status split_subarray(
-      void* subarray,
-      Layout layout,
-      void** subarray_1,
-      void** subarray_2) const;
-
-  /**
-   * Splits the input subarray in half, in a way that the input layout is
-   * respected. This means that if the two resulting subarrays were to
-   * be issued as consecutive queries with the input layout, the retrieved
-   * result would be correct (i.e., the resulting cells would respect the
-   * input layout).
-   *
-   * @tparam T The domain type.
-   * @param subarray The input subarray.
-   * @param layout The query layout.
-   * @param subarray_1 The first subarray resulting from the split.
-   * @param subarray_2 The second subarray resulting from the split.
-   * @return Status
-   */
-  template <class T>
-  Status split_subarray(
-      void* subarray,
-      Layout layout,
-      void** subarray_1,
-      void** subarray_2) const;
-
-  /**
-   * Splits the input subarray in half, in a way that the global layout is
-   * respected. This means that if the two resulting subarrays were to
-   * be issued as consecutive queries with the input layout, the retrieved
-   * result would be correct (i.e., the resulting cells would respect the
-   * global layout).
-   *
-   * @tparam T The domain type.
-   * @param subarray The input subarray.
-   * @param layout The query layout.
-   * @param subarray_1 The first subarray resulting from the split.
-   * @param subarray_2 The second subarray resulting from the split.
-   * @return Status
-   */
-  template <class T>
-  Status split_subarray_global(
-      void* subarray, void** subarray_1, void** subarray_2) const;
-
-  /**
-   * Splits the input subarray in half, in a way that the input cell layout is
-   * respected. This means that if the two resulting subarrays were to
-   * be issued as consecutive queries with the cell layout, the retrieved
-   * result would be correct (i.e., the resulting cells would respect the
-   * input layout).
-   *
-   * @tparam T The domain type.
-   * @param subarray The input subarray.
-   * @param cell_layout The cell layout.
-   * @param subarray_1 The first subarray resulting from the split.
-   * @param subarray_2 The second subarray resulting from the split.
-   * @return Status
-   */
-  template <class T>
-  Status split_subarray_cell(
-      void* subarray,
-      Layout cell_layout,
-      void** subarray_1,
-      void** subarray_2) const;
-
   /**
    * Adds a dimension to the domain.
    *
@@ -178,32 +101,20 @@ class Domain {
    */
   Status add_dimension(const Dimension* dim);
 
-  /**
-   * Returns the number of cells in the input domain. Note that this is
-   * applicable only to integer array domains (otherwise the output is 0).
-   * Also note that it is assummed that the input domain is expanded
-   * such that it aligns with the tile extents.
-   *
-   * @param domain The domain to be checked.
-   * @return The number of cells in the domain.
-   *
-   * @note The function returns 0 in case `domain` is huge, leading to more
-   *      cells than `uint64_t` can hold.
-   */
-  uint64_t cell_num(const void* domain) const;
+  /** Returns true if all dimensions have fixed-sized domain datatypes. */
+  bool all_dims_fixed() const;
 
-  /**
-   * Returns the number of cells in the input domain. Note that this is
-   * applicable only to integer array domains (otherwise the output is 0).
-   * Also note that it is assummed that the input domain is expanded
-   * such that it aligns with the tile extents.
-   *
-   * @tparam T The domain type.
-   * @param domain The domain to be checked.
-   * @return The number of cells in the domain.
-   */
-  template <class T>
-  uint64_t cell_num(const T* domain) const;
+  /** Returns true if all dimensions have integer domain types. */
+  bool all_dims_int() const;
+
+  /** Returns true if all dimensions have string domain types. */
+  bool all_dims_string() const;
+
+  /** Returns true if all dimensions have real domain types. */
+  bool all_dims_real() const;
+
+  /** Returns true if all dimensions have the same type. */
+  bool all_dims_same_type() const;
 
   /** Returns the number of cells per tile (only for the dense case). */
   uint64_t cell_num_per_tile() const;
@@ -214,62 +125,84 @@ class Domain {
    * same regular tile.
    *
    * @tparam T The coordinates type.
-   * @param coords_a The first input coordinates.
-   * @param coords_b The second input coordinates.
+   * @param dim_idx The dimension to compare the coordinates on.
+   * @param a The first input coordinates.
+   * @param b The second input coordinates.
    * @return One of the following:
-   *    - -1 if the first coordinates precede the second
+   *    - -1 if the first coordinate precedes the second
    *    -  0 if the two coordinates are identical
-   *    - +1 if the first coordinates succeed the second
+   *    - +1 if the first coordinate succeeds the second
+   */
+  int cell_order_cmp(
+      unsigned dim_idx, const ResultCoords& a, const ResultCoords& b) const;
+
+  /**
+   * Checks the cell order of the input coordinates. Since the coordinates
+   * are given for a single dimension, this function simply checks which
+   * coordinate is larger.
+   *
+   * @param coord_a The first coordinate.
+   * @param coord_b The second coordinate.
+   * @return One of the following:
+   *    - -1 if the first coordinate is smaller than the second
+   *    -  0 if the two coordinates have the same cell order
+   *    - +1 if the first coordinate is larger than the second
    */
   template <class T>
-  int cell_order_cmp(const T* coords_a, const T* coords_b) const;
+  static int cell_order_cmp_2(const void* coord_a, const void* coord_b);
+
+  /**
+   * Checks the cell order of the input coordinates. Since the coordinates
+   * are given for a single dimension, this function simply checks which
+   * coordinate is larger.
+   *
+   * @param dim The dimension to check the coordinates on.
+   * @param buff The buffer that stores all coordinates.
+   * @param a The position of the first coordinate in the buffer to check.
+   * @param b The position of the second coordinate in the buffer to check.
+   * @return One of the following:
+   *    - -1 if the first coordinate is smaller than the second
+   *    -  0 if the two coordinates have the same cell order
+   *    - +1 if the first coordinate is larger than the second
+   */
+  template <class T>
+  static int cell_order_cmp(
+      const Dimension* dim, const QueryBuffer* buff, uint64_t a, uint64_t b);
+
+  /**
+   * Checks the cell order of the input coordinates.
+   *
+   * @param coord_buffs The input coordinates, given n separate buffers,
+   *     one per dimension. The buffers are sorted in the same order of the
+   *     dimensions as defined in the array schema.
+   * @param a The position of the first coordinate tuple across all buffers.
+   * @param b The position of the second coordinate tuple across all buffers.
+   * @return One of the following:
+   *    - -1 if the first coordinates precede the second on the cell order
+   *    -  0 if the two coordinates have the same cell order
+   *    - +1 if the first coordinates succeed the second on the cell order
+   */
+  int cell_order_cmp(
+      const std::vector<const QueryBuffer*>& coord_buffs,
+      uint64_t a,
+      uint64_t b) const;
 
   /**
    * Populates the object members from the data in the input binary buffer.
    *
    * @param buff The buffer to deserialize from.
+   * @param version The array schema version.
    * @return Status
    */
-  Status deserialize(ConstBuffer* buff);
+  Status deserialize(ConstBuffer* buff, uint32_t version);
 
   /** Returns the cell order. */
   Layout cell_order() const;
 
   /**
-   * Crops the input domain such that it does not exceed the array domain.
-   *
-   * @param domain The domain to be cropped.
-   * @return void
+   * Crops the input ND range such that it does not exceed the array domain.
    */
-  void crop_domain(void* domain) const;
-
-  /**
-   * Crops the input domain such that it does not exceed the array domain.
-   *
-   * @param domain The domain to be cropped.
-   * @return void
-   */
-  template <
-      class T,
-      typename std::enable_if<std::is_integral<T>::value, T>::type* = nullptr>
-  void crop_domain(T* domain) const {
-    auto array_domain = static_cast<const T*>(domain_);
-
-    for (unsigned int i = 0; i < dim_num_; ++i) {
-      if (domain[2 * i] < array_domain[2 * i])
-        domain[2 * i] = array_domain[2 * i];
-      if (domain[2 * i + 1] > array_domain[2 * i + 1])
-        domain[2 * i + 1] = array_domain[2 * i + 1];
-    }
-  }
-
-  /** No-op for float/double domains. */
-  template <
-      class T,
-      typename std::enable_if<!std::is_integral<T>::value, T>::type* = nullptr>
-  void crop_domain(T* domain) const {
-    (void)domain;
-  }
+  void crop_ndrange(NDRange* ndrange) const;
 
   /** Returns the tile order. */
   Layout tile_order() const;
@@ -277,86 +210,31 @@ class Domain {
   /** Returns the number of dimensions. */
   unsigned int dim_num() const;
 
-  /** Returns the domain (serialized dimension domains). */
-  const void* domain() const;
+  /** Returns the domain along the i-th dimension. */
+  const Range& domain(unsigned i) const;
 
-  /** returns the domain along the i-th dimension (nullptr upon error). */
-  const void* domain(unsigned int i) const;
+  /** Returns the domain as a N-dimensional range object. */
+  NDRange domain() const;
 
   /** Returns the i-th dimensions (nullptr upon error). */
   const Dimension* dimension(unsigned int i) const;
 
   /** Returns the dimension given a name (nullptr upon error). */
-  const Dimension* dimension(std::string name) const;
+  const Dimension* dimension(const std::string& name) const;
 
   /** Dumps the domain in ASCII format in the selected output. */
   void dump(FILE* out) const;
 
-  /**
-   * Expands the input domain such that it coincides with the boundaries of
-   * the array's regular tiles (i.e., it maps it on the regular tile grid).
-   * If the array has no regular tile grid, the function does not do anything.
-   *
-   * @param domain The domain to be expanded.
-   * @return void
-   */
-  void expand_domain(void* domain) const;
+  /** Expands ND range `r2` using ND range `r1`. */
+  void expand_ndrange(const NDRange& r1, NDRange* r2) const;
 
   /**
    * Expands the input domain such that it coincides with the boundaries of
    * the array's regular tiles (i.e., it maps it on the regular tile grid).
-   * If the array has no regular tile grid, the function does not do anything.
-   *
-   * @tparam The domain type.
-   * @param domain The domain to be expanded.
-   * @return void
+   * If the array has no regular tile grid or real domain, the function
+   * does not do anything.
    */
-  template <
-      class T,
-      typename std::enable_if<std::is_integral<T>::value, T>::type* = nullptr>
-  void expand_domain(T* domain) const {
-    // Applicable only to regular tiles
-    if (tile_extents_ == nullptr)
-      return;
-
-    auto tile_extents = static_cast<const T*>(tile_extents_);
-    auto array_domain = static_cast<const T*>(domain_);
-
-    for (unsigned int i = 0; i < dim_num_; ++i) {
-      // This will always make the first bound coincide with a tile
-      domain[2 * i] = ((domain[2 * i] - array_domain[2 * i]) / tile_extents[i] *
-                       tile_extents[i]) +
-                      array_domain[2 * i];
-
-      domain[2 * i + 1] =
-          ((domain[2 * i + 1] - array_domain[2 * i]) / tile_extents[i] + 1) *
-              tile_extents[i] -
-          1 + array_domain[2 * i];
-    }
-  }
-
-  /** No-op for float/double domains. */
-  template <
-      class T,
-      typename std::enable_if<!std::is_integral<T>::value, T>::type* = nullptr>
-  void expand_domain(T* domain) const {
-    (void)domain;
-  }
-
-  /**
-   * Returns the position of the input coordinates inside its corresponding
-   * tile, based on the array cell order. Applicable only to **dense** arrays.
-   *
-   * @tparam T The coordinates type.
-   * @param coords The input coordindates, which are expressed as global
-   *     coordinates in the array domain.
-   * @pos The position of the cell coordinates in the array cell order
-   *     within its corresponding tile.
-   * @return Status
-   *
-   */
-  template <class T>
-  Status get_cell_pos(const T* coords, uint64_t* pos) const;
+  void expand_to_tiles(NDRange* ndrange) const;
 
   /**
    * Retrieves the tile coordinates of the input cell coordinates.
@@ -386,42 +264,6 @@ class Domain {
   template <class T>
   void get_end_of_cell_slab(T* subarray, T* start, Layout layout, T* end) const;
 
-  /**
-   * Retrieves the next coordinates along the array cell order within a given
-   * domain (desregarding whether the domain is split into tiles or not).
-   * Applicable only to **dense** arrays, and focusing on **column-major**
-   * cell order.
-   *
-   * @tparam T The coordinates type.
-   * @param domain The targeted domain.
-   * @param cell_coords The input cell coordinates, which the function modifies
-   *     to store the next coordinates at termination.
-   * @param coords_retrieved Will store true if the retrieved coordinates are
-   *     inside the domain, and false otherwise.
-   * @return void
-   */
-  template <class T>
-  void get_next_cell_coords_col(
-      const T* domain, T* cell_coords, bool* coords_retrieved) const;
-
-  /**
-   * Retrieves the next coordinates along the array cell order within a given
-   * domain (desregarding whether the domain is split into tiles or not).
-   * Applicable only to **dense** arrays, and focusing on **row-major**
-   * cell order.
-   *
-   * @tparam T The coordinates type.
-   * @param domain The targeted domain.
-   * @param cell_coords The input cell coordinates, which the function modifies
-   *     to store the next coordinates at termination.
-   * @param coords_retrieved Will store true if the retrieved coordinates are
-   *     inside the domain, and false otherwise.
-   * @return void
-   */
-  template <class T>
-  void get_next_cell_coords_row(
-      const T* domain, T* cell_coords, bool* coords_retrieved) const;
-
   /**
    * Retrieves the next tile coordinates along the array tile order within a
    * given tile domain. Applicable only to **dense** arrays.
@@ -460,18 +302,6 @@ class Domain {
   template <class T>
   void get_tile_domain(const T* subarray, T* tile_subarray) const;
 
-  /**
-   * Returns the tile position along the array tile order within the input
-   * domain. Applicable only to **dense** arrays.
-   *
-   * @tparam T The domain type.
-   * @param tile_coords The tile coordinates.
-   * @return The tile position of *tile_coords* along the tile order of the
-   *     array inside the array domain.
-   */
-  template <class T>
-  uint64_t get_tile_pos(const T* tile_coords) const;
-
   /**
    * Returns the tile position along the array tile order within the input
    * domain. Applicable only to **dense** arrays.
@@ -522,6 +352,15 @@ class Domain {
    */
   Status has_dimension(const std::string& name, bool* has_dim) const;
 
+  /**
+   * Gets the index in the domain of a given dimension name
+   *
+   * @param name Name of dimension to check for
+   * @param dim_idx The index of this dimension in the domain
+   * @return Status
+   */
+  Status get_dimension_index(const std::string& name, unsigned* dim_idx) const;
+
   /**
    * Initializes the domain.
    *
@@ -538,9 +377,10 @@ class Domain {
    * Serializes the object members into a binary buffer.
    *
    * @param buff The buffer to serialize the data into.
+   * @param version The array schema version.
    * @return Status
    */
-  Status serialize(Buffer* buff);
+  Status serialize(Buffer* buff, uint32_t version);
 
   /**
    * For every dimension that has a null tile extent, it sets
@@ -558,126 +398,84 @@ class Domain {
   template <class T>
   uint64_t stride(Layout subarray_layout) const;
 
+  /** Returns the tile extent along the i-th dimension. */
+  const ByteVecValue& tile_extent(unsigned i) const;
+
   /** Returns the tile extents. */
-  const void* tile_extents() const;
-
-  /** returns the tile extent along the i-th dimension (nullptr upon error). */
-  const void* tile_extent(unsigned int i) const;
+  std::vector<ByteVecValue> tile_extents() const;
 
   /**
-   * Returns the number of tiles contained in the input range.
-   *
-   * @note Applicable only to integer domains.
+   * Returns the number of tiles intersecting the input ND range.
+   * Returns 0 if even a single dimension has non-integral type.
    */
-  uint64_t tile_num(const void* range) const;
+  uint64_t tile_num(const NDRange& ndrange) const;
 
   /**
-   * Returns the number of tiles contained in the input range.
-   *
-   * @note Applicable only to integer domains.
+   * Returns the number of cells in the input range.
+   * If there is an overflow, then the function returns MAX_UINT64.
+   * If at least one dimension had a non-integer domain, the
+   * function returns MAX_UINT64.
    */
-  template <
-      class T,
-      typename std::enable_if<std::is_integral<T>::value, T>::type* = nullptr>
-  uint64_t tile_num(const T* range) const {
-    // For easy reference
-    auto tile_extents = static_cast<const T*>(tile_extents_);
-    auto domain = static_cast<const T*>(domain_);
+  uint64_t cell_num(const NDRange& ndrange) const;
 
-    uint64_t ret = 1;
-    for (unsigned int i = 0; i < dim_num_; ++i) {
-      uint64_t start = (range[2 * i] - domain[2 * i]) / tile_extents[i];
-      uint64_t end = (range[2 * i + 1] - domain[2 * i]) / tile_extents[i];
-      ret *= (end - start + 1);
-    }
-    return ret;
-  }
+  /** Returns true if r1 is fully covered by r2. */
+  bool covered(const NDRange& r1, const NDRange& r2) const;
+
+  /** Returns true if the two ND ranges overlap. */
+  bool overlap(const NDRange& r1, const NDRange& r2) const;
+
+  /**
+   * Return ratio of the overalp of the two input ND ranges over
+   * the volume of `r2`.
+   */
+  double overlap_ratio(const NDRange& r1, const NDRange& r2) const;
+
+  /**
+   * Checks the tile order of the input coordinates on the given dimension.
+   *
+   * @param The dimension to compare on.
+   * @param coord_a The first coordinate.
+   * @param coord_b The second coordinate.
+   * @return One of the following:
+   *    - -1 if the first coordinate precedes the second on the tile order
+   *    -  0 if the two coordinates have the same tile order
+   *    - +1 if the first coordinate succeeds the second on the tile order
+   */
+  template <class T>
+  static int tile_order_cmp(
+      const Dimension* dim, const void* coord_a, const void* coord_b);
 
   /**
    * Checks the tile order of the input coordinates.
    *
-   * @tparam T The coordinates type.
-   * @param coords_a The first input coordinates.
-   * @param coords_b The second input coordinates.
+   * @param coord_buffs The input coordinates, given n separate buffers,
+   *     one per dimension. The buffers are sorted in the same order of the
+   *     dimensions as defined in the array schema.
+   * @param a The position of the first coordinate tuple across all buffers.
+   * @param b The position of the second coordinate tuple across all buffers.
    * @return One of the following:
    *    - -1 if the first coordinates precede the second on the tile order
    *    -  0 if the two coordinates have the same tile order
    *    - +1 if the first coordinates succeed the second on the tile order
    */
-  template <class T>
-  int tile_order_cmp(const T* coords_a, const T* coords_b) const;
+  int tile_order_cmp(
+      const std::vector<const QueryBuffer*>& coord_buffs,
+      uint64_t a,
+      uint64_t b) const;
 
   /**
-   * Checks the tile order of the input tile coordinates.
+   * Checks the tile order of the input coordinates for a given dimension.
    *
-   * @tparam T The coordinates type.
-   * @param coords_a The first tile's coordinates.
-   * @param coords_b The second tile's coordinates.
+   * @param dim_idx The index of the dimension to focus on.
+   * @param coord_a The first coordinate on the given dimension.
+   * @param coord_b The second coordinate on the given dimension.
    * @return One of the following:
-   *    - -1 if the first coordinates precede the second on the tile order
+   *    - -1 if the first coordinate precedes the second on the tile order
    *    -  0 if the two coordinates have the same tile order
-   *    - +1 if the first coordinates succeed the second on the tile order
+   *    - +1 if the first coordinate succeeds the second on the tile order
    */
-  template <class T>
-  int tile_order_cmp_tile_coords(
-      const T* tile_coords_a, const T* tile_coords_b) const;
-
-  /** Returns the dimensions type. */
-  Datatype type() const;
-
-  /**
-   * Returns the position of the input coordinates inside its corresponding
-   * tile, based on the array cell order. Applicable only to **dense** arrays,
-   * and focusing on the **column-major** cell order.
-   *
-   * @tparam T The coordinates type.
-   * @param coords The input coordindates, which are expressed as global
-   *     coordinates in the array domain.
-   * @return The position of the cell coordinates in the array cell order
-   *     within its corresponding tile.
-   */
-  template <class T>
-  uint64_t get_cell_pos_col(const T* coords) const;
-
-  /**
-   * Returns the position of the input coordinates inside the input subarray.
-   * Applicable only to **dense** arrays,
-   * and focusing on the **column-major** cell order.
-   *
-   * @tparam T The coordinates type.
-   * @param subarray The input subarray, expressed in global coordinates.
-   * @param coords The input coordindates, expressed in global coordinates.
-   * @return The position of the cell coordinates in the subarray.
-   */
-  template <class T>
-  uint64_t get_cell_pos_col(const T* subarray, const T* coords) const;
-
-  /**
-   * Returns the position of the input coordinates inside its corresponding
-   * tile, based on the array cell order. Applicable only to **dense** arrays,
-   * and focusing on the **row-major** cell order.
-   *
-   * @tparam T The coordinates type.
-   * @param coords The input coordindates, which are expressed as global
-   *     coordinates in the array domain.
-   * @return The position of the cell coordinates in the array cell order
-   *     within its corresponding tile.
-   */
-  template <class T>
-  uint64_t get_cell_pos_row(const T* coords) const;
-
-  /**
-   * Returns the position of the input coordinates inside the input subarray.
-   * Applicable only to **dense** arrays,
-   * and focusing on the **row-major** cell order.
-   *
-   * @tparam T The coordinates type.
-   * @param subarray The input subarray, expressed in global coordinates.
-   * @param coords The input coordindates, expressed in global coordinates.
-   * @return The position of the cell coordinates in the subarray.
-   */
-  template <class T>
-  uint64_t get_cell_pos_row(const T* subarray, const T* coords) const;
+  int tile_order_cmp(
+      unsigned dim_idx, const void* coord_a, const void* coord_b) const;
 
  private:
   /* ********************************* */
@@ -691,50 +489,45 @@ class Domain {
   Layout cell_order_;
 
   /** The domain dimensions. */
-  std::vector<Dimension*> dimensions_;
+  std::vector<tdb_unique_ptr<Dimension>> dimensions_;
 
   /** The number of dimensions. */
-  unsigned int dim_num_;
-
-  /**
-   * The array domain, represented by serializing the dimension domains.
-   * It should contain one [lower, upper] pair per dimension.
-   * The type of the values stored in this buffer should match the dimensions
-   * type.
-   */
-  void* domain_;
-
-  /**
-   * The array domain. It should contain one [lower, upper] pair per dimension.
-   * The type of the values stored in this buffer should match the dimensions
-   * type.
-   */
-  void* tile_domain_;
-
-  /**
-   * The tile extents. There should be one value for each dimension. The type
-   * of the values stored in this buffer should match the dimensions type. If
-   * it is NULL, then it means that the array is sparse.
-   */
-  void* tile_extents_;
-
-  /**
-   * Offsets for calculating tile positions and ids for the column-major
-   * tile order.
-   */
-  std::vector<uint64_t> tile_offsets_col_;
-
-  /**
-   * Offsets for calculating tile positions and ids for the row-major
-   * tile order.
-   */
-  std::vector<uint64_t> tile_offsets_row_;
+  unsigned dim_num_;
 
   /** The tile order of the array the domain belongs to. */
   Layout tile_order_;
 
-  /** The type of dimensions. */
-  Datatype type_;
+  /**
+   * Vector of functions, one per dimension, for comparing the cell order of
+   * two coordinates. The inputs to the function are:
+   *
+   * - dim: The dimension to check the coordinates on.
+   * - buff: The buffer that stores all coorinates;
+   * - a, b: The positions of the two coordinates in the buffer to compare.
+   */
+  std::vector<int (*)(
+      const Dimension* dim, const QueryBuffer* buff, uint64_t a, uint64_t b)>
+      cell_order_cmp_func_;
+
+  /**
+   * Vector of functions, one per dimension, for comparing the cell order of
+   * two coordinates. The inputs to the function are:
+   *
+   * - coord_a, coord_b: The two coordinates to compare.
+   */
+  std::vector<int (*)(const void* coord_a, const void* coord_b)>
+      cell_order_cmp_func_2_;
+
+  /**
+   * Vector of functions, one per dimension, for comparing the tile order of
+   * two coordinates. The inputs to the function are:
+   *
+   * - dim: The dimension to compare on.
+   * - coord_a, coord_b: The two coordinates to compare.
+   */
+  std::vector<int (*)(
+      const Dimension* dim, const void* coord_a, const void* coord_b)>
+      tile_order_cmp_func_;
 
   /* ********************************* */
   /*           PRIVATE METHODS         */
@@ -752,36 +545,8 @@ class Domain {
   template <class T>
   void compute_cell_num_per_tile();
 
-  /** Computes the tile domain. */
-  void compute_tile_domain();
-
-  /**
-   * Computes the tile domain.
-   *
-   * @tparam T The domain type.
-   * @return void
-   */
-  template <class T>
-  void compute_tile_domain();
-
-  /**
-   * Computes tile offsets neccessary when computing tile positions and ids.
-   *
-   * @return void
-   */
-  void compute_tile_offsets();
-
-  /**
-   * Computes tile offsets neccessary when computing tile positions and ids.
-   *
-   * @tparam T The coordinates type.
-   * @return void
-   */
-  template <class T>
-  void compute_tile_offsets();
-
-  /** Returns the default name constructed for the i-th dimension. */
-  std::string default_dimension_name(unsigned int i) const;
+  /** Prepares the comparator functions for each dimension. */
+  void set_tile_cell_order_cmp_funcs();
 
   /**
    * Retrieves the next tile coordinates along the array tile order within a
@@ -843,19 +608,6 @@ class Domain {
   void get_next_tile_coords_row(
       const T* domain, T* tile_coords, bool* in) const;
 
-  /**
-   * Returns the tile position along the array tile order within the input
-   * domain. Applicable only to **dense** arrays, and focusing on the
-   * **column-major** tile order.
-   *
-   * @tparam T The domain type.
-   * @param tile_coords The tile coordinates.
-   * @return The tile position of *tile_coords* along the tile order of the
-   *     array inside the array domain.
-   */
-  template <class T>
-  uint64_t get_tile_pos_col(const T* tile_coords) const;
-
   /**
    * Returns the tile position along the array tile order within the input
    * domain. Applicable only to **dense** arrays, and focusing on the
@@ -872,19 +624,6 @@ class Domain {
   template <class T>
   uint64_t get_tile_pos_col(const T* domain, const T* tile_coords) const;
 
-  /**
-   * Returns the tile position along the array tile order within the input
-   * domain. Applicable only to **dense** arrays, and focusing on the
-   * **row-major** tile order.
-   *
-   * @tparam T The domain type.
-   * @param tile_coords The tile coordinates.
-   * @return The tile position of *tile_coords* along the tile order of the
-   *     array inside the array domain.
-   */
-  template <class T>
-  uint64_t get_tile_pos_row(const T* tile_coords) const;
-
   /**
    * Returns the tile position along the array tile order within the input
    * domain. Applicable only to **dense** arrays, and focusing on the
