@@ -86,8 +86,6 @@ const std::string Config::SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_QUERY_CONDITION =
     "0.25";
 const std::string Config::SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_TILE_RANGES = "0.1";
 const std::string Config::SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_ARRAY_DATA = "0.1";
-const std::string Config::SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_RESULT_TILES =
-    "0.05";
 const std::string Config::SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_RCS = "0.05";
 const std::string Config::SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_COORDS =
     "0.5";
@@ -97,9 +95,6 @@ const std::string Config::SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_TILE_RANGES =
     "0.1";
 const std::string Config::SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_ARRAY_DATA =
     "0.1";
-const std::string Config::SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_RESULT_TILES =
-    "0.05";
-const std::string Config::SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_RCS = "0.05";
 const std::string Config::SM_ENABLE_SIGNAL_HANDLERS = "true";
 const std::string Config::SM_COMPUTE_CONCURRENCY_LEVEL =
     utils::parse::to_str(std::thread::hardware_concurrency());
@@ -253,8 +248,6 @@ Config::Config() {
       SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_TILE_RANGES;
   param_values_["sm.mem.reader.sparse_global_order.ratio_array_data"] =
       SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_ARRAY_DATA;
-  param_values_["sm.mem.reader.sparse_global_order.ratio_result_tiles"] =
-      SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_RESULT_TILES;
   param_values_["sm.mem.reader.sparse_global_order.ratio_rcs"] =
       SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_RCS;
   param_values_["sm.mem.reader.sparse_unordered_with_dups.ratio_coords"] =
@@ -266,10 +259,6 @@ Config::Config() {
       SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_TILE_RANGES;
   param_values_["sm.mem.reader.sparse_unordered_with_dups.ratio_array_data"] =
       SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_ARRAY_DATA;
-  param_values_["sm.mem.reader.sparse_unordered_with_dups.ratio_result_tiles"] =
-      SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_RESULT_TILES;
-  param_values_["sm.mem.reader.sparse_unordered_with_dups.ratio_rcs"] =
-      SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_RCS;
   param_values_["sm.enable_signal_handlers"] = SM_ENABLE_SIGNAL_HANDLERS;
   param_values_["sm.compute_concurrency_level"] = SM_COMPUTE_CONCURRENCY_LEVEL;
   param_values_["sm.io_concurrency_level"] = SM_IO_CONCURRENCY_LEVEL;
@@ -472,7 +461,13 @@ Status Config::get(const std::string& param, T* value, bool* found) const {
     return Status::Ok();
 
   // Parameter found, retrieve value
-  return utils::parse::convert(val, value);
+  auto status = utils::parse::convert(val, value);
+  if (!status.ok()) {
+    return Status::ConfigError(
+        std::string("Failed to parse config value '") + std::string(val) +
+        std::string("' for key '") + param + "' due to: " + status.to_string());
+  }
+  return Status::Ok();
 }
 
 const std::map<std::string, std::string>& Config::param_values() const {
@@ -551,9 +546,6 @@ Status Config::unset(const std::string& param) {
   } else if (param == "sm.mem.reader.sparse_global_order.ratio_array_data") {
     param_values_["sm.mem.reader.sparse_global_order.ratio_array_data"] =
         SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_ARRAY_DATA;
-  } else if (param == "sm.mem.reader.sparse_global_order.ratio_result_tiles") {
-    param_values_["sm.mem.reader.sparse_global_order.ratio_result_tiles"] =
-        SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_RESULT_TILES;
   } else if (param == "sm.mem.reader.sparse_global_order.ratio_rcs") {
     param_values_["sm.mem.reader.sparse_global_order.ratio_rcs"] =
         SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_RCS;
@@ -575,14 +567,6 @@ Status Config::unset(const std::string& param) {
       param == "sm.mem.reader.sparse_unordered_with_dups.ratio_array_data") {
     param_values_["sm.mem.reader.sparse_unordered_with_dups.ratio_array_data"] =
         SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_ARRAY_DATA;
-  } else if (
-      param == "sm.mem.reader.sparse_unordered_with_dups.ratio_result_tiles") {
-    param_values_
-        ["sm.mem.reader.sparse_unordered_with_dups.ratio_result_tiles"] =
-            SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_RESULT_TILES;
-  } else if (param == "sm.mem.reader.sparse_unordered_with_dups.ratio_rcs") {
-    param_values_["sm.mem.reader.sparse_unordered_with_dups.ratio_rcs"] =
-        SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_RCS;
   } else if (param == "sm.enable_signal_handlers") {
     param_values_["sm.enable_signal_handlers"] = SM_ENABLE_SIGNAL_HANDLERS;
   } else if (param == "sm.compute_concurrency_level") {
