@@ -176,9 +176,6 @@ class Config {
   /** Ratio of the sparse global order reader budget used for array data. */
   static const std::string SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_ARRAY_DATA;
 
-  /** Ratio of the sparse global order reader budget used for result tiles. */
-  static const std::string SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_RESULT_TILES;
-
   /** Ratio of the sparse global order reader budget used for result cell slabs.
    */
   static const std::string SM_MEM_SPARSE_GLOBAL_ORDER_RATIO_RCS;
@@ -205,18 +202,6 @@ class Config {
    */
   static const std::string SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_ARRAY_DATA;
 
-  /**
-   * Ratio of the sparse unordered with dups reader budget used for result
-   * tiles.
-   */
-  static const std::string SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_RESULT_TILES;
-
-  /**
-   * Ratio of the sparse unordered with dups reader budget used for result
-   * cell slabs.
-   */
-  static const std::string SM_MEM_SPARSE_UNORDERED_WITH_DUPS_RATIO_RCS;
-
   /** Whether or not the signal handlers are installed. */
   static const std::string SM_ENABLE_SIGNAL_HANDLERS;
 
