@@ -1,5 +1,5 @@
 /**
- * @file   tiledb_cpp_api_array.h
+ * @file   array.h
  *
  * @author Ravi Gaddipati
  *
@@ -7,7 +7,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -38,6 +38,7 @@
 #include "array_schema.h"
 #include "context.h"
 #include "tiledb.h"
+#include "tiledb_experimental.h"
 #include "type.h"
 
 #include <unordered_map>
@@ -87,7 +88,17 @@ class Array {
       const Context& ctx,
       const std::string& array_uri,
       tiledb_query_type_t query_type)
-      : Array(ctx, array_uri, query_type, TILEDB_NO_ENCRYPTION, nullptr, 0) {
+      : ctx_(ctx)
+      , schema_(ArraySchema(ctx, (tiledb_array_schema_t*)nullptr)) {
+    tiledb_ctx_t* c_ctx = ctx.ptr().get();
+    tiledb_array_t* array;
+    ctx.handle_error(tiledb_array_alloc(c_ctx, array_uri.c_str(), &array));
+    array_ = std::shared_ptr<tiledb_array_t>(array, deleter_);
+    ctx.handle_error(tiledb_array_open(c_ctx, array, query_type));
+
+    tiledb_array_schema_t* array_schema;
+    ctx.handle_error(tiledb_array_get_schema(c_ctx, array, &array_schema));
+    schema_ = ArraySchema(ctx, array_schema);
   }
 
   /**
@@ -112,6 +123,7 @@ class Array {
    * @param encryption_key The encryption key to use.
    * @param key_length Length in bytes of the encryption key.
    */
+  TILEDB_DEPRECATED
   Array(
       const Context& ctx,
       const std::string& array_uri,
@@ -146,6 +158,7 @@ class Array {
    * @param encryption_key The encryption key to use.
    */
   // clang-format on
+  TILEDB_DEPRECATED
   Array(
       const Context& ctx,
       const std::string& array_uri,
@@ -189,6 +202,7 @@ class Array {
    * @param query_type Query type to open the array for.
    * @param timestamp The timestamp to open the array at.
    */
+  TILEDB_DEPRECATED
   Array(
       const Context& ctx,
       const std::string& array_uri,
@@ -232,6 +246,7 @@ class Array {
    * @param timestamp The timestamp to open the array at.
    */
   // clang-format on
+  TILEDB_DEPRECATED
   Array(
       const Context& ctx,
       const std::string& array_uri,
@@ -246,6 +261,7 @@ class Array {
     tiledb_array_t* array;
     ctx.handle_error(tiledb_array_alloc(c_ctx, array_uri.c_str(), &array));
     array_ = std::shared_ptr<tiledb_array_t>(array, deleter_);
+
     ctx.handle_error(tiledb_array_open_at_with_key(
         c_ctx,
         array,
@@ -267,6 +283,7 @@ class Array {
    * See @ref Array::Array(const Context&,const std::string&,tiledb_query_type_t,tiledb_encryption_type_t,const void*,uint32_t,uint64_t) "Array::Array"
    */
   // clang-format on
+  TILEDB_DEPRECATED
   Array(
       const Context& ctx,
       const std::string& array_uri,
@@ -284,6 +301,54 @@ class Array {
             timestamp) {
   }
 
+  /**
+   * @brief Constructor. This sets the array config.
+   *
+   * **Example:**
+   *
+   * @code{.cpp}
+   * tiledb::Context ctx;
+   * tiledb_config_t* config;
+   * @endcode
+   *
+   * @param ctx TileDB context.
+   * @param carray The array.
+   * @param config The array's config.
+   */
+  Array(const Context& ctx, tiledb_array_t* carray, tiledb_config_t* config)
+      : ctx_(ctx)
+      , schema_(ArraySchema(ctx, (tiledb_array_schema_t*)nullptr)) {
+    tiledb_ctx_t* c_ctx = ctx.ptr().get();
+    ctx.handle_error(tiledb_array_set_config(c_ctx, carray, config));
+  }
+
+  /**
+   * Constructor. Creates a TileDB Array instance wrapping the given pointer.
+   * @param ctx tiledb::Context
+   * @param own=true If false, disables underlying cleanup upon destruction.
+   * @throws TileDBError if construction fails
+   */
+  Array(const Context& ctx, tiledb_array_t* carray, bool own = true)
+      : ctx_(ctx)
+      , schema_(ArraySchema(ctx, (tiledb_array_schema_t*)nullptr)) {
+    if (carray == nullptr)
+      throw TileDBError(
+          "[TileDB::C++API] Error: Failed to create Array from null pointer");
+
+    tiledb_ctx_t* c_ctx = ctx.ptr().get();
+
+    tiledb_array_schema_t* array_schema;
+    ctx.handle_error(tiledb_array_get_schema(c_ctx, carray, &array_schema));
+    schema_ = ArraySchema(ctx, array_schema);
+    owns_c_ptr_ = own;
+
+    array_ = std::shared_ptr<tiledb_array_t>(carray, [own](tiledb_array_t* p) {
+      if (own) {
+        tiledb_array_free(&p);
+      }
+    });
+  }
+
   Array(const Array&) = default;
   Array(Array&&) = default;
   Array& operator=(const Array&) = default;
@@ -291,7 +356,9 @@ class Array {
 
   /** Destructor; calls `close()`. */
   ~Array() {
-    close();
+    if (owns_c_ptr_) {
+      close();
+    }
   }
 
   /** Checks if the array is open. */
@@ -351,7 +418,13 @@ class Array {
    * @throws TileDBError if the array is already open or other error occurred.
    */
   void open(tiledb_query_type_t query_type) {
-    open(query_type, TILEDB_NO_ENCRYPTION, nullptr, 0);
+    auto& ctx = ctx_.get();
+    tiledb_ctx_t* c_ctx = ctx.ptr().get();
+    ctx.handle_error(tiledb_array_open(c_ctx, array_.get(), query_type));
+    tiledb_array_schema_t* array_schema;
+    ctx.handle_error(
+        tiledb_array_get_schema(c_ctx, array_.get(), &array_schema));
+    schema_ = ArraySchema(ctx, array_schema);
   }
 
   /**
@@ -374,6 +447,7 @@ class Array {
    * @param encryption_key The encryption key to use.
    * @param key_length Length in bytes of the encryption key.
    */
+  TILEDB_DEPRECATED
   void open(
       tiledb_query_type_t query_type,
       tiledb_encryption_type_t encryption_type,
@@ -438,6 +512,7 @@ class Array {
    * @param timestamp The timestamp to open the array at.
    * @throws TileDBError if the array is already open or other error occurred.
    */
+  TILEDB_DEPRECATED
   void open(tiledb_query_type_t query_type, uint64_t timestamp) {
     open(query_type, TILEDB_NO_ENCRYPTION, nullptr, 0, timestamp);
   }
@@ -458,7 +533,8 @@ class Array {
    * // Close and open again for reading.
    * array.close();
    * // Get some `timestamp` in milliseconds here
-   * array.open(TILEDB_READ, TILEDB_AES_256_GCM, key, sizeof(key), timestamp);
+   * array.open(TILEDB_READ, TILEDB_AES_256_GCM, key, sizeof(key),
+   * timestamp);
    * @endcode
    *
    * @param query_type The type of queries the array object will be receiving.
@@ -467,6 +543,7 @@ class Array {
    * @param key_length Length in bytes of the encryption key.
    * @param timestamp The timestamp to open the array at.
    */
+  TILEDB_DEPRECATED
   void open(
       tiledb_query_type_t query_type,
       tiledb_encryption_type_t encryption_type,
@@ -553,6 +630,7 @@ class Array {
    * @throws TileDBError if the array was not already open or other error
    * occurred.
    */
+  TILEDB_DEPRECATED
   void reopen_at(uint64_t timestamp) {
     auto& ctx = ctx_.get();
     tiledb_ctx_t* c_ctx = ctx.ptr().get();
@@ -563,7 +641,12 @@ class Array {
     schema_ = ArraySchema(ctx, array_schema);
   }
 
-  /** Returns the timestamp at which the array was opened. */
+  /**
+   * Returns the timestamp at which the array was opened.
+   *
+   * This has been deprecated, use `open_timestamp_end()` instead.
+   */
+  TILEDB_DEPRECATED
   uint64_t timestamp() const {
     auto& ctx = ctx_.get();
     uint64_t timestamp;
@@ -572,8 +655,57 @@ class Array {
     return timestamp;
   }
 
+  /** Sets the inclusive starting timestamp when opening this array. */
+  void set_open_timestamp_start(uint64_t timestamp_start) const {
+    auto& ctx = ctx_.get();
+    ctx.handle_error(tiledb_array_set_open_timestamp_start(
+        ctx.ptr().get(), array_.get(), timestamp_start));
+  }
+
+  /** Sets the inclusive ending timestamp when opening this array. */
+  void set_open_timestamp_end(uint64_t timestamp_end) const {
+    auto& ctx = ctx_.get();
+    ctx.handle_error(tiledb_array_set_open_timestamp_end(
+        ctx.ptr().get(), array_.get(), timestamp_end));
+  }
+
+  /** Retrieves the inclusive starting timestamp. */
+  uint64_t open_timestamp_start() const {
+    auto& ctx = ctx_.get();
+    uint64_t timestamp_start;
+    ctx.handle_error(tiledb_array_get_open_timestamp_start(
+        ctx.ptr().get(), array_.get(), &timestamp_start));
+    return timestamp_start;
+  }
+
+  /** Retrieves the inclusive ending timestamp. */
+  uint64_t open_timestamp_end() const {
+    auto& ctx = ctx_.get();
+    uint64_t timestamp_end;
+    ctx.handle_error(tiledb_array_get_open_timestamp_end(
+        ctx.ptr().get(), array_.get(), &timestamp_end));
+    return timestamp_end;
+  }
+
+  /** Sets the array config. */
+  void set_config(const Config& config) const {
+    auto& ctx = ctx_.get();
+    ctx.handle_error(tiledb_array_set_config(
+        ctx.ptr().get(), array_.get(), config.ptr().get()));
+  }
+
+  /** Retrieves the array config. */
+  Config config() const {
+    auto& ctx = ctx_.get();
+    tiledb_config_t* config = nullptr;
+    ctx.handle_error(
+        tiledb_array_get_config(ctx.ptr().get(), array_.get(), &config));
+    return Config(&config);
+  }
+
   /**
-   * Closes the array. The destructor calls this automatically.
+   * Closes the array. The destructor calls this automatically
+   * if the underlying pointer is owned.
    *
    * **Example:**
    * @code{.cpp}
@@ -606,7 +738,8 @@ class Array {
       const Context& ctx,
       const std::string& uri,
       Config* const config = nullptr) {
-    consolidate(ctx, uri, TILEDB_NO_ENCRYPTION, nullptr, 0, config);
+    ctx.handle_error(tiledb_array_consolidate(
+        ctx.ptr().get(), uri.c_str(), config ? config->ptr().get() : nullptr));
   }
 
   /**
@@ -636,6 +769,7 @@ class Array {
    * @param key_length Length in bytes of the encryption key.
    * @param config Configuration parameters for the consolidation.
    */
+  TILEDB_DEPRECATED
   static void consolidate(
       const Context& ctx,
       const std::string& uri,
@@ -652,6 +786,28 @@ class Array {
         config ? config->ptr().get() : nullptr));
   }
 
+  /**
+   * Cleans up the array, such as consolidated fragments and array metadata.
+   * Note that this will coarsen the granularity of time traveling (see docs
+   * for more information).
+   *
+   * **Example:**
+   * @code{.cpp}
+   * tiledb::Array::vacuum(ctx, "s3://bucket-name/array-name");
+   * @endcode
+   *
+   * @param ctx TileDB context
+   * @param array_uri The URI of the TileDB array to be vacuumed.
+   * @param config Configuration parameters for the vacuuming.
+   */
+  static void vacuum(
+      const Context& ctx,
+      const std::string& uri,
+      Config* const config = nullptr) {
+    ctx.handle_error(tiledb_array_vacuum(
+        ctx.ptr().get(), uri.c_str(), config ? config->ptr().get() : nullptr));
+  }
+
   // clang-format off
   /**
    * @copybrief Array::consolidate(const Context&,const std::string&,tiledb_encryption_type_t,const void*,uint32_t,const Config&)
@@ -697,7 +853,65 @@ class Array {
    * @param schema The array schema.
    */
   static void create(const std::string& uri, const ArraySchema& schema) {
-    create(uri, schema, TILEDB_NO_ENCRYPTION, nullptr, 0);
+    auto& ctx = schema.context();
+    tiledb_ctx_t* c_ctx = ctx.ptr().get();
+    ctx.handle_error(tiledb_array_schema_check(c_ctx, schema.ptr().get()));
+    ctx.handle_error(
+        tiledb_array_create(c_ctx, uri.c_str(), schema.ptr().get()));
+  }
+
+  /**
+   * Loads the array schema from an array.
+   *
+   * **Example:**
+   * @code{.cpp}
+   * auto schema = tiledb::Array::load_schema(ctx,
+   * "s3://bucket-name/array-name");
+   * @endcode
+   *
+   * @param ctx The TileDB context.
+   * @param uri The array URI.
+   * @return The loaded ArraySchema object.
+   */
+  static ArraySchema load_schema(const Context& ctx, const std::string& uri) {
+    tiledb_array_schema_t* schema;
+    ctx.handle_error(
+        tiledb_array_schema_load(ctx.ptr().get(), uri.c_str(), &schema));
+    return ArraySchema(ctx, schema);
+  }
+
+  /**
+   * Loads the array schema from an encrypted array.
+   *
+   * **Example:**
+   * @code{.cpp}
+   * auto schema = tiledb::Array::load_schema(ctx,
+   * "s3://bucket-name/array-name", key_type, key, key_len);
+   * @endcode
+   *
+   * @param ctx The TileDB context.
+   * @param uri The array URI.
+   * @param encryption_type The encryption type to use.
+   * @param encryption_key The encryption key to use.
+   * @param key_length Length in bytes of the encryption key.
+   * @return The loaded ArraySchema object.
+   */
+  TILEDB_DEPRECATED
+  static ArraySchema load_schema(
+      const Context& ctx,
+      const std::string& uri,
+      tiledb_encryption_type_t encryption_type,
+      const void* encryption_key,
+      uint32_t key_length) {
+    tiledb_array_schema_t* schema;
+    ctx.handle_error(tiledb_array_schema_load_with_key(
+        ctx.ptr().get(),
+        uri.c_str(),
+        encryption_type,
+        encryption_key,
+        key_length,
+        &schema));
+    return ArraySchema(ctx, schema);
   }
 
   /**
@@ -717,6 +931,7 @@ class Array {
    * @param encryption_key The encryption key to use.
    * @param key_length Length in bytes of the encryption key.
    */
+  TILEDB_DEPRECATED
   static void create(
       const std::string& uri,
       const ArraySchema& schema,
@@ -828,79 +1043,173 @@ class Array {
   }
 
   /**
-   * Compute an upper bound on the buffer elements needed to read a subarray.
+   * Retrieves the non-empty domain from the array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   *
    *
    * **Example:**
    * @code{.cpp}
    * tiledb::Context ctx;
    * tiledb::Array array(ctx, "s3://bucket-name/array-name", TILEDB_READ);
-   * std::vector<int> subarray = {0, 2, 0, 2};
-   * auto max_elements = array.max_buffer_elements(subarray);
-   *
-   * // For fixed-sized attributes, `.second` is the max number of elements
-   * // that can be read for the attribute. Use it to size the vector.
-   * std::vector<int> data_a1(max_elements["a1"].second);
-   *
-   * // In sparse reads, coords are also fixed-sized attributes.
-   * std::vector<int> coords(max_elements[TILEDB_COORDS].second);
-   *
-   * // In variable attributes, e.g. std::string type, need two buffers,
-   * // one for offsets and one for cell data.
-   * std::vector<uint64_t> offsets_a1(max_elements["a2"].first);
-   * std::vector<char> data_a1(max_elements["a2"].second);
+   * // Specify the dimension type (example uint32_t)
+   * auto non_empty = array.non_empty_domain<uint32_t>(0);
    * @endcode
    *
-   * @tparam T The domain datatype
-   * @param subarray Targeted subarray.
-   * @return A map of attribute name (including `TILEDB_COORDS`) to
-   *     the maximum number of elements that can be read in the given subarray.
-   *     For each attribute, a pair of numbers are returned. The first,
-   *     for variable-length attributes, is the maximum number of offsets
-   *     for that attribute in the given subarray. For fixed-length attributes
-   *     and coordinates, the first is always 0. The second is the maximum
-   *     number of elements for that attribute in the given subarray.
+   * @tparam T Dimension datatype
+   * @param idx The dimension index.
+   * @return The {lower, upper} pair of the non-empty domain (inclusive)
+   *         on the input dimension.
    */
   template <typename T>
-  std::unordered_map<std::string, std::pair<uint64_t, uint64_t>>
-  max_buffer_elements(const std::vector<T>& subarray) {
-    auto ctx = ctx_.get();
-    tiledb_ctx_t* c_ctx = ctx.ptr().get();
-    impl::type_check<T>(schema_.domain().type(), 1);
+  std::pair<T, T> non_empty_domain(unsigned idx) {
+    auto dim = schema_.domain().dimension(idx);
+    impl::type_check<T>(dim.type());
+    std::pair<T, T> ret;
+    std::vector<T> buf(2);
+    int empty;
 
-    // Handle attributes
-    std::unordered_map<std::string, std::pair<uint64_t, uint64_t>> ret;
-    auto schema_attrs = schema_.attributes();
-    uint64_t attr_size, type_size;
+    auto& ctx = ctx_.get();
+    ctx.handle_error(tiledb_array_get_non_empty_domain_from_index(
+        ctx.ptr().get(), array_.get(), idx, buf.data(), &empty));
 
-    for (const auto& a : schema_attrs) {
-      auto var = a.second.cell_val_num() == TILEDB_VAR_NUM;
-      auto name = a.second.name();
-      type_size = tiledb_datatype_size(a.second.type());
+    if (empty)
+      return ret;
 
-      if (var) {
-        uint64_t size_off, size_val;
-        ctx.handle_error(tiledb_array_max_buffer_size_var(
-            c_ctx,
-            array_.get(),
-            name.c_str(),
-            subarray.data(),
-            &size_off,
-            &size_val));
-        ret[a.first] = std::pair<uint64_t, uint64_t>(
-            size_off / TILEDB_OFFSET_SIZE, size_val / type_size);
-      } else {
-        ctx.handle_error(tiledb_array_max_buffer_size(
-            c_ctx, array_.get(), name.c_str(), subarray.data(), &attr_size));
-        ret[a.first] = std::pair<uint64_t, uint64_t>(0, attr_size / type_size);
-      }
-    }
+    ret = std::pair<T, T>(buf[0], buf[1]);
+    return ret;
+  }
 
-    // Handle coordinates
-    type_size = tiledb_datatype_size(schema_.domain().type());
-    ctx.handle_error(tiledb_array_max_buffer_size(
-        c_ctx, array_.get(), TILEDB_COORDS, subarray.data(), &attr_size));
-    ret[TILEDB_COORDS] =
-        std::pair<uint64_t, uint64_t>(0, attr_size / type_size);
+  /**
+   * Retrieves the non-empty domain from the array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   *
+   *
+   * **Example:**
+   * @code{.cpp}
+   * tiledb::Context ctx;
+   * tiledb::Array array(ctx, "s3://bucket-name/array-name", TILEDB_READ);
+   * // Specify the dimension type (example uint32_t)
+   * auto non_empty = array.non_empty_domain<uint32_t>("d1");
+   * @endcode
+   *
+   * @tparam T Dimension datatype
+   * @param name The dimension name.
+   * @return The {lower, upper} pair of the non-empty domain (inclusive)
+   *         on the input dimension.
+   */
+  template <typename T>
+  std::pair<T, T> non_empty_domain(const std::string& name) {
+    auto dim = schema_.domain().dimension(name);
+    impl::type_check<T>(dim.type());
+    std::pair<T, T> ret;
+    std::vector<T> buf(2);
+    int empty;
+
+    auto& ctx = ctx_.get();
+    ctx.handle_error(tiledb_array_get_non_empty_domain_from_name(
+        ctx.ptr().get(), array_.get(), name.c_str(), buf.data(), &empty));
+
+    if (empty)
+      return ret;
+
+    ret = std::pair<T, T>(buf[0], buf[1]);
+    return ret;
+  }
+
+  /**
+   * Retrieves the non-empty domain from the array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   * Applicable only to var-sized dimensions.
+   *
+   * **Example:**
+   * @code{.cpp}
+   * tiledb::Context ctx;
+   * tiledb::Array array(ctx, "s3://bucket-name/array-name", TILEDB_READ);
+   * // Specify the dimension type (example uint32_t)
+   * auto non_empty = array.non_empty_domain_var(0);
+   * @endcode
+   *
+   * @param idx The dimension index.
+   * @return The {lower, upper} pair of the non-empty domain (inclusive)
+   *         on the input dimension.
+   */
+  std::pair<std::string, std::string> non_empty_domain_var(unsigned idx) {
+    auto dim = schema_.domain().dimension(idx);
+    impl::type_check<char>(dim.type());
+    std::pair<std::string, std::string> ret;
+
+    // Get range sizes
+    uint64_t start_size, end_size;
+    int empty;
+    auto& ctx = ctx_.get();
+    ctx.handle_error(tiledb_array_get_non_empty_domain_var_size_from_index(
+        ctx.ptr().get(), array_.get(), idx, &start_size, &end_size, &empty));
+
+    if (empty)
+      return ret;
+
+    // Get ranges
+    ret.first.resize(start_size);
+    ret.second.resize(end_size);
+    ctx.handle_error(tiledb_array_get_non_empty_domain_var_from_index(
+        ctx.ptr().get(),
+        array_.get(),
+        idx,
+        &(ret.first[0]),
+        &(ret.second[0]),
+        &empty));
+
+    return ret;
+  }
+
+  /**
+   * Retrieves the non-empty domain from the array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   * Applicable only to var-sized dimensions.
+   *
+   * **Example:**
+   * @code{.cpp}
+   * tiledb::Context ctx;
+   * tiledb::Array array(ctx, "s3://bucket-name/array-name", TILEDB_READ);
+   * // Specify the dimension type (example uint32_t)
+   * auto non_empty = array.non_empty_domain_var("d1");
+   * @endcode
+   *
+   * @param name The dimension name.
+   * @return The {lower, upper} pair of the non-empty domain (inclusive)
+   *         on the input dimension.
+   */
+  std::pair<std::string, std::string> non_empty_domain_var(
+      const std::string& name) {
+    auto dim = schema_.domain().dimension(name);
+    impl::type_check<char>(dim.type());
+    std::pair<std::string, std::string> ret;
+
+    // Get range sizes
+    uint64_t start_size, end_size;
+    int empty;
+    auto& ctx = ctx_.get();
+    ctx.handle_error(tiledb_array_get_non_empty_domain_var_size_from_name(
+        ctx.ptr().get(),
+        array_.get(),
+        name.c_str(),
+        &start_size,
+        &end_size,
+        &empty));
+
+    if (empty)
+      return ret;
+
+    // Get ranges
+    ret.first.resize(start_size);
+    ret.second.resize(end_size);
+    ctx.handle_error(tiledb_array_get_non_empty_domain_var_from_name(
+        ctx.ptr().get(),
+        array_.get(),
+        name.c_str(),
+        &(ret.first[0]),
+        &(ret.second[0]),
+        &empty));
 
     return ret;
   }
@@ -935,7 +1244,14 @@ class Array {
       const Context& ctx,
       const std::string& uri,
       Config* const config = nullptr) {
-    consolidate_metadata(ctx, uri, TILEDB_NO_ENCRYPTION, nullptr, 0, config);
+    Config local_cfg;
+    Config* config_aux = config;
+    if (!config_aux) {
+      config_aux = &local_cfg;
+    }
+
+    (*config_aux)["sm.consolidation.mode"] = "array_meta";
+    consolidate(ctx, uri, config_aux);
   }
 
   /**
@@ -965,6 +1281,7 @@ class Array {
    * @param key_length Length in bytes of the encryption key.
    * @param config Configuration parameters for the consolidation.
    */
+  TILEDB_DEPRECATED
   static void consolidate_metadata(
       const Context& ctx,
       const std::string& uri,
@@ -972,18 +1289,20 @@ class Array {
       const void* encryption_key,
       uint32_t key_length,
       Config* const config = nullptr) {
-    ctx.handle_error(tiledb_array_consolidate_metadata_with_key(
-        ctx.ptr().get(),
-        uri.c_str(),
-        encryption_type,
-        encryption_key,
-        key_length,
-        config ? config->ptr().get() : nullptr));
+    Config local_cfg;
+    Config* config_aux = config;
+    if (!config_aux) {
+      config_aux = &local_cfg;
+    }
+
+    (*config_aux)["sm.consolidation.mode"] = "array_meta";
+    consolidate(
+        ctx, uri, encryption_type, encryption_key, key_length, config_aux);
   }
 
   // clang-format off
   /**
-   * @copybrief Array::consolidate_metadata(const Context&,const std::string&,tiledb_encryption_type_t,const void*,uint32_t,const Config&)
+   * @copybrief Array::consolidate_metadata(const Context&, const std::string&, tiledb_encryption_type_t, const void*,uint32_t, const Config&)
    *
    * See @ref Array::consolidate_metadata(
    *     const Context&,
@@ -1006,13 +1325,43 @@ class Array {
       tiledb_encryption_type_t encryption_type,
       const std::string& encryption_key,
       Config* const config = nullptr) {
-    return consolidate_metadata(
+    Config local_cfg;
+    Config* config_aux = config;
+    if (!config_aux) {
+      config_aux = &local_cfg;
+    }
+
+    (*config_aux)["sm.consolidation.mode"] = "array_meta";
+    consolidate(
         ctx,
         uri,
         encryption_type,
         encryption_key.data(),
         (uint32_t)encryption_key.size(),
-        config);
+        config_aux);
+  }
+
+  /**
+   * @brief Upgrades an array to the latest format version.
+   *
+   *
+   * **Example:**
+   * @code{.cpp}
+   * tiledb::Array::upgrade_version(ctx, "array_name");
+   * @endcode
+   *
+   * @param ctx TileDB context
+   * @param array_uri The URI of the TileDB array to be upgraded.
+   * @param config Configuration parameters for the upgrade.
+   */
+  static void upgrade_version(
+      const Context& ctx,
+      const std::string& array_uri,
+      Config* const config = nullptr) {
+    ctx.handle_error(tiledb_array_upgrade_version(
+        ctx.ptr().get(),
+        array_uri.c_str(),
+        config ? config->ptr().get() : nullptr));
   }
 
   /**
@@ -1167,6 +1516,9 @@ class Array {
   /** Pointer to the TileDB C array object. */
   std::shared_ptr<tiledb_array_t> array_;
 
+  /** Flag indicating ownership of the TileDB C array object */
+  bool owns_c_ptr_ = true;
+
   /** The array schema. */
   ArraySchema schema_;
 
