@@ -1,5 +1,5 @@
 /**
- * @file   tiledb_cpp_api_config.h
+ * @file   config.h
  *
  * @author Ravi Gaddipati
  *
@@ -7,7 +7,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -38,8 +38,10 @@
 #include "tiledb.h"
 #include "utils.h"
 
+#include <map>
 #include <memory>
 #include <string>
+#include <unordered_map>
 
 namespace tiledb {
 
@@ -47,10 +49,13 @@ class Config;  // Forward decl for impl classes
 
 namespace impl {
 
-class ConfigIter : public std::iterator<
-                       std::forward_iterator_tag,
-                       const std::pair<std::string, std::string>> {
+class ConfigIter {
  public:
+  using iterator_category = std::forward_iterator_tag;
+  using value_type = const std::pair<std::string, std::string>;
+  using difference_type = std::ptrdiff_t;
+  using pointer = value_type*;
+  using reference = value_type&;
   /* ********************************* */
   /*     CONSTRUCTORS & DESTRUCTORS    */
   /* ********************************* */
@@ -197,6 +202,31 @@ class Config {
     }
   }
 
+  /**
+   * Constructor that takes as input a STL map that stores the config parameters
+   *
+   * @param config
+   */
+  explicit Config(const std::map<std::string, std::string>& config) {
+    create_config();
+    for (const auto& kv : config) {
+      set(kv.first, kv.second);
+    }
+  }
+
+  /**
+   * Constructor that takes as input a STL unordered_map that stores the config
+   * parameters
+   *
+   * @param config
+   */
+  explicit Config(const std::unordered_map<std::string, std::string>& config) {
+    create_config();
+    for (const auto& kv : config) {
+      set(kv.first, kv.second);
+    }
+  }
+
   /* ********************************* */
   /*                API                */
   /* ********************************* */
@@ -208,6 +238,24 @@ class Config {
     impl::check_config_error(err);
   }
 
+  /** Compares configs for equality. */
+  bool operator==(const Config& rhs) const {
+    uint8_t equal;
+    int rc = tiledb_config_compare(config_.get(), rhs.config_.get(), &equal);
+    if (rc != TILEDB_OK)
+      throw std::runtime_error(
+          "[TileDB::C++API] Error: Failed to compare configurations");
+    if (equal == 1)
+      return true;
+    else
+      return false;
+  }
+
+  /** Compares configs for inequality. */
+  bool operator!=(const Config& rhs) const {
+    return !(*this == rhs);
+  }
+
   /** Returns the pointer to the TileDB C config object. */
   std::shared_ptr<tiledb_config_t> ptr() const {
     return config_;
@@ -219,20 +267,24 @@ class Config {
    *
    * - `sm.dedup_coords` <br>
    *    If `true`, cells with duplicate coordinates will be removed during
-   *    sparse array writes. Note that ties during deduplication are
-   *    arbitrary. <br>
+   *    sparse fragment writes. Note that ties during deduplication are broken
+   *    arbitrarily. <br>
    *    **Default**: false
    * - `sm.check_coord_dups` <br>
    *    This is applicable only if `sm.dedup_coords` is `false`.
    *    If `true`, an error will be thrown if there are cells with duplicate
-   *    coordinates during sparse array writes. If `false` and there are
-   *    duplicates, the duplicates will be written without errors, but the
-   *    TileDB behavior could be unpredictable. <br>
+   *    coordinates during sparse fragment writes. If `false` and there are
+   *    duplicates, the duplicates will be written without errors. <br>
    *    **Default**: true
    * - `sm.check_coord_oob` <br>
    *    If `true`, an error will be thrown if there are cells with coordinates
-   *    falling outside the array domain during sparse array writes. <br>
+   *    falling outside the array domain during sparse fragment writes. <br>
    *    **Default**: true
+   *    `sm.read_range_oob` <br>
+   *    If `error`, this will check ranges for read with out-of-bounds on the
+   *    dimension domain's and error. If `warn`, the ranges will be capped at
+   *    the dimension's domain and a warning logged. <br>
+   *    **Default**: warn
    * - `sm.check_global_order` <br>
    *    Checks if the coordinates obey the global array order. Applicable only
    *    to sparse writes in global order.
@@ -242,32 +294,40 @@ class Config {
    *    **Default**: 10,000,000
    * - `sm.array_schema_cache_size` <br>
    *    Array schema cache size in bytes. Any `uint64_t` value is acceptable.
-   * <br>
-   *    **Default**: 10,000,000
-   * - `sm.fragment_metadata_cache_size` <br>
-   *    The fragment metadata cache size in bytes. Any `uint64_t` value is
-   *    acceptable. <br>
+   *    <br>
    *    **Default**: 10,000,000
    * - `sm.enable_signal_handlers` <br>
    *    Whether or not TileDB will install signal handlers. <br>
    *    **Default**: true
-   * - `sm.num_async_threads` <br>
-   *    The number of threads allocated for async queries. <br>
-   *    **Default**: 1
-   * - `sm.num_reader_threads` <br>
-   *    The number of threads allocated for issuing reads to VFS in
-   *    parallel. <br>
-   *    **Default**: 1
-   * - `sm.num_writer_threads` <br>
-   *    The number of threads allocated for issuing writes to VFS in
-   *    parallel.<br>
-   *    **Default**: 1
-   * - `sm.num_tbb_threads` <br>
-   *    The number of threads allocated for the TBB thread pool (if TBB is
-   *    enabled). Note: this is a whole-program setting. Usually this should not
-   *    be modified from the default. See also the documentation for TBB's
-   *    `task_scheduler_init` class.<br>
-   *    **Default**: TBB automatic
+   * - `sm.compute_concurrency_level` <br>
+   *    Upper-bound on number of threads to allocate for compute-bound tasks.
+   * <br>
+   *    **Default*: # cores
+   * - `sm.io_concurrency_level` <br>
+   *    Upper-bound on number of threads to allocate for IO-bound tasks. <br>
+   *    **Default*: # cores
+   * - `sm.vacuum.mode` <br>
+   *    The vacuuming mode, one of `fragments` (remove consolidated fragments),
+   *    `fragment_meta` (remove only consolidated fragment metadata), or
+   *    `array_meta` (remove consolidated array metadata files). <br>
+   *    **Default**: fragments
+   * - `sm.vacuum.timestamp_start` <br>
+   *    **Experimental** <br>
+   *    When set, an array will be vacuumed between this value and
+   *    `sm.vacuum.timestamp_end` (inclusive). <br>
+   *    Only for `fragments` and `array_meta` vacuum mode. <br>
+   *    **Default**: 0
+   * - `sm.vacuum.timestamp_end` <br>
+   *    **Experimental** <br>
+   *    When set, an array will be vacuumed between `sm.vacuum.timestamp_start`
+   *    and this value (inclusive). <br>
+   *    Only for `fragments` and `array_meta` vacuum mode. <br>
+   *    **Default**: UINT64_MAX
+   * - `sm.consolidation_mode` <br>
+   *    The consolidation mode, one of `fragments` (consolidate all fragments),
+   *    `fragment_meta` (consolidate only fragment metadata footers to a single
+   *    file), or `array_meta` (consolidate array metadata only). <br>
+   *    **Default**: "fragments"
    * - `sm.consolidation.amplification` <br>
    *    The factor by which the size of the dense fragment resulting
    *    from consolidating a set of fragments (containing at least one
@@ -295,6 +355,18 @@ class Config {
    *    The size ratio that two ("adjacent") fragments must satisfy to be
    *    considered for consolidation in a single step.<br>
    *    **Default**: 0.0
+   * - `sm.consolidation.timestamp_start` <br>
+   *    **Experimental** <br>
+   *    When set, an array will be consolidated between this value and
+   *    `sm.consolidation.timestamp_end` (inclusive). <br>
+   *    Only for `fragments` and `array_meta` consolidation mode. <br>
+   *    **Default**: 0
+   * - `sm.consolidation.timestamp_end` <br>
+   *    **Experimental** <br>
+   *    When set, an array will be consolidated between
+   *    `sm.consolidation.timestamp_start` and this value (inclusive). <br>
+   *    Only for `fragments` and `array_meta` consolidation mode. <br>
+   *    **Default**: UINT64_MAX
    * - `sm.memory_budget` <br>
    *    The memory budget for tiles of fixed-sized attributes (or offsets for
    *    var-sized attributes) to be fetched during reads.<br>
@@ -303,10 +375,78 @@ class Config {
    *    The memory budget for tiles of var-sized attributes
    *    to be fetched during reads.<br>
    *    **Default**: 10GB
-   * - `vfs.num_threads` <br>
-   *    The number of threads allocated for VFS operations (any backend), per
-   *    VFS instance. <br>
-   *    **Default**: number of cores
+   * - `sm.var_offsets.bitsize` <br>
+   *    The size of offsets in bits to be used for offset buffers of var-sized
+   *    attributes<br>
+   *    **Default**: 64
+   * - `sm.var_offsets.extra_element` <br>
+   *    Add an extra element to the end of the offsets buffer of var-sized
+   *    attributes which will point to the end of the values buffer.<br>
+   *    **Default**: false
+   * - `sm.var_offsets.mode` <br>
+   *    The offsets format (`bytes` or `elements`) to be used for
+   *    var-sized attributes.<br>
+   *    **Default**: bytes
+   * - `sm.query.dense.reader` <br>
+   *    Which reader to use for dense queries. "refactored" or "legacy".<br>
+   *    **Default**: lagacy
+   * - `sm.query.sparse_global_order.reader` <br>
+   *    Which reader to use for sparse global order queries. "refactored"
+   *    or "legacy".<br>
+   *    **Default**: legacy
+   * - `sm.query.sparse_unordered_with_dups.reader` <br>
+   *    Which reader to use for sparse unordered with dups queries.
+   *    "refactored" or "legacy".<br>
+   *    **Default**: refactored
+   * - `sm.mem.malloc_trim` <br>
+   *    Should malloc_trim be called on context and query destruction? This
+   *    might reduce residual memory usage. <br>
+   *    **Default**: true
+   * - `sm.mem.total_budget` <br>
+   *    Memory budget for readers and writers. <br>
+   *    **Default**: 10GB
+   * - `sm.mem.reader.sparse_global_order.ratio_coords` <br>
+   *    Ratio of the budget allocated for coordinates in the sparse global
+   *    order reader. <br>
+   *    **Default**: 0.5
+   * - `sm.mem.reader.sparse_global_order.ratio_query_condition` <br>
+   *    Ratio of the budget allocated for the query condition in the sparse
+   *    global order reader. <br>
+   *    **Default**: 0.25
+   * - `sm.mem.reader.sparse_global_order.ratio_tile_ranges` <br>
+   *    Ratio of the budget allocated for tile ranges in the sparse global
+   *    order reader. <br>
+   *    **Default**: 0.1
+   * - `sm.mem.reader.sparse_global_order.ratio_array_data` <br>
+   *    Ratio of the budget allocated for array data in the sparse global
+   *    order reader. <br>
+   *    **Default**: 0.1
+   * - `sm.mem.reader.sparse_global_order.ratio_rcs` <br>
+   *    Ratio of the budget allocated for result cell slabs in the sparse
+   *    global order reader. <br>
+   *    **Default**: 0.05
+   * - `sm.mem.reader.sparse_unordered_with_dups.ratio_coords` <br>
+   *    Ratio of the budget allocated for coordinates in the sparse unordered
+   *    with duplicates reader. <br>
+   *    **Default**: 0.5
+   * - `sm.mem.reader.sparse_unordered_with_dups.ratio_query_condition` <br>
+   *    Ratio of the budget allocated for the query condition in the sparse
+   *    unordered with duplicates reader. <br>
+   *    **Default**: 0.25
+   * - `sm.mem.reader.sparse_unordered_with_dups.ratio_tile_ranges` <br>
+   *    Ratio of the budget allocated for tile ranges in the sparse unordered
+   *    with duplicates reader. <br>
+   *    **Default**: 0.1
+   * - `sm.mem.reader.sparse_unordered_with_dups.ratio_array_data` <br>
+   *    Ratio of the budget allocated for array data in the sparse unordered
+   *    with duplicates reader. <br>
+   *    **Default**: 0.1
+   *    The maximum byte size to read-ahead from the backend. <br>
+   *    **Default**: 102400
+   * -  `vfs.read_ahead_cache_size` <br>
+   *    The the total maximum size of the read-ahead cache, which is an LRU.
+   *    <br>
+   *    **Default**: 10485760
    * - `vfs.min_parallel_size` <br>
    *    The minimum number of bytes in a parallel VFS operation
    *    (except parallel S3 writes, which are controlled by
@@ -318,14 +458,67 @@ class Config {
    * - `vfs.min_batch_gap` <br>
    *    The minimum number of bytes between two VFS read batches.<br>
    *    **Default**: 500KB
+   * - `vfs.file.posix_file_permissions` <br>
+   *    permissions to use for posix file system with file or dir creation.<br>
+   *    **Default**: 644
+   * - `vfs.file.posix_directory_permissions` <br>
+   *    permissions to use for posix file system with file or dir creation.<br>
+   *    **Default**: 755
    * - `vfs.file.max_parallel_ops` <br>
    *    The maximum number of parallel operations on objects with `file:///`
    *    URIs. <br>
-   *    **Default**: `vfs.num_threads`
+   *    **Default**: `sm.io_concurrency_level`
    * - `vfs.file.enable_filelocks` <br>
    *    If set to `false`, file locking operations are no-ops for `file:///`
    *    URIs in VFS. <br>
    *    **Default**: `true`
+   * - `vfs.azure.storage_account_name` <br>
+   *    Set the Azure Storage Account name. <br>
+   *    **Default**: ""
+   * - `vfs.azure.storage_account_key` <br>
+   *    Set the Azure Storage Account key. <br>
+   *    **Default**: ""
+   * - `vfs.azure.storage_sas_token` <br>
+   *    Set the Azure Storage SAS (shared access signature) token. <br>
+   *    **Default**: ""
+   * - `vfs.azure.blob_endpoint` <br>
+   *    Overrides the default Azure Storage Blob endpoint. If empty, the
+   *    endpoint will be constructed from the storage account name. This
+   *    should not include an http:// or https:// prefix. <br>
+   *    **Default**: ""
+   * - `vfs.azure.block_list_block_size` <br>
+   *    The block size (in bytes) used in Azure blob block list writes.
+   *    Any `uint64_t` value is acceptable. Note:
+   *    `vfs.azure.block_list_block_size * vfs.azure.max_parallel_ops` bytes
+   *    will be buffered before issuing block uploads in parallel. <br>
+   *    **Default**: "5242880"
+   * - `vfs.azure.max_parallel_ops` <br>
+   *    The maximum number of Azure backend parallel operations. <br>
+   *    **Default**: `sm.io_concurrency_level`
+   * - `vfs.azure.use_block_list_upload` <br>
+   *    Determines if the Azure backend can use chunked block uploads. <br>
+   *    **Default**: "true"
+   * - `vfs.azure.use_https` <br>
+   *    Determines if the blob endpoint should use HTTP or HTTPS.
+   *    **Default**: "true"
+   * - `vfs.gcs.project_id` <br>
+   *    Set the GCS project id. <br>
+   *    **Default**: ""
+   * - `vfs.gcs.multi_part_size` <br>
+   *    The part size (in bytes) used in GCS multi part writes.
+   *    Any `uint64_t` value is acceptable. Note:
+   *    `vfs.gcs.multi_part_size * vfs.gcs.max_parallel_ops` bytes will
+   *    be buffered before issuing part uploads in parallel. <br>
+   *    **Default**: "5242880"
+   * - `vfs.gcs.max_parallel_ops` <br>
+   *    The maximum number of GCS backend parallel operations. <br>
+   *    **Default**: `sm.io_concurrency_level`
+   * - `vfs.gcs.use_multi_part_upload` <br>
+   *    Determines if the GCS backend can use chunked part uploads. <br>
+   *    **Default**: "true"
+   * - `vfs.gcs.request_timeout_ms` <br>
+   *    The maximum amount of time to retry network requests to GCS. <br>
+   *    **Default**: "3000"
    * - `vfs.s3.region` <br>
    *    The S3 region, if S3 is enabled. <br>
    *    **Default**: us-east-1
@@ -338,6 +531,23 @@ class Config {
    * - `vfs.s3.aws_session_token` <br>
    *    Set the AWS_SESSION_TOKEN <br>
    *    **Default**: ""
+   * - `vfs.s3.aws_role_arn` <br>
+   *    Determines the role that we want to assume.
+   *    Set the AWS_ROLE_ARN <br>
+   *    **Default**: ""
+   * - `vfs.s3.aws_external_id` <br>
+   *    Third party access ID to your resources when assuming a role.
+   *    Set the AWS_EXTERNAL_ID <br>
+   *    **Default**: ""
+   * - `vfs.s3.aws_load_frequency` <br>
+   *    Session time limit when assuming a role.
+   *    Set the AWS_LOAD_FREQUENCY <br>
+   *    **Default**: ""
+   * - `vfs.s3.aws_session_name` <br>
+   *    (Optional) session name when assuming a role.
+   *    Can be used for tracing and bookkeeping.
+   *    Set the AWS_SESSION_NAME <br>
+   *    **Default**: ""
    * - `vfs.s3.scheme` <br>
    *    The S3 scheme (`http` or `https`), if S3 is enabled. <br>
    *    **Default**: https
@@ -348,13 +558,16 @@ class Config {
    *    The S3 use of virtual addressing (`true` or `false`), if S3 is
    *    enabled. <br>
    *    **Default**: true
-   * - `vfs.s3.use_virtual_addressing` <br>
-   *    The S3 use of virtual addressing (`true` or `false`), if S3 is
+   * - `vfs.s3.skip_init` <br>
+   *    Skip Aws::InitAPI for the S3 layer (`true` or `false`) <br>
+   *    **Default**: false
+   * - `vfs.s3.use_multipart_upload` <br>
+   *    The S3 use of multi-part upload requests (`true` or `false`), if S3 is
    *    enabled. <br>
    *    **Default**: true
    * - `vfs.s3.max_parallel_ops` <br>
    *    The maximum number of S3 backend parallel operations. <br>
-   *    **Default**: `vfs.num_threads`
+   *    **Default**: `sm.io_concurrency_level`
    * - `vfs.s3.multipart_part_size` <br>
    *    The part size (in bytes) used in S3 multipart writes.
    *    Any `uint64_t` value is acceptable. Note: `vfs.s3.multipart_part_size *
@@ -389,6 +602,9 @@ class Config {
    * - `vfs.s3.request_timeout_ms` <br>
    *    The request timeout in ms. Any `long` value is acceptable. <br>
    *    **Default**: 3000
+   * - `vfs.s3.requester_pays` <br>
+   *    The requester pays for the S3 access charges. <br>
+   *    **Default**: false
    * - `vfs.s3.proxy_host` <br>
    *    The proxy host. <br>
    *    **Default**: ""
@@ -397,7 +613,7 @@ class Config {
    *    **Default**: 0
    * - `vfs.s3.proxy_scheme` <br>
    *    The proxy scheme. <br>
-   *    **Default**: "https"
+   *    **Default**: "http"
    * - `vfs.s3.proxy_username` <br>
    *    The proxy username. Note: this parameter is not serialized by
    *    `tiledb_config_save_to_file`. <br>
@@ -408,8 +624,32 @@ class Config {
    *    **Default**: ""
    * - `vfs.s3.verify_ssl` <br>
    *    Enable HTTPS certificate verification. <br>
-   *    **Default**: true""
-   * - `vfs.hdfs.name_node"` <br>
+   *    **Default**: true
+   * - `vfs.s3.sse` <br>
+   *    The server-side encryption algorithm to use. Supported non-empty
+   *    values are "aes256" and "kms" (AWS key management service). <br>
+   *    **Default**: ""
+   * - `vfs.s3.bucket_canned_acl` <br>
+   *    Names of values found in Aws::S3::Model::BucketCannedACL enumeration.
+   *    "NOT_SET"
+   *    "private_"
+   *    "public_read"
+   *    "public_read_write"
+   *    "authenticated_read"
+   *    **Default**: "NOT_SET"
+   * - `vfs.s3.object_canned_acl` <br>
+   *    Names of values found in Aws::S3::Model::ObjectCannedACL enumeration.
+   *    (The first 5 are the same as for "vfs.s3.bucket_canned_acl".)
+   *    "NOT_SET"
+   *    "private_"
+   *    "public_read"
+   *    "public_read_write"
+   *    "authenticated_read"
+   *    (The following three items are found only in
+   *     Aws::S3::Model::ObjectCannedACL.) "aws_exec_read" "owner_read"
+   *    "bucket_owner_full_control"
+   *    **Default**: "NOT_SET"
+   * - `vfs.hdfs.name_node_uri"` <br>
    *    Name node for HDFS. <br>
    *    **Default**: ""
    * - `vfs.hdfs.username` <br>
@@ -418,6 +658,61 @@ class Config {
    * - `vfs.hdfs.kerb_ticket_cache_path` <br>
    *    HDFS kerb ticket cache path. <br>
    *    **Default**: ""
+   * - `config.env_var_prefix` <br>
+   *    Prefix of environmental variables for reading configuration
+   *    parameters. <br>
+   *    **Default**: "TILEDB_"
+   * - `config.logging_level` <br>
+   *    The logging level configured, possible values: "0": fatal, "1": error,
+   *    "2": warn, "3": info "4": debug, "5": trace <br>
+   *    **Default**: "1" if --enable-verbose bootstrap flag is provided,
+   *    "0" otherwise <br>
+   * - `config.logging_format` <br>
+   *    The logging format configured (DEFAULT or JSON)
+   *    **Default**: "DEFAULT"
+   * - `rest.server_address` <br>
+   *    URL for REST server to use for remote arrays. <br>
+   *    **Default**: "https://api.tiledb.com"
+   * - `rest.server_serialization_format` <br>
+   *    Serialization format to use for remote array requests (CAPNP or
+   *    JSON). <br>
+   *    **Default**: "CAPNP"
+   * - `rest.username` <br>
+   *    Username for login to REST server. <br>
+   *    **Default**: ""
+   * - `rest.password` <br>
+   *    Password for login to REST server. <br>
+   *    **Default**: ""
+   * - `rest.token` <br>
+   *    Authentication token for REST server (used instead of
+   *    username/password). <br>
+   *    **Default**: ""
+   * - `rest.resubmit_incomplete` <br>
+   *    If true, incomplete queries received from server are automatically
+   *    resubmitted before returning to user control. <br>
+   *    **Default**: "true"
+   * - `rest.ignore_ssl_validation` <br>
+   *    Have curl ignore ssl peer and host validation for REST server. <br>
+   *    **Default**: false
+   * - `rest.creation_access_credentials_name` <br>
+   *    The name of the registered access key to use for creation of the REST
+   *    server. <br>
+   *    **Default**: no default set
+   * -  `rest.retry_http_codes` <br>
+   *    CSV list of http status codes to automatically retry a REST request for
+   *    <br>
+   *    **Default**: "503"
+   * - `rest.retry_count` <br>
+   *    Number of times to retry failed REST requests <br>
+   *    **Default**: 3
+   * - `rest.retry_initial_delay_ms` <br>
+   *    Initial delay in milliseconds to wait until retrying a REST request
+   *    <br>
+   *    **Default**: 500
+   * - `rest.retry_delay_factor` <br>
+   *    The delay factor to exponentially wait until further retries of a
+   *    failed REST request <br>
+   *    **Default**: 1.25
    */
   Config& set(const std::string& param, const std::string& value) {
     tiledb_error_t* err;
