@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -41,16 +41,22 @@
 #ifdef HAVE_HDFS
 
 #include "tiledb/sm/filesystem/hdfs_filesystem.h"
+#include "tiledb/common/logger.h"
+#include "tiledb/sm/config/config.h"
 #include "tiledb/sm/misc/constants.h"
-#include "tiledb/sm/misc/logger.h"
+#include "tiledb/sm/misc/uri.h"
 #include "tiledb/sm/misc/utils.h"
 
 #include <dlfcn.h>
 #include <cassert>
+#include <climits>
 #include <fstream>
+#include <functional>
 #include <iostream>
 #include <sstream>
 
+using namespace tiledb::common;
+
 namespace tiledb {
 namespace sm {
 
@@ -356,6 +362,11 @@ Status HDFS::is_file(const URI& uri, bool* is_file) {
 }
 
 Status HDFS::touch(const URI& uri) {
+  if (uri.to_string().back() == '/') {
+    return LOG_STATUS(Status::HDFSError(std::string(
+        "Cannot create file; URI is a directory: " + uri.to_string())));
+  }
+
   hdfsFS fs = nullptr;
   RETURN_NOT_OK(connect(&fs));
   hdfsFile writeFile =
