@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -33,19 +33,24 @@
 #ifndef _WIN32
 
 #include "tiledb/sm/filesystem/posix.h"
+#include "tiledb/common/logger.h"
+#include "tiledb/common/thread_pool.h"
 #include "tiledb/sm/misc/constants.h"
-#include "tiledb/sm/misc/logger.h"
-#include "tiledb/sm/misc/stats.h"
 #include "tiledb/sm/misc/utils.h"
 
 #include <dirent.h>
-
+#include <fcntl.h>
 #include <limits.h>
-
-#include <ftw.h>
+#include <sys/stat.h>
+#include <unistd.h>
 
 #include <fstream>
+#include <future>
 #include <iostream>
+#include <queue>
+#include <sstream>
+
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
@@ -107,6 +112,23 @@ void Posix::adjacent_slashes_dedup(std::string* path) {
 }
 
 std::string Posix::abs_path(const std::string& path) {
+  std::string resolved_path = abs_path_internal(path);
+
+  // Ensure the returned has the same postfix slash as 'path'.
+  if (utils::parse::ends_with(path, "/")) {
+    if (!utils::parse::ends_with(resolved_path, "/")) {
+      resolved_path = resolved_path + "/";
+    }
+  } else {
+    if (utils::parse::ends_with(resolved_path, "/")) {
+      resolved_path = resolved_path.substr(0, resolved_path.length() - 1);
+    }
+  }
+
+  return resolved_path;
+}
+
+std::string Posix::abs_path_internal(const std::string& path) {
   // Initialize current, home and root
   std::string current = current_dir();
   auto env_home_ptr = getenv("HOME");
@@ -148,7 +170,11 @@ Status Posix::create_dir(const std::string& path) const {
         std::string("Cannot create directory '") + path +
         "'; Directory already exists"));
   }
-  if (mkdir(path.c_str(), S_IRWXU) != 0) {
+
+  uint32_t permissions = 0;
+  RETURN_NOT_OK(get_posix_directory_permissions(&permissions));
+
+  if (mkdir(path.c_str(), permissions) != 0) {
     return LOG_STATUS(Status::IOError(
         std::string("Cannot create directory '") + path + "'; " +
         strerror(errno)));
@@ -157,7 +183,10 @@ Status Posix::create_dir(const std::string& path) const {
 }
 
 Status Posix::touch(const std::string& filename) const {
-  int fd = ::open(filename.c_str(), O_WRONLY | O_CREAT | O_SYNC, S_IRWXU);
+  uint32_t permissions = 0;
+  RETURN_NOT_OK(get_posix_file_permissions(&permissions));
+
+  int fd = ::open(filename.c_str(), O_WRONLY | O_CREAT | O_SYNC, permissions);
   if (fd == -1 || ::close(fd) != 0) {
     return LOG_STATUS(Status::IOError(
         std::string("Failed to create file '") + filename + "'; " +
@@ -240,7 +269,7 @@ Status Posix::filelock_lock(
   fl.l_pid = getpid();
 
   // Open the file
-  *fd = ::open(filename.c_str(), O_RDWR);
+  *fd = ::open(filename.c_str(), shared ? O_RDONLY : O_WRONLY);
   if (*fd == -1) {
     return LOG_STATUS(Status::IOError(
         "Cannot open filelock '" + filename + "'; " + strerror(errno)));
@@ -294,8 +323,8 @@ Status Posix::ls(
   while ((next_path = readdir(dir)) != nullptr) {
     if (!strcmp(next_path->d_name, ".") || !strcmp(next_path->d_name, ".."))
       continue;
-    auto abspath = path + "/" + next_path->d_name;
-    paths->push_back(abspath);
+    std::string abspath = path + "/" + next_path->d_name;
+    paths->emplace_back(abspath);
   }
   // close parent directory
   if (closedir(dir) != 0) {
@@ -314,6 +343,45 @@ Status Posix::move_path(
   return Status::Ok();
 }
 
+Status Posix::copy_file(
+    const std::string& old_path, const std::string& new_path) {
+  std::ifstream src(old_path, std::ios::binary);
+  std::ofstream dst(new_path, std::ios::binary);
+  dst << src.rdbuf();
+  return Status::Ok();
+}
+
+Status Posix::copy_dir(
+    const std::string& old_path, const std::string& new_path) {
+  RETURN_NOT_OK(create_dir(new_path));
+  std::vector<std::string> paths;
+  RETURN_NOT_OK(ls(old_path, &paths));
+
+  std::queue<std::string> path_queue;
+  for (auto& path : paths)
+    path_queue.emplace(std::move(path));
+
+  while (!path_queue.empty()) {
+    std::string file_name_abs = path_queue.front();
+    std::string file_name = file_name_abs.substr(old_path.length() + 1);
+    path_queue.pop();
+
+    if (is_dir(file_name_abs)) {
+      RETURN_NOT_OK(create_dir(new_path + "/" + file_name));
+      std::vector<std::string> child_paths;
+      RETURN_NOT_OK(ls(file_name_abs, &child_paths));
+      for (auto& path : child_paths)
+        path_queue.emplace(std::move(path));
+    } else {
+      assert(is_file(file_name_abs));
+      RETURN_NOT_OK(
+          copy_file(old_path + "/" + file_name, new_path + "/" + file_name));
+    }
+  }
+
+  return Status::Ok();
+}
+
 void Posix::purge_dots_from_path(std::string* path) {
   // Trivial case
   if (path == nullptr)
@@ -411,13 +479,17 @@ Status Posix::read(
 }
 
 Status Posix::sync(const std::string& path) {
+  uint32_t permissions = 0;
+
   // Open file
   int fd = -1;
-  if (is_dir(path))  // DIRECTORY
-    fd = open(path.c_str(), O_RDONLY, S_IRWXU);
-  else if (is_file(path))  // FILE
-    fd = open(path.c_str(), O_WRONLY | O_APPEND | O_CREAT, S_IRWXU);
-  else
+  if (is_dir(path)) {  // DIRECTORY
+    RETURN_NOT_OK(get_posix_directory_permissions(&permissions));
+    fd = open(path.c_str(), O_RDONLY, permissions);
+  } else if (is_file(path)) {  // FILE
+    RETURN_NOT_OK(get_posix_file_permissions(&permissions));
+    fd = open(path.c_str(), O_WRONLY | O_APPEND | O_CREAT, permissions);
+  } else
     return Status::Ok();  // If file does not exist, exit
 
   // Handle error
@@ -457,6 +529,9 @@ Status Posix::write(
       "vfs.file.max_parallel_ops", &max_parallel_ops, &found));
   assert(found);
 
+  uint32_t permissions = 0;
+  RETURN_NOT_OK(get_posix_file_permissions(&permissions));
+
   // Get file offset (equal to file size)
   Status st;
   uint64_t file_offset = 0;
@@ -470,7 +545,7 @@ Status Posix::write(
   }
 
   // Open or create file.
-  int fd = open(path.c_str(), O_WRONLY | O_CREAT, S_IRWXU);
+  int fd = open(path.c_str(), O_WRONLY | O_CREAT, permissions);
   if (fd == -1) {
     return LOG_STATUS(Status::IOError(
         std::string("Cannot open file '") + path + "'; " + strerror(errno)));
@@ -489,8 +564,7 @@ Status Posix::write(
       return LOG_STATUS(Status::IOError(errmsg.str()));
     }
   } else {
-    STATS_COUNTER_ADD(vfs_posix_write_num_parallelized, 1);
-    std::vector<std::future<Status>> results;
+    std::vector<ThreadPool::Task> results;
     uint64_t thread_write_nbytes = utils::math::ceil(buffer_size, num_ops);
     for (uint64_t i = 0; i < num_ops; i++) {
       uint64_t begin = i * thread_write_nbytes,
@@ -499,7 +573,7 @@ Status Posix::write(
       uint64_t thread_nbytes = end - begin + 1;
       uint64_t thread_file_offset = file_offset + begin;
       auto thread_buffer = reinterpret_cast<const char*>(buffer) + begin;
-      results.push_back(vfs_thread_pool_->enqueue(
+      results.emplace_back(vfs_thread_pool_->execute(
           [fd, thread_file_offset, thread_buffer, thread_nbytes]() {
             return write_at(
                 fd, thread_file_offset, thread_buffer, thread_nbytes);
@@ -551,6 +625,32 @@ Status Posix::write_at(
   return Status::Ok();
 }
 
+Status Posix::get_posix_file_permissions(uint32_t* permissions) const {
+  // Get config params
+  bool found = false;
+  std::string posix_permissions =
+      config_.get().get("vfs.file.posix_file_permissions", &found);
+  assert(found);
+
+  // Permissions are passed in octal notation by the user
+  *permissions = std::strtol(posix_permissions.c_str(), NULL, 8);
+
+  return Status::Ok();
+}
+
+Status Posix::get_posix_directory_permissions(uint32_t* permissions) const {
+  // Get config params
+  bool found = false;
+  std::string posix_permissions =
+      config_.get().get("vfs.file.posix_directory_permissions", &found);
+  assert(found);
+
+  // Permissions are passed in octal notation by the user
+  *permissions = std::strtol(posix_permissions.c_str(), NULL, 8);
+
+  return Status::Ok();
+}
+
 }  // namespace sm
 }  // namespace tiledb
 
