@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -34,18 +34,21 @@
 #define TILEDB_S3_H
 
 #ifdef HAVE_S3
+#include "tiledb/common/rwlock.h"
+#include "tiledb/common/status.h"
+#include "tiledb/common/thread_pool.h"
 #include "tiledb/sm/buffer/buffer.h"
 #include "tiledb/sm/config/config.h"
 #include "tiledb/sm/filesystem/s3_thread_pool_executor.h"
 #include "tiledb/sm/misc/constants.h"
-#include "tiledb/sm/misc/status.h"
-#include "tiledb/sm/misc/thread_pool.h"
 #include "tiledb/sm/misc/uri.h"
+#include "tiledb/sm/stats/global_stats.h"
+#include "tiledb/sm/stats/stats.h"
 
 #include <aws/core/Aws.h>
 #include <aws/core/auth/AWSCredentialsProviderChain.h>
 #include <aws/core/client/ClientConfiguration.h>
-#include <aws/core/client/DefaultRetryStrategy.h>
+#include <aws/core/client/RetryStrategy.h>
 #include <aws/core/http/HttpClient.h>
 #include <aws/core/utils/HashingUtils.h>
 #include <aws/core/utils/Outcome.h>
@@ -53,7 +56,9 @@
 #include <aws/core/utils/UUID.h>
 #include <aws/core/utils/memory/stl/AWSStringStream.h>
 #include <aws/core/utils/ratelimiter/DefaultRateLimiter.h>
+#include <aws/core/utils/stream/PreallocatedStreamBuf.h>
 #include <aws/core/utils/threading/Executor.h>
+#include <aws/identity-management/auth/STSAssumeRoleCredentialsProvider.h>
 #include <aws/s3/S3Client.h>
 #include <aws/s3/model/CompleteMultipartUploadRequest.h>
 #include <aws/s3/model/CopyObjectRequest.h>
@@ -72,6 +77,8 @@
 #include <string>
 #include <vector>
 
+using namespace tiledb::common;
+
 namespace tiledb {
 namespace sm {
 
@@ -91,6 +98,9 @@ class S3 {
   /** Destructor. */
   ~S3();
 
+  DISABLE_COPY_AND_COPY_ASSIGN(S3);
+  DISABLE_MOVE_AND_MOVE_ASSIGN(S3);
+
   /* ********************************* */
   /*                 API               */
   /* ********************************* */
@@ -98,11 +108,15 @@ class S3 {
   /**
    * Initializes and connects an S3 client.
    *
+   * @param parent_stats The parent stats.
    * @param config Configuration parameters.
    * @param thread_pool The parent VFS thread pool.
    * @return Status
    */
-  Status init(const Config& config, ThreadPool* thread_pool);
+  Status init(
+      stats::Stats* parent_stats,
+      const Config& config,
+      ThreadPool* thread_pool);
 
   /**
    * Creates a bucket.
@@ -137,9 +151,11 @@ class S3 {
    * Check if a bucket exists.
    *
    * @param bucket The name of the bucket.
-   * @return bool
+   * @param exists Mutates to `true` if `uri` is an existing bucket,
+   *   and `false` otherwise.
+   * @return Status
    */
-  bool is_bucket(const URI& uri) const;
+  Status is_bucket(const URI& uri, bool* exists) const;
 
   /**
    * Checks if there is an object with prefix `uri/`. For instance, suppose
@@ -165,9 +181,11 @@ class S3 {
    * Checks if the given URI is an existing S3 object.
    *
    * @param uri The URI of the object to be checked.
-   * @return `True` if `uri` is an existing object, and `false` otherwise.
+   * @param exists Mutates to `true` if `uri` is an existing object,
+   *   and `false` otherwise.
+   * @return Status
    */
-  bool is_object(const URI& uri) const;
+  Status is_object(const URI& uri, bool* exists) const;
 
   /**
    * Lists the objects that start with `prefix`. Full URI paths are
@@ -183,7 +201,7 @@ class S3 {
    * and the delimiter is `/`, the returned URIs will be
    *
    * - `foo/boo`
-   * - `foo/bar/`
+   * - `foo/bar`
    *
    * @param prefix The prefix URI.
    * @param paths Pointer of a vector of URIs to store the retrieved paths.
@@ -219,6 +237,24 @@ class S3 {
    */
   Status move_dir(const URI& old_uri, const URI& new_uri);
 
+  /**
+   * Copies a file.
+   *
+   * @param old_uri The URI of the old path.
+   * @param new_uri The URI of the new path.
+   * @return Status
+   */
+  Status copy_file(const URI& old_uri, const URI& new_uri);
+
+  /**
+   * Copies a directory. All subdirectories and files are copied.
+   *
+   * @param old_uri The URI of the old path.
+   * @param new_uri The URI of the new path.
+   * @return Status
+   */
+  Status copy_dir(const URI& old_uri, const URI& new_uri);
+
   /**
    * Returns the size of the input object with a given URI in bytes.
    *
@@ -235,10 +271,17 @@ class S3 {
    * @param offset The offset in the object from which the read will start.
    * @param buffer The buffer into which the data will be written.
    * @param length The size of the data to be read from the object.
+   * @param read_ahead_length The additional length to read ahead.
+   * @param length_returned Returns the total length read into `buffer`.
    * @return Status
    */
   Status read(
-      const URI& uri, off_t offset, void* buffer, uint64_t length) const;
+      const URI& uri,
+      off_t offset,
+      void* buffer,
+      uint64_t length,
+      uint64_t read_ahead_length,
+      uint64_t* length_returned) const;
 
   /**
    * Deletes a bucket.
@@ -273,7 +316,7 @@ class S3 {
    * - `s3://some_bucket/foo/bar2/bar3
    * - `s3://some_bucket/foo/bar4
    *
-   * In contrast, `remove("s3://some_bucket/foo2")` will not delete anything;
+   * In contrast, `remove("s3://some_bucket/foo2")` will not delete anything,
    * the function internally appends `/` to the end of the URI, and therefore
    * there is not object with prefix "s3://some_bucket/foo2/" in this example.
    *
@@ -311,6 +354,83 @@ class S3 {
    */
   enum State { UNINITIALIZED, INITIALIZED, DISCONNECTED };
 
+  /**
+   * The retry strategy for S3 request failures.
+   */
+  class S3RetryStrategy : public Aws::Client::RetryStrategy {
+   public:
+    /** Constructor. */
+    S3RetryStrategy(
+        stats::Stats* const s3_stats,
+        const uint64_t max_retries,
+        const uint64_t scale_factor)
+        : s3_stats_(s3_stats)
+        , max_retries_(max_retries)
+        , scale_factor_(scale_factor) {
+    }
+
+    /*
+     * Returns true if the error can be retried given the error and
+     * the number of times already tried.
+     */
+    bool ShouldRetry(
+        const Aws::Client::AWSError<Aws::Client::CoreErrors>& error,
+        long attempted_retries) const override {
+      // Unconditionally retry on 'SLOW_DOWN' errors. The request
+      // will eventually succeed.
+      if (error.GetErrorType() == Aws::Client::CoreErrors::SLOW_DOWN) {
+        // With an average retry interval of 1.5 seconds, 100 retries
+        // would be a 2.5 minute hang, which is unreasonably long. Error
+        // out in this scenario, as we probably have encountered a
+        // programming error.
+        if (attempted_retries == 100) {
+          return false;
+        }
+
+        s3_stats_->add_counter("vfs_s3_slow_down_retries", 1);
+
+        return true;
+      }
+
+      if (static_cast<uint64_t>(attempted_retries) >= max_retries_)
+        return false;
+
+      return error.ShouldRetry();
+    }
+
+    /**
+     * Calculates the time in milliseconds the client should sleep
+     * before attempting another request based on the error and attempted
+     * retries.
+     */
+    long CalculateDelayBeforeNextRetry(
+        const Aws::Client::AWSError<Aws::Client::CoreErrors>& error,
+        long attempted_retries) const override {
+      // Unconditionally retry 'SLOW_DOWN' errors between 1.25 and 1.75
+      // seconds. The random 0.5 second range is to reduce the likelyhood
+      // of starving a single client when multiple clients are performing
+      // requests on the same prefix.
+      if (error.GetErrorType() == Aws::Client::CoreErrors::SLOW_DOWN) {
+        return 1250 + rand() % 500;
+      }
+
+      if (attempted_retries == 0)
+        return 0;
+
+      return (1L << attempted_retries) * static_cast<long>(scale_factor_);
+    }
+
+   private:
+    /** The S3 `stats_`. */
+    stats::Stats* s3_stats_;
+
+    /** The maximum number of retries after an error. */
+    uint64_t max_retries_;
+
+    /** The scale of each exponential backoff delay. */
+    uint64_t scale_factor_;
+  };
+
   /**
    * This struct wraps the context state of a pending multipart upload request.
    */
@@ -334,6 +454,9 @@ class S3 {
 
   /** Contains all state associated with a multipart upload transaction. */
   struct MultiPartUploadState {
+    MultiPartUploadState()
+        : part_number(0)
+        , st(Status::Ok()){};
     /** Constructor. */
     MultiPartUploadState(
         const int in_part_number,
@@ -346,11 +469,39 @@ class S3 {
         , key(std::move(in_key))
         , upload_id(std::move(in_upload_id))
         , completed_parts(std::move(in_completed_parts))
-        , st(Status::Ok()) {
+        , st(Status::Ok()){};
+
+    MultiPartUploadState(MultiPartUploadState&& other) noexcept {
+      this->part_number = other.part_number;
+      this->bucket = std::move(other.bucket);
+      this->key = std::move(other.key);
+      this->upload_id = std::move(other.upload_id);
+      this->completed_parts = std::move(other.completed_parts);
+      this->st = other.st;
+    }
+
+    // Copy initialization
+    MultiPartUploadState(const MultiPartUploadState& other) {
+      this->part_number = other.part_number;
+      this->bucket = other.bucket;
+      this->key = other.key;
+      this->upload_id = other.upload_id;
+      this->completed_parts = other.completed_parts;
+      this->st = other.st;
+    }
+
+    MultiPartUploadState& operator=(const MultiPartUploadState& other) {
+      this->part_number = other.part_number;
+      this->bucket = other.bucket;
+      this->key = other.key;
+      this->upload_id = other.upload_id;
+      this->completed_parts = other.completed_parts;
+      this->st = other.st;
+      return *this;
     }
 
     /** The current part number. */
-    int part_number;
+    uint64_t part_number;
 
     /** The AWS bucket. */
     Aws::String bucket;
@@ -366,12 +517,44 @@ class S3 {
 
     /** The overall status of the multipart request. */
     Status st;
+
+    /** Mutex for thread safety */
+    mutable std::mutex mtx;
+  };
+
+  /**
+   * Used to stream results from the GetObject request into
+   * a pre-allocated buffer.
+   */
+  class PreallocatedIOStream : public Aws::IOStream {
+   public:
+    /**
+     * Value constructor. Does not take ownership of the
+     * input buffer.
+     *
+     * @param buffer The pre-allocated underlying buffer.
+     * @param size The maximum size of the underlying buffer.
+     */
+    PreallocatedIOStream(void* const buffer, const size_t size)
+        : Aws::IOStream(new Aws::Utils::Stream::PreallocatedStreamBuf(
+              reinterpret_cast<unsigned char*>(buffer), size)) {
+    }
+
+    /** Destructor. */
+    ~PreallocatedIOStream() {
+      // Delete the unmanaged `Aws::Utils::Stream::PreallocatedStreamBuf`
+      // that was allocated in the constructor.
+      delete rdbuf();
+    }
   };
 
   /* ********************************* */
   /*         PRIVATE ATTRIBUTES        */
   /* ********************************* */
 
+  /** The class stats. */
+  stats::Stats* stats_;
+
   /** The current state. */
   State state_;
 
@@ -379,7 +562,11 @@ class S3 {
    * The lazily-initialized S3 client. This is mutable so that nominally const
    * functions can call init_client().
    */
-  mutable std::shared_ptr<Aws::S3::S3Client> client_;
+  mutable tdb_shared_ptr<Aws::S3::S3Client> client_;
+
+  /** The AWS credetial provider. */
+  mutable tdb_shared_ptr<Aws::Auth::AWSCredentialsProvider>
+      credentials_provider_;
 
   /**
    * Mutex protecting client initialization. This is mutable so that nominally
@@ -388,13 +575,10 @@ class S3 {
   mutable std::mutex client_init_mtx_;
 
   /** Configuration object used to initialize the client. */
-  std::unique_ptr<Aws::Client::ClientConfiguration> client_config_;
+  mutable tdb_unique_ptr<Aws::Client::ClientConfiguration> client_config_;
 
-  /** The executor  used by 'client_'. */
-  std::shared_ptr<S3ThreadPoolExecutor> s3_tp_executor_;
-
-  /** Credentials object used to initialize the client. */
-  std::unique_ptr<Aws::Auth::AWSCredentials> client_creds_;
+  /** The executor used by 'client_'. */
+  mutable std::shared_ptr<S3ThreadPoolExecutor> s3_tp_executor_;
 
   /** The size of the file buffers used in multipart uploads. */
   uint64_t file_buffer_size_;
@@ -407,7 +591,7 @@ class S3 {
       multipart_upload_states_;
 
   /** Protects 'multipart_upload_states_'. */
-  std::mutex multipart_upload_mtx_;
+  RWLock multipart_upload_rwlock_;
 
   /** The maximum number of parallel operations issued. */
   uint64_t max_parallel_ops_;
@@ -430,6 +614,27 @@ class S3 {
   /** Whether or not to use multipart upload. */
   bool use_multipart_upload_;
 
+  /** Config stored from init for lazy client_init. */
+  Config config_;
+
+  /** Set the request payer for a s3 request. */
+  Aws::S3::Model::RequestPayer request_payer_;
+
+  /** The server-side encryption algorithm. */
+  Aws::S3::Model::ServerSideEncryption sse_;
+
+  /** The server-side encryption kms key. */
+  std::string sse_kms_key_id_;
+
+  /** Protects file_buffers map */
+  std::mutex file_buffers_mtx_;
+
+  /** If !NOT_SET assign to object requests supporting SetACL() */
+  Aws::S3::Model::ObjectCannedACL object_canned_acl_;
+
+  /** If !NOT_SET assign to bucket requests supporting SetACL() */
+  Aws::S3::Model::BucketCannedACL bucket_canned_acl_;
+
   /* ********************************* */
   /*          PRIVATE METHODS          */
   /* ********************************* */
@@ -479,6 +684,12 @@ class S3 {
    */
   std::string remove_front_slash(const std::string& path) const;
 
+  /**
+   * Returns the input `path` after removing a potential `/` character
+   * from the end if it exists.
+   */
+  std::string remove_trailing_slash(const std::string& path) const;
+
   /**
    * Writes the contents of the input buffer to the S3 object given by
    * the input `uri` as a new series of multipart uploads. It then
@@ -505,7 +716,8 @@ class S3 {
    * Initiates a new multipart upload request for the input URI. Note: the
    * caller must hold the multipart data structure mutex.
    */
-  Status initiate_multipart_request(Aws::Http::URI aws_uri);
+  Status initiate_multipart_request(
+      Aws::Http::URI aws_uri, MultiPartUploadState* state);
 
   /**
    * Return the given authority and path strings joined with a '/'
@@ -514,16 +726,22 @@ class S3 {
   std::string join_authority_and_path(
       const std::string& authority, const std::string& path) const;
 
+  /** Checks if the given object exists on S3. */
+  Status is_object(
+      const Aws::String& bucket_name,
+      const Aws::String& object_key,
+      bool* exists) const;
+
   /** Waits for the input object to be propagated. */
-  bool wait_for_object_to_propagate(
+  Status wait_for_object_to_propagate(
       const Aws::String& bucketName, const Aws::String& objectKey) const;
 
   /** Waits for the input object to be deleted. */
-  bool wait_for_object_to_be_deleted(
+  Status wait_for_object_to_be_deleted(
       const Aws::String& bucketName, const Aws::String& objectKey) const;
 
   /** Waits for the bucket to be created. */
-  bool wait_for_bucket_to_be_created(const URI& bucket_uri) const;
+  Status wait_for_bucket_to_be_created(const URI& bucket_uri) const;
 
   /**
    * Builds and returns a CompleteMultipartUploadRequest for completing an
@@ -576,7 +794,7 @@ class S3 {
    * Writes the input buffer to a file by issuing one or more multipart upload
    * requests. If the file does not exist, then it is created. If the file
    * exists then it is appended to. This command will upload chunks of an
-   * in-progress write each time the parallelisation buffer size is exceeded
+   * in-progress write each time the parallelization buffer size is exceeded
    * (calculated as product of  `multipart_part_size`` * ``max_parallel_ops``
    *  configuration options).
    *
