@@ -78,6 +78,8 @@ using namespace tiledb::common;
 namespace tiledb {
 namespace sm {
 
+class Tile;
+
 enum class Filesystem : uint8_t;
 enum class VFSMode : uint8_t;
 
@@ -364,7 +366,7 @@ class VFS {
    */
   Status read_all(
       const URI& uri,
-      const std::vector<std::tuple<uint64_t, void*, uint64_t>>& regions,
+      const std::vector<std::tuple<uint64_t, Tile*, uint64_t>>& regions,
       ThreadPool* thread_pool,
       std::vector<ThreadPool::Task>* tasks,
       bool use_read_ahead = true);
@@ -432,7 +434,7 @@ class VFS {
    */
   struct BatchedRead {
     /** Construct a BatchedRead consisting of the single given region. */
-    BatchedRead(const std::tuple<uint64_t, void*, uint64_t>& region) {
+    BatchedRead(const std::tuple<uint64_t, Tile*, uint64_t>& region) {
       offset = std::get<0>(region);
       nbytes = std::get<2>(region);
       regions.push_back(region);
@@ -448,7 +450,7 @@ class VFS {
      * Original regions making up the batch. Vector of tuples of the form
      * (offset, dest_buffer, nbytes).
      */
-    std::vector<std::tuple<uint64_t, void*, uint64_t>> regions;
+    std::vector<std::tuple<uint64_t, Tile*, uint64_t>> regions;
   };
 
   /**
@@ -674,7 +676,7 @@ class VFS {
    * @return Status
    */
   Status compute_read_batches(
-      const std::vector<std::tuple<uint64_t, void*, uint64_t>>& regions,
+      const std::vector<std::tuple<uint64_t, Tile*, uint64_t>>& regions,
       std::vector<BatchedRead>* batches) const;
 
   /**
