@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -32,17 +32,24 @@
 #ifdef _WIN32
 
 #include "tiledb/sm/filesystem/win.h"
+#include "tiledb/common/heap_memory.h"
+#include "tiledb/common/logger.h"
 #include "tiledb/sm/misc/constants.h"
-#include "tiledb/sm/misc/logger.h"
-#include "tiledb/sm/misc/stats.h"
 #include "tiledb/sm/misc/utils.h"
 
+#if !defined(NOMINMAX)
+#define NOMINMAX  // suppress definition of min/max macros in Windows headers
+#endif
 #include <Shlwapi.h>
 #include <Windows.h>
 #include <wininet.h>  // For INTERNET_MAX_URL_LENGTH
+#include <algorithm>
 #include <cassert>
 #include <fstream>
 #include <iostream>
+#include <sstream>
+
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
@@ -127,13 +134,13 @@ Status Win::touch(const std::string& filename) const {
 std::string Win::current_dir() {
   std::string dir;
   unsigned long length = GetCurrentDirectory(0, nullptr);
-  char* path = (char*)std::malloc(length * sizeof(char));
+  char* path = (char*)tdb_malloc(length * sizeof(char));
   if (path == nullptr || GetCurrentDirectory(length, path) == 0) {
     LOG_STATUS(
         Status::IOError(std::string("Failed to get current directory.")));
   }
   dir = path;
-  std::free(path);
+  tdb_free(path);
   return dir;
 }
 
@@ -245,7 +252,7 @@ Status Win::filelock_lock(
     return LOG_STATUS(Status::IOError(
         std::string("Failed to lock '" + filename + "'; CreateFile error")));
   }
-  OVERLAPPED overlapped = {0};
+  OVERLAPPED overlapped = {0, 0, {{0, 0}}, 0};
   if (LockFileEx(
           file_h,
           shared ? 0 : LOCKFILE_EXCLUSIVE_LOCK,
@@ -264,7 +271,7 @@ Status Win::filelock_lock(
 }
 
 Status Win::filelock_unlock(filelock_t fd) const {
-  OVERLAPPED overlapped = {0};
+  OVERLAPPED overlapped = {0, 0, {{0, 0}}, 0};
   if (UnlockFileEx(fd, 0, MAXDWORD, MAXDWORD, &overlapped) == 0) {
     CloseHandle(fd);
     return LOG_STATUS(
@@ -470,8 +477,7 @@ Status Win::write(
           Status::IOError(std::string("Cannot write to file '") + path));
     }
   } else {
-    STATS_COUNTER_ADD(vfs_win32_write_num_parallelized, 1);
-    std::vector<std::future<Status>> results;
+    std::vector<ThreadPool::Task> results;
     uint64_t thread_write_nbytes = utils::math::ceil(buffer_size, num_ops);
     for (uint64_t i = 0; i < num_ops; i++) {
       uint64_t begin = i * thread_write_nbytes,
@@ -480,7 +486,7 @@ Status Win::write(
       uint64_t thread_nbytes = end - begin + 1;
       uint64_t thread_file_offset = file_offset + begin;
       auto thread_buffer = reinterpret_cast<const char*>(buffer) + begin;
-      results.push_back(vfs_thread_pool_->enqueue(
+      results.push_back(vfs_thread_pool_->execute(
           [file_h, thread_file_offset, thread_buffer, thread_nbytes]() {
             return write_at(
                 file_h, thread_file_offset, thread_buffer, thread_nbytes);
@@ -518,7 +524,7 @@ Status Win::write_at(
   while (buffer_size > constants::max_write_bytes) {
     LARGE_INTEGER offset;
     offset.QuadPart = file_offset;
-    OVERLAPPED ov = {0};
+    OVERLAPPED ov = {0, 0, {{0, 0}}, 0};
     ov.Offset = offset.LowPart;
     ov.OffsetHigh = offset.HighPart;
     if (WriteFile(
@@ -538,7 +544,7 @@ Status Win::write_at(
   }
   LARGE_INTEGER offset;
   offset.QuadPart = file_offset;
-  OVERLAPPED ov = {0};
+  OVERLAPPED ov = {0, 0, {{0, 0}}, 0};
   ov.Offset = offset.LowPart;
   ov.OffsetHigh = offset.HighPart;
   if (WriteFile(
@@ -559,7 +565,7 @@ std::string Win::uri_from_path(const std::string& path) {
   unsigned long uri_length = INTERNET_MAX_URL_LENGTH;
   char uri[INTERNET_MAX_URL_LENGTH];
   std::string str_uri;
-  if (UrlCreateFromPath(path.c_str(), uri, &uri_length, NULL) != S_OK) {
+  if (UrlCreateFromPath(path.c_str(), uri, &uri_length, 0) != S_OK) {
     LOG_STATUS(Status::IOError(
         std::string("Failed to convert path '" + path + "' to URI.")));
   }
@@ -573,12 +579,18 @@ std::string Win::path_from_uri(const std::string& uri) {
   }
 
   std::string uri_with_scheme =
-      utils::parse::starts_with(uri, "file:///") ? uri : "file:///" + uri;
+      (utils::parse::starts_with(uri, "file://") ||
+       // also accept 'file:/x...'
+       (utils::parse::starts_with(uri, "file:/") && uri.substr(6, 1) != "/")) ?
+          uri
+          // else treat as file: item on 'localhost' (empty host name)
+          :
+          "file:///" + uri;
 
   unsigned long path_length = MAX_PATH;
   char path[MAX_PATH];
   std::string str_path;
-  if (PathCreateFromUrl(uri_with_scheme.c_str(), path, &path_length, NULL) !=
+  if (PathCreateFromUrl(uri_with_scheme.c_str(), path, &path_length, 0) !=
       S_OK) {
     LOG_STATUS(Status::IOError(std::string(
         "Failed to convert URI '" + uri_with_scheme + "' to path.")));
