@@ -258,7 +258,8 @@ Status FilterPipeline::filter_chunks_reverse(
     total_size += std::get<2>(input[i]);
   }
 
-  RETURN_NOT_OK(output->realloc(total_size));
+  if (total_size != output->alloced_size())
+    RETURN_NOT_OK(output->realloc(total_size));
 
   // Run each chunk through the entire pipeline.
   auto status = parallel_for(compute_tp, 0, input.size(), [&](uint64_t i) {
@@ -311,6 +312,8 @@ Status FilterPipeline::filter_chunks_reverse(
             output_chunk_buffer, orig_chunk_len));
       }
 
+      f->init_resource_pool(compute_tp->concurrency_level());
+
       RETURN_NOT_OK(f->run_reverse(
           &input_metadata,
           &input_data,
@@ -541,7 +544,7 @@ void FilterPipeline::swap(FilterPipeline& other) {
 Status FilterPipeline::append_encryption_filter(
     FilterPipeline* pipeline, const EncryptionKey& encryption_key) {
   switch (encryption_key.encryption_type()) {
-    case EncryptionType ::NO_ENCRYPTION:
+    case EncryptionType::NO_ENCRYPTION:
       return Status::Ok();
     case EncryptionType::AES_256_GCM:
       return pipeline->add_filter(EncryptionAES256GCMFilter(encryption_key));
