@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -71,6 +71,9 @@ const unsigned rtree_fanout = 10;
 /** The array schema file name. */
 const std::string array_schema_filename = "__array_schema.tdb";
 
+/** The array schema folder name. */
+const std::string array_schema_folder_name = "__schema";
+
 /** The array metadata folder name. */
 const std::string array_metadata_folder_name = "__meta";
 
@@ -86,7 +89,13 @@ const uint64_t cell_var_offset_size = sizeof(uint64_t);
 /** The type of a variable cell offset. */
 const Datatype cell_var_offset_type = Datatype::UINT64;
 
-/** A special value indicating varibale size. */
+/** The size of a validity cell. */
+const uint64_t cell_validity_size = sizeof(uint8_t);
+
+/** The type of a validity cell. */
+const Datatype cell_validity_type = Datatype::UINT8;
+
+/** A special value indicating variable size. */
 const uint64_t var_size = std::numeric_limits<uint64_t>::max();
 
 /** The default compressor for the offsets of variable-sized cells. */
@@ -95,6 +104,12 @@ Compressor cell_var_offsets_compression = Compressor::ZSTD;
 /** The default compression level for the offsets of variable-sized cells. */
 int cell_var_offsets_compression_level = -1;
 
+/** The default compressor for the validity value cells. */
+Compressor cell_validity_compression = Compressor::RLE;
+
+/** The default compression level for the validity value cells. */
+int cell_validity_compression_level = -1;
+
 /** Special name reserved for the coordinates attribute. */
 const std::string coords = "__coords";
 
@@ -167,6 +182,15 @@ const uint8_t empty_any = 0;
 /** The file suffix used in TileDB. */
 const std::string file_suffix = ".tdb";
 
+/** Suffix for the special vacuum files used in TileDB. */
+const std::string vacuum_file_suffix = ".vac";
+
+/** Suffix for the special ok files used in TileDB. */
+const std::string ok_file_suffix = ".ok";
+
+/** Suffix for the special metadata files used in TileDB. */
+const std::string meta_file_suffix = ".meta";
+
 /** Default datatype for a generic tile. */
 const Datatype generic_tile_datatype = Datatype::CHAR;
 
@@ -204,6 +228,7 @@ const std::string query_type_read_str = "READ";
 
 /** TILEDB_WRITE Query String **/
 const std::string query_type_write_str = "WRITE";
+
 /** TILEDB_FAILED Query String **/
 const std::string query_status_failed_str = "FAILED";
 
@@ -219,6 +244,33 @@ const std::string query_status_incomplete_str = "INCOMPLETE";
 /** TILEDB_UNINITIALIZED Query String **/
 const std::string query_status_uninitialized_str = "UNINITIALIZED";
 
+/** TILEDB_LT Query Condition Op String **/
+const std::string query_condition_op_lt_str = "LT";
+
+/** TILEDB_LE Query Condition Op String **/
+const std::string query_condition_op_le_str = "LE";
+
+/** TILEDB_GT Query Condition Op String **/
+const std::string query_condition_op_gt_str = "GT";
+
+/** TILEDB_GE Query Condition Op String **/
+const std::string query_condition_op_ge_str = "GE";
+
+/** TILEDB_EQ Query Condition Op String **/
+const std::string query_condition_op_eq_str = "EQ";
+
+/** TILEDB_NE Query Condition Op String **/
+const std::string query_condition_op_ne_str = "NE";
+
+/** TILEDB_AND Query Condition Combination Op String **/
+const std::string query_condition_combination_op_and_str = "AND";
+
+/** TILEDB_OR Query Condition Combination Op String **/
+const std::string query_condition_combination_op_or_str = "OR";
+
+/** TILEDB_NOT Query Condition Combination Op String **/
+const std::string query_condition_combination_op_not_str = "NOT";
+
 /** TILEDB_COMPRESSION Filter type string */
 const std::string filter_type_compression_str = "COMPRESSION";
 
@@ -378,6 +430,33 @@ const std::string datetime_fs_str = "DATETIME_FS";
 /** The string representation for type DATETIME_AS. */
 const std::string datetime_as_str = "DATETIME_AS";
 
+/** The string representation for type TIME_HR. */
+const std::string time_hr_str = "TIME_HR";
+
+/** The string representation for type TIME_MIN. */
+const std::string time_min_str = "TIME_MIN";
+
+/** The string representation for type TIME_SEC. */
+const std::string time_sec_str = "TIME_SEC";
+
+/** The string representation for type TIME_MS. */
+const std::string time_ms_str = "TIME_MS";
+
+/** The string representation for type TIME_US. */
+const std::string time_us_str = "TIME_US";
+
+/** The string representation for type TIME_NS. */
+const std::string time_ns_str = "TIME_NS";
+
+/** The string representation for type TIME_PS. */
+const std::string time_ps_str = "TIME_PS";
+
+/** The string representation for type TIME_FS. */
+const std::string time_fs_str = "TIME_FS";
+
+/** The string representation for type TIME_AS. */
+const std::string time_as_str = "TIME_AS";
+
 /** The string representation for the dense array type. */
 const std::string dense_str = "dense";
 
@@ -396,6 +475,9 @@ const std::string global_order_str = "global-order";
 /** The string representation for the unordered layout. */
 const std::string unordered_str = "unordered";
 
+/** The string representation for the Hilbert layout. */
+const std::string hilbert_str = "hilbert";
+
 /** The string representation of null. */
 const std::string null_str = "null";
 
@@ -414,6 +496,15 @@ const std::string filesystem_type_hdfs_str = "HDFS";
 /** The string representation for filesystem type s3. */
 const std::string filesystem_type_s3_str = "S3";
 
+/** The string representation for filesystem type Azure. */
+const std::string filesystem_type_azure_str = "AZURE";
+
+/** The string representation for filesystem type GCS. */
+const std::string filesystem_type_gcs_str = "GCS";
+
+/** The string representation for in-memory filesystem */
+const std::string filesystem_type_mem_str = "MEM";
+
 /** The string representation for WalkOrder preorder. */
 const std::string walkorder_preorder_str = "PREORDER";
 
@@ -434,46 +525,38 @@ const int32_t library_version[3] = {
     TILEDB_VERSION_MAJOR, TILEDB_VERSION_MINOR, TILEDB_VERSION_PATCH};
 
 /** The TileDB serialization format version number. */
-const uint32_t format_version = 4;
+const uint32_t format_version = 10;
+
+/** The lowest version supported for back compat writes. */
+const uint32_t back_compat_writes_min_format_version = 7;
 
 /** The maximum size of a tile chunk (unit of compression) in bytes. */
 const uint64_t max_tile_chunk_size = 64 * 1024;
 
-/** The default attribute name prefix. */
-const std::string default_attr_name = "__attr";
-
-/** The default dimension name prefix. */
-const std::string default_dim_name = "__dim";
-
-/** The key attribute name. */
-const std::string key_attr_name = "__key";
-
-/** The key attribute type. */
-Datatype key_attr_type = Datatype::ANY;
-
-/** The key attribute compressor. */
-Compressor key_attr_compressor = Compressor::ZSTD;
-
 /**
- * The name of the first key dimension (recall that a key in a
- * key-value store is hashed into a 16-byte MD5 digest, which
- * is represented as a 2-dimensional uint64_t value.
+ * Maximum splitting depth on a string dimension. For a maximum depth
+ * of D, the maximum number of splits is 2^D.
  */
-const std::string key_dim_1 = "__key_dim_1";
-
-/**
- * The name of the second key dimension (recall that a key in a
- * key-value store is hashed into a 16-byte MD5 digest, which
- * is represented as a 2-dimensional uint64_t value.
- */
-const std::string key_dim_2 = "__key_dim_2";
+const uint64_t max_string_dim_split_depth = 7;
 
 /** Maximum number of attempts to wait for an S3 response. */
-const unsigned int s3_max_attempts = 1000;
+const unsigned int s3_max_attempts = 100;
 
 /** Milliseconds of wait time between S3 attempts. */
 const unsigned int s3_attempt_sleep_ms = 100;
 
+/** Maximum number of attempts to wait for an Azure response. */
+const unsigned int azure_max_attempts = 10;
+
+/** Milliseconds of wait time between Azure attempts. */
+const unsigned int azure_attempt_sleep_ms = 1000;
+
+/** Maximum number of attempts to wait for a GCS response. */
+const unsigned int gcs_max_attempts = 100;
+
+/** Milliseconds of wait time between GCS attempts. */
+const unsigned int gcs_attempt_sleep_ms = 1000;
+
 /** An allocation tag used for logging. */
 const std::string s3_allocation_tag = "TileDB";
 
@@ -483,6 +566,12 @@ const std::string special_name_prefix = "__";
 /** Number of milliseconds between watchdog thread wakeups. */
 const unsigned watchdog_thread_sleep_ms = 1000;
 
+/** Maximum number of concurrent attribute reads. */
+const unsigned concurrent_attr_reads = 2;
+
+/** The redirection header key in REST response. */
+extern const std::string redirection_header_key = "location";
+
 const void* fill_value(Datatype type) {
   switch (type) {
     case Datatype::INT8:
@@ -534,6 +623,15 @@ const void* fill_value(Datatype type) {
     case Datatype::DATETIME_PS:
     case Datatype::DATETIME_FS:
     case Datatype::DATETIME_AS:
+    case Datatype::TIME_HR:
+    case Datatype::TIME_MIN:
+    case Datatype::TIME_SEC:
+    case Datatype::TIME_MS:
+    case Datatype::TIME_US:
+    case Datatype::TIME_NS:
+    case Datatype::TIME_PS:
+    case Datatype::TIME_FS:
+    case Datatype::TIME_AS:
       return &constants::empty_int64;
   }
 
@@ -554,6 +652,8 @@ const std::array<std::string, 6> cert_files_linux = {
     "/etc/ssl/cert.pem"                                   // Alpine Linux
 };
 #endif
+
+const std::string config_delimiter = ",";
 }  // namespace constants
 
 }  // namespace sm
