@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -65,18 +65,27 @@ extern const std::string filelock_name;
 /** The array schema file name. */
 extern const std::string array_schema_filename;
 
+/** The array schema folder name. */
+extern const std::string array_schema_folder_name;
+
 /** The array metadata folder name. */
 extern const std::string array_metadata_folder_name;
 
 /** The default tile capacity. */
 extern const uint64_t capacity;
 
-/** The size of a variable cell offset. */
+/** The size of a variable offset cell. */
 extern const uint64_t cell_var_offset_size;
 
-/** The type of a variable cell offset. */
+/** The type of a variable offset cell. */
 extern const Datatype cell_var_offset_type;
 
+/** The size of a validity cell. */
+extern const uint64_t cell_validity_size;
+
+/** The type of a validity cell. */
+extern const Datatype cell_validity_type;
+
 /** A special value indicating varibale size. */
 extern const uint64_t var_size;
 
@@ -86,6 +95,12 @@ extern Compressor cell_var_offsets_compression;
 /** The default compression level for the offsets of variable-sized cells. */
 extern int cell_var_offsets_compression_level;
 
+/** The default compressor for the validity value cells. */
+extern Compressor cell_validity_compression;
+
+/** The default compression level for the validity value cells. */
+extern int cell_validity_compression_level;
+
 /** The default compressor for the coordinates. */
 extern Compressor coords_compression;
 
@@ -155,6 +170,15 @@ extern const uint8_t empty_any;
 /** The file suffix used in TileDB. */
 extern const std::string file_suffix;
 
+/** Suffix for the special vacuum files used in TileDB. */
+extern const std::string vacuum_file_suffix;
+
+/** Suffix for the special ok files used in TileDB. */
+extern const std::string ok_file_suffix;
+
+/** Suffix for the special metadata files used in TileDB. */
+extern const std::string meta_file_suffix;
+
 /** The fragment metadata file name. */
 extern const std::string fragment_metadata_filename;
 
@@ -209,6 +233,33 @@ extern const std::string query_status_incomplete_str;
 /** TILEDB_UNINITIALIZED Query String **/
 extern const std::string query_status_uninitialized_str;
 
+/** TILEDB_LT Query Condition Op String **/
+extern const std::string query_condition_op_lt_str;
+
+/** TILEDB_LE Query Condition Op String **/
+extern const std::string query_condition_op_le_str;
+
+/** TILEDB_GT Query Condition Op String **/
+extern const std::string query_condition_op_gt_str;
+
+/** TILEDB_GE Query Condition Op String **/
+extern const std::string query_condition_op_ge_str;
+
+/** TILEDB_EQ Query Condition Op String **/
+extern const std::string query_condition_op_eq_str;
+
+/** TILEDB_NE Query Condition Op String **/
+extern const std::string query_condition_op_ne_str;
+
+/** TILEDB_AND Query Condition Combination Op String **/
+extern const std::string query_condition_combination_op_and_str;
+
+/** TILEDB_OR Query Condition Combination Op String **/
+extern const std::string query_condition_combination_op_or_str;
+
+/** TILEDB_NOT Query Condition Combination Op String **/
+extern const std::string query_condition_combination_op_not_str;
+
 /** TILEDB_COMPRESSION Filter type string */
 extern const std::string filter_type_compression_str;
 
@@ -366,6 +417,33 @@ extern const std::string datetime_fs_str;
 /** The string representation for type DATETIME_AS. */
 extern const std::string datetime_as_str;
 
+/** The string representation for type TIME_HR. */
+extern const std::string time_hr_str;
+
+/** The string representation for type TIME_MIN. */
+extern const std::string time_min_str;
+
+/** The string representation for type TIME_SEC. */
+extern const std::string time_sec_str;
+
+/** The string representation for type TIME_MS. */
+extern const std::string time_ms_str;
+
+/** The string representation for type TIME_US. */
+extern const std::string time_us_str;
+
+/** The string representation for type TIME_NS. */
+extern const std::string time_ns_str;
+
+/** The string representation for type TIME_PS. */
+extern const std::string time_ps_str;
+
+/** The string representation for type TIME_FS. */
+extern const std::string time_fs_str;
+
+/** The string representation for type TIME_AS. */
+extern const std::string time_as_str;
+
 /** The string representation for the dense array type. */
 extern const std::string dense_str;
 
@@ -384,6 +462,9 @@ extern const std::string global_order_str;
 /** The string representation for the unordered layout. */
 extern const std::string unordered_str;
 
+/** The string representation for the Hilbert layout. */
+extern const std::string hilbert_str;
+
 /** The string representation of null. */
 extern const std::string null_str;
 
@@ -402,6 +483,15 @@ extern const std::string filesystem_type_hdfs_str;
 /** The string representation for filesystem type s3. */
 extern const std::string filesystem_type_s3_str;
 
+/** The string representation for filesystem type azure. */
+extern const std::string filesystem_type_azure_str;
+
+/** The string representation for filesystem type gcs. */
+extern const std::string filesystem_type_gcs_str;
+
+/** The string representation for in-memory filesystem. */
+extern const std::string filesystem_type_mem_str;
+
 /** The string representation for WalkOrder preorder. */
 extern const std::string walkorder_preorder_str;
 
@@ -423,37 +513,17 @@ extern const int32_t library_version[3];
 /** The TileDB serialization format version number. */
 extern const uint32_t format_version;
 
+/** The lowest version supported for back compat writes. */
+extern const uint32_t back_compat_writes_min_format_version;
+
 /** The maximum size of a tile chunk (unit of compression) in bytes. */
 extern const uint64_t max_tile_chunk_size;
 
-/** The default attribute name prefix. */
-extern const std::string default_attr_name;
-
-/** The default dimension name prefix. */
-extern const std::string default_dim_name;
-
-/** The key attribute name. */
-extern const std::string key_attr_name;
-
-/** The key attribute type. */
-extern Datatype key_attr_type;
-
-/** The key attribute compressor. */
-extern Compressor key_attr_compressor;
-
 /**
- * The name of the first key dimension (recall that a key in a
- * key-value store is hashed into a 16-byte MD5 digest, which
- * is represented as a 2-dimensional uint64_t value.
+ * Maximum splitting depth on a string dimension. For a maximum depth
+ * of D, the maximum number of splits is 2^D.
  */
-extern const std::string key_dim_1;
-
-/**
- * The name of the second key dimension (recall that a key in a
- * key-value store is hashed into a 16-byte MD5 digest, which
- * is represented as a 2-dimensional uint64_t value.
- */
-extern const std::string key_dim_2;
+extern const uint64_t max_string_dim_split_depth;
 
 /** Maximum number of attempts to wait for an S3 response. */
 extern const unsigned int s3_max_attempts;
@@ -461,6 +531,22 @@ extern const unsigned int s3_max_attempts;
 /** Milliseconds of wait time between S3 attempts. */
 extern const unsigned int s3_attempt_sleep_ms;
 
+/** Maximum number of attempts to wait for an Azure response. */
+extern const unsigned int azure_max_attempts;
+
+/**
+ * Milliseconds of wait time between Azure attempts. Currently,
+ * the Azure SDK only supports retries with a second-granularity.
+ * This number will be floored to the nearest second.
+ */
+extern const unsigned int azure_attempt_sleep_ms;
+
+/** Maximum number of attempts to wait for an GCS response. */
+extern const unsigned int gcs_max_attempts;
+
+/** Milliseconds of wait time between GCS attempts. */
+extern const unsigned int gcs_attempt_sleep_ms;
+
 /** An allocation tag used for logging. */
 extern const std::string s3_allocation_tag;
 
@@ -470,13 +556,22 @@ extern const std::string special_name_prefix;
 /** Number of milliseconds between watchdog thread wakeups. */
 extern const unsigned watchdog_thread_sleep_ms;
 
+/** Maximum number of concurrent attribute reads. */
+extern const unsigned concurrent_attr_reads;
+
 /** Returns the empty fill value based on the input datatype. */
 const void* fill_value(Datatype type);
 
+/** The redirection header key in REST response. */
+extern const std::string redirection_header_key;
+
 #ifdef __linux__
 /** List of possible certificates files for libcurl */
 extern const std::array<std::string, 6> cert_files_linux;
 #endif
+
+/** Delimiter for lists passed as config parameter */
+extern const std::string config_delimiter;
 }  // namespace constants
 
 }  // namespace sm
