@@ -452,7 +452,8 @@ std::string to_str(const void* value, Datatype type) {
   return ss.str();
 }
 
-uint64_t common_prefix_size(const std::string& a, const std::string& b) {
+uint64_t common_prefix_size(
+    const std::string_view& a, const std::string_view& b) {
   auto size = std::min(a.size(), b.size());
   for (size_t i = 0; i < size; ++i) {
     if (a[i] != b[i])
