@@ -44,6 +44,7 @@
 #include "tiledb/sm/array_schema/array_schema.h"
 #include "tiledb/sm/array_schema/dimension.h"
 #include "tiledb/sm/array_schema/domain.h"
+#include "tiledb/sm/enums/query_status_details.h"
 #include "tiledb/sm/fragment/written_fragment_info.h"
 #include "tiledb/sm/misc/utils.h"
 #include "tiledb/sm/query/iquery_strategy.h"
@@ -159,6 +160,16 @@ class Query {
   Status add_range(
       unsigned dim_idx, const void* start, const void* end, const void* stride);
 
+  /**
+   * @brief Set point ranges from an array
+   *
+   * @param dim_idx Dimension index
+   * @param start Pointer to start of the array
+   * @param count Number of elements to add
+   * @return Status
+   */
+  Status add_point_ranges(unsigned dim_idx, const void* start, uint64_t count);
+
   /**
    * Adds a variable-sized range to the (read/write) query on the input
    * dimension by index, in the form of (start, end).
@@ -848,6 +859,9 @@ class Query {
   /** Returns the query status. */
   QueryStatus status() const;
 
+  /** Returns the query status incomplete reason. */
+  QueryStatusDetailsReason status_incomplete_reason() const;
+
   /** Returns the query type. */
   QueryType type() const;
 
@@ -866,6 +880,9 @@ class Query {
   /** Use the refactored sparse unordered with dups reader or not. */
   bool use_refactored_sparse_unordered_with_dups_reader();
 
+  /** Returns if all ranges for this query are non overlapping. */
+  std::tuple<Status, bool> non_overlapping_ranges();
+
  private:
   /* ********************************* */
   /*         PRIVATE ATTRIBUTES        */
