@@ -126,6 +126,11 @@ bool Reader::incomplete() const {
   return read_state_.overflowed_ || !read_state_.done();
 }
 
+QueryStatusDetailsReason Reader::status_incomplete_reason() const {
+  return incomplete() ? QueryStatusDetailsReason::REASON_USER_BUFFER_SIZE :
+                        QueryStatusDetailsReason::REASON_NONE;
+}
+
 Status Reader::init() {
   // Sanity checks
   if (storage_manager_ == nullptr)
@@ -248,6 +253,43 @@ void Reader::reset() {
 /*         PRIVATE METHODS        */
 /* ****************************** */
 
+Status Reader::apply_query_condition(
+    std::vector<ResultCellSlab>* const result_cell_slabs,
+    std::vector<ResultTile*>* result_tiles,
+    Subarray* subarray,
+    uint64_t stride) {
+  if (condition_.empty() || result_cell_slabs->empty())
+    return Status::Ok();
+
+  // To evaluate the query condition, we need to read tiles for the
+  // attributes used in the query condition. Build a map of attribute
+  // names to read.
+  std::unordered_map<std::string, ProcessTileFlags> names;
+  for (const auto& condition_name : condition_.field_names()) {
+    names[condition_name] = ProcessTileFlag::READ;
+  }
+
+  // Each element in `names` has been flagged with `ProcessTileFlag::READ`.
+  // This will read the tiles, but will not copy them into the user buffers.
+  RETURN_NOT_OK(process_tiles(
+      &names,
+      result_tiles,
+      result_cell_slabs,
+      subarray,
+      stride,
+      std::numeric_limits<uint64_t>::max()));
+
+  // The `UINT64_MAX` is a sentinel value to indicate that we do not
+  // use a stride in the cell index calculation. To simplify our logic,
+  // assign this to `1`.
+  if (stride == UINT64_MAX)
+    stride = 1;
+
+  RETURN_NOT_OK(condition_.apply(array_schema_, result_cell_slabs, stride));
+
+  return Status::Ok();
+}
+
 Status Reader::compute_result_cell_slabs(
     const std::vector<ResultCoords>& result_coords,
     std::vector<ResultCellSlab>* result_cell_slabs) const {
@@ -549,7 +591,6 @@ Status Reader::compute_sparse_result_tiles(
   auto timer_se = stats_->start_timer("compute_sparse_result_tiles");
 
   // For easy reference
-  auto domain = array_schema_->domain();
   auto& partitioner = read_state_.partitioner_;
   const auto& subarray = partitioner.current();
   auto range_num = subarray.range_num();
@@ -577,7 +618,8 @@ Status Reader::compute_sparse_result_tiles(
           auto pair = std::pair<unsigned, uint64_t>(f, t);
           // Add tile only if it does not already exist
           if (result_tile_map->find(pair) == result_tile_map->end()) {
-            result_tiles->emplace_back(f, t, domain);
+            result_tiles->emplace_back(
+                f, t, fragment_metadata_[f]->array_schema());
             (*result_tile_map)[pair] = result_tiles->size() - 1;
           }
           // Always check range for multiple fragments
@@ -595,7 +637,8 @@ Status Reader::compute_sparse_result_tiles(
         auto pair = std::pair<unsigned, uint64_t>(f, t);
         // Add tile only if it does not already exist
         if (result_tile_map->find(pair) == result_tile_map->end()) {
-          result_tiles->emplace_back(f, t, domain);
+          result_tiles->emplace_back(
+              f, t, fragment_metadata_[f]->array_schema());
           (*result_tile_map)[pair] = result_tiles->size() - 1;
         }
         // Always check range for multiple fragments
@@ -752,11 +795,18 @@ Status Reader::compute_result_coords(
   // ignore fragments with a version < 5.
   const auto dim_num = array_schema_->dim_num();
   std::vector<std::string> dim_names;
+  std::vector<std::string> var_size_dim_names;
   dim_names.reserve(dim_num);
-  for (unsigned d = 0; d < dim_num; ++d)
-    dim_names.emplace_back(array_schema_->dimension(d)->name());
+  for (unsigned d = 0; d < dim_num; ++d) {
+    const auto& name = array_schema_->dimension(d)->name();
+    dim_names.emplace_back(name);
+    if (array_schema_->var_size(name))
+      var_size_dim_names.emplace_back(name);
+  }
   RETURN_CANCEL_OR_ERROR(
       load_tile_offsets(read_state_.partitioner_.subarray(), &dim_names));
+  RETURN_CANCEL_OR_ERROR(load_tile_var_sizes(
+      read_state_.partitioner_.subarray(), &var_size_dim_names));
 
   // Read and unfilter zipped coordinate tiles. Note that
   // this will ignore fragments with a version >= 5.
