@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2018-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2018-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -30,19 +30,33 @@
  * This file declares a REST client class.
  */
 
+// clang-format off
+#ifdef TILEDB_SERIALIZATION
+#include "tiledb/sm/serialization/capnp_utils.h"
+#include "tiledb/sm/serialization/array_schema_evolution.h"
+#include "tiledb/sm/serialization/query.h"
+#include "tiledb/sm/serialization/tiledb-rest.h"
+#include "tiledb/sm/rest/curl.h" // must be included last to avoid Windows.h
+#endif
+// clang-format on
+
+// Something, somewhere seems to be defining TIME_MS as a macro
+#if defined(TIME_MS)
+#undef TIME_MS
+#endif
+
 #include <cassert>
 
-#include "tiledb/sm/misc/stats.h"
+#include "tiledb/common/logger.h"
+#include "tiledb/sm/array/array.h"
+#include "tiledb/sm/enums/query_type.h"
+#include "tiledb/sm/misc/constants.h"
 #include "tiledb/sm/misc/utils.h"
+#include "tiledb/sm/query/query.h"
 #include "tiledb/sm/rest/rest_client.h"
 #include "tiledb/sm/serialization/array_schema.h"
-#include "tiledb/sm/serialization/capnp_utils.h"
-#include "tiledb/sm/serialization/query.h"
 
-#ifdef TILEDB_SERIALIZATION
-#include "tiledb/sm/rest/curl.h"
-#include "tiledb/sm/serialization/tiledb-rest.capnp.h"
-#endif
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
@@ -50,7 +64,9 @@ namespace sm {
 #ifdef TILEDB_SERIALIZATION
 
 RestClient::RestClient()
-    : config_(nullptr)
+    : stats_(nullptr)
+    , config_(nullptr)
+    , compute_tp_(nullptr)
     , resubmit_incomplete_(true) {
   auto st = utils::parse::convert(
       Config::REST_SERIALIZATION_DEFAULT_FORMAT, &serialization_type_);
@@ -58,12 +74,18 @@ RestClient::RestClient()
   (void)st;
 }
 
-Status RestClient::init(const Config* config) {
+Status RestClient::init(
+    stats::Stats* const parent_stats,
+    const Config* config,
+    ThreadPool* compute_tp) {
   if (config == nullptr)
     return LOG_STATUS(
         Status::RestError("Error initializing rest client; config is null."));
 
+  stats_ = parent_stats->create_child("RestClient");
+
   config_ = config;
+  compute_tp_ = compute_tp;
 
   const char* c_str;
   RETURN_NOT_OK(config_->get("rest.server_address", &c_str));
@@ -92,90 +114,107 @@ Status RestClient::set_header(
 
 Status RestClient::get_array_schema_from_rest(
     const URI& uri, ArraySchema** array_schema) {
-  STATS_FUNC_IN(rest_array_get_schema);
-
   // Init curl and form the URL
   Curl curlc;
-  RETURN_NOT_OK(curlc.init(config_, extra_headers_));
   std::string array_ns, array_uri;
   RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
-  std::string url = rest_server_ + "/v1/arrays/" + array_ns + "/" +
-                    curlc.url_escape(array_uri);
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  const std::string url = redirect_uri(cache_key) + "/v1/arrays/" + array_ns +
+                          "/" + curlc.url_escape(array_uri);
 
   // Get the data
   Buffer returned_data;
-  RETURN_NOT_OK(curlc.get_data(url, serialization_type_, &returned_data));
+  RETURN_NOT_OK(curlc.get_data(
+      stats_, url, serialization_type_, &returned_data, cache_key));
   if (returned_data.data() == nullptr || returned_data.size() == 0)
     return LOG_STATUS(Status::RestError(
         "Error getting array schema from REST; server returned no data."));
 
   return serialization::array_schema_deserialize(
       array_schema, serialization_type_, returned_data);
-
-  STATS_FUNC_OUT(rest_array_get_schema);
 }
 
 Status RestClient::post_array_schema_to_rest(
     const URI& uri, ArraySchema* array_schema) {
-  STATS_FUNC_IN(rest_array_create);
-
   Buffer buff;
   RETURN_NOT_OK(serialization::array_schema_serialize(
-      array_schema, serialization_type_, &buff));
+      array_schema, serialization_type_, &buff, false));
   // Wrap in a list
   BufferList serialized;
   RETURN_NOT_OK(serialized.add_buffer(std::move(buff)));
 
+  bool found = false;
+  const std::string creation_access_credentials_name =
+      config_->get("rest.creation_access_credentials_name", &found);
+  if (found)
+    RETURN_NOT_OK(set_header(
+        "X-TILEDB-CLOUD-ACCESS-CREDENTIALS-NAME",
+        creation_access_credentials_name));
+
   // Init curl and form the URL
   Curl curlc;
-  RETURN_NOT_OK(curlc.init(config_, extra_headers_));
   std::string array_ns, array_uri;
   RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
-  std::string url = rest_server_ + "/v1/arrays/" + array_ns + "/" +
-                    curlc.url_escape(array_uri);
-
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  auto deduced_url = redirect_uri(cache_key) + "/v1/arrays/" + array_ns + "/" +
+                     curlc.url_escape(array_uri);
   Buffer returned_data;
-  return curlc.post_data(url, serialization_type_, &serialized, &returned_data);
-
-  STATS_FUNC_OUT(rest_array_create);
+  const Status sc = curlc.post_data(
+      stats_,
+      deduced_url,
+      serialization_type_,
+      &serialized,
+      &returned_data,
+      cache_key);
+  return sc;
 }
 
 Status RestClient::deregister_array_from_rest(const URI& uri) {
   // Init curl and form the URL
   Curl curlc;
-  RETURN_NOT_OK(curlc.init(config_, extra_headers_));
   std::string array_ns, array_uri;
   RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
-  std::string url = rest_server_ + "/v1/arrays/" + array_ns + "/" +
-                    curlc.url_escape(array_uri) + "/deregister";
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  const std::string url = redirect_uri(cache_key) + "/v1/arrays/" + array_ns +
+                          "/" + curlc.url_escape(array_uri) + "/deregister";
 
   Buffer returned_data;
-  return curlc.delete_data(url, serialization_type_, &returned_data);
+  return curlc.delete_data(
+      stats_, url, serialization_type_, &returned_data, cache_key);
 }
 
 Status RestClient::get_array_non_empty_domain(
-    Array* array, void* domain, bool* is_empty) {
+    Array* array, uint64_t timestamp_start, uint64_t timestamp_end) {
   if (array == nullptr)
     return LOG_STATUS(
         Status::RestError("Cannot get array non-empty domain; array is null"));
-  if (array->array_schema() == nullptr)
-    return LOG_STATUS(Status::RestError(
-        "Cannot get array non-empty domain; array schema is null"));
   if (array->array_uri().to_string().empty())
     return LOG_STATUS(Status::RestError(
         "Cannot get array non-empty domain; array URI is empty"));
 
   // Init curl and form the URL
   Curl curlc;
-  RETURN_NOT_OK(curlc.init(config_, extra_headers_));
   std::string array_ns, array_uri;
   RETURN_NOT_OK(array->array_uri().get_rest_components(&array_ns, &array_uri));
-  std::string url = rest_server_ + "/v1/arrays/" + array_ns + "/" +
-                    curlc.url_escape(array_uri) + "/non_empty_domain";
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  const std::string url = redirect_uri(cache_key) + "/v2/arrays/" + array_ns +
+                          "/" + curlc.url_escape(array_uri) +
+                          "/non_empty_domain?" +
+                          "start_timestamp=" + std::to_string(timestamp_start) +
+                          "&end_timestamp=" + std::to_string(timestamp_end);
 
   // Get the data
   Buffer returned_data;
-  RETURN_NOT_OK(curlc.get_data(url, serialization_type_, &returned_data));
+  RETURN_NOT_OK(curlc.get_data(
+      stats_, url, serialization_type_, &returned_data, cache_key));
 
   if (returned_data.data() == nullptr || returned_data.size() == 0)
     return LOG_STATUS(
@@ -184,7 +223,7 @@ Status RestClient::get_array_non_empty_domain(
 
   // Deserialize data returned
   return serialization::nonempty_domain_deserialize(
-      array, returned_data, serialization_type_, domain, is_empty);
+      array, returned_data, serialization_type_);
 }
 
 Status RestClient::get_array_max_buffer_sizes(
@@ -201,16 +240,19 @@ Status RestClient::get_array_max_buffer_sizes(
 
   // Init curl and form the URL
   Curl curlc;
-  RETURN_NOT_OK(curlc.init(config_, extra_headers_));
   std::string array_ns, array_uri;
   RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
-  std::string url = rest_server_ + "/v1/arrays/" + array_ns + "/" +
-                    curlc.url_escape(array_uri) + "/max_buffer_sizes" +
-                    subarray_query_param;
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  const std::string url = redirect_uri(cache_key) + "/v1/arrays/" + array_ns +
+                          "/" + curlc.url_escape(array_uri) +
+                          "/max_buffer_sizes" + subarray_query_param;
 
   // Get the data
   Buffer returned_data;
-  RETURN_NOT_OK(curlc.get_data(url, serialization_type_, &returned_data));
+  RETURN_NOT_OK(curlc.get_data(
+      stats_, url, serialization_type_, &returned_data, cache_key));
 
   if (returned_data.data() == nullptr || returned_data.size() == 0)
     return LOG_STATUS(
@@ -223,23 +265,31 @@ Status RestClient::get_array_max_buffer_sizes(
 }
 
 Status RestClient::get_array_metadata_from_rest(
-    const URI& uri, uint64_t timestamp, Array* array) {
+    const URI& uri,
+    uint64_t timestamp_start,
+    uint64_t timestamp_end,
+    Array* array) {
   if (array == nullptr)
     return LOG_STATUS(Status::RestError(
         "Error getting array metadata from REST; array is null."));
 
   // Init curl and form the URL
   Curl curlc;
-  RETURN_NOT_OK(curlc.init(config_, extra_headers_));
   std::string array_ns, array_uri;
   RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
-  std::string url = rest_server_ + "/v1/arrays/" + array_ns + "/" +
-                    curlc.url_escape(array_uri) +
-                    "/array_metadata?timestamp=" + std::to_string(timestamp);
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  const std::string url = redirect_uri(cache_key) + "/v1/arrays/" + array_ns +
+                          "/" + curlc.url_escape(array_uri) +
+                          "/array_metadata?" +
+                          "start_timestamp=" + std::to_string(timestamp_start) +
+                          "&end_timestamp=" + std::to_string(timestamp_end);
 
   // Get the data
   Buffer returned_data;
-  RETURN_NOT_OK(curlc.get_data(url, serialization_type_, &returned_data));
+  RETURN_NOT_OK(curlc.get_data(
+      stats_, url, serialization_type_, &returned_data, cache_key));
   if (returned_data.data() == nullptr || returned_data.size() == 0)
     return LOG_STATUS(Status::RestError(
         "Error getting array metadata from REST; server returned no data."));
@@ -248,7 +298,11 @@ Status RestClient::get_array_metadata_from_rest(
       array, serialization_type_, returned_data);
 }
 
-Status RestClient::post_array_metadata_to_rest(const URI& uri, Array* array) {
+Status RestClient::post_array_metadata_to_rest(
+    const URI& uri,
+    uint64_t timestamp_start,
+    uint64_t timestamp_end,
+    Array* array) {
   if (array == nullptr)
     return LOG_STATUS(Status::RestError(
         "Error posting array metadata to REST; array is null."));
@@ -262,20 +316,24 @@ Status RestClient::post_array_metadata_to_rest(const URI& uri, Array* array) {
 
   // Init curl and form the URL
   Curl curlc;
-  RETURN_NOT_OK(curlc.init(config_, extra_headers_));
   std::string array_ns, array_uri;
   RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
-  std::string url = rest_server_ + "/v1/arrays/" + array_ns + "/" +
-                    curlc.url_escape(array_uri) + "/array_metadata";
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  const std::string url = redirect_uri(cache_key) + "/v1/arrays/" + array_ns +
+                          "/" + curlc.url_escape(array_uri) +
+                          "/array_metadata?" +
+                          "start_timestamp=" + std::to_string(timestamp_start) +
+                          "&end_timestamp=" + std::to_string(timestamp_end);
 
   // Put the data
   Buffer returned_data;
-  return curlc.post_data(url, serialization_type_, &serialized, &returned_data);
+  return curlc.post_data(
+      stats_, url, serialization_type_, &serialized, &returned_data, cache_key);
 }
 
 Status RestClient::submit_query_to_rest(const URI& uri, Query* query) {
-  STATS_FUNC_IN(rest_query_submit);
-
   // Local state tracking for the current offsets into the user's query buffers.
   // This allows resubmission of incomplete queries while appending to the
   // same user buffers.
@@ -288,8 +346,6 @@ Status RestClient::submit_query_to_rest(const URI& uri, Query* query) {
   RETURN_NOT_OK(update_attribute_buffer_sizes(copy_state, query));
 
   return Status::Ok();
-
-  STATS_FUNC_OUT(rest_query_submit);
 }
 
 Status RestClient::post_query_submit(
@@ -301,6 +357,14 @@ Status RestClient::post_query_submit(
         Status::RestError("Error submitting query to REST; null array."));
   }
 
+  auto rest_scratch = query->rest_scratch();
+
+  if (rest_scratch->size() > 0) {
+    bool skip;
+    query_post_call_back(
+        false, nullptr, 0, &skip, rest_scratch, query, copy_state);
+  }
+
   // Serialize query to send
   BufferList serialized;
   RETURN_NOT_OK(serialization::query_serialize(
@@ -308,34 +372,41 @@ Status RestClient::post_query_submit(
 
   // Init curl and form the URL
   Curl curlc;
-  RETURN_NOT_OK(curlc.init(config_, extra_headers_));
   std::string array_ns, array_uri;
   RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
-  std::string url = rest_server_ + "/v2/arrays/" + array_ns + "/" +
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  std::string url = redirect_uri(cache_key) + "/v2/arrays/" + array_ns + "/" +
                     curlc.url_escape(array_uri) +
                     "/query/submit?type=" + query_type_str(query->type()) +
                     "&read_all=" + (resubmit_incomplete_ ? "true" : "false");
 
   // Remote array reads always supply the timestamp.
-  if (query->type() == QueryType::READ)
-    url += "&open_at=" + std::to_string(array->timestamp());
+  url += "&start_timestamp=" + std::to_string(array->timestamp_start());
+  url += "&end_timestamp=" + std::to_string(array->timestamp_end());
 
   // Create the callback that will process the response buffers as they
   // are received.
-  Buffer scratch;
   auto write_cb = std::bind(
-      &RestClient::post_data_write_cb,
+      &RestClient::query_post_call_back,
       this,
       std::placeholders::_1,
       std::placeholders::_2,
       std::placeholders::_3,
       std::placeholders::_4,
-      &scratch,
+      rest_scratch,
       query,
       copy_state);
 
   const Status st = curlc.post_data(
-      url, serialization_type_, &serialized, std::move(write_cb));
+      stats_,
+      url,
+      serialization_type_,
+      &serialized,
+      rest_scratch.get(),
+      std::move(write_cb),
+      cache_key);
 
   if (!st.ok() && copy_state->empty()) {
     return LOG_STATUS(Status::RestError(
@@ -345,15 +416,15 @@ Status RestClient::post_query_submit(
         st.message()));
   }
 
-  return Status::Ok();
+  return st;
 }
 
-size_t RestClient::post_data_write_cb(
+size_t RestClient::query_post_call_back(
     const bool reset,
     void* const contents,
     const size_t content_nbytes,
     bool* const skip_retries,
-    Buffer* const scratch,
+    tdb_shared_ptr<Buffer> scratch,
     Query* query,
     serialization::CopyState* copy_state) {
   // All return statements in this function must pass through this wrapper.
@@ -388,7 +459,7 @@ size_t RestClient::post_data_write_cb(
   // from within the Curl object.
   if (reset) {
     scratch->set_size(0);
-    scratch->set_offset(0);
+    scratch->reset_offset();
     copy_state->clear();
   }
 
@@ -444,7 +515,6 @@ size_t RestClient::post_data_write_cb(
       st = aux.write(scratch->cur_data(), query_size);
       if (!st.ok()) {
         scratch->set_offset(scratch->offset() - 8);
-        scratch->set_size(scratch->offset());
         return return_wrapper(bytes_processed);
       }
 
@@ -454,10 +524,9 @@ size_t RestClient::post_data_write_cb(
       // error status.
       aux.reset_offset();
       st = serialization::query_deserialize(
-          aux, serialization_type_, true, copy_state, query);
+          aux, serialization_type_, true, copy_state, query, compute_tp_);
       if (!st.ok()) {
         scratch->set_offset(scratch->offset() - 8);
-        scratch->set_size(scratch->offset());
         return return_wrapper(bytes_processed);
       }
     } else {
@@ -466,10 +535,9 @@ size_t RestClient::post_data_write_cb(
       // data when deserializing read queries, this will return an
       // error status.
       st = serialization::query_deserialize(
-          *scratch, serialization_type_, true, copy_state, query);
+          *scratch, serialization_type_, true, copy_state, query, compute_tp_);
       if (!st.ok()) {
         scratch->set_offset(scratch->offset() - 8);
-        scratch->set_size(scratch->offset());
         return return_wrapper(bytes_processed);
       }
     }
@@ -483,9 +551,9 @@ size_t RestClient::post_data_write_cb(
   // consumption by overwriting the serialized query objects that we
   // have already processed.
   const uint64_t length = scratch->size() - scratch->offset();
-  if (scratch->offset() != 0) {
+  if (scratch->offset() != 0 && length != 0) {
     const uint64_t offset = scratch->offset();
-    scratch->set_offset(0);
+    scratch->reset_offset();
 
     // When the length of the remaining bytes is less than offset,
     // we can safely read the remaining bytes from 'scratch' and
@@ -497,16 +565,21 @@ size_t RestClient::post_data_write_cb(
     // there will be an overlap in the memory of the source and
     // destination.
     if (length <= offset) {
+      scratch->reset_size();
       st = scratch->write(scratch->data(offset), length);
     } else {
       Buffer aux;
       st = aux.write(scratch->data(offset), length);
       if (st.ok()) {
+        scratch->reset_size();
         st = scratch->write(aux.data(), aux.size());
       }
     }
 
     assert(st.ok());
+    if (!st.ok()) {
+      LOG_STATUS(st);
+    }
     assert(scratch->size() == length);
   }
 
@@ -524,32 +597,39 @@ Status RestClient::finalize_query_to_rest(const URI& uri, Query* query) {
 
   // Init curl and form the URL
   Curl curlc;
-  RETURN_NOT_OK(curlc.init(config_, extra_headers_));
   std::string array_ns, array_uri;
   RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
-  std::string url = rest_server_ + "/v1/arrays/" + array_ns + "/" +
-                    curlc.url_escape(array_uri) +
-                    "/query/finalize?type=" + query_type_str(query->type());
-
-  Buffer returned_data;
+  const std::string cache_key = array_ns + ":" + array_uri;
   RETURN_NOT_OK(
-      curlc.post_data(url, serialization_type_, &serialized, &returned_data));
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  const std::string url =
+      redirect_uri(cache_key) + "/v1/arrays/" + array_ns + "/" +
+      curlc.url_escape(array_uri) +
+      "/query/finalize?type=" + query_type_str(query->type());
+  Buffer returned_data;
+  RETURN_NOT_OK(curlc.post_data(
+      stats_,
+      url,
+      serialization_type_,
+      &serialized,
+      &returned_data,
+      cache_key));
 
   if (returned_data.data() == nullptr || returned_data.size() == 0)
     return LOG_STATUS(
         Status::RestError("Error finalizing query; server returned no data."));
 
   // Deserialize data returned
-  returned_data.set_offset(0);
+  returned_data.reset_offset();
   return serialization::query_deserialize(
-      returned_data, serialization_type_, true, nullptr, query);
+      returned_data, serialization_type_, true, nullptr, query, compute_tp_);
 }
 
 Status RestClient::subarray_to_str(
     const ArraySchema* schema,
     const void* subarray,
     std::string* subarray_str) {
-  const auto coords_type = schema->coords_type();
+  const auto coords_type = schema->dimension(0)->type();
   const auto dim_num = schema->dim_num();
   const auto subarray_nelts = 2 * dim_num;
 
@@ -592,6 +672,15 @@ Status RestClient::subarray_to_str(
       case Datatype::DATETIME_PS:
       case Datatype::DATETIME_FS:
       case Datatype::DATETIME_AS:
+      case Datatype::TIME_HR:
+      case Datatype::TIME_MIN:
+      case Datatype::TIME_SEC:
+      case Datatype::TIME_MS:
+      case Datatype::TIME_US:
+      case Datatype::TIME_NS:
+      case Datatype::TIME_PS:
+      case Datatype::TIME_FS:
+      case Datatype::TIME_AS:
       case Datatype::INT64:
         ss << ((const int64_t*)subarray)[i];
         break;
@@ -624,60 +713,109 @@ Status RestClient::update_attribute_buffer_sizes(
   if (query->type() != QueryType::READ)
     return Status::Ok();
 
-  const auto schema = query->array_schema();
-  if (schema == nullptr)
-    return LOG_STATUS(Status::RestError(
-        "Error updating attribute buffer sizes; array schema is null"));
-
-  const auto attrs = query->attributes();
-  std::set<std::string> attr_names;
-  attr_names.insert(constants::coords);
-  for (const auto& name : attrs)
-    attr_names.insert(name);
-
-  for (const auto& attr_name : attr_names) {
-    const bool is_coords = attr_name == constants::coords;
-    const auto* attr = schema->attribute(attr_name);
-    if (!is_coords && attr == nullptr)
-      return LOG_STATUS(Status::RestError(
-          "Error updating attribute buffer sizes; no attribute object for '" +
-          attr_name + "'"));
-
-    // Skip attributes that were not a part of the copy process.
-    auto copy_state_it = copy_state.find(attr_name);
-    if (copy_state_it == copy_state.end())
-      continue;
-    auto attr_state = copy_state_it->second;
-
-    const bool var_size = !is_coords && attr->var_size();
-    if (var_size) {
-      uint64_t* offset_buffer = nullptr;
-      uint64_t* offset_buffer_size = nullptr;
-      void* buffer = nullptr;
-      uint64_t* buffer_size = nullptr;
-      RETURN_NOT_OK(query->get_buffer(
-          attr_name.c_str(),
-          &offset_buffer,
-          &offset_buffer_size,
-          &buffer,
-          &buffer_size));
-      if (offset_buffer_size != nullptr)
-        *offset_buffer_size = attr_state.offset_size;
-      if (buffer_size != nullptr)
-        *buffer_size = attr_state.data_size;
-    } else {
-      void* buffer = nullptr;
-      uint64_t* buffer_size = nullptr;
-      RETURN_NOT_OK(
-          query->get_buffer(attr_name.c_str(), &buffer, &buffer_size));
-      if (buffer_size != nullptr)
-        *buffer_size = attr_state.data_size;
-    }
+  for (const auto& cit : copy_state) {
+    const auto& name = cit.first;
+    auto state = cit.second;
+    auto query_buffer = query->buffer(name);
+    if (query_buffer.buffer_var_size_ != nullptr) {
+      *query_buffer.buffer_var_size_ = state.data_size;
+      *query_buffer.buffer_size_ = state.offset_size;
+    } else if (query_buffer.buffer_size_ != nullptr)
+      *query_buffer.buffer_size_ = state.data_size;
   }
 
   return Status::Ok();
 }
 
+Status RestClient::get_query_est_result_sizes(const URI& uri, Query* query) {
+  if (query == nullptr)
+    return LOG_STATUS(Status::RestError(
+        "Error getting query estimated result size from REST; Query is null."));
+
+  // Get array
+  const Array* array = query->array();
+  if (array == nullptr) {
+    return LOG_STATUS(Status::RestError(
+        "Error festing query estimated result size from REST; null array."));
+  }
+
+  // Serialize query to send
+  BufferList serialized;
+  RETURN_NOT_OK(serialization::query_serialize(
+      query, serialization_type_, true, &serialized));
+
+  // Init curl and form the URL
+  Curl curlc;
+  std::string array_ns, array_uri;
+  RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  std::string url =
+      redirect_uri(cache_key) + "/v1/arrays/" + array_ns + "/" +
+      curlc.url_escape(array_uri) +
+      "/query/est_result_sizes?type=" + query_type_str(query->type());
+
+  // Remote array reads always supply the timestamp.
+  if (query->type() == QueryType::READ) {
+    url += "&start_timestamp=" + std::to_string(array->timestamp_start());
+    url += "&end_timestamp=" + std::to_string(array->timestamp_end());
+  }
+
+  // Get the data
+  Buffer returned_data;
+  RETURN_NOT_OK(curlc.post_data(
+      stats_,
+      url,
+      serialization_type_,
+      &serialized,
+      &returned_data,
+      cache_key));
+  if (returned_data.data() == nullptr || returned_data.size() == 0)
+    return LOG_STATUS(Status::RestError(
+        "Error getting array metadata from REST; server returned no data."));
+
+  return serialization::query_est_result_size_deserialize(
+      query, serialization_type_, true, returned_data);
+}
+
+std::string RestClient::redirect_uri(const std::string& cache_key) {
+  std::unique_lock<std::mutex> rd_lck(redirect_mtx_);
+  std::unordered_map<std::string, std::string>::const_iterator cache_it =
+      redirect_meta_.find(cache_key);
+
+  return (cache_it == redirect_meta_.end()) ? rest_server_ : cache_it->second;
+}
+
+Status RestClient::post_array_schema_evolution_to_rest(
+    const URI& uri, ArraySchemaEvolution* array_schema_evolution) {
+  Buffer buff;
+  RETURN_NOT_OK(serialization::array_schema_evolution_serialize(
+      array_schema_evolution, serialization_type_, &buff, false));
+  // Wrap in a list
+  BufferList serialized;
+  RETURN_NOT_OK(serialized.add_buffer(std::move(buff)));
+
+  // Init curl and form the URL
+  Curl curlc;
+  std::string array_ns, array_uri;
+  RETURN_NOT_OK(uri.get_rest_components(&array_ns, &array_uri));
+  const std::string cache_key = array_ns + ":" + array_uri;
+  RETURN_NOT_OK(
+      curlc.init(config_, extra_headers_, &redirect_meta_, &redirect_mtx_));
+  auto deduced_url = redirect_uri(cache_key) + "/v1/arrays/" + array_ns + "/" +
+                     curlc.url_escape(array_uri) + "/evolve";
+  Buffer returned_data;
+  const Status sc = curlc.post_data(
+      stats_,
+      deduced_url,
+      serialization_type_,
+      &serialized,
+      &returned_data,
+      cache_key);
+  return sc;
+}
+
 #else
 
 RestClient::RestClient() {
@@ -686,7 +824,7 @@ RestClient::RestClient() {
   (void)serialization_type_;
 }
 
-Status RestClient::init(const Config*) {
+Status RestClient::init(stats::Stats*, const Config*, ThreadPool*) {
   return LOG_STATUS(
       Status::RestError("Cannot use rest client; serialization not enabled."));
 }
@@ -711,7 +849,7 @@ Status RestClient::deregister_array_from_rest(const URI&) {
       Status::RestError("Cannot use rest client; serialization not enabled."));
 }
 
-Status RestClient::get_array_non_empty_domain(Array*, void*, bool*) {
+Status RestClient::get_array_non_empty_domain(Array*, uint64_t, uint64_t) {
   return LOG_STATUS(
       Status::RestError("Cannot use rest client; serialization not enabled."));
 }
@@ -725,12 +863,14 @@ Status RestClient::get_array_max_buffer_sizes(
       Status::RestError("Cannot use rest client; serialization not enabled."));
 }
 
-Status RestClient::get_array_metadata_from_rest(const URI&, uint64_t, Array*) {
+Status RestClient::get_array_metadata_from_rest(
+    const URI&, uint64_t, uint64_t, Array*) {
   return LOG_STATUS(
       Status::RestError("Cannot use rest client; serialization not enabled."));
 }
 
-Status RestClient::post_array_metadata_to_rest(const URI&, Array*) {
+Status RestClient::post_array_metadata_to_rest(
+    const URI&, uint64_t, uint64_t, Array*) {
   return LOG_STATUS(
       Status::RestError("Cannot use rest client; serialization not enabled."));
 }
@@ -745,6 +885,17 @@ Status RestClient::finalize_query_to_rest(const URI&, Query*) {
       Status::RestError("Cannot use rest client; serialization not enabled."));
 }
 
+Status RestClient::get_query_est_result_sizes(const URI&, Query*) {
+  return LOG_STATUS(
+      Status::RestError("Cannot use rest client; serialization not enabled."));
+}
+
+Status RestClient::post_array_schema_evolution_to_rest(
+    const URI&, ArraySchemaEvolution*) {
+  return LOG_STATUS(
+      Status::RestError("Cannot use rest client; serialization not enabled."));
+}
+
 #endif  // TILEDB_SERIALIZATION
 
 }  // namespace sm
