@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2018-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2018-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -33,22 +33,33 @@
 #ifndef TILEDB_REST_CLIENT_H
 #define TILEDB_REST_CLIENT_H
 
-#include "tiledb/sm/array_schema/array_schema.h"
-#include "tiledb/sm/enums/serialization_type.h"
-#include "tiledb/sm/misc/status.h"
-#include "tiledb/sm/query/query.h"
+#include <string>
+#include <unordered_map>
+
+#include "tiledb/common/status.h"
+#include "tiledb/common/thread_pool.h"
 #include "tiledb/sm/serialization/query.h"
+#include "tiledb/sm/stats/stats.h"
+
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
 
+class ArraySchema;
+class Config;
+class Query;
+
+enum class SerializationType : uint8_t;
+
 class RestClient {
  public:
   /** Constructor. */
   RestClient();
 
   /** Initialize the REST client with the given config. */
-  Status init(const Config* config);
+  Status init(
+      stats::Stats* parent_stats, const Config* config, ThreadPool* compute_tp);
 
   /** Sets a header that will be attached to all requests. */
   Status set_header(const std::string& name, const std::string& value);
@@ -82,14 +93,13 @@ class RestClient {
   /**
    * Get array's non_empty domain from rest server
    *
-   * @param uri of array being loaded
-   * @param domain The domain to be retrieved.
-   * @param is_empty The function sets it to `1` if the non-empty domain is
-   *     empty (i.e., the array does not contain any data yet), and `0`
-   * otherwise.
+   * @param array Array model to fetch and set non empty domain on
+   * @param timestamp_start Inclusive starting timestamp at which to open array
+   * @param timestamp_end Inclusive ending timestamp at which to open array
    * @return Status Ok() on success Error() on failures
    */
-  Status get_array_non_empty_domain(Array* array, void* domain, bool* is_empty);
+  Status get_array_non_empty_domain(
+      Array* array, uint64_t timestamp_start, uint64_t timestamp_end);
 
   /**
    * Get array's max buffer sizes from rest server.
@@ -112,21 +122,31 @@ class RestClient {
    * Metadata of the array to match the returned values).
    *
    * @param uri Array URI
-   * @param timestamp Timestamp at which to open array
+   * @param timestamp_start Inclusive starting timestamp at which to open array
+   * @param timestamp_end Inclusive ending timestamp at which to open array
    * @param array Array to fetch metadata for
    * @return Status
    */
   Status get_array_metadata_from_rest(
-      const URI& uri, uint64_t timestamp, Array* array);
+      const URI& uri,
+      uint64_t timestamp_start,
+      uint64_t timestamp_end,
+      Array* array);
 
   /**
    * Posts the array's metadata to the REST server.
    *
    * @param uri Array URI
+   * @param timestamp_start Inclusive starting timestamp at which to open array
+   * @param timestamp_end Inclusive ending timestamp at which to open array
    * @param array Array to update/post metadata for.
    * @return Status
    */
-  Status post_array_metadata_to_rest(const URI& uri, Array* array);
+  Status post_array_metadata_to_rest(
+      const URI& uri,
+      uint64_t timestamp_start,
+      uint64_t timestamp_end,
+      Array* array);
 
   /**
    * Post a data query to rest server
@@ -141,19 +161,43 @@ class RestClient {
    * Post a data query to rest server
    *
    * @param uri of array being queried
-   * @param query to send to server and store results in, this qill be modified
+   * @param query to send to server and store results in, this will be modified
    * @return Status Ok() on success Error() on failures
    */
   Status finalize_query_to_rest(const URI& uri, Query* query);
 
+  /**
+   * Get array's non_empty domain from rest server
+   *
+   * @param array Array model to fetch and set non empty domain on
+   * @return Status Ok() on success Error() on failures
+   */
+  Status get_query_est_result_sizes(const URI& uri, Query* query);
+
+  /**
+   * Post array schema evolution to rest server
+   *
+   * @param uri of array being queried
+   * @param array_schema_evolution to send to server
+   * @return Status Ok() on success Error() on failures
+   */
+  Status post_array_schema_evolution_to_rest(
+      const URI& uri, ArraySchemaEvolution* array_schema_evolution);
+
  private:
   /* ********************************* */
   /*        PRIVATE ATTRIBUTES         */
   /* ********************************* */
 
+  /** The class stats. */
+  stats::Stats* stats_;
+
   /** The TileDB config options (contains server and auth info). */
   const Config* config_;
 
+  /** The thread pool for compute-bound tasks. */
+  ThreadPool* compute_tp_;
+
   /** Rest server config param. */
   std::string rest_server_;
 
@@ -162,7 +206,7 @@ class RestClient {
 
   /**
    * If true (the default), automatically resubmit incomplete queries on the
-   * server-side. This garauntees that the user only receive a complete query
+   * server-side. This guarantees that the user only receive a complete query
    * result from the server.
    *
    * When this is turned on, it is currently an error if the user buffers on the
@@ -174,6 +218,12 @@ class RestClient {
   /** Collection of extra headers that are attached to REST requests. */
   std::unordered_map<std::string, std::string> extra_headers_;
 
+  /** Array URI to redirected server mapping. */
+  std::unordered_map<std::string, std::string> redirect_meta_;
+
+  /** Mutex for thread-safety. */
+  mutable std::mutex redirect_mtx_;
+
   /* ********************************* */
   /*         PRIVATE METHODS           */
   /* ********************************* */
@@ -220,12 +270,12 @@ class RestClient {
    *    map is updated accordingly.
    * @return Number of acknowledged bytes
    */
-  size_t post_data_write_cb(
-      bool reset,
-      void* contents,
-      size_t content_nbytes,
-      bool* skip_retries,
-      Buffer* scratch,
+  size_t query_post_call_back(
+      const bool reset,
+      void* constcontents,
+      const size_t content_nbytes,
+      bool* constskip_retries,
+      tdb_shared_ptr<Buffer> scratch,
       Query* query,
       serialization::CopyState* copy_state);
 
@@ -254,6 +304,16 @@ class RestClient {
    */
   Status update_attribute_buffer_sizes(
       const serialization::CopyState& copy_state, Query* query) const;
+
+  /**
+   * Helper function encapsulating the functionality of looking up for cached
+   * redirected rest server addresses to avoid the redirection overhead
+   *
+   * @param array_ns Array namespace
+   * @param array_uri Array URI
+   * @return Returns the redirection URI if exists and empty string otherwise
+   */
+  std::string redirect_uri(const std::string& cache_key);
 };
 
 }  // namespace sm
