@@ -431,7 +431,7 @@ Status domain_from_capnp(
 }
 
 Status array_schema_to_capnp(
-    const ArraySchema* array_schema,
+    ArraySchema* array_schema,
     capnp::ArraySchema::Builder* array_schema_builder,
     const bool client_side) {
   if (array_schema == nullptr)
@@ -441,6 +441,8 @@ Status array_schema_to_capnp(
   // Only set the URI if client side
   if (client_side)
     array_schema_builder->setUri(array_schema->array_uri().to_string());
+
+  array_schema_builder->setName(array_schema->name());
   auto v = kj::heapArray<int32_t>(1);
   v[0] = array_schema->version();
   array_schema_builder->setVersion(v);
@@ -505,6 +507,9 @@ Status array_schema_from_capnp(
   if (schema_reader.hasUri())
     (*array_schema)->set_array_uri(URI(schema_reader.getUri().cStr()));
 
+  if (schema_reader.hasName())
+    (*array_schema)->set_name(schema_reader.getName().cStr());
+
   (*array_schema)->set_cell_order(layout);
   (*array_schema)->set_capacity(schema_reader.getCapacity());
   (*array_schema)->set_allows_dups(schema_reader.getAllowsDuplicates());
