@@ -597,38 +597,42 @@ Status index_read_state_from_capnp(
     const capnp::ReadStateIndex::Reader& read_state_reader,
     SparseIndexReaderBase* reader) {
   auto read_state = reader->read_state();
-  const auto* domain = schema->domain();
 
   read_state->done_adding_result_tiles_ =
       read_state_reader.getDoneAddingResultTiles();
 
-  assert(read_state_reader.hasResultCellSlab());
-  RETURN_NOT_OK(reader->clear_result_tiles());
-  read_state->result_cell_slabs_.clear();
+  // Only sparse global order reader has a result cell slab.
+  if (read_state_reader.hasResultCellSlab()) {
+    SparseGlobalOrderReader* sparse_global_order_reader =
+        (SparseGlobalOrderReader*)reader;
+    RETURN_NOT_OK(sparse_global_order_reader->clear_result_tiles());
+    read_state->result_cell_slabs_.clear();
 
-  std::unordered_map<
-      std::pair<unsigned, uint64_t>,
-      ResultTile*,
-      tiledb::sm::utils::hash::pair_hash>
-      result_tile_map;
-  for (const auto rcs : read_state_reader.getResultCellSlab()) {
-    auto start = rcs.getStart();
-    auto length = rcs.getLength();
-    auto frag_idx = rcs.getFragIdx();
-    auto tile_idx = rcs.getTileIdx();
+    std::unordered_map<
+        std::pair<unsigned, uint64_t>,
+        ResultTile*,
+        tiledb::sm::utils::hash::pair_hash>
+        result_tile_map;
+    for (const auto rcs : read_state_reader.getResultCellSlab()) {
+      auto start = rcs.getStart();
+      auto length = rcs.getLength();
+      auto frag_idx = rcs.getFragIdx();
+      auto tile_idx = rcs.getTileIdx();
 
-    ResultTile* rt = nullptr;
-    auto it =
-        result_tile_map.find(std::pair<unsigned, uint64_t>(frag_idx, tile_idx));
-    if (it != result_tile_map.end()) {
-      rt = it->second;
-    } else {
-      rt = reader->add_result_tile_unsafe(frag_idx, tile_idx, domain);
-      result_tile_map.emplace(
-          std::pair<unsigned, uint64_t>(frag_idx, tile_idx), rt);
+      ResultTile* rt = nullptr;
+      auto it = result_tile_map.find(
+          std::pair<unsigned, uint64_t>(frag_idx, tile_idx));
+      if (it != result_tile_map.end()) {
+        rt = it->second;
+      } else {
+        rt = sparse_global_order_reader->add_result_tile_unsafe(
+            frag_idx, tile_idx, schema);
+        result_tile_map.emplace(
+            std::pair<unsigned, uint64_t>(frag_idx, tile_idx), rt);
+      }
+
+      read_state->result_cell_slabs_.emplace_back(rt, start, length);
     }
-
-    read_state->result_cell_slabs_.emplace_back(rt, start, length);
   }
 
   assert(read_state_reader.hasFragTileIdx());
@@ -1134,13 +1138,28 @@ Status query_to_capnp(Query& query, capnp::Query::Builder* query_builder) {
         !schema->dense() && layout == Layout::UNORDERED &&
         schema->allows_dups()) {
       auto builder = query_builder->initReaderIndex();
-      auto reader = (SparseUnorderedWithDupsReader*)query.strategy();
 
-      query_builder->setVarOffsetsMode(reader->offsets_mode());
-      query_builder->setVarOffsetsAddExtraElement(
-          reader->offsets_extra_element());
-      query_builder->setVarOffsetsBitsize(reader->offsets_bitsize());
-      RETURN_NOT_OK(index_reader_to_capnp(query, *reader, &builder));
+      auto&& [st, non_overlapping_ranges]{query.non_overlapping_ranges()};
+      RETURN_NOT_OK(st);
+
+      if (non_overlapping_ranges) {
+        auto reader = (SparseUnorderedWithDupsReader<uint8_t>*)query.strategy();
+
+        query_builder->setVarOffsetsMode(reader->offsets_mode());
+        query_builder->setVarOffsetsAddExtraElement(
+            reader->offsets_extra_element());
+        query_builder->setVarOffsetsBitsize(reader->offsets_bitsize());
+        RETURN_NOT_OK(index_reader_to_capnp(query, *reader, &builder));
+      } else {
+        auto reader =
+            (SparseUnorderedWithDupsReader<uint64_t>*)query.strategy();
+
+        query_builder->setVarOffsetsMode(reader->offsets_mode());
+        query_builder->setVarOffsetsAddExtraElement(
+            reader->offsets_extra_element());
+        query_builder->setVarOffsetsBitsize(reader->offsets_bitsize());
+        RETURN_NOT_OK(index_reader_to_capnp(query, *reader, &builder));
+      }
     } else if (
         query.use_refactored_sparse_global_order_reader() && !schema->dense() &&
         (layout == Layout::GLOBAL_ORDER ||
@@ -1714,26 +1733,54 @@ Status query_from_capnp(
       query->clear_strategy();
       RETURN_NOT_OK(query->set_layout_unsafe(layout));
 
+      auto&& [st, non_overlapping_ranges]{query->non_overlapping_ranges()};
+      RETURN_NOT_OK(st);
+
       auto reader_reader = query_reader.getReaderIndex();
-      auto reader = (SparseUnorderedWithDupsReader*)query->strategy();
 
-      if (query_reader.hasVarOffsetsMode()) {
+      if (non_overlapping_ranges) {
+        auto reader =
+            (SparseUnorderedWithDupsReader<uint8_t>*)query->strategy();
+
+        if (query_reader.hasVarOffsetsMode()) {
+          RETURN_NOT_OK(
+              reader->set_offsets_mode(query_reader.getVarOffsetsMode()));
+        }
+
+        RETURN_NOT_OK(reader->set_offsets_extra_element(
+            query_reader.getVarOffsetsAddExtraElement()));
+
+        if (query_reader.getVarOffsetsBitsize() > 0) {
+          RETURN_NOT_OK(
+              reader->set_offsets_bitsize(query_reader.getVarOffsetsBitsize()));
+        }
+
+        RETURN_NOT_OK(reader->initialize_memory_budget());
+
         RETURN_NOT_OK(
-            reader->set_offsets_mode(query_reader.getVarOffsetsMode()));
-      }
+            index_reader_from_capnp(schema, reader_reader, query, reader));
+      } else {
+        auto reader =
+            (SparseUnorderedWithDupsReader<uint64_t>*)query->strategy();
 
-      RETURN_NOT_OK(reader->set_offsets_extra_element(
-          query_reader.getVarOffsetsAddExtraElement()));
+        if (query_reader.hasVarOffsetsMode()) {
+          RETURN_NOT_OK(
+              reader->set_offsets_mode(query_reader.getVarOffsetsMode()));
+        }
+
+        RETURN_NOT_OK(reader->set_offsets_extra_element(
+            query_reader.getVarOffsetsAddExtraElement()));
+
+        if (query_reader.getVarOffsetsBitsize() > 0) {
+          RETURN_NOT_OK(
+              reader->set_offsets_bitsize(query_reader.getVarOffsetsBitsize()));
+        }
+
+        RETURN_NOT_OK(reader->initialize_memory_budget());
 
-      if (query_reader.getVarOffsetsBitsize() > 0) {
         RETURN_NOT_OK(
-            reader->set_offsets_bitsize(query_reader.getVarOffsetsBitsize()));
+            index_reader_from_capnp(schema, reader_reader, query, reader));
       }
-
-      RETURN_NOT_OK(reader->initialize_memory_budget());
-
-      RETURN_NOT_OK(
-          index_reader_from_capnp(schema, reader_reader, query, reader));
     } else if (query_reader.hasDenseReader()) {
       // Strategy needs to be cleared here to create the correct reader.
       query->clear_strategy();
