@@ -426,11 +426,11 @@ Status StorageManager::array_load_fragments(
     }
     RETURN_NOT_OK(it->second->set_encryption_key(enc_key));
     open_array = it->second;
-
-    // Lock the array
-    open_array->mtx_lock();
   }
 
+  // Lock the array
+  open_array->mtx_lock();
+
   std::unordered_map<std::string, uint64_t> offsets;
 
   // Get fragment metadata in the case of reads, if not fetched already
@@ -486,11 +486,11 @@ StorageManager::array_reopen(
       return {st, std::nullopt, std::nullopt, std::nullopt};
 
     open_array = it->second;
-
-    // Lock the array
-    open_array->mtx_lock();
   }
 
+  // Lock the array
+  open_array->mtx_lock();
+
   // Determine which fragments to load
   std::vector<TimestampedURI> fragments_to_load;
   std::vector<URI> fragment_uris;
@@ -1534,6 +1534,9 @@ Status StorageManager::get_fragment_info(
           array_type == ArrayType::SPARSE ? timestamp_start : 0,
           timestamp_end);
 
+  (void)array_schemas_opt;
+  (void)array_schema_opt;  // -Wno-unused GCC 7.3
+
   if (!st.ok())
     return st;
 
@@ -2026,7 +2029,8 @@ StorageManager::load_all_array_schemas(
             std::nullopt};
 
   std::vector<URI> schema_uris;
-  RETURN_NOT_OK_TUPLE(get_array_schema_uris(array_uri, &schema_uris));
+  RETURN_NOT_OK_TUPLE(
+      get_array_schema_uris(array_uri, &schema_uris), std::nullopt);
   if (schema_uris.empty()) {
     return {
         logger_->status(Status::StorageManagerError(
@@ -2047,7 +2051,7 @@ StorageManager::load_all_array_schemas(
         schema_vector[schema_ith] = array_schema;
         return Status::Ok();
       });
-  RETURN_NOT_OK_TUPLE(status);
+  RETURN_NOT_OK_TUPLE(status, std::nullopt);
 
   std::unordered_map<std::string, tdb_shared_ptr<ArraySchema>> array_schemas;
   for (const auto& array_schema : schema_vector) {
@@ -2075,11 +2079,11 @@ Status StorageManager::load_array_metadata(
     auto it = open_arrays_for_reads_.find(array_uri.to_string());
     assert(it != open_arrays_for_reads_.end());
     open_array = it->second;
-
-    // Lock the array
-    open_array->mtx_lock();
   }
 
+  // Lock the array
+  open_array->mtx_lock();
+
   // Determine which array metadata to load
   std::vector<TimestampedURI> array_metadata_to_load;
   std::vector<URI> array_metadata_uris;
@@ -2125,23 +2129,26 @@ Status StorageManager::object_type(const URI& uri, ObjectType* type) const {
     }
   }
 
-  std::vector<URI> child_uris;
-  RETURN_NOT_OK(vfs_->ls(dir_uri, &child_uris));
+  bool exists = false;
+  RETURN_NOT_OK(vfs_->is_dir(
+      dir_uri.join_path(constants::array_schema_folder_name), &exists));
+  if (exists) {
+    *type = ObjectType::ARRAY;
+    return Status::Ok();
+  }
 
-  for (const auto& child_uri : child_uris) {
-    auto uri_str = child_uri.to_string();
-    if (utils::parse::ends_with(uri_str, constants::group_filename)) {
-      *type = ObjectType::GROUP;
-      return Status::Ok();
-    } else if (utils::parse::ends_with(
-                   uri_str, constants::array_schema_filename)) {
-      *type = ObjectType::ARRAY;
-      return Status::Ok();
-    } else if (utils::parse::ends_with(
-                   uri_str, constants::array_schema_folder_name)) {
-      *type = ObjectType::ARRAY;
-      return Status::Ok();
-    }
+  RETURN_NOT_OK(vfs_->is_file(
+      dir_uri.join_path(constants::array_schema_filename), &exists));
+  if (exists) {
+    *type = ObjectType::ARRAY;
+    return Status::Ok();
+  }
+
+  RETURN_NOT_OK(
+      vfs_->is_file(dir_uri.join_path(constants::group_filename), &exists));
+  if (exists) {
+    *type = ObjectType::GROUP;
+    return Status::Ok();
   }
 
   *type = ObjectType::INVALID;
