@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  * @copyright Copyright (c) 2016 MIT and Intel Corporation
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -34,6 +34,7 @@
 #ifndef TILEDB_STORAGE_MANAGER_H
 #define TILEDB_STORAGE_MANAGER_H
 
+#include <atomic>
 #include <condition_variable>
 #include <functional>
 #include <list>
@@ -42,29 +43,43 @@
 #include <queue>
 #include <string>
 #include <thread>
+#include <unordered_map>
 
-#include "tiledb/sm/array_schema/array_schema.h"
-#include "tiledb/sm/cache/lru_cache.h"
+#include "tiledb/common/heap_memory.h"
+#include "tiledb/common/logger_public.h"
+#include "tiledb/common/status.h"
+#include "tiledb/common/thread_pool.h"
 #include "tiledb/sm/config/config.h"
-#include "tiledb/sm/encryption/encryption.h"
-#include "tiledb/sm/encryption/encryption_key_validation.h"
-#include "tiledb/sm/enums/object_type.h"
 #include "tiledb/sm/enums/walk_order.h"
-#include "tiledb/sm/filesystem/vfs.h"
-#include "tiledb/sm/fragment/fragment_info.h"
-#include "tiledb/sm/misc/status.h"
-#include "tiledb/sm/misc/thread_pool.h"
+#include "tiledb/sm/filesystem/filelock.h"
+#include "tiledb/sm/fragment/single_fragment_info.h"
+#include "tiledb/sm/misc/cancelable_tasks.h"
 #include "tiledb/sm/misc/uri.h"
-#include "tiledb/sm/query/query.h"
-#include "tiledb/sm/storage_manager/consolidator.h"
-#include "tiledb/sm/storage_manager/open_array.h"
+#include "tiledb/sm/stats/global_stats.h"
+
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
 
 class Array;
+class ArraySchema;
+class ArraySchemaEvolution;
+class Buffer;
+class BufferLRUCache;
 class Consolidator;
+class EncryptionKey;
+class FragmentMetadata;
+class FragmentInfo;
+class Metadata;
+class OpenArray;
+class OpenArrayMemoryTracker;
+class Query;
 class RestClient;
+class VFS;
+
+enum class EncryptionType : uint8_t;
+enum class ObjectType : uint8_t;
 
 /** The storage manager that manages pretty much everything in TileDB. */
 class StorageManager {
@@ -98,11 +113,18 @@ class StorageManager {
   /* ********************************* */
 
   /** Constructor. */
-  StorageManager();
+  StorageManager(
+      ThreadPool* compute_tp,
+      ThreadPool* io_tp,
+      stats::Stats* parent_stats,
+      tdb_shared_ptr<Logger> logger);
 
   /** Destructor. */
   ~StorageManager();
 
+  DISABLE_COPY_AND_COPY_ASSIGN(StorageManager);
+  DISABLE_MOVE_AND_MOVE_ASSIGN(StorageManager);
+
   /* ********************************* */
   /*                API                */
   /* ********************************* */
@@ -133,56 +155,88 @@ class StorageManager {
    * Opens an array for reads at a timestamp. All the metadata of the
    * fragments created before or at the input timestamp are retrieved.
    *
-   * @param array_uri The array URI.
-   * @param timestamp The timestamp at which the array will be opened.
-   *     In TileDB, timestamps are in ms elapsed since
-   *     1970-01-01 00:00:00 +0000 (UTC).
-   * @param encryption_key The encryption key to use.
-   * @param array_schema The array schema to be retrieved after the
-   *     array is opened.
-   * @param fragment_metadata The fragment metadata to be retrieved
-   *     after the array is opened.
-   * @return Status
-   */
-  Status array_open_for_reads(
-      const URI& array_uri,
-      uint64_t timestamp,
-      const EncryptionKey& encryption_key,
-      ArraySchema** array_schema,
-      std::vector<FragmentMetadata*>* fragment_metadata);
-
-  /**
-   * Opens an array for reads, focusing only on a given list of fragments.
-   * Only the metadata of the input fragments are retrieved.
+   * If a timestamp_start is provided, this API will open the array between
+   * `timestamp_start` and `timestamp_end`.
    *
    * @param array_uri The array URI.
-   * @param fragments The fragments to open the array with.
-   * @param encryption_key The encryption key to use.
-   * @param array_schema The array schema to be retrieved after the
-   *     array is opened.
-   * @param fragment_metadata The fragment metadata to be retrieved
-   *     after the array is opened.
-   * @return Status
+   * @param enc_key The encryption key to use.
+   * @param timestamp_start The (optional) starting timestamp to open the array
+   * between, starting at this timestamp and ending at `timestamp_end`.
+   * @param timestamp_end The timestamp at which the array will be opened.
+   *     In TileDB, timestamps are in ms elapsed since
+   *     1970-01-01 00:00:00 +0000 (UTC).
+   * @return tuple of Status, latest ArraySchema, map of all array schemas and
+   * vector of FragmentMetadata
+   *        Status Ok on success, else error
+   *        ArraySchema The array schema to be retrieved after the
+   *           array is opened.
+   *        ArraySchemaMap Map of all array schemas found keyed by name
+   *        fragment_metadata The fragment metadata to be retrieved
+   *           after the array is opened.
    */
-  Status array_open_for_reads(
+  std::tuple<
+      Status,
+      std::optional<ArraySchema*>,
+      std::optional<
+          std::unordered_map<std::string, tdb_shared_ptr<ArraySchema>>>,
+      std::optional<std::vector<tdb_shared_ptr<FragmentMetadata>>>>
+  array_open_for_reads(
       const URI& array_uri,
-      const std::vector<FragmentInfo>& fragments,
-      const EncryptionKey& encryption_key,
-      ArraySchema** array_schema,
-      std::vector<FragmentMetadata*>* fragment_metadata);
+      const EncryptionKey& enc_key,
+      uint64_t timestamp_start,
+      uint64_t timestamp_end);
+
+  /**
+   * Opens an array for reads without fragments.
+   *
+   * @param array_uri The array URI.
+   * @param enc_key The encryption key to use.
+   * @return tuple of Status, latest ArraySchema and map of all array schemas
+   *        Status Ok on success, else error
+   *        ArraySchema The array schema to be retrieved after the
+   *          array is opened.
+   *        ArraySchemaMap Map of all array schemas found keyed by name
+   */
+  std::tuple<
+      Status,
+      std::optional<ArraySchema*>,
+      std::optional<
+          std::unordered_map<std::string, tdb_shared_ptr<ArraySchema>>>>
+  array_open_for_reads_without_fragments(
+      const URI& array_uri, const EncryptionKey& enc_key);
 
   /** Opens an array for writes.
    *
    * @param array_uri The array URI.
-   * @param encryption_key The encryption key.
-   * @param array_schema The array schema to be retrieved after the
-   *     array is opened.
-   * @return
+   * @param enc_key The encryption key.
+   * @return tuple of Status, latest ArraySchema and map of all array schemas
+   *        Status Ok on success, else error
+   *        ArraySchema The array schema to be retrieved after the
+   *          array is opened.
+   *        ArraySchemaMap Map of all array schemas found keyed by name
    */
-  Status array_open_for_writes(
+  std::tuple<
+      Status,
+      std::optional<ArraySchema*>,
+      std::optional<
+          std::unordered_map<std::string, tdb_shared_ptr<ArraySchema>>>>
+  array_open_for_writes(const URI& array_uri, const EncryptionKey& enc_key);
+
+  /**
+   * Load fragments for an already open array.
+   *
+   * @param array_uri The array URI.
+   * @param enc_key The encryption key to use.
+   * @param fragment_metadata The fragment metadata to be retrieved
+   *          after the array is opened.
+   * @param fragment_info The list of fragment info.
+   * @return Status
+   */
+  Status array_load_fragments(
       const URI& array_uri,
-      const EncryptionKey& encryption_key,
-      ArraySchema** array_schema);
+      const EncryptionKey& enc_key,
+      std::vector<tdb_shared_ptr<FragmentMetadata>>* fragment_metadata,
+      const std::vector<TimestampedURI>& fragment_info);
 
   /**
    * Reopens an already open array at a potentially new timestamp,
@@ -190,22 +244,32 @@ class StorageManager {
    * in the array.
    *
    * @param array_uri The array URI.
-   * @param timestamp The timestamp at which the array will be opened.
+   * @param enc_key The encryption key to use.
+   * @param timestamp_start The optional first timestamp between which the
+   *     array will be opened.
+   * @param timestamp_end The timestamp at which the array will be opened.
    *     In TileDB, timestamps are in ms elapsed since
    *     1970-01-01 00:00:00 +0000 (UTC).
-   * @param encryption_key The encryption key to use.
-   * @param array_schema The array schema to be retrieved after the
-   *     array is opened.
-   * @param fragment_metadata The fragment metadata to be retrieved
-   *     after the array is opened.
-   * @return Status
+   * @return tuple of Status, latest ArraySchema, map of all array schemas and
+   * vector of FragmentMetadata
+   *        Status Ok on success, else error
+   *        ArraySchema The array schema to be retrieved after the
+   *          array is opened.
+   *        ArraySchemaMap Map of all array schemas found keyed by name
+   *        FragmentMetadata The fragment metadata to be retrieved
+   *          after the array is opened.
    */
-  Status array_reopen(
+  std::tuple<
+      Status,
+      std::optional<ArraySchema*>,
+      std::optional<
+          std::unordered_map<std::string, tdb_shared_ptr<ArraySchema>>>,
+      std::optional<std::vector<tdb_shared_ptr<FragmentMetadata>>>>
+  array_reopen(
       const URI& array_uri,
-      uint64_t timestamp,
-      const EncryptionKey& encryption_key,
-      ArraySchema** array_schema,
-      std::vector<FragmentMetadata*>* fragment_metadata);
+      const EncryptionKey& enc_key,
+      uint64_t timestamp_start,
+      uint64_t timestamp_end);
 
   /**
    * Consolidates the fragments of an array into a single one.
@@ -227,6 +291,49 @@ class StorageManager {
       uint32_t key_length,
       const Config* config);
 
+  /**
+   * Cleans up the array, such as its consolidated fragments and array
+   * metadata. Note that this will coarsen the granularity of time traveling
+   * (see docs for more information).
+   *
+   * @param array_name The name of the array to be vacuumed.
+   * @param config Configuration parameters for vacuuming.
+   * @return Status
+   */
+  Status array_vacuum(const char* array_name, const Config* config);
+
+  /**
+   * Cleans up fragments that took part in consolidation. Note that this
+   * will coarsen the granularity of time traveling (see docs for more
+   * information).
+   *
+   * @param array_name The name of the array to be vacuumed.
+   * @param timestamp_start The timestamp to start vacuuming at.
+   * @param timestamp_end The timestamp to end vacuuming at.
+   * @return Status
+   */
+  Status array_vacuum_fragments(
+      const char* array_name, uint64_t timestamp_start, uint64_t timestamp_end);
+
+  /**
+   * Cleans up consolidated fragment metadata (all except the last one).
+   *
+   * @param array_name The name of the array to be consolidated.
+   * @return Status
+   */
+  Status array_vacuum_fragment_meta(const char* array_name);
+
+  /**
+   * Cleans up consolidated array metadata.
+   *
+   * @param array_name The name of the array to be consolidated.
+   * @param timestamp_start The timestamp to start vacuuming at.
+   * @param timestamp_end The timestamp to end vacuuming at.
+   * @return Status
+   */
+  Status array_vacuum_array_meta(
+      const char* array_name, uint64_t timestamp_start, uint64_t timestamp_end);
+
   /**
    * Consolidates the metadata of an array into a single file.
    *
@@ -261,6 +368,53 @@ class StorageManager {
       ArraySchema* array_schema,
       const EncryptionKey& encryption_key);
 
+  /**
+   * Evolve a TileDB array schema and store its new schema.
+   *
+   * @param array_uri The URI of the array to be evolved.
+   * @param schema_evolution The schema evolution.
+   * @param encryption_key The encryption key to use.
+   * @return Status
+   */
+  Status array_evolve_schema(
+      const URI& array_uri,
+      ArraySchemaEvolution* array_schema,
+      const EncryptionKey& encryption_key);
+
+  /**
+   * Upgrade a TileDB array to latest format version.
+   *
+   * @param array_uri The URI of the array to be upgraded.
+   * @param config Configuration parameters for the upgrade
+   *     (`nullptr` means default, which will use the config associated with
+   *      this instance).
+   * @return Status
+   */
+  Status array_upgrade_version(const URI& array_uri, const Config* config);
+
+  /**
+   * Gets the memory tracker for an open array.
+   *
+   * @param array_uri The array URI.
+   * @param top_level Specifies if the call is the top level call in recursion.
+   * @return The memory tracker.
+   */
+  OpenArrayMemoryTracker* array_memory_tracker(
+      const URI& array_uri, bool top_level = true);
+
+  /**
+   * Retrieves the non-empty domain from an array. This is the union of the
+   * non-empty domains of the array fragments.
+   *
+   * @param array An open array object (must be already open).
+   * @param domain The domain to be retrieved.
+   * @param is_empty `true` if the non-empty domain is empty (the array
+   *     is empty).
+   * @return Status
+   */
+  Status array_get_non_empty_domain(
+      Array* array, NDRange* domain, bool* is_empty);
+
   /**
    * Retrieves the non-empty domain from an array. This is the union of the
    * non-empty domains of the array fragments.
@@ -273,6 +427,106 @@ class StorageManager {
    */
   Status array_get_non_empty_domain(Array* array, void* domain, bool* is_empty);
 
+  /**
+   * Retrieves the non-empty domain from an array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   *
+   * @param array An open array object (must be already open).
+   * @param idx The dimension index.
+   * @param domain The domain to be retrieved.
+   * @param is_empty `ture` if the non-empty domain is empty (the array
+   *     is empty).
+   * @return Status
+   */
+  Status array_get_non_empty_domain_from_index(
+      Array* array, unsigned idx, void* domain, bool* is_empty);
+
+  /**
+   * Retrieves the non-empty domain from an array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   *
+   * @param array An open array object (must be already open).
+   * @param name The dimension name.
+   * @param domain The domain to be retrieved.
+   * @param is_empty `ture` if the non-empty domain is empty (the array
+   *     is empty).
+   * @return Status
+   */
+  Status array_get_non_empty_domain_from_name(
+      Array* array, const char* name, void* domain, bool* is_empty);
+
+  /**
+   * Retrieves the non-empty domain size from an array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   * Applicable only to var-sized dimensions.
+   *
+   * @param array An open array object (must be already open).
+   * @param idx The dimension index.
+   * @param start_size The size in bytes of the range start.
+   * @param end_size The size in bytes of the range end.
+   * @param is_empty `ture` if the non-empty domain is empty (the array
+   *     is empty).
+   * @return Status
+   */
+  Status array_get_non_empty_domain_var_size_from_index(
+      Array* array,
+      unsigned idx,
+      uint64_t* start_size,
+      uint64_t* end_size,
+      bool* is_empty);
+
+  /**
+   * Retrieves the non-empty domain size from an array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   * Applicable only to var-sized dimensions.
+   *
+   * @param array An open array object (must be already open).
+   * @param name The dimension name.
+   * @param start_size The size in bytes of the range start.
+   * @param end_size The size in bytes of the range end.
+   * @param is_empty `ture` if the non-empty domain is empty (the array
+   *     is empty).
+   * @return Status
+   */
+  Status array_get_non_empty_domain_var_size_from_name(
+      Array* array,
+      const char* name,
+      uint64_t* start_size,
+      uint64_t* end_size,
+      bool* is_empty);
+
+  /**
+   * Retrieves the non-empty domain from an array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   * Applicable only to var-sized dimensions.
+   *
+   * @param array An open array object (must be already open).
+   * @param idx The dimension index.
+   * @param start The domain range start to set.
+   * @param end The domain range end to set.
+   * @param is_empty `ture` if the non-empty domain is empty (the array
+   *     is empty).
+   * @return Status
+   */
+  Status array_get_non_empty_domain_var_from_index(
+      Array* array, unsigned idx, void* start, void* end, bool* is_empty);
+
+  /**
+   * Retrieves the non-empty domain from an array on the given dimension.
+   * This is the union of the non-empty domains of the array fragments.
+   * Applicable only to var-sized dimensions.
+   *
+   * @param array An open array object (must be already open).
+   * @param name The dimension name.
+   * @param start The domain range start to set.
+   * @param end The domain range end to set.
+   * @param is_empty `ture` if the non-empty domain is empty (the array
+   *     is empty).
+   * @return Status
+   */
+  Status array_get_non_empty_domain_var_from_name(
+      Array* array, const char* name, void* start, void* end, bool* is_empty);
+
   /**
    * Retrieves the encryption type from an array.
    *
@@ -331,34 +585,45 @@ class StorageManager {
    * Gets the fragment information for a given array at a particular
    * timestamp.
    *
-   * @param array_schema The array schema.
-   * @param timestamp The function will consider fragments created
+   * @param array The array.
+   * @param timestamp_start The function will consider fragments created
+   *     at or after this timestamp.
+   * @param timestamp_end The function will consider fragments created
    *     at or before this timestamp.
    * @param fragment_info The fragment information to be retrieved.
    *     The fragments are sorted in chronological creation order.
-   * @param encryption_key The encryption key in case the array is encrypted.
+   * @param get_to_vacuum Whether or not to receive information about
+   *     fragments to vacuum.
    * @return Status
    */
   Status get_fragment_info(
-      const ArraySchema* array_schema,
-      uint64_t timestamp,
-      const EncryptionKey& encryption_key,
-      std::vector<FragmentInfo>* fragment_info);
+      const Array& array,
+      uint64_t timestamp_start,
+      uint64_t timestamp_end,
+      FragmentInfo* fragment_info,
+      bool get_to_vacuum = false);
 
   /**
    * Gets the fragment info for a single fragment URI.
    *
-   * @param array_schema The array schema.
-   * @param encryption_key The encryption key.
+   * @param array The array.
    * @param fragment_uri The fragment URI.
    * @param fragment_info The fragment info to retrieve.
    * @return Status
    */
   Status get_fragment_info(
-      const ArraySchema* array_schema,
-      const EncryptionKey& encryption_key,
+      const Array& array,
       const URI& fragment_uri,
-      FragmentInfo* fragment_info);
+      SingleFragmentInfo* fragment_info);
+
+  /**
+   * Retrieves all the fragment URIs of an array, along with the latest
+   * consolidated fragment metadata URI `meta_uri`.
+   */
+  Status get_fragment_uris(
+      const URI& array_uri,
+      std::vector<URI>* fragment_uris,
+      URI* meta_uri) const;
 
   /** Returns the current map of any set tags. */
   const std::unordered_map<std::string, std::string>& tags() const;
@@ -379,6 +644,12 @@ class StorageManager {
    */
   Status init(const Config* config);
 
+  /** Returns the thread pool for compute-bound tasks. */
+  ThreadPool* compute_tp();
+
+  /** Returns the thread pool for io-bound tasks. */
+  ThreadPool* io_tp();
+
   /**
    * If the storage manager was configured with a REST server, return the
    * client instance. Else, return nullptr.
@@ -405,14 +676,22 @@ class StorageManager {
   Status is_dir(const URI& uri, bool* is_dir) const;
 
   /**
-   * Checks if the input URI represents a fragment.
+   * Checks if the input URI represents a fragment. The functions takes into
+   * account the fragment version, which is retrived directly from the URI.
+   * For versions >= 5, the function checks whether the URI is included
+   * in `ok_uris`. For versions < 5, `ok_uris` is empty so the function
+   * checks for the existence of the fragment metadata file in the fragment
+   * URI directory. Therefore, the function is more expensive for earlier
+   * fragment versions.
    *
    * @param The URI to be checked.
-   * @param is_fragment Set to `true` if the URI is a fragment and `false`
+   * @param ok_uris For checking URI existence of versions >= 5.
+   * @param is_fragment Set to `1` if the URI is a fragment and `0`
    *     otherwise.
    * @return Status
    */
-  Status is_fragment(const URI& uri, bool* is_fragment) const;
+  Status is_fragment(
+      const URI& uri, const std::set<URI>& ok_uris, int* is_fragment) const;
 
   /**
    * Checks if the input URI represents a group.
@@ -434,6 +713,48 @@ class StorageManager {
    */
   Status is_file(const URI& uri, bool* is_file) const;
 
+  /**
+   * Check if a URI is a vacuum file or not based on the file suffix
+   * @param uri
+   * @return true is vacuum file, false otherwise
+   */
+  bool is_vacuum_file(const URI& uri) const;
+
+  /**
+   * Retrieve all array schemas for an array uri under its __schema directory.
+   *
+   * @param array_uri The URI path of the array.
+   * @param uris The vector of array schema URIS sorted from earliest to the
+   * latest.
+   * @return Status
+   */
+  Status get_array_schema_uris(
+      const URI& array_uri, std::vector<URI>* schema_uris) const;
+
+  /**
+   * Get latest array schema for an array uri.
+   *
+   * @param array_uri The URI path of the array.
+   * @param uri The latest array schema URI.
+   * @return Status
+   */
+  Status get_latest_array_schema_uri(
+      const URI& array_uri, URI* schema_uri) const;
+
+  /**
+   * Loads the schema of a schema uri from persistent storage into memory.
+   *
+   * @param array_schema_uri The URI path of the array schema.
+   * @param array_uri The URI path of the array.
+   * @param encryption_key The encryption key to use.
+   * @param array_schema The array schema to be retrieved.
+   * @return Status
+   */
+  Status load_array_schema_from_uri(
+      const URI& array_schema_uri,
+      const EncryptionKey& encryption_key,
+      ArraySchema** array_schema);
+
   /**
    * Loads the schema of an array from persistent storage into memory.
    *
@@ -447,14 +768,32 @@ class StorageManager {
       const EncryptionKey& encryption_key,
       ArraySchema** array_schema);
 
+  /**
+   * Loads all schemas of an array from persistent storage into memory.
+   *
+   * @param array_uri The URI path of the array.
+   * @param encryption_key The encryption key to use.
+   * @return tuple of Status and optional unordered map. If Status is an error
+   * the unordered_map will be nullopt
+   *        Status Ok on success, else error
+   *        ArraySchemaMap Map of all array schemas found keyed by name
+   */
+  std::tuple<
+      Status,
+      std::optional<
+          std::unordered_map<std::string, tdb_shared_ptr<ArraySchema>>>>
+  load_all_array_schemas(
+      const URI& array_uri, const EncryptionKey& encryption_key);
+
   /**
    * Loads the array metadata from persistent storage that were created
-   * at or before `timestamp`.
+   * at or before `timestamp_end` and at or after `timestamp_start`.
    */
   Status load_array_metadata(
       const URI& array_uri,
       const EncryptionKey& encryption_key,
-      uint64_t timestamp,
+      uint64_t timestamp_start,
+      uint64_t timestamp_end,
       Metadata* metadata);
 
   /** Removes a TileDB object (group, array). */
@@ -582,9 +921,6 @@ class StorageManager {
       uint64_t nbytes,
       bool* in_cache) const;
 
-  /** Returns the Reader thread pool. */
-  ThreadPool* reader_thread_pool();
-
   /**
    * Reads from a file into the input buffer.
    *
@@ -598,6 +934,18 @@ class StorageManager {
   Status read(
       const URI& uri, uint64_t offset, Buffer* buffer, uint64_t nbytes) const;
 
+  /**
+   * Reads from a file into the raw input buffer.
+   *
+   * @param uri The URI file to read from.
+   * @param offset The offset in the file the read will start from.
+   * @param buffer The buffer to write into.
+   * @param nbytes The number of bytes to read.
+   * @return Status.
+   */
+  Status read(
+      const URI& uri, uint64_t offset, void* buffer, uint64_t nbytes) const;
+
   /**
    * Sets a string/string KV "tag" on the storage manager instance.
    *
@@ -622,7 +970,7 @@ class StorageManager {
   /**
    * Stores the array metadata into persistent storage.
    *
-   * @param array_uri The URI of the array.
+   * @param array_uri The array URI.
    * @param encryption_key The encryption key to use.
    * @param array_metadata The array metadata.
    * @return Status
@@ -638,9 +986,6 @@ class StorageManager {
   /** Syncs a file or directory, flushing its contents to persistent storage. */
   Status sync(const URI& uri);
 
-  /** Returns the Writer thread pool. */
-  ThreadPool* writer_thread_pool();
-
   /** Returns the virtual filesystem object. */
   VFS* vfs() const;
 
@@ -677,6 +1022,12 @@ class StorageManager {
    */
   Status write(const URI& uri, void* data, uint64_t size) const;
 
+  /** Returns `stats_`. */
+  stats::Stats* stats();
+
+  /** Returns the internal logger object. */
+  tdb_shared_ptr<Logger> logger() const;
+
  private:
   /* ********************************* */
   /*        PRIVATE DATATYPES          */
@@ -709,6 +1060,12 @@ class StorageManager {
   /*        PRIVATE ATTRIBUTES         */
   /* ********************************* */
 
+  /** The class stats. */
+  stats::Stats* stats_;
+
+  /** The class logger. */
+  tdb_shared_ptr<Logger> logger_;
+
   /** Set to true when tasks are being cancelled. */
   bool cancellation_in_progress_;
 
@@ -755,14 +1112,11 @@ class StorageManager {
   /** Guards queries_in_progress_ counter. */
   std::condition_variable queries_in_progress_cv_;
 
-  /** The storage manager's thread pool for async queries. */
-  ThreadPool async_thread_pool_;
+  /** The thread pool for compute-bound tasks. */
+  ThreadPool* const compute_tp_;
 
-  /** The storage manager's thread pool for Readers. */
-  ThreadPool reader_thread_pool_;
-
-  /** The storage manager's thread pool for Writers. */
-  ThreadPool writer_thread_pool_;
+  /** The thread pool for io-bound tasks. */
+  ThreadPool* const io_tp_;
 
   /** Tracks all scheduled tasks that can be safely cancelled before execution.
    */
@@ -772,7 +1126,7 @@ class StorageManager {
   std::unordered_map<std::string, std::string> tags_;
 
   /** A tile cache. */
-  LRUCache* tile_cache_;
+  tdb_unique_ptr<BufferLRUCache> tile_cache_;
 
   /**
    * Virtual filesystem handler. It directs queries to the appropriate
@@ -781,27 +1135,12 @@ class StorageManager {
   VFS* vfs_;
 
   /** The rest client (may be null if none was configured). */
-  std::unique_ptr<RestClient> rest_client_;
+  tdb_unique_ptr<RestClient> rest_client_;
 
   /* ********************************* */
   /*         PRIVATE METHODS           */
   /* ********************************* */
 
-  /**
-   * Retrieves the non-empty domain from the input fragment metadata. This is
-   * the union of the non-empty domains of the fragments.
-   *
-   * @param metadata The metadata of all fragments in the array.
-   * @param dim_num The number of dimensions in the domain.
-   * @param domain The domain to be retrieved.
-   * @return void
-   */
-  template <class T>
-  void array_get_non_empty_domain(
-      const std::vector<FragmentMetadata*>& metadata,
-      unsigned dim_num,
-      T* domain);
-
   /**
    * This is an auxiliary function to the other `array_open*` functions.
    * It opens the array, retrieves an `OpenArray` instance, acquires
@@ -823,10 +1162,6 @@ class StorageManager {
   /** Decrement the count of in-progress queries. */
   void decrement_in_progress();
 
-  /** Retrieves all the fragment URI's of an array. */
-  Status get_fragment_uris(
-      const URI& array_uri, std::vector<URI>* fragment_uris) const;
-
   /** Retrieves all the array metadata URI's of an array. */
   Status get_array_metadata_uris(
       const URI& array_uri, std::vector<URI>* array_metadata_uris) const;
@@ -878,6 +1213,12 @@ class StorageManager {
    * @param open_array The open array object.
    * @param encryption_key The encryption key to use.
    * @param fragments_to_load The fragments whose metadata to load.
+   * @param meta_buff A buffer that contains the consolidated fragment
+   *     metadata.
+   * @param offsets A map from a fragment name to an offset in `meta_buff`
+   *     where the basic metadata can be found. If the offset cannot be
+   *     found, then the metadata of that fragment will be loaded from
+   *     storage instead.
    * @param fragment_metadata The fragment metadata retrieved in a
    *     vector.
    * @return Status
@@ -886,21 +1227,62 @@ class StorageManager {
       OpenArray* open_array,
       const EncryptionKey& encryption_key,
       const std::vector<TimestampedURI>& fragments_to_load,
-      std::vector<FragmentMetadata*>* fragment_metadata);
+      Buffer* meta_buff,
+      const std::unordered_map<std::string, uint64_t>& offsets,
+      std::vector<tdb_shared_ptr<FragmentMetadata>>* fragment_metadata);
+
+  /**
+   * Loads the latest consolidated fragment metadata from storage.
+   *
+   * @param uri The URI of the consolidated fragment metadata.
+   * @param enc_key The encryption key that may be needed to access the file.
+   * @param f_buff The buffer to hold the consolidated fragment metadata.
+   * @param offsets A map from the fragment name to the offset in `f_buff` where
+   *     the basic fragment metadata starts.
+   * @return Status
+   */
+  Status load_consolidated_fragment_meta(
+      const URI& uri,
+      const EncryptionKey& enc_key,
+      Buffer* f_buff,
+      std::unordered_map<std::string, uint64_t>* offsets);
+
+  /**
+   * Retrieves the URI of the latest consolidated fragment metadata,
+   * among the URIs in `uris`.
+   */
+  Status get_consolidated_fragment_meta_uri(
+      const std::vector<URI>& uris, URI* meta_uri) const;
 
   /**
    * Applicable to fragment and array metadata URIs.
    *
    * Gets the sorted URIs in ascending first timestamp order,
    * breaking ties with lexicographic
-   * sorting of UUID. Only the URIs with timestamp smaller than or
-   * equal to `timestamp` are considered. The sorted URIs are
+   * sorting of UUID. Only the URIs with timestamp between `timestamp_start`
+   * and `timestamp_end` (inclusive) are considered. The sorted URIs are
    * stored in the last input, including their timestamps.
    */
   Status get_sorted_uris(
       const std::vector<URI>& uris,
-      uint64_t timestamp,
-      std::vector<TimestampedURI>* sorted_uris) const;
+      std::vector<TimestampedURI>* sorted_uris,
+      uint64_t timestamp_start,
+      uint64_t timestamp_end) const;
+
+  /**
+   * It computes the URIs `to_vacuum` from the input `uris`, considering
+   * only the URIs whose first timestamp is greater than or equal to
+   * `timestamp_start` and second timestamp is smaller than or equal to
+   * `timestamp_end`. The function also retrieves the `vac_uris` (files with
+   * `.vac` suffix) that were used to compute `to_vacuum`.
+   */
+  Status get_uris_to_vacuum(
+      const std::vector<URI>& uris,
+      uint64_t timestamp_start,
+      uint64_t timestamp_end,
+      std::vector<URI>* to_vacuum,
+      std::vector<URI>* vac_uris,
+      bool allow_partial = true) const;
 
   /** Block until there are zero in-progress queries. */
   void wait_for_zero_in_progress();
@@ -908,15 +1290,6 @@ class StorageManager {
   /** Initializes a REST client, if one was configured. */
   Status init_rest_client();
 
-  /**
-   * Retrieves a new array metadata URI, incorporating the input timestamp
-   * range.
-   */
-  Status new_array_metadata_uri(
-      const URI& array_uri,
-      const std::pair<uint64_t, uint64_t>& timestamp_range,
-      URI* new_uri) const;
-
   /** Sets default tag values on this StorageManager. */
   Status set_default_tags();
 };
